/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// baseset.c
///  Definition of basesets
//   Pierre G. Richard 87/07/02 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include "syntax.eh"
#include "baseset.h"

#define CH_CONTROL     S_SHUNNABLE | S_CONTROLS  /* controls */
#define CH_USED        S_SHUNNABLE               /* graphic - mindata */
#define CH_UNUSED      S_NONSGML

/*STATIC----------------------------------------------------------find_baseset-+
| Returns a pointer to a SetDescriptor corresponding to the identifier.        |
| If not found, returns a NULL pointer.                                        |
|                                                                              |
| Each baseset is defined by an unique identifier (parameter literal).         |
| The SetDescriptor structure describes, for each code position, the type of   |
| character associated, its translation onto the system character set,         |
| the translation of the upper case form onto the system character set.        |
|                                                                              |
| As the two last values are system-dependent, to allow a greater system       |
| independency of this code:                                                   |
|    - if possible, the character described is entered directly,               |
|      so the translation between system is done automatically                 |
|      when this program is compiled,                                          |
|    - if not, the integer value must be entered.  As it is not the            |
|      same, C pre-processor conditionals must be used:                        |
|      For example, #if (0x20 == ' ') will correspond to ASCII system set.     |
|                                                                              |
| Implementation note:                                                         |
|   - only ASCII or EBCDIC system character set are implemented so far         |
|   - on the 8086, a loadds is done to access to the table,                    |
|     because ds is no more at sp.                                             |
+-----------------------------------------------------------------------------*/
const SetDescriptor * find_baseset(const char *pszPubid)
{
   /*-------------------------------------------------------------+
   | ISO 646-1983//CHARSET International Reference Version (IRV)  |
   +-------------------------------------------------------------*/
   static SetDescriptor ISO_646_1983_chtab[MAX_CHARSET] = {
   /*  00 NUL */  { CH_CONTROL   , 0x00  , 0x00  },  // same
   /*  01 SOH */  { CH_CONTROL   , 0x01  , 0x01  },  // same
   /*  02 STX */  { CH_CONTROL   , 0x02  , 0x02  },  // same
   /*  03 ETX */  { CH_CONTROL   , 0x03  , 0x03  },  // same
   #if (0x20 == ' ')
   /*  04 EOT */  { CH_CONTROL   , 0x04  , 0x04  },  // ASCII
   /*  05 ENQ */  { CH_CONTROL   , 0x05  , 0x05  },  // ASCII
   /*  06 ACK */  { CH_CONTROL   , 0x06  , 0x06  },  // ASCII
   #else
   /*  04 EOT */  { CH_CONTROL   , 0x37  , 0x37  },  // EBCDIC
   /*  05 ENQ */  { CH_CONTROL   , 0x2D  , 0x2D  },  // EBCDIC
   /*  06 ACK */  { CH_CONTROL   , 0x2E  , 0x2E  },  // EBCDIC
   #endif
   /*  07 BEL */  { CH_CONTROL   , '\a'  , '\a'  },
   /*  08 BS  */  { CH_CONTROL   , '\b'  , '\b'  },
   /*  09 HT  */  { CH_CONTROL   , '\t'  , '\t'  },
   /*  0A LF  */  { CH_CONTROL   , '\n'  , '\n'  },
   /*  0B VT  */  { CH_CONTROL   , '\v'  , '\v'  },
   /*  0C FF  */  { CH_CONTROL   , '\f'  , '\f'  },
   /*  0D CR  */  { CH_CONTROL   , '\r'  , '\r'  },
   /*  0E SO  */  { CH_CONTROL   , 0x0E  , 0x0E  },  // same
   /*  0F SI  */  { CH_CONTROL   , 0x0F  , 0x0F  },  // same
   /*  10 DLE */  { CH_CONTROL   , 0x10  , 0x10  },  // same
   /*  11 DC1 */  { CH_CONTROL   , 0x11  , 0x11  },  // same
   /*  12 DC2 */  { CH_CONTROL   , 0x12  , 0x12  },  // same
   /*  13 DC3 */  { CH_CONTROL   , 0x13  , 0x13  },  // same
   #if (0x20 == ' ')
   /*  14 DC4 */  { CH_CONTROL   , 0x14  , 0x14  },  // ASCII
   /*  15 NAK */  { CH_CONTROL   , 0x15  , 0x15  },  // ASCII
   /*  16 SYN */  { CH_CONTROL   , 0x16  , 0x16  },  // ASCII
   /*  17 ETB */  { CH_CONTROL   , 0x17  , 0x17  },  // ASCII
   #else
   /*  14 DC4 */  { CH_CONTROL   , 0x3C  , 0x3C  },  // EBCDIC
   /*  15 NAK */  { CH_CONTROL   , 0x3D  , 0x3D  },  // EBCDIC
   /*  16 SYN */  { CH_CONTROL   , 0x32  , 0x32  },  // EBCDIC
   /*  17 ETB */  { CH_CONTROL   , 0x26  , 0x26  },  // EBCDIC
   #endif
   /*  18 CAN */  { CH_CONTROL   , 0x18  , 0x18  },  // same
   /*  19 EM  */  { CH_CONTROL   , 0x19  , 0x19  },  // same
   #if (0x20 == ' ')
   /*  1A SUB */  { CH_CONTROL   , 0x1A  , 0x1A  },  // ASCII
   /*  1B ESC */  { CH_CONTROL   , 0x1B  , 0x1B  },  // ASCII
   /*  1C FS  */  { CH_CONTROL   , 0x1C  , 0x1C  },  // ASCII
   #else
   /*  1A SUB */  { CH_CONTROL   , 0x3F  , 0x3F  },  // EBCDIC
   /*  1B ESC */  { CH_CONTROL   , 0x27  , 0x27  },  // EBCDIC
   /*  1C FS  */  { CH_CONTROL   , 0x22  , 0x22  },  // EBCDIC
   #endif
   /*  1D GS  */  { CH_CONTROL   , 0x1D  , 0x1D  },  // same ?
   /*  1E RS  */  { CH_CONTROL   , 0x1E  , 0x1E  },  // same ?
   /*  1F US  */  { CH_CONTROL   , 0x1F  , 0x1F  },  // same ?
   /*  20 SPC */  { CH_USED      , ' '   , ' '   },  // Space
   /*  21 !   */  { CH_USED      , '!'   , '!'   },  // Exclamation mark
   /*  22 "   */  { CH_USED      , '"'   , '"'   },  // Double quote
   /*  23 #   */  { CH_USED      , '#'   , '#'   },  // Number sign
   /*  24 CUR */  { CH_USED      , ''   , ''   },  // Currency sign
   /*  25 %   */  { CH_USED      , '%'   , '%'   },  // Percent sign
   /*  26 &   */  { CH_USED      , '&'   , '&'   },  // Ampersand
   /*  27 '   */  { S_SPECIAL    , '\''  , '\''  },  // Single quote
   /*  28 (   */  { S_SPECIAL    , '('   , '('   },  // Left parenthesis
   /*  29 )   */  { S_SPECIAL    , ')'   , ')'   },  // Right parenthesis
   /*  2A *   */  { CH_USED      , '*'   , '*'   },  // Asterisk
   /*  2B +   */  { S_SPECIAL    , '+'   , '+'   },  // Plus sign
   /*  2C ,   */  { S_SPECIAL    , ','   , ','   },  // Comma
   /*  2D -   */  { S_SPECIAL    , '-'   , '-'   },  // Dash
   /*  2E .   */  { S_SPECIAL    , '.'   , '.'   },  // Period
   /*  2F /   */  { S_SPECIAL    , '/'   , '/'   },  // Solidus
   /*  30 0   */  { S_DIGIT      , '0'   , '0'   },  // Digit 0
   /*  31 1   */  { S_DIGIT      , '1'   , '1'   },  // Digit 1
   /*  32 2   */  { S_DIGIT      , '2'   , '2'   },  // Digit 2
   /*  33 3   */  { S_DIGIT      , '3'   , '3'   },  // Digit 3
   /*  34 4   */  { S_DIGIT      , '4'   , '4'   },  // Digit 4
   /*  35 5   */  { S_DIGIT      , '5'   , '5'   },  // Digit 5
   /*  36 6   */  { S_DIGIT      , '6'   , '6'   },  // Digit 6
   /*  37 7   */  { S_DIGIT      , '7'   , '7'   },  // Digit 7
   /*  38 8   */  { S_DIGIT      , '8'   , '8'   },  // Digit 8
   /*  39 9   */  { S_DIGIT      , '9'   , '9'   },  // Digit 9
   /*  3A :   */  { S_SPECIAL    , ':'   , ':'   },  // Colon
   /*  3B ;   */  { CH_USED      , ';'   , ';'   },  // Semi-colon
   /*  3C <   */  { CH_USED      , '<'   , '<'   },  // Less than sign
   /*  3D =   */  { S_SPECIAL    , '='   , '='   },  // Equal
   /*  3E >   */  { CH_USED      , '>'   , '>'   },  // Greater than sign
   /*  3F ?   */  { S_SPECIAL    , '?'   , '?'   },  // Interrogation mark
   /*  40 @   */  { CH_USED      , '@'   , '@'   },  // At sign
   /*  41 A   */  { S_LETTER     , 'A'   , 'A'   },  // Capital letter A
   /*  42 B   */  { S_LETTER     , 'B'   , 'B'   },  // Capital letter B
   /*  43 C   */  { S_LETTER     , 'C'   , 'C'   },  // Capital letter C
   /*  44 D   */  { S_LETTER     , 'D'   , 'D'   },  // Capital letter D
   /*  45 E   */  { S_LETTER     , 'E'   , 'E'   },  // Capital letter E
   /*  46 F   */  { S_LETTER     , 'F'   , 'F'   },  // Capital letter F
   /*  47 G   */  { S_LETTER     , 'G'   , 'G'   },  // Capital letter G
   /*  48 H   */  { S_LETTER     , 'H'   , 'H'   },  // Capital letter H
   /*  49 I   */  { S_LETTER     , 'I'   , 'I'   },  // Capital letter I
   /*  4A J   */  { S_LETTER     , 'J'   , 'J'   },  // Capital letter J
   /*  4B K   */  { S_LETTER     , 'K'   , 'K'   },  // Capital letter K
   /*  4C L   */  { S_LETTER     , 'L'   , 'L'   },  // Capital letter L
   /*  4D M   */  { S_LETTER     , 'M'   , 'M'   },  // Capital letter M
   /*  4E N   */  { S_LETTER     , 'N'   , 'N'   },  // Capital letter N
   /*  4F O   */  { S_LETTER     , 'O'   , 'O'   },  // Capital letter O
   /*  50 P   */  { S_LETTER     , 'P'   , 'P'   },  // Capital letter P
   /*  51 Q   */  { S_LETTER     , 'Q'   , 'Q'   },  // Capital letter Q
   /*  52 R   */  { S_LETTER     , 'R'   , 'R'   },  // Capital letter R
   /*  53 S   */  { S_LETTER     , 'S'   , 'S'   },  // Capital letter S
   /*  54 T   */  { S_LETTER     , 'T'   , 'T'   },  // Capital letter T
   /*  55 U   */  { S_LETTER     , 'U'   , 'U'   },  // Capital letter U
   /*  56 V   */  { S_LETTER     , 'V'   , 'V'   },  // Capital letter V
   /*  57 W   */  { S_LETTER     , 'W'   , 'W'   },  // Capital letter W
   /*  58 X   */  { S_LETTER     , 'X'   , 'X'   },  // Capital letter X
   /*  59 Y   */  { S_LETTER     , 'Y'   , 'Y'   },  // Capital letter Y
   /*  5A Z   */  { S_LETTER     , 'Z'   , 'Z'   },  // Capital letter Z
   /*  5B [   */  { CH_USED      , '['   , '['   },  // Left bracket
   /*  5C \   */  { CH_USED      , '\\'  , '\\'  },  // Back solidus
   /*  5D ]   */  { CH_USED      , ']'   , ']'   },  // Right bracket
   /*  5E ^   */  { CH_USED      , '^'   , '^'   },  // Circumflex accent
   /*  5F _   */  { CH_USED      , '_'   , '_'   },  // Underscore
   /*  60 `   */  { CH_USED      , '`'   , '`'   },  // Back quote
   /*  61 a   */  { S_LETTER     , 'a'   , 'A'   },  // Small letter a
   /*  62 b   */  { S_LETTER     , 'b'   , 'B'   },  // Small letter b
   /*  63 c   */  { S_LETTER     , 'c'   , 'C'   },  // Small letter c
   /*  64 d   */  { S_LETTER     , 'd'   , 'D'   },  // Small letter d
   /*  65 e   */  { S_LETTER     , 'e'   , 'E'   },  // Small letter e
   /*  66 f   */  { S_LETTER     , 'f'   , 'F'   },  // Small letter f
   /*  67 g   */  { S_LETTER     , 'g'   , 'G'   },  // Small letter g
   /*  68 h   */  { S_LETTER     , 'h'   , 'H'   },  // Small letter h
   /*  69 i   */  { S_LETTER     , 'i'   , 'I'   },  // Small letter i
   /*  6A j   */  { S_LETTER     , 'j'   , 'J'   },  // Small letter j
   /*  6B k   */  { S_LETTER     , 'k'   , 'K'   },  // Small letter k
   /*  6C l   */  { S_LETTER     , 'l'   , 'L'   },  // Small letter l
   /*  6D m   */  { S_LETTER     , 'm'   , 'M'   },  // Small letter m
   /*  6E n   */  { S_LETTER     , 'n'   , 'N'   },  // Small letter n
   /*  6F o   */  { S_LETTER     , 'o'   , 'O'   },  // Small letter o
   /*  70 p   */  { S_LETTER     , 'p'   , 'P'   },  // Small letter p
   /*  71 q   */  { S_LETTER     , 'q'   , 'Q'   },  // Small letter q
   /*  72 r   */  { S_LETTER     , 'r'   , 'R'   },  // Small letter r
   /*  73 s   */  { S_LETTER     , 's'   , 'S'   },  // Small letter s
   /*  74 t   */  { S_LETTER     , 't'   , 'T'   },  // Small letter t
   /*  75 u   */  { S_LETTER     , 'u'   , 'U'   },  // Small letter u
   /*  76 v   */  { S_LETTER     , 'v'   , 'V'   },  // Small letter v
   /*  77 w   */  { S_LETTER     , 'w'   , 'W'   },  // Small letter w
   /*  78 x   */  { S_LETTER     , 'x'   , 'X'   },  // Small letter x
   /*  79 y   */  { S_LETTER     , 'y'   , 'Y'   },  // Small letter y
   /*  7A z   */  { S_LETTER     , 'z'   , 'Z'   },  // Small letter z
   /*  7B {   */  { CH_USED      , '{'   , '{'   },  // Left brace
   /*  7C |   */  { CH_USED      , '|'   , '|'   },  // Vertical bar
   /*  7D }   */  { CH_USED      , '}'   , '}'   },  // Right brace
   /*  7E ~   */  { CH_USED      , '~'   , '~'   },  // Tilda
   #if (0x20 == ' ')
   /*  7F DEL */  { CH_CONTROL   , 0x7F  , 0x7F  },  // ASCII
   #else
   /*  7F DEL */  { CH_CONTROL   , 0x07  , 0x07  },  // EBCDIC
   #endif
   /*  80    */  { CH_UNUSED    , ''   , ''   },
   /*  81    */  { CH_UNUSED    , ''   , ''   },
   /*  82    */  { CH_UNUSED    , ''   , ''   },
   /*  83    */  { CH_UNUSED    , ''   , ''   },
   /*  84    */  { CH_UNUSED    , ''   , ''   },
   /*  85    */  { CH_UNUSED    , ''   , ''   },
   /*  86    */  { CH_UNUSED    , ''   , ''   },
   /*  87    */  { CH_UNUSED    , ''   , ''   },
   /*  88    */  { CH_UNUSED    , ''   , ''   },
   /*  89    */  { CH_UNUSED    , ''   , ''   },
   /*  8A    */  { CH_UNUSED    , ''   , ''   },
   /*  8B    */  { CH_UNUSED    , ''   , ''   },
   /*  8C    */  { CH_UNUSED    , ''   , ''   },
   /*  8D    */  { CH_UNUSED    , ''   , ''   },
   /*  8E    */  { CH_UNUSED    , ''   , ''   },
   /*  8F    */  { CH_UNUSED    , ''   , ''   },
   /*  90    */  { CH_UNUSED    , ''   , ''   },
   /*  91    */  { CH_UNUSED    , ''   , ''   },
   /*  92    */  { CH_UNUSED    , ''   , ''   },
   /*  93    */  { CH_UNUSED    , ''   , ''   },
   /*  94    */  { CH_UNUSED    , ''   , ''   },
   /*  95    */  { CH_UNUSED    , ''   , ''   },
   /*  96    */  { CH_UNUSED    , ''   , ''   },
   /*  97    */  { CH_UNUSED    , ''   , ''   },
   /*  98    */  { CH_UNUSED    , ''   , ''   },
   /*  99    */  { CH_UNUSED    , ''   , ''   },
   /*  9A    */  { CH_UNUSED    , ''   , ''   },
   /*  9B    */  { CH_UNUSED    , ''   , ''   },
   /*  9C    */  { CH_UNUSED    , ''   , ''   },
   /*  9D    */  { CH_UNUSED    , ''   , ''   },
   /*  9E    */  { CH_UNUSED    , ''   , ''   },
   /*  9F    */  { CH_UNUSED    , ''   , ''   },
   /*  A0    */  { CH_UNUSED    , ''   , ''   },
   /*  A1    */  { CH_UNUSED    , ''   , ''   },
   /*  A2    */  { CH_UNUSED    , ''   , ''   },
   /*  A3    */  { CH_UNUSED    , ''   , ''   },
   /*  A4    */  { CH_UNUSED    , ''   , ''   },
   /*  A5    */  { CH_UNUSED    , ''   , ''   },
   /*  A6    */  { CH_UNUSED    , ''   , ''   },
   /*  A7    */  { CH_UNUSED    , ''   , ''   },
   /*  A8    */  { CH_UNUSED    , ''   , ''   },
   /*  A9    */  { CH_UNUSED    , ''   , ''   },
   /*  AA    */  { CH_UNUSED    , ''   , ''   },
   /*  AB    */  { CH_UNUSED    , ''   , ''   },
   /*  AC    */  { CH_UNUSED    , ''   , ''   },
   /*  AD    */  { CH_UNUSED    , ''   , ''   },
   /*  AE    */  { CH_UNUSED    , ''   , ''   },
   /*  AF    */  { CH_UNUSED    , ''   , ''   },
   /*  B0    */  { CH_UNUSED    , ''   , ''   },
   /*  B1    */  { CH_UNUSED    , ''   , ''   },
   /*  B2    */  { CH_UNUSED    , ''   , ''   },
   /*  B3    */  { CH_UNUSED    , ''   , ''   },
   #if (0x20 == ' ')
   /*  B4    */  { CH_UNUSED    , 0xB4  ,0xB4   },  // ASCII
   #else
   /*  B4    */  { CH_UNUSED    , 0x15  ,0x15   },  // EBCDIC
   #endif
   /*  B5    */  { CH_UNUSED    , ''   , ''   },
   /*  B6    */  { CH_UNUSED    , ''   , ''   },
   /*  B7    */  { CH_UNUSED    , ''   , ''   },
   /*  B8    */  { CH_UNUSED    , ''   , ''   },
   /*  B9    */  { CH_UNUSED    , ''   , ''   },
   /*  BA    */  { CH_UNUSED    , ''   , ''   },
   /*  BB    */  { CH_UNUSED    , ''   , ''   },
   /*  BC    */  { CH_UNUSED    , ''   , ''   },
   /*  BD    */  { CH_UNUSED    , ''   , ''   },
   /*  BE    */  { CH_UNUSED    , ''   , ''   },
   /*  BF    */  { CH_UNUSED    , ''   , ''   },
   /*  C0    */  { CH_UNUSED    , ''   , ''   },
   /*  C1    */  { CH_UNUSED    , ''   , ''   },
   /*  C2    */  { CH_UNUSED    , ''   , ''   },
   /*  C3    */  { CH_UNUSED    , ''   , ''   },
   /*  C4    */  { CH_UNUSED    , ''   , ''   },
   #if (0x20 == ' ')
   /*  C5    */  { CH_UNUSED    , ''   , ''   },
   #else
   /*  C5    */  { CH_UNUSED    , 0x15  , 0x15  },  // EBCDIC
   #endif
   /*  C6    */  { CH_UNUSED    , ''   , ''   },
   /*  C7    */  { CH_UNUSED    , ''   , ''   },
   /*  C8    */  { CH_UNUSED    , ''   , ''   },
   /*  C9    */  { CH_UNUSED    , ''   , ''   },
   /*  CA    */  { CH_UNUSED    , ''   , ''   },
   /*  CB    */  { CH_UNUSED    , ''   , ''   },
   /*  CC    */  { CH_UNUSED    , ''   , ''   },
   /*  CD    */  { CH_UNUSED    , ''   , ''   },
   /*  CE    */  { CH_UNUSED    , ''   , ''   },
   /*  CF    */  { CH_UNUSED    , ''   , ''   },
   /*  D0    */  { CH_UNUSED    , ''   , ''   },
   /*  D1    */  { CH_UNUSED    , ''   , ''   },
   /*  D2    */  { CH_UNUSED    , ''   , ''   },
   /*  D3    */  { CH_UNUSED    , ''   , ''   },
   /*  D4    */  { CH_UNUSED    , ''   , ''   },
   /*  D5    */  { CH_UNUSED    , ''   , ''   },
   /*  D6    */  { CH_UNUSED    , ''   , ''   },
   /*  D7    */  { CH_UNUSED    , ''   , ''   },
   /*  D8    */  { CH_UNUSED    , ''   , ''   },
   /*  D9    */  { CH_UNUSED    , ''   , ''   },
   /*  DA    */  { CH_UNUSED    , ''   , ''   },
   /*  DB    */  { CH_UNUSED    , ''   , ''   },
   /*  DC    */  { CH_UNUSED    , ''   , ''   },
   /*  DD    */  { CH_UNUSED    , ''   , ''   },
   /*  DE    */  { CH_UNUSED    , ''   , ''   },
   /*  DF    */  { CH_UNUSED    , ''   , ''   },
   /*  E0    */  { CH_UNUSED    , ''   , ''   },
   /*  E1    */  { CH_UNUSED    , ''   , ''   },
   /*  E2    */  { CH_UNUSED    , ''   , ''   },
   /*  E3    */  { CH_UNUSED    , ''   , ''   },
   /*  E4    */  { CH_UNUSED    , ''   , ''   },
   /*  E5    */  { CH_UNUSED    , ''   , ''   },
   /*  E6    */  { CH_UNUSED    , ''   , ''   },
   /*  E7    */  { CH_UNUSED    , ''   , ''   },
   /*  E8    */  { CH_UNUSED    , ''   , ''   },
   /*  E9    */  { CH_UNUSED    , ''   , ''   },
   /*  EA    */  { CH_UNUSED    , ''   , ''   },
   /*  EB    */  { CH_UNUSED    , ''   , ''   },
   /*  EC    */  { CH_UNUSED    , ''   , ''   },
   /*  ED    */  { CH_UNUSED    , ''   , ''   },
   /*  EE    */  { CH_UNUSED    , ''   , ''   },
   /*  EF    */  { CH_UNUSED    , ''   , ''   },
   /*  F0    */  { CH_UNUSED    , ''   , ''   },
   /*  F1    */  { CH_UNUSED    , ''   , ''   },
   /*  F2    */  { CH_UNUSED    , ''   , ''   },
   /*  F3    */  { CH_UNUSED    , ''   , ''   },
   /*  F4    */  { CH_UNUSED    , ''   , ''   },
   /*  F5    */  { CH_UNUSED    , ''   , ''   },
   /*  F6    */  { CH_UNUSED    , ''   , ''   },
   /*  F7    */  { CH_UNUSED    , ''   , ''   },
   /*  F8    */  { CH_UNUSED    , ''   , ''   },
   /*  F9    */  { CH_UNUSED    , ''   , ''   },
   /*  FA    */  { CH_UNUSED    , ''   , ''   },
   /*  FB    */  { CH_UNUSED    , ''   , ''   },
   /*  FC    */  { CH_UNUSED    , ''   , ''   },
   /*  FD    */  { CH_UNUSED    , ''   , ''   },
   /*  FE    */  { CH_UNUSED    , ''   , ''   },
   /*  FF     */  { CH_CONTROL   , 0xFF  , 0xFF  }   // same ?
   };

   /*-------------------------------------------------------------+
   | ISO 8859:1987//CHARSET International Reference Version (IRV) |
   +-------------------------------------------------------------*/
   static SetDescriptor ISO_8859_chtab[MAX_CHARSET] = {
   /*  00 N/A */  { CH_UNUSED    , 0x00  , 0x00  },
   /*  01 N/A */  { CH_UNUSED    , 0x01  , 0x01  },
   /*  02 N/A */  { CH_UNUSED    , 0x02  , 0x02  },
   /*  03 N/A */  { CH_UNUSED    , 0x03  , 0x03  },
   /*  04 N/A */  { CH_UNUSED    , 0x04  , 0x04  },
   /*  05 N/A */  { CH_UNUSED    , 0x05  , 0x05  },
   /*  06 N/A */  { CH_UNUSED    , 0x06  , 0x06  },
   /*  07 N/A */  { CH_UNUSED    , 0x07  , 0x07  },
   /*  08 N/A */  { CH_UNUSED    , 0x08  , 0x08  },
   /*  09 N/A */  { CH_UNUSED    , 0x09  , 0x09  },
   /*  0A N/A */  { CH_UNUSED    , 0x0A  , 0x0A  },
   /*  0B N/A */  { CH_UNUSED    , 0x0B  , 0x0B  },
   /*  0C N/A */  { CH_UNUSED    , 0x0C  , 0x0C  },
   /*  0D N/A */  { CH_UNUSED    , 0x0D  , 0x0D  },
   /*  0E N/A */  { CH_UNUSED    , 0x0E  , 0x0E  },
   /*  0F N/A */  { CH_UNUSED    , 0x0F  , 0x0F  },
   /*  10 N/A */  { CH_UNUSED    , 0x10  , 0x10  },
   /*  11 N/A */  { CH_UNUSED    , 0x11  , 0x11  },
   /*  12 N/A */  { CH_UNUSED    , 0x12  , 0x12  },
   /*  13 N/A */  { CH_UNUSED    , 0x13  , 0x13  },
   /*  14 N/A */  { CH_UNUSED    , 0x14  , 0x14  },
   /*  15 N/A */  { CH_UNUSED    , 0x15  , 0x15  },
   /*  16 N/A */  { CH_UNUSED    , 0x16  , 0x16  },
   /*  17 N/A */  { CH_UNUSED    , 0x17  , 0x17  },
   /*  18 N/A */  { CH_UNUSED    , 0x18  , 0x18  },
   /*  19 N/A */  { CH_UNUSED    , 0x19  , 0x19  },
   /*  1A N/A */  { CH_UNUSED    , 0x1A  , 0x1A  },
   /*  1B N/A */  { CH_UNUSED    , 0x1B  , 0x1B  },
   /*  1C N/A */  { CH_UNUSED    , 0x1C  , 0x1C  },
   /*  1D N/A */  { CH_UNUSED    , 0x1D  , 0x1D  },
   /*  1E N/A */  { CH_UNUSED    , 0x1E  , 0x1E  },
   /*  1F N/A */  { CH_UNUSED    , 0x1F  , 0x1F  },
   /*  20 SPC */  { CH_USED      , ' '   , ' '   },  // Space
   /*  21 !   */  { CH_USED      , '!'   , '!'   },  // Exclamation mark
   /*  22 "   */  { CH_USED      , '"'   , '"'   },  // Double quote
   /*  23 #   */  { CH_USED      , '#'   , '#'   },  // Number sign
   /*  24 $   */  { CH_USED      , '$'   , '$'   },  // Dollar sign
   /*  25 %   */  { CH_USED      , '%'   , '%'   },  // Percent sign
   /*  26 &   */  { CH_USED      , '&'   , '&'   },  // Ampersand
   /*  27 '   */  { S_SPECIAL    , '\''  , '\''  },  // Single quote
   /*  28 (   */  { S_SPECIAL    , '('   , '('   },  // Left parenthesis
   /*  29 )   */  { S_SPECIAL    , ')'   , ')'   },  // Right parenthesis
   /*  2A *   */  { CH_USED      , '*'   , '*'   },  // Asterisk
   /*  2B +   */  { S_SPECIAL    , '+'   , '+'   },  // Plus sign
   /*  2C ,   */  { S_SPECIAL    , ','   , ','   },  // Comma
   /*  2D -   */  { S_SPECIAL    , '-'   , '-'   },  // Dash
   /*  2E .   */  { S_SPECIAL    , '.'   , '.'   },  // Period
   /*  2F /   */  { S_SPECIAL    , '/'   , '/'   },  // Solidus
   /*  30 0   */  { S_DIGIT      , '0'   , '0'   },  // Digit 0
   /*  31 1   */  { S_DIGIT      , '1'   , '1'   },  // Digit 1
   /*  32 2   */  { S_DIGIT      , '2'   , '2'   },  // Digit 2
   /*  33 3   */  { S_DIGIT      , '3'   , '3'   },  // Digit 3
   /*  34 4   */  { S_DIGIT      , '4'   , '4'   },  // Digit 4
   /*  35 5   */  { S_DIGIT      , '5'   , '5'   },  // Digit 5
   /*  36 6   */  { S_DIGIT      , '6'   , '6'   },  // Digit 6
   /*  37 7   */  { S_DIGIT      , '7'   , '7'   },  // Digit 7
   /*  38 8   */  { S_DIGIT      , '8'   , '8'   },  // Digit 8
   /*  39 9   */  { S_DIGIT      , '9'   , '9'   },  // Digit 9
   /*  3A :   */  { S_SPECIAL    , ':'   , ':'   },  // Colon
   /*  3B ;   */  { CH_USED      , ';'   , ';'   },  // Semi-colon
   /*  3C <   */  { CH_USED      , '<'   , '<'   },  // Less than sign
   /*  3D =   */  { S_SPECIAL    , '='   , '='   },  // Equal
   /*  3E >   */  { CH_USED      , '>'   , '>'   },  // Greater than sign
   /*  3F ?   */  { S_SPECIAL    , '?'   , '?'   },  // Interrogation mark
   /*  40 @   */  { CH_USED      , '@'   , '@'   },  // At sign
   /*  41 A   */  { S_LETTER     , 'A'   , 'A'   },  // Capital letter A
   /*  42 B   */  { S_LETTER     , 'B'   , 'B'   },  // Capital letter B
   /*  43 C   */  { S_LETTER     , 'C'   , 'C'   },  // Capital letter C
   /*  44 D   */  { S_LETTER     , 'D'   , 'D'   },  // Capital letter D
   /*  45 E   */  { S_LETTER     , 'E'   , 'E'   },  // Capital letter E
   /*  46 F   */  { S_LETTER     , 'F'   , 'F'   },  // Capital letter F
   /*  47 G   */  { S_LETTER     , 'G'   , 'G'   },  // Capital letter G
   /*  48 H   */  { S_LETTER     , 'H'   , 'H'   },  // Capital letter H
   /*  49 I   */  { S_LETTER     , 'I'   , 'I'   },  // Capital letter I
   /*  4A J   */  { S_LETTER     , 'J'   , 'J'   },  // Capital letter J
   /*  4B K   */  { S_LETTER     , 'K'   , 'K'   },  // Capital letter K
   /*  4C L   */  { S_LETTER     , 'L'   , 'L'   },  // Capital letter L
   /*  4D M   */  { S_LETTER     , 'M'   , 'M'   },  // Capital letter M
   /*  4E N   */  { S_LETTER     , 'N'   , 'N'   },  // Capital letter N
   /*  4F O   */  { S_LETTER     , 'O'   , 'O'   },  // Capital letter O
   /*  50 P   */  { S_LETTER     , 'P'   , 'P'   },  // Capital letter P
   /*  51 Q   */  { S_LETTER     , 'Q'   , 'Q'   },  // Capital letter Q
   /*  52 R   */  { S_LETTER     , 'R'   , 'R'   },  // Capital letter R
   /*  53 S   */  { S_LETTER     , 'S'   , 'S'   },  // Capital letter S
   /*  54 T   */  { S_LETTER     , 'T'   , 'T'   },  // Capital letter T
   /*  55 U   */  { S_LETTER     , 'U'   , 'U'   },  // Capital letter U
   /*  56 V   */  { S_LETTER     , 'V'   , 'V'   },  // Capital letter V
   /*  57 W   */  { S_LETTER     , 'W'   , 'W'   },  // Capital letter W
   /*  58 X   */  { S_LETTER     , 'X'   , 'X'   },  // Capital letter X
   /*  59 Y   */  { S_LETTER     , 'Y'   , 'Y'   },  // Capital letter Y
   /*  5A Z   */  { S_LETTER     , 'Z'   , 'Z'   },  // Capital letter Z
   /*  5B [   */  { CH_USED      , '['   , '['   },  // Left bracket
   /*  5C \   */  { CH_USED      , '\\'  , '\\'  },  // Back solidus
   /*  5D ]   */  { CH_USED      , ']'   , ']'   },  // Right bracket
   /*  5E ^   */  { CH_USED      , '^'   , '^'   },  // Circumflex accent
   /*  5F _   */  { CH_USED      , '_'   , '_'   },  // Underscore
   /*  60 `   */  { CH_USED      , '`'   , '`'   },  // Back quote
   /*  61 a   */  { S_LETTER     , 'a'   , 'A'   },  // Small letter a
   /*  62 b   */  { S_LETTER     , 'b'   , 'B'   },  // Small letter b
   /*  63 c   */  { S_LETTER     , 'c'   , 'C'   },  // Small letter c
   /*  64 d   */  { S_LETTER     , 'd'   , 'D'   },  // Small letter d
   /*  65 e   */  { S_LETTER     , 'e'   , 'E'   },  // Small letter e
   /*  66 f   */  { S_LETTER     , 'f'   , 'F'   },  // Small letter f
   /*  67 g   */  { S_LETTER     , 'g'   , 'G'   },  // Small letter g
   /*  68 h   */  { S_LETTER     , 'h'   , 'H'   },  // Small letter h
   /*  69 i   */  { S_LETTER     , 'i'   , 'I'   },  // Small letter i
   /*  6A j   */  { S_LETTER     , 'j'   , 'J'   },  // Small letter j
   /*  6B k   */  { S_LETTER     , 'k'   , 'K'   },  // Small letter k
   /*  6C l   */  { S_LETTER     , 'l'   , 'L'   },  // Small letter l
   /*  6D m   */  { S_LETTER     , 'm'   , 'M'   },  // Small letter m
   /*  6E n   */  { S_LETTER     , 'n'   , 'N'   },  // Small letter n
   /*  6F o   */  { S_LETTER     , 'o'   , 'O'   },  // Small letter o
   /*  70 p   */  { S_LETTER     , 'p'   , 'P'   },  // Small letter p
   /*  71 q   */  { S_LETTER     , 'q'   , 'Q'   },  // Small letter q
   /*  72 r   */  { S_LETTER     , 'r'   , 'R'   },  // Small letter r
   /*  73 s   */  { S_LETTER     , 's'   , 'S'   },  // Small letter s
   /*  74 t   */  { S_LETTER     , 't'   , 'T'   },  // Small letter t
   /*  75 u   */  { S_LETTER     , 'u'   , 'U'   },  // Small letter u
   /*  76 v   */  { S_LETTER     , 'v'   , 'V'   },  // Small letter v
   /*  77 w   */  { S_LETTER     , 'w'   , 'W'   },  // Small letter w
   /*  78 x   */  { S_LETTER     , 'x'   , 'X'   },  // Small letter x
   /*  79 y   */  { S_LETTER     , 'y'   , 'Y'   },  // Small letter y
   /*  7A z   */  { S_LETTER     , 'z'   , 'Z'   },  // Small letter z
   /*  7B {   */  { CH_USED      , '{'   , '{'   },  // Left brace
   /*  7C |   */  { CH_USED      , '|'   , '|'   },  // Vertical bar
   /*  7D }   */  { CH_USED      , '}'   , '}'   },  // Right brace
   /*  7E ~   */  { CH_USED      , '~'   , '~'   },  // Tilda
   /*  7F     */  { CH_UNUSED    , 0x7F  , 0x7F  },
   /*  80 N/A */  { CH_UNUSED    , 0x80  , 0x80  },
   /*  81 N/A */  { CH_UNUSED    , 0x81  , 0x81  },
   /*  82 N/A */  { CH_UNUSED    , 0x82  , 0x82  },
   /*  83 N/A */  { CH_UNUSED    , 0x83  , 0x83  },
   /*  84 N/A */  { CH_UNUSED    , 0x84  , 0x84  },
   /*  85 N/A */  { CH_UNUSED    , 0x85  , 0x85  },
   /*  86 N/A */  { CH_UNUSED    , 0x86  , 0x86  },
   /*  87 N/A */  { CH_UNUSED    , 0x87  , 0x87  },
   /*  88 N/A */  { CH_UNUSED    , 0x88  , 0x88  },
   /*  89 N/A */  { CH_UNUSED    , 0x89  , 0x89  },
   /*  8A N/A */  { CH_UNUSED    , 0x8A  , 0x8A  },
   /*  8B N/A */  { CH_UNUSED    , 0x8B  , 0x8B  },
   /*  8C N/A */  { CH_UNUSED    , 0x8C  , 0x8C  },
   /*  8D N/A */  { CH_UNUSED    , 0x8D  , 0x8D  },
   /*  8E N/A */  { CH_UNUSED    , 0x8E  , 0x8E  },
   /*  8F N/A */  { CH_UNUSED    , 0x8F  , 0x8F  },
   /*  90 N/A */  { CH_UNUSED    , 0x90  , 0x90  },
   /*  91 N/A */  { CH_UNUSED    , 0x91  , 0x91  },
   /*  92 N/A */  { CH_UNUSED    , 0x92  , 0x92  },
   /*  93 N/A */  { CH_UNUSED    , 0x93  , 0x93  },
   /*  94 N/A */  { CH_UNUSED    , 0x94  , 0x94  },
   /*  95 N/A */  { CH_UNUSED    , 0x95  , 0x95  },
   /*  96 N/A */  { CH_UNUSED    , 0x96  , 0x96  },
   /*  97 N/A */  { CH_UNUSED    , 0x97  , 0x97  },
   /*  98 N/A */  { CH_UNUSED    , 0x98  , 0x98  },
   /*  99 N/A */  { CH_UNUSED    , 0x99  , 0x99  },
   /*  9A N/A */  { CH_UNUSED    , 0x9A  , 0x9A  },
   /*  9B N/A */  { CH_UNUSED    , 0x9B  , 0x9B  },
   /*  9C N/A */  { CH_UNUSED    , 0x9C  , 0x9C  },
   /*  9D N/A */  { CH_UNUSED    , 0x9D  , 0x9D  },
   /*  9E N/A */  { CH_UNUSED    , 0x9E  , 0x9E  },
   /*  9F N/A */  { CH_UNUSED    , 0x9F  , 0x9F  },
   /*  A0    */  { CH_USED      , 0xFF  , 0xFF  },  // NBSP: Required blank ???
   /*  A1    */  { CH_USED      , ''   , ''   },  // Inverted exclamation mark
   /*  A2    */  { CH_USED      , ''   , ''   },  // Cent sign
   /*  A3    */  { CH_USED      , ''   , ''   },  // Pound sign
   /*  A4    */  { CH_USED      , ''   , ''   },  // Currency sign
   /*  A5    */  { CH_USED      , ''   , ''   },  // Yen sign
   /*  A6    */  { CH_USED      , ''   , ''   },  // Broken bar
   /*  A7    */  { CH_USED      , ''   , ''   },  // Paragraph Sign, section Sign
   /*  A8    */  { CH_USED      , ''   , ''   },  // Diaeresis
   /*  A9    */  { CH_USED      , ''   , ''   },  // Copyright sign
   /*  AA    */  { CH_USED      , ''   , ''   },  // Feminine ordinal indicator
   /*  AB    */  { CH_USED      , ''   , ''   },  // Left angle quotation mark
   /*  AC    */  { CH_USED      , ''   , ''   },  // Not sign
   /*  AD    */  { CH_USED      , ''   , ''   },  // SHY: Soft hyphen ???
   /*  AE    */  { CH_USED      , ''   , ''   },  // Registered Trade Mark sign
   /*  AF    */  { CH_USED      , ''   , ''   },  // Macron
   /*  B0    */  { CH_USED      , ''   , ''   },  // Ring above, degree sign
   /*  B1    */  { CH_USED      , ''   , ''   },  // Plus-minus sign
   /*  B2    */  { CH_USED      , ''   , ''   },  // Superscript 2
   /*  B3    */  { CH_USED      , ''   , ''   },  // Superscript 3
   /*  B4    */  { CH_USED      , ''   , ''   },  // Acute accent
   /*  B5    */  { CH_USED      , ''   , ''   },  // Micro sign
   /*  B6    */  { CH_USED      , ''   , ''   },  // Pilcrow sign
   /*  B7    */  { CH_USED      , ''   , ''   },  // Middle dot
   /*  B8    */  { CH_USED      , ''   , ''   },  // Cedilla
   /*  B9    */  { CH_USED      , ''   , ''   },  // Superscript 1
   /*  BA    */  { CH_USED      , ''   , ''   },  // Masculine ordinal indicator
   /*  BB    */  { CH_USED      , ''   , ''   },  // Right angle quotation mark
   /*  BC    */  { CH_USED      , ''   , ''   },  // Vulgar fraction one quarter
   /*  BD    */  { CH_USED      , ''   , ''   },  // Vulgar fraction one half
   /*  BE    */  { CH_USED      , ''   , ''   },  // Vulgar fraction three quarters
   /*  BF    */  { CH_USED      , ''   , ''   },  // Inverted question mark
   /*  C0    */  { CH_USED      , ''   , ''   },  // Capital letter A with grave accent
   /*  C1    */  { CH_USED      , ''   , ''   },  // Capital letter A with acute accent
   /*  C2    */  { CH_USED      , ''   , ''   },  // Capital letter A with circumflex accent
   /*  C3    */  { CH_USED      , ''   , ''   },  // Capital letter A with tilde
   /*  C4    */  { CH_USED      , ''   , ''   },  // Capital letter A with diaeresis
   /*  C5    */  { CH_USED      , ''   , ''   },  // Capital letter A with ring above
   /*  C6    */  { CH_USED      , ''   , ''   },  // Capital diphtong A with E
   /*  C7    */  { CH_USED      , ''   , ''   },  // Capital letter C with cedilla
   /*  C8    */  { CH_USED      , ''   , ''   },  // Capital letter E with grave accent
   /*  C9    */  { CH_USED      , ''   , ''   },  // Capital letter E with acute accent
   /*  CA    */  { CH_USED      , ''   , ''   },  // Capital letter E with circumflex accent
   /*  CB    */  { CH_USED      , ''   , ''   },  // Capital letter E with diaeresis
   /*  CC    */  { CH_USED      , ''   , ''   },  // Capital letter I with grave accent
   /*  CD    */  { CH_USED      , ''   , ''   },  // Capital letter I with acute accent
   /*  CE    */  { CH_USED      , ''   , ''   },  // Capital letter I with circumflex accent
   /*  CF    */  { CH_USED      , ''   , ''   },  // Capital letter I with diaeresis
   /*  D0    */  { CH_USED      , ''   , ''   },  // Capital icelandic letter ETH
   /*  D1    */  { CH_USED      , ''   , ''   },  // Capital letter N with tilde
   /*  D2    */  { CH_USED      , ''   , ''   },  // Capital letter O with grave accent
   /*  D3    */  { CH_USED      , ''   , ''   },  // Capital letter O with acute accent
   /*  D4    */  { CH_USED      , ''   , ''   },  // Capital letter O with circumflex accent
   /*  D5    */  { CH_USED      , ''   , ''   },  // Capital letter O with tilde
   /*  D6    */  { CH_USED      , ''   , ''   },  // Capital letter O with diaeresis
   /*  D7    */  { CH_USED      , ''   , ''   },  // Multiplication sign
   /*  D8    */  { CH_USED      , ''   , ''   },  // Capital letter O with oblique stroke
   /*  D9    */  { CH_USED      , ''   , ''   },  // Capital letter U with grave accent
   /*  DA    */  { CH_USED      , ''   , ''   },  // Capital letter U with acute accent
   /*  DB    */  { CH_USED      , ''   , ''   },  // Capital letter U with circumflex accent
   /*  DC    */  { CH_USED      , ''   , ''   },  // Capital letter U with diaeresis
   /*  DD    */  { CH_USED      , ''   , ''   },  // Capital letter Y with acute accent
   /*  DE    */  { CH_USED      , ''   , ''   },  // Capital icelandic letter THORN
   /*  DF    */  { CH_USED      , ''   , ''   },  // Small german letter sharp S
   /*  E0    */  { CH_USED      , ''   , ''   },  // Small letter A with grave accent
   /*  E1    */  { CH_USED      , ''   , ''   },  // Small letter A with acute accent
   /*  E2    */  { CH_USED      , ''   , ''   },  // Small letter A with circumflex accent
   /*  E3    */  { CH_USED      , ''   , ''   },  // Small letter A with tilde
   /*  E4    */  { CH_USED      , ''   , ''   },  // Small letter A with diaeresis
   /*  E5    */  { CH_USED      , ''   , ''   },  // Small letter A with ring above
   /*  E6    */  { CH_USED      , ''   , ''   },  // Small diphtong A with E
   /*  E7    */  { CH_USED      , ''   , ''   },  // Small letter C with cedilla
   /*  E8    */  { CH_USED      , ''   , ''   },  // Small letter E with grave accent
   /*  E9    */  { CH_USED      , ''   , ''   },  // Small letter E with acute accent
   /*  EA    */  { CH_USED      , ''   , ''   },  // Small letter E with circumflex accent
   /*  EB    */  { CH_USED      , ''   , ''   },  // Small letter E with diaeresis
   /*  EC    */  { CH_USED      , ''   , ''   },  // Small letter I with grave accent
   /*  ED    */  { CH_USED      , ''   , ''   },  // Small letter I with acute accent
   /*  EE    */  { CH_USED      , ''   , ''   },  // Small letter I with circumflex accent
   /*  EF    */  { CH_USED      , ''   , ''   },  // Small letter I with diaeresis
   /*  F0    */  { CH_USED      , ''   , ''   },  // Small icelandic letter ETH
   /*  F1    */  { CH_USED      , ''   , ''   },  // Small letter N with tilde
   /*  F2    */  { CH_USED      , ''   , ''   },  // Small letter O with grave accent
   /*  F3    */  { CH_USED      , ''   , ''   },  // Small letter O with acute accent
   /*  F4    */  { CH_USED      , ''   , ''   },  // Small letter O with circumflex accent
   /*  F5    */  { CH_USED      , ''   , ''   },  // Small letter O with tilde
   /*  F6    */  { CH_USED      , ''   , ''   },  // Small letter O with diaeresis
   /*  F7    */  { CH_USED      , ''   , ''   },  // Division sign
   /*  F8    */  { CH_USED      , ''   , ''   },  // Small letter O with oblique stroke
   /*  F9    */  { CH_USED      , ''   , ''   },  // Small letter U with grave accent
   /*  FA    */  { CH_USED      , ''   , ''   },  // Small letter U with acute accent
   /*  FB    */  { CH_USED      , ''   , ''   },  // Small letter U with circumflex accent
   /*  FC    */  { CH_USED      , ''   , ''   },  // Small letter U with diaeresis
   /*  FD    */  { CH_USED      , ''   , ''   },  // Small letter Y with acute accent
   /*  FE    */  { CH_USED      , ''   , ''   },  // Small icelandic letter THORN
   /*  FF     */  { CH_USED      , ''   , ''   }   // Small letter Y with diaeresis
   };

   /*-------------------------------------------------------------+
   | IBM 395                                                      |
   +-------------------------------------------------------------*/
   static SetDescriptor IBM_395_chtab[MAX_CHARSET] = {
   /*  00 N/A */  { CH_UNUSED    , 0x00  , 0x00  },
   /*  01 N/A */  { CH_UNUSED    , 0x01  , 0x01  },
   /*  02 N/A */  { CH_UNUSED    , 0x02  , 0x02  },
   /*  03 N/A */  { CH_UNUSED    , 0x03  , 0x03  },
   /*  04 N/A */  { CH_UNUSED    , 0x04  , 0x04  },
   /*  05 N/A */  { CH_UNUSED    , 0x05  , 0x05  },
   /*  06 N/A */  { CH_UNUSED    , 0x06  , 0x06  },
   /*  07 N/A */  { CH_UNUSED    , 0x07  , 0x07  },
   /*  08 N/A */  { CH_UNUSED    , 0x08  , 0x08  },
   /*  09 N/A */  { CH_UNUSED    , 0x09  , 0x09  },
   /*  0A N/A */  { CH_UNUSED    , 0x0A  , 0x0A  },
   /*  0B N/A */  { CH_UNUSED    , 0x0B  , 0x0B  },
   /*  0C N/A */  { CH_UNUSED    , 0x0C  , 0x0C  },
   /*  0D N/A */  { CH_UNUSED    , 0x0D  , 0x0D  },
   /*  0E N/A */  { CH_UNUSED    , 0x0E  , 0x0E  },
   /*  0F N/A */  { CH_UNUSED    , 0x0F  , 0x0F  },
   /*  10 N/A */  { CH_UNUSED    , 0x10  , 0x10  },
   /*  11 N/A */  { CH_UNUSED    , 0x11  , 0x11  },
   /*  12 N/A */  { CH_UNUSED    , 0x12  , 0x12  },
   /*  13 N/A */  { CH_UNUSED    , 0x13  , 0x13  },
   /*  14 N/A */  { CH_UNUSED    , 0x14  , 0x14  },
   /*  15 N/A */  { CH_UNUSED    , 0x15  , 0x15  },
   /*  16 N/A */  { CH_UNUSED    , 0x16  , 0x16  },
   /*  17 N/A */  { CH_UNUSED    , 0x17  , 0x17  },
   /*  18 N/A */  { CH_UNUSED    , 0x18  , 0x18  },
   /*  19 N/A */  { CH_UNUSED    , 0x19  , 0x19  },
   /*  1A N/A */  { CH_UNUSED    , 0x1A  , 0x1A  },
   /*  1B N/A */  { CH_UNUSED    , 0x1B  , 0x1B  },
   /*  1C N/A */  { CH_UNUSED    , 0x1C  , 0x1C  },
   /*  1D N/A */  { CH_UNUSED    , 0x1D  , 0x1D  },
   /*  1E N/A */  { CH_UNUSED    , 0x1E  , 0x1E  },
   /*  1F N/A */  { CH_UNUSED    , 0x1F  , 0x1F  },
   /*  20 N/A */  { CH_UNUSED    , 0x20  , 0x20  },
   /*  21 N/A */  { CH_UNUSED    , 0x21  , 0x21  },
   /*  22 N/A */  { CH_UNUSED    , 0x22  , 0x22  },
   /*  23 N/A */  { CH_UNUSED    , 0x23  , 0x23  },
   /*  24 N/A */  { CH_UNUSED    , 0x24  , 0x24  },
   /*  25 N/A */  { CH_UNUSED    , 0x25  , 0x25  },
   /*  26 N/A */  { CH_UNUSED    , 0x26  , 0x26  },
   /*  27 N/A */  { CH_UNUSED    , 0x27  , 0x27  },
   /*  28 N/A */  { CH_UNUSED    , 0x28  , 0x28  },
   /*  29 N/A */  { CH_UNUSED    , 0x29  , 0x29  },
   /*  2A N/A */  { CH_UNUSED    , 0x2A  , 0x2A  },
   /*  2B N/A */  { CH_UNUSED    , 0x2B  , 0x2B  },
   /*  2C N/A */  { CH_UNUSED    , 0x2C  , 0x2C  },
   /*  2D N/A */  { CH_UNUSED    , 0x2D  , 0x2D  },
   /*  2E N/A */  { CH_UNUSED    , 0x2E  , 0x2E  },
   /*  2F N/A */  { CH_UNUSED    , 0x2F  , 0x2F  },
   /*  30 N/A */  { CH_UNUSED    , 0x30  , 0x30  },
   /*  31 N/A */  { CH_UNUSED    , 0x31  , 0x31  },
   /*  32 N/A */  { CH_UNUSED    , 0x32  , 0x32  },
   /*  33 N/A */  { CH_UNUSED    , 0x33  , 0x33  },
   /*  34 N/A */  { CH_UNUSED    , 0x34  , 0x34  },
   /*  35 N/A */  { CH_UNUSED    , 0x35  , 0x35  },
   /*  36 N/A */  { CH_UNUSED    , 0x36  , 0x36  },
   /*  37 N/A */  { CH_UNUSED    , 0x37  , 0x37  },
   /*  38 N/A */  { CH_UNUSED    , 0x38  , 0x38  },
   /*  39 N/A */  { CH_UNUSED    , 0x39  , 0x39  },
   /*  3A N/A */  { CH_UNUSED    , 0x3A  , 0x3A  },
   /*  3B N/A */  { CH_UNUSED    , 0x3B  , 0x3B  },
   /*  3C N/A */  { CH_UNUSED    , 0x3C  , 0x3C  },
   /*  3D N/A */  { CH_UNUSED    , 0x3D  , 0x3D  },
   /*  3E N/A */  { CH_UNUSED    , 0x3E  , 0x3E  },
   /*  3F N/A */  { CH_UNUSED    , 0x3F  , 0x3F  },
   /*  40 SPC */  { CH_USED      , ' '   , ' '   },  // Space
   #if (0x20 == ' ')                                 // NBSP: Required blank ???
   /*  41     */  { CH_USED      , 0xFF , 0xFF   },  // ASCII
   #else
   /*  41     */  { CH_USED      , 0x41 , 0x41   },  // EBCDIC
   #endif
   /*  42     */  { CH_USED      , ''   , ''   },  // Small letter A with circumflex accent
   /*  43     */  { CH_USED      , ''   , ''   },  // Small letter A with diaeresis
   /*  44     */  { CH_USED      , ''   , ''   },  // Small letter A with grave accent
   /*  45     */  { CH_USED      , ''   , ''   },  // Small letter A with acute accent
   /*  46     */  { CH_USED      , ''   , ''   },  // Small letter A with tilde
   /*  47     */  { CH_USED      , ''   , ''   },  // Small letter A with ring above
   /*  48     */  { CH_USED      , ''   , ''   },  // Small letter C with cedilla
   /*  49     */  { CH_USED      , ''   , ''   },  // Small letter N with tilde
   /*  4A     */  { CH_USED      , ''   , ''   },  // Cent sign
   /*  4B     */  { S_SPECIAL    , '.'   , '.'   },  // Period
   /*  4C     */  { CH_USED      , '<'   , '<'   },  // Less than sign
   /*  4D     */  { S_SPECIAL    , '('   , '('   },  // Left parenthesis
   /*  4E     */  { S_SPECIAL    , '+'   , '+'   },  // Plus sign
   /*  4F     */  { CH_USED      , '|'   , '|'   },  // Vertical bar
   /*  50     */  { CH_USED      , '&'   , '&'   },  // Ampersand
   /*  51     */  { CH_USED      , ''   , ''   },  // Small letter E with acute accent
   /*  52     */  { CH_USED      , ''   , ''   },  // Small letter E with circumflex accent
   /*  53     */  { CH_USED      , ''   , ''   },  // Small letter E with diaeresis
   /*  54     */  { CH_USED      , ''   , ''   },  // Small letter E with grave accent
   /*  55     */  { CH_USED      , ''   , ''   },  // Small letter I with acute accent
   /*  56     */  { CH_USED      , ''   , ''   },  // Small letter I with circumflex accent
   /*  57     */  { CH_USED      , ''   , ''   },  // Small letter I with diaeresis
   /*  58     */  { CH_USED      , ''   , ''   },  // Small letter I with grave accent
   /*  59     */  { CH_USED      , ''   , ''   },  // Small german letter sharp s
   /*  5A     */  { CH_USED      , '!'   , '!'   },  // Exclamation mark
   /*  5B     */  { CH_USED      , '$'   , '$'   },  // Dollar sign
   /*  5C     */  { CH_USED      , '*'   , '*'   },  // Asterisk
   /*  5D     */  { S_SPECIAL    , ')'   , ')'   },  // Right parenthesis
   /*  5E     */  { CH_USED      , ';'   , ';'   },  // Semi-colon
   /*  5F N/A */  { CH_UNUSED    , 0x00  , 0x00  },
   /*  60    */  { CH_USED      , ''   , ''   },  // SHY: Soft hyphen ???
   /*  61     */  { S_SPECIAL    , '/'   , '/'   },  // Solidus
   /*  62     */  { CH_USED      , ''   , ''   },  // Capital letter A with circumflex accent
   /*  63     */  { CH_USED      , ''   , ''   },  // Capital letter A with diaeresis
   /*  64     */  { CH_USED      , ''   , ''   },  // Capital letter A with grave accent
   /*  65     */  { CH_USED      , ''   , ''   },  // Capital letter A with acute accent
   /*  66     */  { CH_USED      , ''   , ''   },  // Capital letter A with tilde
   /*  67     */  { CH_USED      , ''   , ''   },  // Capital letter A with ring above
   /*  68     */  { CH_USED      , ''   , ''   },  // Capital letter C with cedilla
   /*  69     */  { CH_USED      , ''   , ''   },  // Capital letter N with tilde
   /*  6A     */  { CH_USED      , ']'   , ']'   },  // Right bracket
   /*  6B     */  { S_SPECIAL    , ','   , ','   },  // Comma
   /*  6C     */  { CH_USED      , '%'   , '%'   },  // Percent sign
   /*  6D N/A */  { CH_UNUSED    , 0x00  , 0x00  },
   /*  6E     */  { CH_USED      , '>'   , '>'   },  // Greater than sign
   /*  6F     */  { S_SPECIAL    , '?'   , '?'   },  // Interrogation mark
   /*  70     */  { CH_USED      , ''   , ''   },  // Small letter O with oblique stroke
   /*  71     */  { CH_USED      , ''   , ''   },  // Capital letter E with acute accent
   /*  72     */  { CH_USED      , ''   , ''   },  // Capital letter E with circumflex accent
   /*  73     */  { CH_USED      , ''   , ''   },  // Capital letter E with diaeresis
   /*  74     */  { CH_USED      , ''   , ''   },  // Capital letter E with grave accent
   /*  75     */  { CH_USED      , ''   , ''   },  // Capital letter I with acute accent
   /*  76     */  { CH_USED      , ''   , ''   },  // Capital letter I with circumflex accent
   /*  77     */  { CH_USED      , ''   , ''   },  // Capital letter I with diaeresis
   /*  78     */  { CH_USED      , ''   , ''   },  // Capital letter I with grave accent
   /*  79 N/A */  { CH_UNUSED    , 0x00  , 0x00  },
   /*  7A     */  { S_SPECIAL    , ':'   , ':'   },  // Colon
   /*  7B     */  { CH_USED      , '#'   , '#'   },  // Number sign
   /*  7C     */  { CH_USED      , '@'   , '@'   },  // At sign
   /*  7D     */  { S_SPECIAL    , '\''  , '\''  },  // Single quote
   /*  7E =   */  { S_SPECIAL    , '='   , '='   },  // Equal
   /*  7F     */  { CH_USED      , '"'   , '"'   },  // Double quote
   /*  80     */  { CH_USED      , ''   , ''   },  // Capital letter O with oblique stroke
   /*  81 a   */  { S_LETTER     , 'a'   , 'A'   },  // Small letter a
   /*  82 b   */  { S_LETTER     , 'b'   , 'B'   },  // Small letter b
   /*  83 c   */  { S_LETTER     , 'c'   , 'C'   },  // Small letter c
   /*  84 d   */  { S_LETTER     , 'd'   , 'D'   },  // Small letter d
   /*  85 e   */  { S_LETTER     , 'e'   , 'E'   },  // Small letter e
   /*  86 f   */  { S_LETTER     , 'f'   , 'F'   },  // Small letter f
   /*  87 g   */  { S_LETTER     , 'g'   , 'G'   },  // Small letter g
   /*  88 h   */  { S_LETTER     , 'h'   , 'H'   },  // Small letter h
   /*  89 i   */  { S_LETTER     , 'i'   , 'I'   },  // Small letter i
   /*  8A     */  { CH_USED      , ''   , ''   },  // Left angle quotation mark
   /*  8B     */  { CH_USED      , ''   , ''   },  // Right angle quotation mark
   /*  8C     */                                     // Per mill sign
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0x8C  , 0x8C  },  // EBCDIC
   #endif
   /*  8D     */  { S_SPECIAL    , '-'   , '-'   },  // Dash
   /*  8E     */                                     // ij ligature
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0x8E  , 0x8E  },  // EBCDIC
   #endif
   /*  8F     */                                     // ffi ligature
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0x8F  , 0x8F  },  // EBCDIC
   #endif
   /*  90     */  { CH_USED      , ''   , ''   },  // Ring above, degree sign
   /*  91 j   */  { S_LETTER     , 'j'   , 'J'   },  // Small letter j
   /*  92 k   */  { S_LETTER     , 'k'   , 'K'   },  // Small letter k
   /*  93 l   */  { S_LETTER     , 'l'   , 'L'   },  // Small letter l
   /*  94 m   */  { S_LETTER     , 'm'   , 'M'   },  // Small letter m
   /*  95 n   */  { S_LETTER     , 'n'   , 'N'   },  // Small letter n
   /*  96 o   */  { S_LETTER     , 'o'   , 'O'   },  // Small letter o
   /*  97 p   */  { S_LETTER     , 'p'   , 'P'   },  // Small letter p
   /*  98 q   */  { S_LETTER     , 'q'   , 'Q'   },  // Small letter q
   /*  99 r   */  { S_LETTER     , 'r'   , 'R'   },  // Small letter r
   /*  9A     */  { CH_USED      , ''   , ''   },  // Macron
   /*  9B     */                                     // Capital diphtong O with E
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  ,  0x00 },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0x9B  ,  0x9B },  // EBCDIC
   #endif
   /*  9C     */  { CH_USED      , ''   , ''   },  // Small diphtong A with E
   /*  9D     */                                     // Capital letter Y with diaeresis
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  ,  0x00 },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0x9D  ,  0x9D },  // EBCDIC
   #endif
   /*  9E     */  { CH_USED      , ''   , ''   },  // Capital diphtong A with E
   /*  9F     */  { CH_USED      , ''   , ''   },  // Middle dot
   /*  A0     */                                     // ff ligature
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  ,  0x00 },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xA0  ,  0xA0 },  // EBCDIC
   #endif
   /*  A1 N/A */  { CH_UNUSED    , 0x00  , 0x00  },
   /*  A2 s   */  { S_LETTER     , 's'   , 'S'   },  // Small letter s
   /*  A3 t   */  { S_LETTER     , 't'   , 'T'   },  // Small letter t
   /*  A4 u   */  { S_LETTER     , 'u'   , 'U'   },  // Small letter u
   /*  A5 v   */  { S_LETTER     , 'v'   , 'V'   },  // Small letter v
   /*  A6 w   */  { S_LETTER     , 'w'   , 'W'   },  // Small letter w
   /*  A7 x   */  { S_LETTER     , 'x'   , 'X'   },  // Small letter x
   /*  A8 y   */  { S_LETTER     , 'y'   , 'Y'   },  // Small letter y
   /*  A9 z   */  { S_LETTER     , 'z'   , 'Z'   },  // Small letter z
   /*  AA     */  { CH_USED      , ''   , ''   },  // Inverted exclamation mark
   /*  AB     */  { CH_USED      , ''   , ''   },  // Inverted question mark
   /*  AC     */                                     // Small diphtong O with E
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xAC  , 0x9B  },  // EBCDIC
   #endif
   /*  AD     */                                     // ffl ligature
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xAD  , 0xAD  },  // EBCDIC
   #endif
   /*  AE     */                                     // Plus over plus
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xAE  , 0xAE  },  // EBCDIC
   #endif
   /*  AF     */                                     // Low right double quotation mark
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xAF  , 0xAF  },  // EBCDIC
   #endif
   /*  B0     */  { CH_USED      , '['   , '['   },  // Left bracket
   /*  B1     */  { CH_USED      , ''   , ''   },  // Pound sign
   /*  B2     */  { CH_USED      , ''   , ''   },  // Yen sign
   /*  B3     */                                     // fi ligature
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xB3  , 0xB3  },  // EBCDIC
   #endif
   /*  B4     */                                     // fl ligature
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xB4  , 0xB4  },  // EBCDIC
   #endif
   /*  B5     */  { CH_USED      , ''   , ''   },  // Paragraph Sign, section Sign
   /*  B6     */  { CH_USED      , ''   , ''   },  // Pilcrow sign
   /*  B7     */  { CH_USED      , ''   , ''   },  // Vulgar fraction one quarter
   /*  B8     */  { CH_USED      , ''   , ''   },  // Vulgar fraction one half
   /*  B9     */  { CH_USED      , ''   , ''   },  // Vulgar fraction three quarters
   /*  BA     */                                     // Vulgar fraction one eighth
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xBA  , 0xBA  },  // EBCDIC
   #endif
   /*  BB     */                                     // Left single quotation mark
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xBB  , 0xBB  },  // EBCDIC
   #endif
   /*  BC     */                                     // Right single quotation mark
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xBC  , 0xBC  },  // EBCDIC
   #endif
   /*  BD     */                                     // Left double quotation mark
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xBD  , 0xBD  },  // EBCDIC
   #endif
   /*  BE     */                                     // Right double quotation mark
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xBE  , 0xBE  },  // EBCDIC
   #endif
   /*  BF     */                                     // Dagger
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xBF  , 0xBF  },  // EBCDIC
   #endif
   /*  C0     */  { CH_USED      , '{'   , '{'   },  // Left brace
   /*  C1 A   */  { S_LETTER     , 'A'   , 'A'   },  // Capital letter A
   /*  C2 B   */  { S_LETTER     , 'B'   , 'B'   },  // Capital letter B
   /*  C3 C   */  { S_LETTER     , 'C'   , 'C'   },  // Capital letter C
   /*  C4 D   */  { S_LETTER     , 'D'   , 'D'   },  // Capital letter D
   /*  C5 E   */  { S_LETTER     , 'E'   , 'E'   },  // Capital letter E
   /*  C6 F   */  { S_LETTER     , 'F'   , 'F'   },  // Capital letter F
   /*  C7 G   */  { S_LETTER     , 'G'   , 'G'   },  // Capital letter G
   /*  C8 H   */  { S_LETTER     , 'H'   , 'H'   },  // Capital letter H
   /*  C9 I   */  { S_LETTER     , 'I'   , 'I'   },  // Capital letter I
   /*  CA N/A */  { CH_UNUSED    , 0x00  ,  0x00 },
   /*  CB     */  { CH_USED      , ''   , ''   },  // Small letter O with circumflex accent
   /*  CC     */  { CH_USED      , ''   , ''   },  // Small letter O with diaeresis
   /*  CD     */  { CH_USED      , ''   , ''   },  // Small letter O with grave accent
   /*  CE     */  { CH_USED      , ''   , ''   },  // Small letter O with acute accent
   /*  CF     */  { CH_USED      , ''   , ''   },  // Small letter O with tilde
   /*  D0     */  { CH_USED      , '}'   , '}'   },  // Right brace
   /*  D1 J   */  { S_LETTER     , 'J'   , 'J'   },  // Capital letter J
   /*  D2 K   */  { S_LETTER     , 'K'   , 'K'   },  // Capital letter K
   /*  D3 L   */  { S_LETTER     , 'L'   , 'L'   },  // Capital letter L
   /*  D4 M   */  { S_LETTER     , 'M'   , 'M'   },  // Capital letter M
   /*  D5 N   */  { S_LETTER     , 'N'   , 'N'   },  // Capital letter N
   /*  D6 O   */  { S_LETTER     , 'O'   , 'O'   },  // Capital letter O
   /*  D7 P   */  { S_LETTER     , 'P'   , 'P'   },  // Capital letter P
   /*  D8 Q   */  { S_LETTER     , 'Q'   , 'Q'   },  // Capital letter Q
   /*  D9 R   */  { S_LETTER     , 'R'   , 'R'   },  // Capital letter R
   /*  DA     */                                     // Vulgar fraction three eighth
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  ,  0x00 },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xDA  ,  0xDA },  // EBCDIC
   #endif
   /*  DB     */  { CH_USED      , ''   , ''   },  // Small letter U with circumflex accent
   /*  DC     */  { CH_USED      , ''   , ''   },  // Small letter U with diaeresis
   /*  DD     */  { CH_USED      , ''   , ''   },  // Small letter U with grave accent
   /*  DE     */  { CH_USED      , ''   , ''   },  // Small letter U with acute accent
   /*  DF     */                                     // Small letter Y with diaeresis s
   #if (0x20 == ' ')
                  { CH_USED      , ''   , ''   },  // ASCII 850: capital doesn't exist
   #else
                  { CH_USED      , ''   ,  0x9D },  // EBCDIC
   #endif
   /*  E0     */  { CH_USED      , '\\'  , '\\'  },  // Back solidus
   /*  E1 N/A */  { CH_UNUSED    ,  0x00 ,  0x00 },
   /*  E2 S   */  { S_LETTER     , 'S'   , 'S'   },  // Capital letter S
   /*  E3 T   */  { S_LETTER     , 'T'   , 'T'   },  // Capital letter T
   /*  E4 U   */  { S_LETTER     , 'U'   , 'U'   },  // Capital letter U
   /*  E5 V   */  { S_LETTER     , 'V'   , 'V'   },  // Capital letter V
   /*  E6 W   */  { S_LETTER     , 'W'   , 'W'   },  // Capital letter W
   /*  E7 X   */  { S_LETTER     , 'X'   , 'X'   },  // Capital letter X
   /*  E8 Y   */  { S_LETTER     , 'Y'   , 'Y'   },  // Capital letter Y
   /*  E9 Z   */  { S_LETTER     , 'Z'   , 'Z'   },  // Capital letter Z
   /*  EA     */                                     // Vulgar fraction five eighth
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  ,  0x00 },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xEA  ,  0xEA },  // EBCDIC
   #endif
   /*  EB     */  { CH_USED      , ''   , ''   },  // Capital letter O with circumflex accent
   /*  EC     */  { CH_USED      , ''   , ''   },  // Capital letter O with diaeresis
   /*  ED     */  { CH_USED      , ''   , ''   },  // Capital letter O with grave accent
   /*  EE     */  { CH_USED      , ''   , ''   },  // Capital letter O with acute accent
   /*  EF     */  { CH_USED      , ''   , ''   },  // Capital letter O with tilde
   /*  F0 0   */  { S_DIGIT      , '0'   , '0'   },  // Digit 0
   /*  F1 1   */  { S_DIGIT      , '1'   , '1'   },  // Digit 1
   /*  F2 2   */  { S_DIGIT      , '2'   , '2'   },  // Digit 2
   /*  F3 3   */  { S_DIGIT      , '3'   , '3'   },  // Digit 3
   /*  F4 4   */  { S_DIGIT      , '4'   , '4'   },  // Digit 4
   /*  F5 5   */  { S_DIGIT      , '5'   , '5'   },  // Digit 5
   /*  F6 6   */  { S_DIGIT      , '6'   , '6'   },  // Digit 6
   /*  F7 7   */  { S_DIGIT      , '7'   , '7'   },  // Digit 7
   /*  F8 8   */  { S_DIGIT      , '8'   , '8'   },  // Digit 8
   /*  F9 9   */  { S_DIGIT      , '9'   , '9'   },  // Digit 9
   /*  FA     */                                     // Vulgar fraction five eighth
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xFA  , 0xFA  },  // EBCDIC
   #endif
   /*  FB     */  { CH_USED      , ''   , ''   },  // Capital letter U with circumflex accent
   /*  FC     */  { CH_USED      , ''   , ''   },  // Capital letter U with diaeresis
   /*  FD     */  { CH_USED      , ''   , ''   },  // Capital letter U with grave accent
   /*  FE     */  { CH_USED      , ''   , ''   },  // Capital letter U with acute accent
   /*  FF     */                                     // Capital diphtong E with O
   #if (0x20 == ' ')
                  { CH_UNUSED    , 0x00  , 0x00  },  // ASCII 850: doesn't exist
   #else
                  { CH_USED      , 0xFF  , 0xFF  },  // EBCDIC
   #endif
   };

   /*-------------------------------------------------------------+
   | IBM 850                                                      |
   +-------------------------------------------------------------*/
   static SetDescriptor IBM_850_chtab[MAX_CHARSET] = {
   /*  00 NUL */  { CH_CONTROL   , 0x00  ,  0x00 },  // CONTROL
   /*  01    */  { CH_USED      , ''   , ''   },  // White happy face
   /*  02    */  { CH_USED      , ''   , ''   },  // Black happy face
   /*  03    */  { CH_USED      , ''   , ''   },  // Heart
   /*  04    */  { CH_USED      , ''   , ''   },  // Diamond
   /*  05    */  { CH_USED      , ''   , ''   },  // Club
   /*  06    */  { CH_USED      , ''   , ''   },  // Spade
   /*  07    */  { CH_USED      , ''   , ''   },  // Big dot
   /*  08    */  { CH_USED      , ''   , ''   },  // Reverse big dot
   /*  09     */  { CH_USED      , 0x09  ,  0x09 },  // Circle
   /*  0A     */  { CH_USED      , 0x0A  ,  0x0A },  // Reverse circle
   /*  0B    */  { CH_USED      , ''   , ''   },  // Masculine symbol
   /*  0C    */  { CH_USED      , ''   , ''   },  // Feminine symbol
   /*  0D     */  { CH_USED      , 0x0D  ,  0x0D },  // Single music note
   /*  0E    */  { CH_USED      , ''   , ''   },  // Two music notes
   /*  0F    */  { CH_USED      , ''   , ''   },  // Sun
   /*  10    */  { CH_USED      , ''   , ''   },  // Right triangle
   /*  11    */  { CH_USED      , ''   , ''   },  // Left triangle
   /*  12    */  { CH_USED      , ''   , ''   },  // Double vertical arrow
   /*  13    */  { CH_USED      , ''   , ''   },  // Double exclamation mark
   /*  14    */  { CH_USED      , ''   , ''   },  // Pilcrow sign
   /*  15    */  { CH_USED      , ''   , ''   },  // Paragraph sign
   /*  16    */  { CH_USED      , ''   , ''   },  // Fat hyphen
   /*  17    */  { CH_USED      , ''   , ''   },  // Double vertical arrow underlined
   /*  18    */  { CH_USED      , ''   , ''   },  // Up arrow
   /*  19    */  { CH_USED      , ''   , ''   },  // Down arrow
   /*  1A     */  { CH_CONTROL   , 0x1A  ,  0x1A },  // Right arrow
   /*  1B    */  { CH_USED      , ''   , ''   },  // Left arrow
   /*  1C    */  { CH_USED      , ''   , ''   },  // Inverted not sign
   /*  1D    */  { CH_USED      , ''   , ''   },  // Double horizontal arrow
   /*  1E    */  { CH_USED      , ''   , ''   },  // Up triangle
   /*  1F    */  { CH_USED      , ''   , ''   },  // Down triangle
   /*  20 SPC */  { CH_USED      , ' '   , ' '   },  // Space
   /*  21 !   */  { CH_USED      , '!'   , '!'   },  // Exclamation mark
   /*  22 "   */  { CH_USED      , '"'   , '"'   },  // Double quote
   /*  23 #   */  { CH_USED      , '#'   , '#'   },  // Number sign
   /*  24 $   */  { CH_USED      , '$'   , '$'   },  // Dollar sign
   /*  25 %   */  { CH_USED      , '%'   , '%'   },  // Percent sign
   /*  26 &   */  { CH_USED      , '&'   , '&'   },  // Ampersand
   /*  27 '   */  { S_SPECIAL    , '\''  , '\''  },  // Single quote
   /*  28 (   */  { S_SPECIAL    , '('   , '('   },  // Left parenthesis
   /*  29 )   */  { S_SPECIAL    , ')'   , ')'   },  // Right parenthesis
   /*  2A *   */  { CH_USED      , '*'  , '*'    },  // Asterisk
   /*  2B +   */  { S_SPECIAL    , '+'   , '+'   },  // Plus sign
   /*  2C ,   */  { S_SPECIAL    , ','   , ','   },  // Comma
   /*  2D -   */  { S_SPECIAL    , '-'   , '-'   },  // Dash
   /*  2E .   */  { S_SPECIAL    , '.'   , '.'   },  // Period
   /*  2F /   */  { S_SPECIAL    , '/'   , '/'   },  // Solidus
   /*  30 0   */  { S_DIGIT      , '0'   , '0'   },  // Digit 0
   /*  31 1   */  { S_DIGIT      , '1'   , '1'   },  // Digit 1
   /*  32 2   */  { S_DIGIT      , '2'   , '2'   },  // Digit 2
   /*  33 3   */  { S_DIGIT      , '3'   , '3'   },  // Digit 3
   /*  34 4   */  { S_DIGIT      , '4'   , '4'   },  // Digit 4
   /*  35 5   */  { S_DIGIT      , '5'   , '5'   },  // Digit 5
   /*  36 6   */  { S_DIGIT      , '6'   , '6'   },  // Digit 6
   /*  37 7   */  { S_DIGIT      , '7'   , '7'   },  // Digit 7
   /*  38 8   */  { S_DIGIT      , '8'   , '8'   },  // Digit 8
   /*  39 9   */  { S_DIGIT      , '9'   , '9'   },  // Digit 9
   /*  3A :   */  { S_SPECIAL    , ':'   , ':'   },  // Colon
   /*  3B ;   */  { CH_USED      , ';'   , ';'   },  // Semi-colon
   /*  3C <   */  { CH_USED      , '<'   , '<'   },  // Less than sign
   /*  3D =   */  { S_SPECIAL    , '='   , '='   },  // Equal
   /*  3E >   */  { CH_USED      , '>'   , '>'   },  // Greater than sign
   /*  3F ?   */  { S_SPECIAL    , '?'   , '?'   },  // Interrogation mark
   /*  40 @   */  { CH_USED      , '@'   , '@'   },  // At sign
   /*  41 A   */  { S_LETTER     , 'A'   , 'A'   },  // Capital letter A
   /*  42 B   */  { S_LETTER     , 'B'   , 'B'   },  // Capital letter B
   /*  43 C   */  { S_LETTER     , 'C'   , 'C'   },  // Capital letter C
   /*  44 D   */  { S_LETTER     , 'D'   , 'D'   },  // Capital letter D
   /*  45 E   */  { S_LETTER     , 'E'   , 'E'   },  // Capital letter E
   /*  46 F   */  { S_LETTER     , 'F'   , 'F'   },  // Capital letter F
   /*  47 G   */  { S_LETTER     , 'G'   , 'G'   },  // Capital letter G
   /*  48 H   */  { S_LETTER     , 'H'   , 'H'   },  // Capital letter H
   /*  49 I   */  { S_LETTER     , 'I'   , 'I'   },  // Capital letter I
   /*  4A J   */  { S_LETTER     , 'J'   , 'J'   },  // Capital letter J
   /*  4B K   */  { S_LETTER     , 'K'   , 'K'   },  // Capital letter K
   /*  4C L   */  { S_LETTER     , 'L'   , 'L'   },  // Capital letter L
   /*  4D M   */  { S_LETTER     , 'M'   , 'M'   },  // Capital letter M
   /*  4E N   */  { S_LETTER     , 'N'   , 'N'   },  // Capital letter N
   /*  4F O   */  { S_LETTER     , 'O'   , 'O'   },  // Capital letter O
   /*  50 P   */  { S_LETTER     , 'P'   , 'P'   },  // Capital letter P
   /*  51 Q   */  { S_LETTER     , 'Q'   , 'Q'   },  // Capital letter Q
   /*  52 R   */  { S_LETTER     , 'R'   , 'R'   },  // Capital letter R
   /*  53 S   */  { S_LETTER     , 'S'   , 'S'   },  // Capital letter S
   /*  54 T   */  { S_LETTER     , 'T'   , 'T'   },  // Capital letter T
   /*  55 U   */  { S_LETTER     , 'U'   , 'U'   },  // Capital letter U
   /*  56 V   */  { S_LETTER     , 'V'   , 'V'   },  // Capital letter V
   /*  57 W   */  { S_LETTER     , 'W'   , 'W'   },  // Capital letter W
   /*  58 X   */  { S_LETTER     , 'X'   , 'X'   },  // Capital letter X
   /*  59 Y   */  { S_LETTER     , 'Y'   , 'Y'   },  // Capital letter Y
   /*  5A Z   */  { S_LETTER     , 'Z'   , 'Z'   },  // Capital letter Z
   /*  5B [   */  { CH_USED      , '['   , '['   },  // Left bracket
   /*  5C \   */  { CH_USED      , '\\'  , '\\'  },  // Back solidus
   /*  5D ]   */  { CH_USED      , ']'   , ']'   },  // Right bracket
   /*  5E ^   */  { CH_USED      , '^'   , '^'   },  // Circumflex accent
   /*  5F _   */  { CH_USED      , '_'   , '_'   },  // Underscore
   /*  60 `   */  { CH_USED      , '`'   , '`'   },  // Back quote
   /*  61 a   */  { S_LETTER     , 'a'   , 'A'   },  // Small letter A
   /*  62 b   */  { S_LETTER     , 'b'   , 'B'   },  // Small letter B
   /*  63 c   */  { S_LETTER     , 'c'   , 'C'   },  // Small letter C
   /*  64 d   */  { S_LETTER     , 'd'   , 'D'   },  // Small letter D
   /*  65 e   */  { S_LETTER     , 'e'   , 'E'   },  // Small letter E
   /*  66 f   */  { S_LETTER     , 'f'   , 'F'   },  // Small letter F
   /*  67 g   */  { S_LETTER     , 'g'   , 'G'   },  // Small letter G
   /*  68 h   */  { S_LETTER     , 'h'   , 'H'   },  // Small letter H
   /*  69 i   */  { S_LETTER     , 'i'   , 'I'   },  // Small letter I
   /*  6A j   */  { S_LETTER     , 'j'   , 'J'   },  // Small letter J
   /*  6B k   */  { S_LETTER     , 'k'   , 'K'   },  // Small letter K
   /*  6C l   */  { S_LETTER     , 'l'   , 'L'   },  // Small letter L
   /*  6D m   */  { S_LETTER     , 'm'   , 'M'   },  // Small letter M
   /*  6E n   */  { S_LETTER     , 'n'   , 'N'   },  // Small letter N
   /*  6F o   */  { S_LETTER     , 'o'   , 'O'   },  // Small letter O
   /*  70 p   */  { S_LETTER     , 'p'   , 'P'   },  // Small letter P
   /*  71 q   */  { S_LETTER     , 'q'   , 'Q'   },  // Small letter Q
   /*  72 r   */  { S_LETTER     , 'r'   , 'R'   },  // Small letter R
   /*  73 s   */  { S_LETTER     , 's'   , 'S'   },  // Small letter S
   /*  74 t   */  { S_LETTER     , 't'   , 'T'   },  // Small letter T
   /*  75 u   */  { S_LETTER     , 'u'   , 'U'   },  // Small letter U
   /*  76 v   */  { S_LETTER     , 'v'   , 'V'   },  // Small letter V
   /*  77 w   */  { S_LETTER     , 'w'   , 'W'   },  // Small letter W
   /*  78 x   */  { S_LETTER     , 'x'   , 'X'   },  // Small letter X
   /*  79 y   */  { S_LETTER     , 'y'   , 'Y'   },  // Small letter Y
   /*  7A z   */  { S_LETTER     , 'z'   , 'Z'   },  // Small letter Z
   /*  7B {   */  { CH_USED      , '{'   , '{'   },  // Left brace
   /*  7C |   */  { CH_USED      , '|'   , '|'   },  // Vertical bar
   /*  7D }   */  { CH_USED      , '}'   , '}'   },  // Right brace
   /*  7E ~   */  { CH_USED      , '~'   , '~'   },  // Tilda
   /*  7F    */  { CH_USED      , ''   , ''   },  // House
   /*  80    */  { CH_USED      , ''   , ''   },  // Capital letter C with cedilla
   /*  81    */  { CH_USED      , ''   , ''   },  // Small letter U with diaeresis
   /*  82    */  { CH_USED      , ''   , ''   },  // Small letter E with acute accent
   /*  83    */  { CH_USED      , ''   , ''   },  // Small letter A with circumflex accent
   /*  84    */  { CH_USED      , ''   , ''   },  // Small letter A with diaeresis
   /*  85    */  { CH_USED      , ''   , ''   },  // Small letter A with grave accent
   /*  86    */  { CH_USED      , ''   , ''   },  // Small letter A with ring above
   /*  87    */  { CH_USED      , ''   , ''   },  // Small letter C with cedilla
   /*  88    */  { CH_USED      , ''   , ''   },  // Small letter E with circumflex accent
   /*  89    */  { CH_USED      , ''   , ''   },  // Small letter E with diaeresis
   /*  8A    */  { CH_USED      , ''   , ''   },  // Small letter E with grave accent
   /*  8B    */  { CH_USED      , ''   , ''   },  // Small letter I with diaeresis
   /*  8C    */  { CH_USED      , ''   , ''   },  // Small letter I with circumflex accent
   /*  8D    */  { CH_USED      , ''   , ''   },  // Small letter I with grave accent
   /*  8E    */  { CH_USED      , ''   , ''   },  // Capital letter A with diaeresis
   /*  8F    */  { CH_USED      , ''   , ''   },  // Capital letter A with ring above
   /*  90    */  { CH_USED      , ''   , ''   },  // Capital letter E with acute accent
   /*  91    */  { CH_USED      , ''   , ''   },  // Small dipthong A with e
   /*  92    */  { CH_USED      , ''   , ''   },  // Capital dipthong A with E
   /*  93    */  { CH_USED      , ''   , ''   },  // Small letter O with circumflex accent
   /*  94    */  { CH_USED      , ''   , ''   },  // Small letter O with diaeresis
   /*  95    */  { CH_USED      , ''   , ''   },  // Small letter O with grave accent
   /*  96    */  { CH_USED      , ''   , ''   },  // Small letter U with circumflex accent
   /*  97    */  { CH_USED      , ''   , ''   },  // Small letter U with grave accent
   /*  98    */  { CH_USED      , ''   , ''   },  // Small letter Y with diaeresis
   /*  99    */  { CH_USED      , ''   , ''   },  // Capital letter O with diaeresis
   /*  9A    */  { CH_USED      , ''   , ''   },  // Capital letter U with diaeresis
   /*  9B    */  { CH_USED      , ''   , ''   },  // Small letter O with oblique stroke
   /*  9C    */  { CH_USED      , ''   , ''   },  // Pound sign
   /*  9D    */  { CH_USED      , ''   , ''   },  // Capital letter O with oblique stroke
   /*  9E    */  { CH_USED      , ''   , ''   },  // Multiplication sign
   /*  9F    */  { CH_USED      , ''   , ''   },  // Function of
   /*  A0    */  { CH_USED      , ''   , ''   },  // Small letter A with acute accent
   /*  A1    */  { CH_USED      , ''   , ''   },  // Small letter I with acute accent
   /*  A2    */  { CH_USED      , ''   , ''   },  // Small letter O with acute accent
   /*  A3    */  { CH_USED      , ''   , ''   },  // Small letter U with acute accent
   /*  A4    */  { CH_USED      , ''   , ''   },  // Small letter N with tilde
   /*  A5    */  { CH_USED      , ''   , ''   },  // Capital letter N with tilde
   /*  A6    */  { CH_USED      , ''   , ''   },  // Feminine ordinal indicator
   /*  A7    */  { CH_USED      , ''   , ''   },  // Masculine ordinal indicator
   /*  A8    */  { CH_USED      , ''   , ''   },  // Inverted question mark
   /*  A9    */  { CH_USED      , ''   , ''   },  // Registered trade mark sign
   /*  AA    */  { CH_USED      , ''   , ''   },  // Not sign
   /*  AB    */  { CH_USED      , ''   , ''   },  // Vulgar fraction one half
   /*  AC    */  { CH_USED      , ''   , ''   },  // Vulgar fraction one quarter
   /*  AD    */  { CH_USED      , ''   , ''   },  // Inverted exclamation mark
   /*  AE    */  { CH_USED      , ''   , ''   },  // Left angle quotation mark
   /*  AF    */  { CH_USED      , ''   , ''   },  // Right angle quotation mark
   /*  B0    */  { CH_USED      , ''   , ''   },  // 1/4 shaded block
   /*  B1    */  { CH_USED      , ''   , ''   },  // 1/2 shaded block
   /*  B2    */  { CH_USED      , ''   , ''   },  // 3/4 shaded block
   /*  B3    */  { CH_USED      , ''   , ''   },  // boxv - Vertical line
   /*  B4    */  { CH_USED      , 0xB4  , 0xB4  },  // boxvl
   /*  B5    */  { CH_USED      , ''   , ''   },  // Capital letter A with acute accent
   /*  B6    */  { CH_USED      , ''   , ''   },  // Capital letter A with circumflex accent
   /*  B7    */  { CH_USED      , ''   , ''   },  // Capital letter A with grave accent
   /*  B8    */  { CH_USED      , ''   , ''   },  // Copyright Sign
   /*  B9    */  { CH_USED      , ''   , ''   },  // boxVL
   /*  BA    */  { CH_USED      , ''   , ''   },  // boxV
   /*  BB    */  { CH_USED      , ''   , ''   },  // boxDL
   /*  BC    */  { CH_USED      , ''   , ''   },  // boxUL
   /*  BD    */  { CH_USED      , ''   , ''   },  // Cent sign
   /*  BE    */  { CH_USED      , ''   , ''   },  // Yen sign
   /*  BF    */  { CH_USED      , ''   , ''   },  // boxdl
   /*  C0    */  { CH_USED      , ''   , ''   },  // boxur
   /*  C1    */  { CH_USED      , ''   , ''   },  // boxuh
   /*  C2    */  { CH_USED      , ''   , ''   },  // boxdh
   /*  C3    */  { CH_USED      , ''   , ''   },  // boxvr
   /*  C4    */  { CH_USED      , ''   , ''   },  // boxh
   /*  C5    */  { CH_USED      , ''   , ''   },  // boxvh
   /*  C6    */  { CH_USED      , ''   , ''   },  // Small letter A with tilde
   /*  C7    */  { CH_USED      , ''   , ''   },  // Capital letter A with tilde
   /*  C8    */  { CH_USED      , ''   , ''   },  // boxUR
   /*  C9    */  { CH_USED      , ''   , ''   },  // boxDR
   /*  CA    */  { CH_USED      , ''   , ''   },  // boxUH
   /*  CB    */  { CH_USED      , ''   , ''   },  // boxDH
   /*  CC    */  { CH_USED      , ''   , ''   },  // boxVR
   /*  CD    */  { CH_USED      , ''   , ''   },  // boxH
   /*  CE    */  { CH_USED      , ''   , ''   },  // boxVH
   /*  CF    */  { CH_USED      , ''   , ''   },  // Currency sign
   /*  D0    */  { CH_USED      , ''   , ''   },  // Small icelandic letter THORN
   /*  D1    */  { CH_USED      , ''   , ''   },  // Capital icelandic letter THORN
   /*  D2    */  { CH_USED      , ''   , ''   },  // Capital letter E with circumflex accent
   /*  D3    */  { CH_USED      , ''   , ''   },  // Capital letter E with diaeresis
   /*  D4    */  { CH_USED      , ''   , ''   },  // Capital letter E with grave accent
   /*  D5    */  { CH_USED      , ''   , ''   },  // Small letter L reduced (?)
   /*  D6    */  { CH_USED      , ''   , ''   },  // Capital letter I with acute accent
   /*  D7    */  { CH_USED      , ''   , ''   },  // Capital letter I with circumflex accent
   /*  D8    */  { CH_USED      , ''   , ''   },  // Capital letter I with diaeresis
   /*  D9    */  { CH_USED      , ''   , ''   },  // boxul
   /*  DA    */  { CH_USED      , ''   , ''   },  // boxdr
   /*  DB    */  { CH_USED      , ''   , ''   },  // Full block
   /*  DC    */  { CH_USED      , ''   , ''   },  // Lower half block
   /*  DD    */  { CH_USED      , ''   , ''   },  // Broken bar
   /*  DE    */  { CH_USED      , ''   , ''   },  // Capital letter I with grave accent
   /*  DF    */  { CH_USED      , ''   , ''   },  // Upper half block
   /*  E0    */  { CH_USED      , ''   , ''   },  // Capital letter O with acute accent
   /*  E1    */  { CH_USED      , ''   , ''   },  // Small german letter sharp S
   /*  E2    */  { CH_USED      , ''   , ''   },  // Capital letter O with circumflex accent
   /*  E3    */  { CH_USED      , ''   , ''   },  // Capital letter O with grave accent
   /*  E4    */  { CH_USED      , ''   , ''   },  // Small letter O with tilde
   /*  E5    */  { CH_USED      , ''   , ''   },  // Capital letter O with tilde
   /*  E6    */  { CH_USED      , ''   , ''   },  // Micro sign
   /*  E7    */  { CH_USED      , ''   , ''   },  // Capital icelandic letter ETH
   /*  E8    */  { CH_USED      , ''   , ''   },  // Small icelandic letter ETH
   /*  E9    */  { CH_USED      , ''   , ''   },  // Capital letter U with acute accent
   /*  EA    */  { CH_USED      , ''   , ''   },  // Capital letter U with circumflex accent
   /*  EB    */  { CH_USED      , ''   , ''   },  // Capital letter U with grave accent
   /*  EC    */  { CH_USED      , ''   , ''   },  // Small letter Y with acute accent
   /*  ED    */  { CH_USED      , ''   , ''   },  // Capital letter Y with acute accent
   /*  EE    */  { CH_USED      , ''   , ''   },  // Macron
   /*  EF    */  { CH_USED      , ''   , ''   },  // Acute accent
   /*  F0    */  { CH_USED      , ''   , ''   },  // Shy (syllable hyphen)
   /*  F1    */  { CH_USED      , ''   , ''   },  // Plus-minus sign
   /*  F2    */  { CH_USED      , ''   , ''   },  // Double underscore
   /*  F3    */  { CH_USED      , ''   , ''   },  // Vulgar fraction three quarters
   /*  F4    */  { CH_USED      , ''   , ''   },  // Pilcrow sign
   /*  F5    */  { CH_USED      , ''   , ''   },  // Paragraph sign, section sign
   /*  F6    */  { CH_USED      , ''   , ''   },  // Division sign
   /*  F7    */  { CH_USED      , ''   , ''   },  // Cedilla
   /*  F8    */  { CH_USED      , ''   , ''   },  // Ring above, degree sign
   /*  F9    */  { CH_USED      , ''   , ''   },  // Diaeresis
   /*  FA    */  { CH_USED      , ''   , ''   },  // Middle dot
   /*  FB    */  { CH_USED      , ''   , ''   },  // Superscript one
   /*  FC    */  { CH_USED      , ''   , ''   },  // Superscript three
   /*  FD    */  { CH_USED      , ''   , ''   },  // Superscript two
   /*  FE    */  { CH_USED      , ''   , ''   },  // Black square
   /*  FF     */  { CH_USED      , 0xFF  , 0xFF  }   // NBSP (required blank)
   };

   /*-------------------------------------------------------------+
   | ISO Registration Number 109//CHARSET                         |
   |     ECMA-94 Right Part of Latin Alphabet Nr. 3               |
   +-------------------------------------------------------------*/
#if 0                           // NOT YET IMPLEMENTED
   static SetDescriptor ANSI_X3_4_1986_chtab[MAX_CHARSET] = {
   };
#endif

   enum baseset_identifier {
      ISO_646_1983_AHLID  = 16, // ISO 646-1983 (IRV)
      ISO_8859_AHLID      = 17, // ISO 8859-1 Latin alphabet No.1
      ANSI_X3_4_1986_AHLID= 18, // ANSI X3.4-1986 (ASCII)
      IBM_395_AHLID       = 19, // IBM codepage 395
      IBM_850_AHLID       = 20  // IBM codepage 850
   };
   static struct CHSETID {
//***>NOTE: These numbers will be const except Borland 3.1 complains
      int ahlid;
      char *pubid;
   }aChsetidTbl[] = {
      {
         ISO_646_1983_AHLID,
         "ISO 646-1983//CHARSET "
         "International Reference Version (IRV)//ESC 2/5 4/0"
      }, {
         ISO_646_1983_AHLID,
         "ISO 646-1983//CHARSET "
         "International Reference Version (IRV)//ESC 2/8 4/0"
      }, {
         ISO_646_1983_AHLID,
         "ISO 646-1983//CHARSET "
         "International Reference Version (IRV)//ESC 2/9 4/0"
      }, {
         ISO_8859_AHLID,
         "ISO Registration Number 100//CHARSET "
         "ECMA-94 Right-hand Part of Latin Alphabet Nr.1//ESC 2/8 4/2"
      }, {
         ISO_8859_AHLID,
         "ISO Registration Number 100//CHARSET "
         "ECMA-94 Right-hand Part of Latin Alphabet Nr.1//ESC 2/13 4/1"
      }, {
         ISO_8859_AHLID,
         "ISO Registration Number 100//CHARSET "
         "ECMA-94 Right-hand Part of Latin Alphabet Nr.1//ESC 2/14 4/1"
      }, {
         ISO_8859_AHLID,
         "ISO Registration Number 100//CHARSET "
         "ECMA-94 Right-hand Part of Latin Alphabet Nr.1//ESC 2/15 4/1"
      }, {
         ANSI_X3_4_1986_AHLID,
         "+//ANSI//CHARSET "
         "American Standard Code for Information Interchange (ASCII) "
         "ANSI X3.4-1986//ESC 2/8 4/2"
      }, {
         ANSI_X3_4_1986_AHLID,
         "+//ANSI//CHARSET "
         "American Standard Code for Information Interchange (ASCII) "
         "ANSI X3.4-1986//ESC 2/9 4/2"
      }, {
         ANSI_X3_4_1986_AHLID,
         "ANSI X3.4-1986//CHARSET "
         "American Standard Code for Information Interchange (ASCII)//"
         "ESC 2/8 4/2"
      }, {
         IBM_395_AHLID,
         "-//IBM//CHARSET "
         "Codepage 00395//EN"         // was: "Codepage 00395 1984-05//"
      }, {
         IBM_850_AHLID,
         "-//IBM//CHARSET "
         "Codepage 00850 1986-08//"
      }
   };

   for (int i=0; i < sizeof aChsetidTbl /sizeof aChsetidTbl[0]; ++i) {
      if (!strcmp(aChsetidTbl[i].pubid, pszPubid)) {
         switch(aChsetidTbl[i].ahlid) {
         case ISO_646_1983_AHLID:     // ISO 646-1983 (IRV)
            return ISO_646_1983_chtab;
         case ISO_8859_AHLID:         // ISO 8859-1 Latin alphabet No.1
            return ISO_8859_chtab;
         case IBM_395_AHLID:          // IBM codepage 395
            return IBM_395_chtab;
         case IBM_850_AHLID:          // IBM codepage 850
            return IBM_850_chtab;
         case ANSI_X3_4_1986_AHLID:   // ANSI X3.4-1986 (ASCII)
            return 0;                 // ANSI_X3_4_1986_chtab isn't implemented
         default:
            return 0;                 // Unknown
         }
      }
   }
   return 0;                          // Not found
}
/******************************************************************************/
