/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// chkambg.c
///  Check ambiguities of content models
//   Pierre G. Richard 89/06/02 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
/*
|    Check ambiguities using a Finite State Automata model.
*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include "gendef.h"
#include "quantity.h"
#include "prodcl.h"
#include "element.h"
#include "elemdcl.h"
#include "enthdlr.h"
#include "chkambg.h"
#include "errorh.h"

/*------------------------------------------------------------CheckAmbiguities-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
CheckAmbiguities::CheckAmbiguities(
   ModelNode *pMnode,
   int iMaximumNodes,
   boolean isPcdataDuplicateArg,
   EntityHandler *pEhArg,
   const char *pszPcdataNameArg
)
{
   pEh = pEhArg;
   aArcBuf = new Arc[4 * iMaximumNodes];
   pszPcdataName = pszPcdataNameArg;
   graph.pArcStart = aArcBuf;
   graph.pArcFinal = aArcBuf + (2 * iMaximumNodes);
   isPcdataDuplicate = isPcdataDuplicateArg;
   isLocationSaid = FALSE;
   analyse_graph(OR_CONN, pMnode, &graph);
}

/*-----------------------------------------------------------~CheckAmbiguities-+
| Destructor                                                                   |
+-----------------------------------------------------------------------------*/
CheckAmbiguities::~CheckAmbiguities()
{
   delete [] aArcBuf;
}

/*------------------------------------------------------------------enter_arcs-+
| Move arcs from `pArcSourceFrom' to `pArcSourceTo' to target `pArcTargetFrom',|
| excluding the arcs found in the `pArcCompareFrom' to `pArcCompareTo' list.   |
| Returns next target.                                                         |
+-----------------------------------------------------------------------------*/
Arc * CheckAmbiguities::enter_arcs(
   Arc *pArcSourceFrom, Arc *pArcSourceTo,   // Source (arcs to be entered)
   Arc *pArcTargetFrom,                      // Target (where to enter them)
   Arc *pArcCompareFrom, Arc *pArcCompareTo  // Compared to (check redundancy)
) {
   Arc *pArcTarget = pArcTargetFrom;
   for (
      Arc *pArcSource = pArcSourceFrom;
      pArcSource < pArcSourceTo;
      ++pArcSource
   ) {
      for (
         Arc *pArcCompare = pArcCompareFrom;
         ;
         ++pArcCompare
      ) {
         if (pArcCompare == pArcCompareTo) {      // not found in the list:
            *(pArcTarget++) = *pArcSource;        // enter new arc
            break;
         }
         if (*pArcSource == *pArcCompare) break;  // found in the list: break
      }
   }
   return pArcTarget;
}

/*-----------------------------------------------------------check_ambiguities-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void CheckAmbiguities::check_ambiguities(
   ModelNode *pMnode1, ModelNode *pMnode2
)
{
   GenericIdBlock *pGi1, *pGi2;

   if (pMnode1 == pMnode2) return;
   if (pMnode1->is.ambig_found) return;   // Already said

   if (pMnode1->is.datatag) {
      pGi1 = 0;
   }else {
      pGi1 = pMnode1->pGi;
   }
   if (pMnode2->is.datatag) {
      pGi2 = 0;
   }else {
      pGi2 = pMnode2->pGi;
   }
   if (pGi1 == pGi2) {
      if (!isLocationSaid) {
         isLocationSaid = TRUE;
         errorh(AMBMDL, pEh, ERH_E_SEV, pGi1? pGi1->pszName : pszPcdataName);
      }else {                       // Do not locate the error more than twice
         errorh(AMBMDL, ERH_E_SEV, pGi1? pGi1->pszName : pszPcdataName);
      }
      pMnode1->is.ambig_found = TRUE;
   }
}

/*---------------------------------------------------------------analyse_graph-+
| Recursively reduce each model group to a list of:                            |
|    - arcs emaning from start state, (pArcStart)                              |
|    - arcs emaning from final state, (pArcFinal)                              |
|                                                                              |
| A) Given a group, examine sequentially each token.                           |
|    - If the token is an elementary token, add the element to new_pArcStart;  |
|    - If the token is a new group, recurse and get the new_pArcStart and      |
|      new_pArcFinal of this group.                                            |
|                                                                              |
| B) If the token is repeatable, check the REPEAT_AMBG ambiguities: confront   |
|    the arcs emaning from final states with arcs emaning from start states    |
|    (new_St vs new_Fn).                                                       |
|                                                                              |
| C) If the token is omittable, the arcs emaning from final states must        |
|    include those emaning from start state: because the token is omittable,   |
|    the start-state can also be a final-state.                                |
|                                                                              |
| D) Then, update the pArcStart/pArcFinal for the group from the new_pArcStart |
|    and/or new_pArcFinal, depending of the group's connector:                 |
|                                                                              |
| D-1) In an OR group:                                                         |
|    - arcs-from-start-state of the group are                                  |
|      arcs-from-start-state of each token                                     |
|    - arcs-from-final-state of the group are                                  |
|      arcs-from-final-state of each token                                     |
|                                                                              |
| D-2) In an AND group:                                                        |
|    AND group are like OR group, except that, for ALL tokens of the group     |
|    arcs-from-start-state must not duplicate ANY arc-from-final-state.        |
|    Since AND are processed like ORs, this test (check for AND_AMBG           |
|    ambiguities) can be done at the end since the group.St and group.Fn       |
|    are the union of (respectively) each token St and Fn.                     |
|                                                                              |
| D-3) In a SEQ group:                                                         |
|    - check SEQUENCE_AMBG ambiguities: confront arcs-from-start-state         |
|      of each token with current arcs-from-final-state of the group.          |
|    - stop cumulating arcs-from-start-state after the first required token    |
|      is found.  I.e., the first token non OMISSIBLE ('+' or '1') is the      |
|      last token the new_pArcStart of which is added to the pArcStart of      |
|      the group.                                                              |
|    - then, forget about arcs-from-final-state, until the token is required.  |
|      I.e., pArcFinal of the group begins to be updated with the last         |
|      required token, including this one iff it is repeatable ('+').          |
|      Note: this last condition is not explicitely coded at this point:       |
|           if the last required token is also repeatable ('+'), it has        |
|           already been added to new_Fn by C).                                |
|                                                                              |
| E) Detect the other ambiguities (ST_AMBG) by examining the pArcStart.        |
|                                                                              |
|                                                                              |
| Implementation Notes:                                                        |
| -  Arcs of the group (pArcStart/pArcFinal) are stored between group_graph    |
|    and new_graph.                                                            |
|    Arcs of the currently examined token (new_pArcStart/new_pArcFinal)        |
|    are between new_graph and end_graph                                       |
|                                                                              |
| -  To speed up the process, only arcs which are duplicated are stored.       |
+-----------------------------------------------------------------------------*/
void CheckAmbiguities::analyse_graph(
   e_Connector conn, ModelNode *pMnode, Graph * pGraphGroup
)
{
   Graph graphNew;
   Graph graphEnd;                      // working graph for recursion
   boolean isRequiredFound = FALSE;

   graphNew = *pGraphGroup;
   for (;;) {
      graphEnd = graphNew;
      if (pMnode->is.conn) {
         analyse_graph(pMnode->is.conn, pMnode->llink, &graphEnd);
      }else if (pMnode->is.datatag || pMnode->is.pcdata) {
         if (isPcdataDuplicate) *(graphEnd.pArcStart++) = pMnode;
      }else if (pMnode->pGi->isMoreThanOnce) {
         *(graphEnd.pArcStart++) = pMnode;
      }
      if (pMnode->is.occur) {                       // omittable/repeatable
         if (pMnode->is.occur & REPEATABLE) {       // '+' or '*'
            for (
               Arc *pArc1 = graphNew.pArcFinal;
               pArc1 < graphEnd.pArcFinal;
               ++pArc1
            ) {
               for (
                  Arc *pArc2 = graphNew.pArcStart;
                  pArc2 < graphEnd.pArcStart;
                  ++pArc2
               ) {
                  check_ambiguities(*pArc1, *pArc2);
               }
            }
         }
         if (pMnode->is.occur & OMISSIBLE) {
            /*------------------------------------------------+
            | '?' or '*': add new_pArcStart to new_pArcFinal, |
            | not already part of new_pArcFinal               |
            +------------------------------------------------*/
            graphEnd.pArcFinal = enter_arcs(
               graphNew.pArcStart, graphEnd.pArcStart,
               graphEnd.pArcFinal,
               graphNew.pArcFinal, graphEnd.pArcFinal
            );
         }
      }
      switch(conn) {
      case SEQ_CONN:
         /*---------------------------------------------------------------+
         |Check ambiguities between previous pArcFinal vs. new_pArcStart: |
         +---------------------------------------------------------------*/
         {
            for (
               Arc *pArc1 = pGraphGroup->pArcFinal;
               pArc1 < graphNew.pArcFinal;
               ++pArc1
            ) {
               for (
                  Arc *pArc2 = graphNew.pArcStart;
                  pArc2 < graphEnd.pArcStart;
                  ++pArc2
               ) {
                  check_ambiguities(*pArc1, *pArc2);
               }
            }
         }
         if (!isRequiredFound) {
            /*-------------------------------------------------------+
            | Until required is found, continue to add new_pArcStart |
            | to pArcStart, not already enetered in the group        |
            +-------------------------------------------------------*/
            graphNew.pArcStart = enter_arcs(
               graphNew.pArcStart, graphEnd.pArcStart,
               graphNew.pArcStart,
               pGraphGroup->pArcStart, graphNew.pArcStart
            );
         }
         if (!(pMnode->is.occur & OMISSIBLE)) {
            /*--------------------------------------------+
            | If required found, reset group pArcFinal to |
            | just the new_pArcFinal adding to the group. |
            | Compare is useless.                         |
            +--------------------------------------------*/
            isRequiredFound = TRUE;
            graphNew.pArcFinal = enter_arcs(
               graphNew.pArcFinal, graphEnd.pArcFinal,
               pGraphGroup->pArcFinal,
               0, 0
            );
         }else {
            /*-------------------------------------------+
            | Continue to add new_pArcFinal to pArcFinal |
            | not already entered in the group           |
            +-------------------------------------------*/
            graphNew.pArcFinal = enter_arcs(
               graphNew.pArcFinal, graphEnd.pArcFinal,
               graphNew.pArcFinal,
               pGraphGroup->pArcFinal, graphNew.pArcFinal
            );
         }
         break;
      default: /* AND_CONN, OR_CONN */
         /*---------------------------------------------------------+
         | Add new_pArcStart to group pArcStart, avoiding duplicate |
         +---------------------------------------------------------*/
         graphNew.pArcStart = enter_arcs(
            graphNew.pArcStart, graphEnd.pArcStart,
            graphNew.pArcStart,
            pGraphGroup->pArcStart, graphNew.pArcStart
         );
         /*---------------------------------------------------------+
         | Add new_pArcFinal to group pArcFinal, avoiding duplicate |
         +---------------------------------------------------------*/
         graphNew.pArcFinal = enter_arcs(
            graphNew.pArcFinal, graphEnd.pArcFinal,
            graphNew.pArcFinal,
            pGraphGroup->pArcFinal, graphNew.pArcFinal
         );
         break;
      }
      if (!pMnode->is.rtag) break;
      pMnode = pMnode->rlink;
   }
   if (conn == AND_CONN) {
      /*-----------------------------------------------+
      |Check ambiguities between group_St vs. group_Fn |
      +-----------------------------------------------*/
      for (
         Arc *pArc1=pGraphGroup->pArcFinal;
         pArc1 < graphNew.pArcFinal;
         ++pArc1
      ) {
         for (
            Arc *pArc2 = pGraphGroup->pArcStart;
            pArc2 < graphNew.pArcStart;
            ++pArc2
         ) {
            if (*pArc1 != *pArc2) check_ambiguities(*pArc1, *pArc2);
         }
      }
   }
   /*--------------------------------+
   |Confront pGraphGroup->pArcStart  |
   |to get the ambiguities if any.   |
   +--------------------------------*/
   {
      for (
         Arc *pArc1=pGraphGroup->pArcStart;
         pArc1 < graphNew.pArcStart-1;
         ++pArc1
      ) {
         for (
            Arc *pArc2 = pArc1 + 1;
            pArc2 < graphNew.pArcStart;
            ++pArc2
         ) {
            check_ambiguities(*pArc1, *pArc2);
         }
      }
   }
   *pGraphGroup = graphNew;
}
/******************************************************************************/
