/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// dlmvect.c
///  Build delimiter vector array for fast recognition of the SGML delimiters
//   Pierre G. Richard 87/08/20 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include "gendef.h"
#include "enthdlr.h"
#include "dlmvect.h"
#include "dlmstr.h"
#include "samgr.h"
#include "refstr.h"
#include "errorh.h"

/*----------------+
| Local Functions |
+----------------*/
static int compare_delim(
   const void *pItem1, const void *pItem2, const void * const pBase
);

/*-------------------------------------------------------------DelimiterVector-+
| Minimal VectorItem table and dltab of the CORE Concrete Syntax               |
|                                                                              |
| Three minimal dvtab will be built, for LITP_MODE, MD_MODE, and INI_MODE.     |
|  in LITP_MODE: only LIT, LITA, and CRO will be recognized                    |
|  in MD_MODE:   only LIT, LITA, COM, and MDC will be recognized               |
|  in INI_MODE:  only STAGO, MDO, MDO_COM, MDO_MDC will be recognized          |
|                                                                              |
| The final dvtable will look like: (except the pcl is sorted on the           |
|                                                     system character set)    |
|      (*---------------LITP_MODE-----------------*)                           |
|      {         3<<8,  0          , "\"\'&"      },                           |
|      {            0,  LIT_TK     , EMPTY_STRING },                           |
|      {            0,  LITA_TK    , EMPTY_STRING },                           |
|      {            0,  CRO_TK     , "#"          },                           |
|      (*----------------MD_MODE------------------*)                           |
|      {         4<<8,  0          , "\"\'->"     },                           |
|      {            0,  LIT_TK     , EMPTY_STRING },                           |
|      {            0,  LITA_TK    , EMPTY_STRING },                           |
|      {            0,  COM_TK     , "-"          },                           |
|      {            0,  MDC_TK     , EMPTY_STRING },                           |
|      (*---------------INI_MODE------------------*)                           |
|      {         1<<8,  0          , "<"          },                           |
|      {   +---(VI *),  0          , EMPTY_STRING },                           |
|          |                                                                   |
|      {   +-->  1<<8,  STAGO_TK   , "!"          },                           |
|      {   +---(VI *),  0          , EMPTY_STRING },                           |
|          |                                                                   |
|      {   +-->  2<<8,  MDO_TK     , "->"         },                           |
|      {            0,  MDO_COM_TK , "-"          },                           |
|      {            0,  MDO_MDC_TK , EMPTY_STRING }                            |
|                                                                              |
+-----------------------------------------------------------------------------*/
DelimiterVector::DelimiterVector()
{
   struct Bunch {                     // CORE concrete syntax is... in core
      VectorItem aViLitTable[4];
      VectorItem aViMdTable[5];
      VectorItem aViIniTable[7];
      char  pchPclLessthan[1];        // |<|   (Preponderant Character List)
      char  pchPclExclam[1];          // |!|
      char  pchPclLitMode[3];         // |"|'|&|
      char  pchPclMdMode[4];          // |"|'|-|>|
      char  pchPclIniMode[2];         // |-|>|
      char  pszCsPound[1 + 1];        // |#|   (Continuation String)
      char  pszCsDash[1 + 1];         // |-|
      char  pszCsEmpty[1];            // ||
   };

   Bunch *pBunch = (Bunch *)malloc(sizeof(Bunch));
   aViTable = (VectorItem *)pBunch;   // Remember address to be freed
   iMaxVi = (
      (
         sizeof pBunch->aViLitTable +
         sizeof pBunch->aViMdTable +
         sizeof pBunch->aViIniTable
      ) /
      sizeof(VectorItem)
   );
   pBunch->pchPclLessthan[0] = '<';
   pBunch->pchPclExclam[0]   = '!';

   pBunch->pszCsPound[0]     = '#';
   pBunch->pszCsDash[0]      = '-';
   pBunch->pszCsPound[sizeof pBunch->pszCsPound - 1] = '\0';
   pBunch->pszCsDash[sizeof  pBunch->pszCsDash  - 1] = '\0';
   pBunch->pszCsEmpty[sizeof pBunch->pszCsEmpty - 1] = '\0';

   /*-------------------------------------+
   |Initialize primary nodes in min dvtab |
   +-------------------------------------*/
   dvtbent[LITP_MODE].pVi = pBunch->aViLitTable;
   dvtbent[MD_MODE].pVi   = pBunch->aViMdTable;
   dvtbent[INI_MODE].pVi  = pBunch->aViIniTable;

   VectorItem *pViLit = pBunch->aViLitTable;
   pViLit->h.isBseq = FALSE;
   pViLit->h.count = sizeof pBunch->pchPclLitMode;
   pViLit->tk = NOT_FOUND_TK;
   pViLit->str = pBunch->pchPclLitMode;              // will be |"|'|&|

   VectorItem *pViMd = pBunch->aViMdTable;
   pViMd->h.isBseq = FALSE;
   pViMd->h.count = sizeof pBunch->pchPclMdMode;
   pViMd->tk = NOT_FOUND_TK;
   pViMd->str = pBunch->pchPclMdMode;                // will be |"|'|-|>|

   VectorItem *pViIni = pBunch->aViIniTable;
   pViIni->h.isBseq = FALSE;
   pViIni->h.count = sizeof pBunch->pchPclLessthan;
   pViIni->tk = NOT_FOUND_TK;
   pViIni->str = pBunch->pchPclLessthan;        // |<|

   ++pViIni;
   pViIni->next = pViIni + 1;
   pViIni->tk = NOT_FOUND_TK;
   pViIni->str = pBunch->pszCsEmpty;

   ++pViIni;
   pViIni->h.isBseq = FALSE;
   pViIni->h.count = sizeof pBunch->pchPclExclam;
   pViIni->tk = STAGO_TK;
   pViIni->str = pBunch->pchPclExclam;          // |!|

   ++pViIni;
   pViIni->next = pViIni + 1;
   pViIni->tk = NOT_FOUND_TK;
   pViIni->str = pBunch->pszCsEmpty;

   ++pViIni;
   pViIni->h.isBseq = FALSE;
   pViIni->h.count = sizeof pBunch->pchPclIniMode;
   pViIni->tk = MDO_TK;
   pViIni->str = pBunch->pchPclIniMode;         // will be |-|>|


   /*---------------------+
   |Setup secondary nodes |
   +---------------------*/
   int iLitPcl = 0;
   int iMdPcl = 0;
   int iIniPcl = 0;

   for (int i=0; i < MAX_CHARSET; ++i) {
      switch (i) {             // need to do it this way: a pcl is sorted
      case '\"':               // LIT
         pBunch->pchPclLitMode[iLitPcl++] = (char)i;
         ++pViLit;
         pViLit->next = 0;
         pViLit->tk = LIT_TK;
         pViLit->str = pBunch->pszCsEmpty;

         pBunch->pchPclMdMode[iMdPcl++] = (char)i;
         ++pViMd;
         pViMd->next = 0;
         pViMd->tk = LIT_TK;
         pViMd->str = pBunch->pszCsEmpty;

         break;

      case '\'':               // LITA
         pBunch->pchPclLitMode[iLitPcl++] = (char)i;
         ++pViLit;
         pViLit->next = 0;
         pViLit->tk = LITA_TK;
         pViLit->str = pBunch->pszCsEmpty;

         pBunch->pchPclMdMode[iMdPcl++] = (char)i;
         ++pViMd;
         pViMd->next = 0;
         pViMd->tk = LITA_TK;
         pViMd->str = pBunch->pszCsEmpty;

         break;

      case '&':                // CRO
         pBunch->pchPclLitMode[iLitPcl++] = (char)i;
         ++pViLit;
         pViLit->next = 0;
         pViLit->tk = CRO_TK;
         pViLit->str = pBunch->pszCsPound;    // |#|

         break;

      case '-':                // COM
         pBunch->pchPclMdMode[iMdPcl++] = (char)i;
         ++pViMd;
         pViMd->next = 0;
         pViMd->tk = COM_TK;
         pViMd->str = pBunch->pszCsDash;      //  |-|

         pBunch->pchPclIniMode[iIniPcl++] = (char)i;
         ++pViIni;
         pViIni->next = 0;
         pViIni->tk = MDO_COM_TK;
         pViIni->str = pBunch->pszCsDash;     //  |-|

         break;

      case '>':                // MDC
         pBunch->pchPclMdMode[iMdPcl++] = (char)i;
         ++pViMd;
         pViMd->next = 0;
         pViMd->tk = MDC_TK;
         pViMd->str = pBunch->pszCsEmpty;

         pBunch->pchPclIniMode[iIniPcl++] = (char)i;
         ++pViIni;
         pViIni->next = 0;
         pViIni->tk = MDO_MDC_TK;
         pViIni->str = pBunch->pszCsEmpty;

         break;

      default:
         break;
      }
   }

   /*--------------+
   | Set up dltab  |
   +--------------*/
   memset(dltab, 0, sizeof dltab);
   dltab['<'] = (1 << INI_MODE);
   dltab['\"'] = (1 << LITP_MODE) | (1 << MD_MODE);
   dltab['\''] = (1 << LITP_MODE) | (1 << MD_MODE);
   dltab['&'] = (1 << LITP_MODE);
   dltab['-'] = (1 << MD_MODE);
   dltab['>'] = (1 << MD_MODE);
}

/*-------------------------------------------------------------DelimiterVector-+
| DelimiterVector constructor from OCS                                         |
|                                                                              |
|  (S): Relocate string in stringbuf                                           |
|  (V): Relocate VectorItem in the VectorItem table                            |
|                                                                              |
|                           +-------------------+---------+------------------+ |
| For a Primary Node   PDV= |count, isBseq      |  Token  | Pred Char Str(S) | |
|                           +-------------------+---------+------------------+ |
|                                                                              |
|                           +-------------------+---------+------------------+ |
| For a Secondary Node SDV= |Next_PDV(V) or NULL|  Token  |  Contin Str(S)   | |
|                           +-------------------+---------+------------------+ |
|                                                                              |
|                           +-------------------+---------+------------------+ |
| For a Bseq Node      BDV= |Next_PDV(V) or NULL|  Token  |  Contin Str(S)   | |
|                           --------------------+---------+------------------- |
|                           |Prev_BDV(V) or NULL| Bseqlen |  must_recur      | |
|                           +-------------------+---------+------------------+ |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
|                                                                              |
+-----------------------------------------------------------------------------*/
DelimiterVector::DelimiterVector(
   ApplicationHandle ahlAccess,     // OCS Access handle
   char *strbuf                     // string area of this OCS
)
{
   int iMaxVi;

   EntityHandler::read(ahlAccess, dvtbent, sizeof dvtbent);
   EntityHandler::read(ahlAccess, dltab, sizeof dltab);
   iMaxVi = EntityHandler::read_record(
      ahlAccess, (void **)&aViTable
   ) / sizeof aViTable[0];

   /*
   | Solve dvtbent entries
   */
   {
      for (int i=0; i < sizeof dvtbent / sizeof dvtbent[0]; ++i) {
          dvtbent[i].pVi = aViTable + dvtbent[i].rel;
      }
   }

   /*
   | Relocation of the VectorItem array
   */
   {
      boolean isBseqExtension = FALSE;
      int i = 0;

      for (VectorItem *pVi=aViTable; iMaxVi--; ++pVi) {
         if ((i!=1) || (!isBseqExtension)) {  // not a B-seq Extension
            pVi->str = strbuf + pVi->ofsStr;  // continuation string, or PCL
         }
         if (!i) {                            // Primary Node
            i = pVi->h.count;                 // Reset secondary node count
            if (pVi->h.isBseq) {              // B-Seq included ?
               i += 2;                        // Yes: bump count
               isBseqExtension = TRUE;        // Remember a B-Seq follows
            }else {
               isBseqExtension = FALSE;       // No B-Seq in the list
            }
         }else {                              // Secondary or B-Seq Node
            if (pVi->ixVi) {
               pVi->next = aViTable + pVi->ixVi;
            }else {
               pVi->next = 0;
            }
            --i;
         }
      }
   }
}

/*PROTECTED----------------------------------------------------DelimiterVector-+
| Constructor for building delimiter vectors from the SGML declaration         |
+-----------------------------------------------------------------------------*/
DelimiterVector::DelimiterVector(int)
{
   memset(dltab, 0, sizeof dltab);
   aViTable = 0;   // be safe
}

/*------------------------------------------------------------~DelimiterVector-+
| DelimiterVector destructor                                                   |
+-----------------------------------------------------------------------------*/
DelimiterVector::~DelimiterVector()
{
   free(aViTable);
}

/*----------------------------------------------------------NewDelimiterVector-+
| NewDelimiterVector constructor                                               |
+-----------------------------------------------------------------------------*/
NewDelimiterVector::NewDelimiterVector(
   StringKeeper& skArg, NewSyntax& syntaxNewArg
) :
   DelimiterVector(0), sk(skArg), syntaxNew(syntaxNewArg)
{
   pSaDelim = new SortArray(
      sizeof(DelimDesc),         // size of each array item
      compare_delim,             // comparison function
      &iCountDlmdsc,             // where to return count of items entered
      (void **)&aDlmdscTable,    // where to return address of the array
      (void *)this
   );
   DelimDesc dlmdsc;
   dlmdsc.token = NOT_FOUND_TK;
   *(dlmdsc.pszDelim = (char *)malloc(1)) = '\0';
   pSaDelim->enter(&dlmdsc);
}

/*------------------------------------------------------------define_delimiter-+
| Enter a new delimiter:                                                       |
| - Update the new GTYPE to make all its characters belong to the SGML class   |
| - Register the first character into DLTAB (fast recognition)                 |
| - Store its value, prefixed by a letter, into the SaDelim SortedArray.       |
|                                                                              |
| The delimiter is prefixed by a letter: [a..z], corresponding to its mode     |
| recognition mode.  This letter will act as a mjor sort key when the          |
| SaDelim array will be sorted -- this preludes to the creation of the table   |
| of vector items (aViTable) by build_delimiter_vectors.                       |
+-----------------------------------------------------------------------------*/
void NewDelimiterVector::define_delimiter(
   const char *pszValue,
   const e_LexToken token,
   unsigned int allmode,
   const boolean isSingleDelimiter
)
{
   const boolean isShortref = (token >= SR_first)? TRUE : FALSE;
   const int iLen = 1 + strlen(pszValue);
   /*--------------------------------+
   | Store each of the strings       |
   | forming the combined delimiters |
   +--------------------------------*/
   if (isSingleDelimiter) {
      for (int i=0; i < DDLI_MAX_SIZE; ++i) {
         if (token == ReferenceString::delim1_token_of((e_IxDDlmitmRefer)i)) {
            memcpy(aDdlmvstr[i].s1 = (char *)malloc(iLen), pszValue, iLen);
         }
         if (token == ReferenceString::delim2_token_of((e_IxDDlmitmRefer)i)) {
            memcpy(aDdlmvstr[i].s2 = (char *)malloc(iLen), pszValue, iLen);
         }
      }
   }
   for (const char *pch = pszValue; *pch; ++pch) {       // force SGML class
      if ((!isShortref) || (*pch != syntaxNew.chBSEQ)) { // not for BSEQ char
         syntaxNew.gtype[*pch] |= S_SGML;
      }
   }
   if ((isShortref) && (pszValue[0] == syntaxNew.chBSEQ)) {
      for (int i=0; i < MAX_CHARSET; ++i) {
         if (syntaxNew.is_blank(i)) dltab[i] |= allmode;
      }
   }else {
      dltab[pszValue[0]] |= allmode;          // mode in dltab[1st char]
   }

   DelimDesc dlmdsc;
   dlmdsc.token = token;
   for (char c='a'; allmode; allmode >>= 1, ++c) {
      if (allmode & 1) {                   // could be not executed at all
         dlmdsc.pszDelim = (char *)malloc(1+iLen);
         dlmdsc.pszDelim[0] = c;
         memcpy(dlmdsc.pszDelim+1, pszValue, iLen);
         pSaDelim->enter(&dlmdsc);
      }
   }
}

/*----------------------------------------------------------compare_delim_desc-+
| Comparison between two delimiters descriptors.                               |
| The B-sequence (for ShortRef's) will be greater than any other character.    |
| Note: Duplicates are caught in check_duplicate_delimiter, called by get-stub.|
+-----------------------------------------------------------------------------*/
int NewDelimiterVector::compare_delim_desc(
   const DelimDesc& dlmdsc1, const DelimDesc& dlmdsc2
)
{
   const char *psz1;
   const char *psz2;

   for (
      psz1 = dlmdsc1.pszDelim, psz2 = dlmdsc2.pszDelim;
      (*psz1 == *psz2) && (*psz1);             // until it differs or EOS
      ++psz1, ++psz2
   );
   if ((dlmdsc1.token >= SR_first) && (*psz1 == syntaxNew.chBSEQ)) {
      return 1;
   }else if ((dlmdsc2.token >= SR_first) && (*psz2 == syntaxNew.chBSEQ)) {
      return -1;
   }else {
      return *psz1 - *psz2;
   }
}

/*STATIC---------------------------------------------------------compare_delim-+
| Call the `compare_delimiter_desc' to sort the SortArray delimiter items.     |
+-----------------------------------------------------------------------------*/
static int compare_delim(
   const void *pItem1, const void *pItem2, const void * const pBase
)
{
   return ((NewDelimiterVector *)pBase)->compare_delim_desc(
      *((const NewDelimiterVector::DelimDesc *)pItem1),
      *((const NewDelimiterVector::DelimDesc *)pItem2)
   );
}

/*-----------------------------------------------------build_delimiter_vectors-+
| Close the saDelim SortArray, build the dvtbent                               |
+-----------------------------------------------------------------------------*/
void NewDelimiterVector::build_delimiter_vectors(
   Features& featuresNew, NewDelimiterString &dlmstrNew
)
{
   /*--------------------------------+
   | Install the combined delimiters |
   +--------------------------------*/
   for (int i=0; i < DDLI_MAX_SIZE; ++i) {
      boolean isToInstall;

      e_IxDDlmitmRefer ix = (e_IxDDlmitmRefer)i;
      switch(ix) {                              /* contextual constraints */
      case STAGO_TAGC_DDLI_ix:                  /* GI:  */
      case ETAGO_TAGC_DDLI_ix:                  /* GI:  */
         isToInstall = featuresNew.isShorttag;
         break;
      case STAGO_GRPO_DDLI_ix:                  /* GI:  */
      case ETAGO_GRPO_DDLI_ix:                  /* GI:  */
         isToInstall = featuresNew.isConcur;
         break;
      case ERO_GRPO_DDLI_ix:                    /* NMS: */
      case PERO_GRPO_DDLI_ix:                   /* NMS: */
         isToInstall = (
            (featuresNew.isConcur)   || (featuresNew.isSimple)  ||
            (featuresNew.isExplicit) || (featuresNew.isImplicit)
         )? TRUE: FALSE;
         break;
      default:
         isToInstall = TRUE;
         break;
      }
      char *pchPart1 = aDdlmvstr[ix].s1;
      char *pchPart2 = aDdlmvstr[ix].s2;
      if (isToInstall) {
         pchPart1 = (char *)realloc(
            pchPart1, 1 + strlen(pchPart1) + strlen(pchPart2)
         );
         strcat(pchPart1, pchPart2);
         dlmstrNew.set_delimiter_string_table(ix, pchPart1);
         define_delimiter(
            pchPart1,
            ReferenceString::token_of(ix),
            ReferenceString::recognition_modes_of(ix),
            FALSE
         );
      }
      free(pchPart1);
      free(pchPart2);
   }

   /*------------------------------+
   | Get the new dvtab and dvtbent |
   +------------------------------*/
   delete pSaDelim;
   {
      DelimTree TrHead;

      TrHead.ltag = FALSE;
      TrHead.rtag = TRUE;
      TrHead.llink = TrHead.rlink = &TrHead;
      iMaxVi = - (NBR_OF_MODES + 1);
      get_stub(0, 0, 1, iCountDlmdsc, &TrHead, aDlmdscTable[0].pszDelim);
      aViTable = (VectorItem *)malloc(iMaxVi * sizeof(VectorItem));

      pViCurr = aViTable;
      ixViPrevBseq = 0;

      /*-----------------------------------------+
      | For each mode, set up entries in dvtbent |
      +-----------------------------------------*/
      DelimTree *pTrNext = TrHead.llink->rtag?  TrHead.llink->rlink : 0;
      for (int iMode=0; pTrNext; ++iMode) {
         DelimTree *pTr = pTrNext;
         dvtbent[iMode].rel = diffptr(pViCurr, aViTable);
         if (pTr->ltag) prim_postorder(pTr->llink);
         pTrNext = pTr->rtag?  pTr->rlink : 0;
         free(pTr);
      }
      free(TrHead.llink);
   }
   /*
   | Clean up
   */
   for (i=0; i < iCountDlmdsc; ++i) {
      free(aDlmdscTable[i].pszDelim);
   }
   free(aDlmdscTable);
}

/*--------------------------------------------------------------------get_stub-+
| Process a sublist of keywords, and build the tree.                           |
| ENTRY: Sublist coordinates, parent node, prep char                           |
| when RETURN: the branch at parent node is updated                            |
+-----------------------------------------------------------------------------*/
void NewDelimiterVector::get_stub(
   unsigned int iColMin,
   unsigned int iRowMin,
   unsigned int iRowHigh,
   unsigned int iRowMax,
   DelimTree *pTr,
   char *pchStubBeg
)
{
   /*------------------------------------+
   | Create a new child node in the tree |
   +------------------------------------*/
   {
      DelimTree *pTrNew = (DelimTree *)malloc(sizeof(DelimTree)); // new child
      pTrNew->rlink = pTr;                    // set father thread
      pTrNew->llink = pTr->llink;             // inherits brother/uncle thread
      pTrNew->rtag = pTrNew->ltag = FALSE;    // just threads
      pTr->ltag = TRUE;                       // Brand new dad
      pTr->llink = pTrNew;                    // link father to his child
      pTr = pTrNew;                           // activate child
   }

   int iCount;
   /*-------------------------------------------+
   | If not a dead leaf, get a new brother node |
   +-------------------------------------------*/
   if (*pchStubBeg) {
      pTr->token = NOT_FOUND_TK;              // no info into the primary node
      DelimTree *pTrNew = (DelimTree *)malloc(sizeof(DelimTree)); // brother
      pTrNew->llink = pTr;                    // set elder thread
      pTrNew->rlink = pTr->rlink;             // inherits father thread
      pTrNew->rtag = pTrNew->ltag = FALSE;    // just threads
      pTr->rtag = TRUE;                       // Brand new elder
      pTr->rlink = pTrNew;                    // link elder to his brother
      pTr = pTrNew;                           // activate brother
      iCount = 1;
   }else {
      iCount = 0;
   }

   boolean isBseqFound = FALSE;
   /*-----------------------------------------------------+
   | Determines all subclasses of [iRowMin..iRowMax][col] |
   +-----------------------------------------------------*/
   for (;;) {
      boolean isShortref = (
         aDlmdscTable[iRowMin].token >= SR_first
      )? TRUE : FALSE;
      boolean isInBseq = (
         (isShortref) && (*pchStubBeg == syntaxNew.chBSEQ)
      )? TRUE : FALSE;
      int iCol = iColMin;
      char *pchStubEnd = pchStubBeg;
      char c = *pchStubEnd;
      unsigned int iRow;
      /*--------------------------------------------------------------------+
      | Get the longest prefix in the [iRowMin..iRowHigh][iColMin] subclass |
      |                                                                     |
      | On entry:                                                           |
      |   pchStubBeg points at the beginning;                               |
      |   pchStubEnd will point at the end;                                 |
      |                                                                     |
      | When the loop exits:                                                |
      |    [iRowMin..iRow][iCol] defines the first sub-subclass (if any)    |
      |    [iRow][iCol] is the first element of the co-subclass (if any)    |
      |                                                                     |
      | Force a cleavage (break to avoid continuation string) if:           |
      |   - recognition mode changes (column is 1); or,                     |
      |   - a B sequence is found: break on the first B                     |
      +--------------------------------------------------------------------*/
      while (*pchStubEnd) {
         ++iCol;
         c = *++pchStubEnd;
         if (c == '\0') {                     // delimiter found
            check_duplicate_delimiter(iRowMin, iRowHigh, iCol);
            iRow = iRowMin+1;
         }else {
            iRow = iRowMin;
            while (
               (++iRow < iRowHigh) &&
               (c==aDlmdscTable[iRow].pszDelim[iCol]) && (
                  (c != syntaxNew.chBSEQ) || (isShortref) ||
                  (aDlmdscTable[iRow].token < SR_first)
               )
            );
         }
         if (
            (iRow < iRowHigh) ||              // 1st sub-class found
            (
               (!isInBseq) && (isShortref) && (c == syntaxNew.chBSEQ)
            ) ||                              // B-seq cleavage
            (iCol==1)                         // Mode cleavage
         ) {
            break;
         }
      }
      /*-----------------------------------------------+
      | Install the prefix found in the secondary node |
      +-----------------------------------------------*/
      if (pchStubEnd > pchStubBeg) {          // not a primary node:
         if (isInBseq) {
            pTr->isBseq = isBseqFound = TRUE;
            pTr->bseqlen = 1;
            while (
               (*++pchStubBeg == syntaxNew.chBSEQ) && (pchStubBeg < pchStubEnd)
            ) {
               ++pTr->bseqlen;
            }
            ++iCount;                         // need an extra VectorItem
         }else {
            pTr->isBseq = FALSE;
            pTr->chPrp = *pchStubBeg++;       // set preponderant char
         }
         pTr->info = pchStubBeg;
         pTr->len = diffptr(pchStubEnd, pchStubBeg);
      }
      /*-------------------------------------------------+
      | If a sub-subclass exists, process it recursively |
      | Else: set up the token                           |
      +-------------------------------------------------*/
      if (
         (iRowMin+1 < iRowHigh) || (iCol == 1) ||
         ((isShortref) && (c == syntaxNew.chBSEQ))
      ) {
         pTr->token = NOT_FOUND_TK;
         get_stub(iCol, iRowMin, iRow, iRowHigh, pTr, pchStubEnd);
      }else {
         pTr->token = aDlmdscTable[iRowMin].token;
      }
      /*---------------------------------------------------------+
      | Check if we can loop for the next co-subclass at iColMin |
      | if so, set the boundaries in [iRowMin .. row-hi]         |
      +---------------------------------------------------------*/
      if (iRowHigh >= iRowMax) break;
      iRowMin = iRowHigh;
      pchStubBeg = aDlmdscTable[iRowMin].pszDelim + iColMin;
      c = *pchStubBeg;
      while (
         (++iRowHigh < iRowMax) &&
         (c == aDlmdscTable[iRowHigh].pszDelim[iColMin])
      );
      /*---------------------+
      |Get a new brother node|
      +---------------------*/
      {
         DelimTree *pTrNew = (DelimTree *)malloc(sizeof(DelimTree)); // brother
         pTrNew->llink = pTr;                 // set elder thread
         pTrNew->rlink = pTr->rlink;          // inherits father thread
         pTrNew->rtag = pTrNew->ltag = FALSE; // just threads
         pTr->rtag = TRUE;                    // Brand new elder
         pTr->rlink = pTrNew;                 // link elder to his brother
         pTr = pTrNew;                        // activate brother
         ++iCount;
      }
   }
   iMaxVi += 1 + iCount;
   /*-----------------------------------------------------+
   | Set up the Preponderant Characters length in Primary |
   +-----------------------------------------------------*/
   pTr->rlink->llink->len = isBseqFound? iCount - 2 : iCount;
   pTr->rlink->llink->isBseq = isBseqFound;
}

/*---------------------------------------------------check_duplicate_delimiter-+
| Check if two delimiter or delimiter-in-context are differents.               |
| NOTE: If it is a delimiter and delimiter_in-context, we could pursue,        |
| since the latter would prevail.                                              |
+-----------------------------------------------------------------------------*/
void NewDelimiterVector::check_duplicate_delimiter(
   unsigned int iRowMin, unsigned int iRowHigh, int iCol
)
{
   e_LexToken token = aDlmdscTable[iRowMin].token;
   char previous_c = 0;
   unsigned int iRow = iRowMin;

   while (++iRow < iRowHigh) {
      char c = aDlmdscTable[iRow].pszDelim[iCol];
      if (!c) {                        // delimiter is simply duplicated
         errorh(DUPDLM, ERH_T_SEV, aDlmdscTable[iRowMin].pszDelim+1);
      }else {
         switch (token) {
         default:
            return;                    // no need to pursue
         case CRO_TK:
            if (!syntaxNew.is_cref(c)) continue;
            break;
         case STAGO_TK:
         case ETAGO_TK:
         case ERO_TK:
         case PERO_TK:
         case MDO_TK:
            if (!syntaxNew.is_namestrt(c)) continue;
            break;
         }
         if (previous_c != c) {        // avoid to tell it more than once
            previous_c = c;
            errorh(                    // Duplicate delimiter-in-context
               DUPDIC, ERH_E_SEV,      // Warning: YASP can recover
               aDlmdscTable[iRowMin].pszDelim+1, aDlmdscTable[iRow].pszDelim+1
            );
         }
      }
   }
}

/*--------------------------------------------------------------prim_postorder-+
| Process first node of a left branch of the tree, during postorder traversal. |
| ENTRY: First node of a mode.                                                 |
| when RETURN: the branch starting at this node has been traversed, and freed. |
+-----------------------------------------------------------------------------*/
void NewDelimiterVector::prim_postorder(DelimTree *pTr)
{
   pViCurr->h.count = (char)pTr->len;
   pViCurr->h.isBseq = (boolean)pTr->isBseq;
   pViCurr->tk = pTr->token;
   pchPCL = sk.point(
      pViCurr->ofsStr = sk.park(0, pTr->len)
   );
   ++pViCurr;
   secd_postorder(pTr->rlink);
   free(pTr);
}

/*--------------------------------------------------------------secd_postorder-+
| Node visit, and Postorder traversal of the tree.                             |
| The first left node is visited by prim_postorder (predominant char list).    |
| when RETURN: the branch starting at this node has been traversed, and freed. |
+-----------------------------------------------------------------------------*/
void NewDelimiterVector::secd_postorder(DelimTree *pTr)
{
   unsigned int *pIxVi;

   if (!pTr->isBseq) {
      *pchPCL++ = pTr->chPrp;            // update predominant char list
   }
   pViCurr->tk = pTr->token;             // set the token
   pIxVi = &pViCurr->ixVi;               // keep a pointer to left pk index
   if (pTr->len) {
      pViCurr->ofsStr = sk.park(pTr->info, 1+pTr->len);
      (sk.point(pViCurr->ofsStr))[pTr->len] = '\0';
   }else {
      pViCurr->ofsStr = 0;               // point a null string
   }
   if (pTr->isBseq) {
      (++pViCurr)->bseqlen = pTr->bseqlen;
      pViCurr->ixVi = ixViPrevBseq;      // .CCS identicals => .OCS too:
      pViCurr->ofsStr = 0;               // erase LSB (not mandatory)
      if (
         (!pTr->len) &&                  // if there is no continuation
         (pTr->ltag) &&                  // and the primary child node
         (pTr->llink->isBseq)            // has a brother BSEQ
      ) {
         ixViPrevBseq = diffptr(pViCurr, aViTable) - 1;
         pViCurr->must_recur = TRUE;     // turn on recurse flag
      }else {
         pViCurr->must_recur = FALSE;    // turn off recurse flag
         ixViPrevBseq = 0;               // reset ixViPrevBseq
      }
   }
   ++pViCurr;
   if (pTr->rtag) secd_postorder(pTr->rlink);
   if (pTr->ltag) {
      *pIxVi = diffptr(pViCurr, aViTable);
      prim_postorder(pTr->llink);
   }else {
      *pIxVi = 0;                        // leaf: not forward referenced
   }
   free(pTr);                            // free this node
}

/*----------------------------------------------------------------------unload-+
| Write the delimiter vectors to disk                                          |
+-----------------------------------------------------------------------------*/
void NewDelimiterVector::unload(ApplicationHandle ahlAccess)
{
   EntityHandler::write(ahlAccess, dvtbent, sizeof dvtbent);
   EntityHandler::write(ahlAccess, dltab, sizeof dltab);
   EntityHandler::write_record (ahlAccess, aViTable, iMaxVi * sizeof(VectorItem));
}
/******************************************************************************/
