/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// entdcl.c
///  ENTITY declaration
//   Pierre G. Richard 87/08/07 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <new.h>
#include "gendef.h"
#include "prsstate.h"
#include "quantity.h"
#include "dlmstr.h"
#include "sgmldcl.h"
#include "prsmode.h"
#include "rsvname.h"
#include "sort.h"
#include "enthdlr.h"
#include "prodcl.h"
#include "entdcl.h"
#include "notadcl.h"
#include "attldcl.h"

/*-----------------------------------------------------------------------parse-+
| Process ENTITY MD                                                            |
|                                                                              |
| Enter here when ENTITY has been decoded,                                     |
| Returns with curr_char (supposed to ) == MDC                                 |
|                                                                              |
| [101] entity declaration = mdo, "ENTITY",                                    |
|           ps+, entity name, ps+, entity text, ps*, mdc                       |
|                                                                              |
| [102] entity name = general entity name | parameter entity name              |
| [103] general entity name = name | (rni,"DEFAULT")                           |
| [104] parameter entity name = pero, ps+, name                                |
|                                                                              |
| [105] entity text = parameter literal | data text | bracketed text |         |
|                     external entity specification                            |
|                                                                              |
| [106] data text = ("CDATA"|"SDATA"|"PI"), ps+, parameter literal             |
| [107] bracketed text = ("STARTAG"|"ENDTAG"|"MS"|"MD"), ps+, parameter literal|
| [108] external entity specification = external identifier,(ps+, entity type)?|
| [109] entity type = "SUBDOC" |                                               |
|                     ( ("NDATA" | "CDATA" | " SDATA"),                        |
|                        ps+, notation name, data attribute specification?)    |
| [149.2] data attribute specification =                                       |
|                   ps+, dso, attribute specification list, ps*, dsc           |
|                                                                              |
| IMPLEMENTATION NOTE:                                                         |
|   The default entity has an empty name (a name with a length of zero).       |
|   Because of the hash table structure, this ensures the default entity       |
|   will be the first when traversing the hash table.  Because all             |
|   other no-named -> defaulted entities occur after, the algorithm            |
|   will be able to refer to the already solved default entity literal         |
|   or EXID, avoiding duplication of the literal or EXID.                      |
|   This is compatible with paragraph 2 of 10.5.1.2 (capacities are counted    |
|   once towards ENTCHCAP for the default entity only -- not for the           |
|   entities defaulted).                                                       |
|                                                                              |
+-----------------------------------------------------------------------------*/
void EntityDcl::parse()
{
   static const VarntChoiceEntry entk_choices[] = {
      { SDATA_VRN_ix,     SDATA_EFK    },
      { CDATA_VRN_ix,     CDATA_EFK    },
      { PI_VRN_ix,        PI_EFK       },
      { STARTTAG_VRN_ix,  STARTTAG_EFK },
      { ENDTAG_VRN_ix,    ENDTAG_EFK   },
      { MS_VRN_ix,        MS_EFK       },
      { MD_VRN_ix,        MD_EFK       },
      { NOT_FOUND_VRN_ix, SGML_EFK     }  // Assume so
   };
   static const VarntChoiceEntry xtyk_choices[] = {
      { SUBDOC_VRN_ix,    SUBDOC_EFK   },
      { NDATA_VRN_ix,     NDATA_EFK    },
      { CDATA_VRN_ix,     CDATA_EFK    }, // Xternal CDATA
      { SDATA_VRN_ix,     SDATA_EFK    }, // Xternal SDATA
      { NOT_FOUND_VRN_ix, INVALID_EFK  }  // definitively invalid
   };
   EntityDef *pEntNew;
   EntityDef::Flags is;
   char *pszNewName;
   {
      /*-----------------------------+
      | Enter the new entity into    |
      | the appropriate hash table   |
      +-----------------------------*/
      char aszName[NAMELEN_MAX+1];

      pPrsstat->skip_ps();
      if (pPrsstat->get_curr_token() == RNI_TK) {
         pSyntax->match_rsvname_or_die(pEh, DEFAULT_VRN_ix);
         aszName[0] = '\0';         // this is the "name" of the default
      }else {
         if (pPrsstat->get_curr_token() != NOT_FOUND_TK) {
            errorh(NONAME, pEh, ERH_T_SEV);
         }
         const char *pszPeroDelimiter = pSyntax->pDlmstr->value_of(PERO_VDL_ix);
         if (pSyntax->match_string(pEh, pszPeroDelimiter)) { // PERO found
            is.parameter = TRUE;
            pPrsstat->skip_inferable_required_ps();
            if (pPrsstat->get_curr_token() != NOT_FOUND_TK) {
               errorh(NONAME, pEh, ERH_T_SEV);
            }
            pSyntax->read_entity_name(pEh, aszName);
            if ((strlen(aszName) + strlen(pszPeroDelimiter)) > NAMELEN) {
               errorh(BIGNAMEP, pEh, ERH_E_SEV);
            }
         }else {
            pSyntax->read_entity_name(pEh, aszName);
         }
      }

      /*
      | pEntNew->pszName != NULL tells if duplicated
      */
      pszNewName = aszName;
      if (is.parameter) {
         pEntNew = (EntityDef *)expnhashPrmEnt.add_value(
            &pszNewName, sizeof (EntityDef)
         );
      }else {
         pEntNew = (EntityDef *)expnhashGenEnt.add_value(
            &pszNewName, sizeof (EntityDef)
         );
      }
      if (pEntNew->name()) {                        // duplicated:
         errorh(DUPNAMENT, pEh, ERH_I_SEV);         // it's an error
      }
   }

   /*----------------------------------+
   | Find out what kind of entity text |
   |  Search for:                      |
   |  - a LIT/LITA (parm lit),         |
   |  - an entk_choice (data, bracket) |
   |  - SGML_EFK => external entity    |
   +----------------------------------*/
   pPrsstat->skip_ps();                          // may be a lit/lita?
   e_EntityKind kind;
   if (pPrsstat->get_curr_token() != NOT_FOUND_TK) {  // assume so...
      kind = TEXT_EFK;
   }else {                                       // else, keyword?
      kind = (e_EntityKind)pSyntax->match_rsvname_in_list(pEh, entk_choices);
   }

   int iTextLength = 0;
   int iLitLength = 0;
   char *pszLit = 0;
   ExternId *pExid = 0;
   EntityNotation *pNad = 0;

   if (kind != SGML_EFK) {
      /*--------------------------------+
      | Memory Entity:                  |
      | - decode the qualifier (if any),|
      | - get the parameter literal     |
      +--------------------------------*/
      if (kind == TEXT_EFK) {
         is.rcdata = TRUE;
      }else {
         switch (kind) {
         case PI_EFK:
            is.bracketed = TRUE;              // not bracketed for ISO
            break;
         case STARTTAG_EFK:
            is.bracketed = TRUE;
            is.rcdata = TRUE;
            iTextLength = strlen(
               pSyntax->pDlmstr->value_of(STAGO_VDL_ix)
            ) + strlen(
               pSyntax->pDlmstr->value_of(TAGC_VDL_ix)
            );
            break;
         case ENDTAG_EFK:
            is.bracketed = TRUE;
            is.rcdata = TRUE;
            iTextLength = strlen(
               pSyntax->pDlmstr->value_of(ETAGO_VDL_ix)
            ) + strlen(
               pSyntax->pDlmstr->value_of(TAGC_VDL_ix)
            );
            break;
         case MS_EFK:
            is.bracketed = TRUE;
            is.rcdata = TRUE;
            iTextLength = strlen(
               pSyntax->pDlmstr->value_of(MDO_DSO_VDL_ix)
            ) + strlen(
               pSyntax->pDlmstr->value_of(MSC_MDC_VDL_ix)
            );
            break;
         case MD_EFK:
            is.bracketed = TRUE;
            is.rcdata = TRUE;
            iTextLength = strlen(
               pSyntax->pDlmstr->value_of(MDO_VDL_ix)
            ) + strlen(
               pSyntax->pDlmstr->value_of(MDC_VDL_ix)
            );
            break;
         default:  /* case CDATA_EFK SDATA_EFK */
            if (is.parameter) {
               errorh(INVPETKW, pEh, ERH_E_SEV);
               kind = TEXT_EFK;
               is.rcdata = TRUE;
            }
            break;
         }
         pPrsstat->skip_inferable_required_ps();
      }
      iLitLength = pPrsstat->read_parameter_literal(pchLitBuffer, FALSE);
      iTextLength += iLitLength;
      if (iTextLength > LITLEN) errorh(BIGLITBK, pEh, ERH_E_SEV);

      pszLit = (char *)malloc(iLitLength + 1);
      memcpy(pszLit, pchLitBuffer, iLitLength + 1);
      pPrsstat->skip_ps();
   }else {                                       // kind == SGML
      is.external = TRUE;
      is.rcdata = TRUE;                          // RCDATA ON (default)

      /*
      | External entity: get external identifier.
      */
      pExid = new ExternId(pPrsstat, pszNewName); // no class check
      if (pPrsstat->get_curr_token() != NOT_FOUND_TK) {    // assume MDC found
         pExid->assign_class(pEh, TEXT_XFC, pSgmldcl->features.isFormal);
      }else {
         kind = (e_EntityKind)pSyntax->match_rsvname_in_list(pEh, xtyk_choices);
         switch (kind) {
         case INVALID_EFK:
            errorh(INVKWD, pEh, ERH_E_SEV);
            kind = SGML_EFK;
            break;
         case SUBDOC_EFK:
            if ((is.parameter) || (!pSgmldcl->features.isSubdoc)) {
               errorh(INVKWD, pEh, ERH_E_SEV);
               kind = SGML_EFK;
            }else {
               pExid->assign_class(
                  pEh, SUBDOC_XFC, pSgmldcl->features.isFormal
               );
               is.rcdata = FALSE;
            }
            pPrsstat->skip_ps();
            break;
         default:                             // NDATA, CDATA, SDATA
            if (is.parameter) {               // if parameter entity
               errorh(INVPETKW, pEh, ERH_E_SEV);
               kind = SGML_EFK;
            }else {
               pExid->assign_class(
                  pEh, NONSGML_XFC, pSgmldcl->features.isFormal
               );
            }
            pPrsstat->skip_ps();
            {
               char pszName[NAMELEN_MAX+1];   // Get the notation name
               pSyntax->read_general_name(pEh, pszName);
               pNad = new(pszName) EntityNotation;
            }
            pPrsstat->skip_ps();
            if (
               (pEntNew->name()) ||           // if entity duplicated
               (is.parameter)                 // or parameter entity
            ) {
               if (pPrsstat->get_curr_token() == DSO_TK) {
                  parse_data_att_spec(0);     // throw any existing ASL
                  pPrsstat->skip_ps();
               }                              // else assume MDC found
               delete pNad;                   // discard nad
               pNad = 0;
               break;                         // and leave
            }
            is.notation = TRUE;
            is.rcdata = FALSE;
            ++nbrof.nad;

            if (pPrsstat->get_curr_token() == DSO_TK) {
               /*-------------------------------------------------+
               | Valid entity, and ASL given:                     |
               | If an Attribute Definition exists, parse the ASL:|
               |   - put the new Attribute list in                |
               |     pNad->pAttSpeclist extension                 |
               | Else, get rid of ASL, issue an Warning           |
               +-------------------------------------------------*/
               NotationBlock *pNb;

               pNb = (NotationBlock *)expnhashNotation.search_value(
                  pNad->name()
               );
               if ((!pNb) || (!(pNb->pAnode))) {
                  errorh(NOATTDEF, pEh, ERH_E_SEV);
               }
               pNad->set_attspec(parse_data_att_spec(pNb->pAnode));
               pPrsstat->skip_ps();
            }
         }
      }
      iTextLength = pExid->system_id_length() + pExid->public_id_length();
   }

   /*------------------------+
   | Construct the EntityDef |
   +------------------------*/
   if (!pEntNew->name()) {                       // brand new
      if (is.external) {                         // external entities
         ++pPrsstat->iCountOfExid;
         nbrof.exid_text += iTextLength;
      }else {                                    // for non external entities:
         if (is.parameter) {                     // parameter entity
            nbrof.p_text += iTextLength;
         }else {                                 // general entity
            nbrof.g_text += iTextLength;
         }
      }
      new (pEntNew) EntityDef(                   // keep it
         pszNewName, is, kind, pszLit, iLitLength, pExid, pNad
      );
   }else {                                       // duplicate:
      if (is.external) {
         delete pExid;
      }else {
         free(pszLit);
      }
   }
}

/*---------------------------------------------------------parse_data_att_spec-+
| Parse a data attribute specification  (N/C/S DATA external entity),          |
|                                                                              |
| On ENTRY:                                                                    |
|  - curr_token == DSO,                                                        |
|  - pAnode is pointing the attribute node or is zero if we must get rid of    |
|    the ASL                                                                   |
|                                                                              |
| This is almost equivalent to `finish_parse_tag', except we are working with  |
| an "unstringized" Attribute list: the Attribute List need to be "de_offset"  |
| before to call `att_spec_list'.  A copy of the new attlist is done, if       |
| at least one attribute is specified.                                         |
|                                                                              |
|  NDATA, ps+, notation name,  ps+,                                            |
|        (ps+, dso, attribute specification list, ps*, dsc)?                   |
|                                                                              |
+-----------------------------------------------------------------------------*/
AttributeList * EntityDcl::parse_data_att_spec(AttributeNode *pAnode)
{
   AttributeList * pAttlist = 0;

   pPrsmode->push_mode(ASL_MODE);         // Att Spec List starting with DSO
   pPrsstat->skip_S();
   if (pPrsstat->get_curr_token() == DSC_TK) {
      errorh(INVEMPASL, pEh, ERH_E_SEV);  // 11.4.1.2 last para
   }
   if (!pAnode) {                         // Invalid: get rid of the ASL
      unsigned int uiNormedLenVal;
      pPrsstat->att_spec_list(0, DATA_TGF, uiNormedLenVal);
   }else {
      /*--------------------------------------------------+
      | Regain access to the address of parked chunks     |
      | Parse the Specification List into a copy of       |
      | the addressable current Attribute Definition List |
      +--------------------------------------------------*/
      unsigned int *aOfsTab;
      unsigned int uiNormedLenVal;
      boolean isSpecifMustBeKept = FALSE;

      pAttlist = new AttributeList(sk, pAnode->attlist, aOfsTab);
      pPrsstat->att_spec_list(pAttlist, DATA_TGF, uiNormedLenVal);
      nbrof.dclv_text += uiNormedLenVal;

      /*-----------------------------------------------------+
      | Check that non-specified attributes were OK to be so |
      | Update the number of components.                     |
      +-----------------------------------------------------*/
      {
         boolean isAttMustBeSpecified;
         if ((pSgmldcl->features.isShorttag)||(pSgmldcl->features.isOmittag)) {
            isAttMustBeSpecified = FALSE;
         }else {
            isAttMustBeSpecified = TRUE;
         }
         Attribute *pAtt;
         for (int ixAtt=0; (pAtt = (*pAttlist)[ixAtt]) != 0; ++ixAtt) {
            if (!pAtt->is_specified()) {
               switch(pAtt->default_value()) {
               case REQUIRED_AFV:
                  errorh(NORQDATT, pEh, ERH_E_SEV, pAtt->name());
                  break;
               case IMPLIED_AFV:
                  break;                            // IMPLIED needs no specif
               default:
                  if (isAttMustBeSpecified) {
                     errorh(NOATTSPC, pEh, ERH_E_SEV, pAtt->name());
                  }
               }
            }else {
               isSpecifMustBeKept = TRUE;
               nbrof.components += pAtt->component_array_size();
            }
         }
      }
      /*------------------------------------------------+
      | Reset the parked offset, copy the new spec list |
      +------------------------------------------------*/
      pAnode->attlist.re_offset(aOfsTab);
      if (isSpecifMustBeKept) {                  // changes done
         nbrof.attributes += pAttlist->count();
         pAttlist->dislocate(sk, pAnode->attlist);
      }else {
         delete pAttlist;
         pAttlist = 0;
      }
   }
   pPrsmode->pop_mode();
   if (pPrsstat->get_curr_token() != DSC_TK) errorh(NODSC, pEh, ERH_E_SEV);
   return pAttlist;
}
/******************************************************************************/
