/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// enthdlr.c
///  EntityDef handler - layer above the EntityDef Manager
//   Pierre G. Richard 87/04/06 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
/*
| Manages SGML entities.
|
| A ring-buffer is used to allow quick unread; when a new entity is opened
| in the nest, it's impossible to unread the previous one before this point.
| This does not affect the usability of the code, due to the Ee rules..
|
*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include "gendef.h"
#include "enthdlr.h"
#include "externid.h"
#include "entity.h"
#include "notation.h"
#include "sgmldcl.h"
#include "prsmode.h"
#include "quantity.h"
#include "dlmstr.h"
#include "syntax.h"
#include "errorh.h"

/*-----------------------------------------------------------------Constructor-+
|                                                                              |
+-----------------------------------------------------------------------------*/
EntityHandler::EntityHandler(
   const Document& doc, ParseMode *pPrsmodeArg
) : Server(doc)
{
   iBufPos = 0;                       // Current position in circular buffer
   iBufHead = 0;                      // Head of circular buffer
   iEntLvl = 0;                       // entity level (1==PRIMARY)
   pYbCurr = 0;                       // Current EntityDef Control Block
   pPrsmode = pPrsmodeArg;            // Shadow from ParseState member
}

/*---------------------------------------------------------------shadow_syntax-+
| Copy a few pointers in private area when those have been changed             |
+-----------------------------------------------------------------------------*/
void EntityHandler::shadow_syntax(Syntax *pSyntaxArg)
{
   pSyntax = pSyntaxArg;              // Shadow from SGMLdcl member
   pDlmstr = pSyntax->pDlmstr;        // Shadow
   pQuantity = pSyntax->pQuantity;    // Shadow
}

/*PRIVATE--------------------------------------------------------diff_head_pos-+
| Compute the number of characters between the head and the current position   |
| in the circular buffer.                                                      |
+-----------------------------------------------------------------------------*/
inline int EntityHandler::diff_head_pos()
{
   if (iBufHead < iBufPos) {
      return BUFSIZE - (iBufPos - iBufHead) + 1;
   }else {
      return (iBufHead - iBufPos) + 1;
   }
}

/*------------------------------------------------------------------------open-+
| Open an entity.                                                              |
|                                                                              |
| First, check if the reference is allowed.                                    |
| If so and the entity is parseable (hence "openable"), open it.               |
|                                                                              |
| Return:                                                                      |
|    TRUE: the entity reference was allowed (even if not openable);            |
|    FALSE: the entity reference was not allowed, or, if openable, not opened  |
|                                                                              |
| I.e., if `check_allowed_reference' returns:                                  |
|   - INVALID_ERC : the entity is not valid.                                   |
|         Don't open the entity (of course).                                   |
|         Return FALSE.                                                        |
|   - OPENABLE_ERC : Entity is parseable ("openable")                          |
|         Open the entity.                                                     |
|         If successful, return TRUE.  Otherwise FALSE.                        |
|                                                                              |
|   - UNPARSEABLE_ERC:  Entity is not openable, but perfectly valid            |
|         Don't open the entity -- caller will certainly tell the Application  |
|         an entity was found, after analyzis of the entity kind.              |
|         Return TRUE.                                                         |
|                                                                              |
| Note: To N/S/C DATA external entities, a NOTATION is always associated.      |
| - In LITC_MODE the notation is not handled because the entity will be opened |
|   and expanded by the parser.  Parser is not able to handle the notation,    |
|   and (when the AVS is declared in ODT) the notation is not even solved.     |
|   These entitites will be read like data-char (see 9.3.1 in ISO8879).        |
|                                                                              |
| - In CON_MODE the entity is passed asis to the application via entity        |
|   control blocks, pointing a notation control block.                         |
+-----------------------------------------------------------------------------*/
boolean EntityHandler::open(EntityDef *pEnt)
{
   static char dflt_pszApplInfo[] = "** No application info **";
   ApplicationHandle ahlAccess;              // access handle
   EntityControlBlock *pYb;

   switch (pEnt->check_ref_validity(this, pPrsmode->get_mode())) {
   case EntityDef::INVALID_ERC:        //  not openable, because it's invalid
      return FALSE;
   case EntityDef::UNPARSEABLE_ERC:    //  not openable, but perfectly valid
      return TRUE;
   default: /* OPENABLE_ERC */         //  openable
      break;
   }
   if (pEnt->is_external()) {
      /*--------------------------------+
      | RC  6:  File is already open    |
      | RC  8:  Bad handle              |
      | RC > 100: CMS return code + 100 |
      +--------------------------------*/
      switch (sr_xenop(pEnt->external_id()->poem_id_handle(), ahlAccess)) {
      case 0:
         break;
      default:
         if (iEntLvl) {         // NOT for PRIMARY entity
            errorh(BADOPEN, this, ERH_E_SEV, pEnt->name());
         }else {
            errorh(BADOPPRIM, ERH_T_SEV);
         }
         return FALSE;
      }
   }

   /*------------------------------------------------+
   | Here the entity needs to be expanded: it can be |
   | a non parseable (general) entity for AVS.       |
   | If it is an external entity, it has been opened.|
   | Prepare the entity stack for the new entity.    |
   +------------------------------------------------*/
   if (iEntLvl) {      // NOT for PRIMARY entity
      if (iEntLvl > pQuantity->value_of(ENTLVL_QTY_ix)) {
         errorh(BIGENTLVL, this, ERH_E_SEV);
         return FALSE;
      }
      /*
      |Do a single copy if no unreads in circular buffer,
      |or save all items to unread in a temporary mallocated area
      */
      if (iBufHead == iBufPos) {     // just current item is to save
         pYbCurr->iForeItemCount = 1;
         pYbCurr->oneFore.ch = achBuffer[iBufPos];
         pYbCurr->oneFore.koc = akocBuffer[iBufPos];
      }else {
         pYbCurr->iForeItemCount = diff_head_pos();
         pYbCurr->manyFore.ach = (char *)malloc(
            pYbCurr->iForeItemCount * (
               sizeof pYbCurr->manyFore.ach[0] + sizeof pYbCurr->manyFore.akoc[0]
            )
         );
         pYbCurr->manyFore.akoc = pYbCurr->manyFore.ach + pYbCurr->iForeItemCount;
         if (iBufHead > iBufPos) {
            memcpy(
               pYbCurr->manyFore.ach, achBuffer+iBufPos,
               pYbCurr->iForeItemCount * sizeof pYbCurr->manyFore.ach[0]
            );
            memcpy(
               pYbCurr->manyFore.akoc, akocBuffer+iBufPos,
               pYbCurr->iForeItemCount * sizeof pYbCurr->manyFore.akoc[0]
            );
         }else {
            memcpy(
               pYbCurr->manyFore.ach, achBuffer+iBufPos,
               (BUFSIZE-iBufPos) * sizeof pYbCurr->manyFore.ach[0]
            );
            memcpy(
               pYbCurr->manyFore.ach+BUFSIZE-iBufPos, achBuffer,
               (iBufHead + 1) * sizeof pYbCurr->manyFore.ach[0]
            );
            memcpy(
               pYbCurr->manyFore.akoc, akocBuffer+iBufPos,
               (BUFSIZE-iBufPos) * sizeof pYbCurr->manyFore.akoc[0]
            );
            memcpy(
               pYbCurr->manyFore.akoc+BUFSIZE-iBufPos, akocBuffer,
               (iBufHead + 1) * sizeof pYbCurr->manyFore.akoc[0]
            );
         }
      }
   }
   pYb = (EntityControlBlock *)calloc(      // Create a new YB
      1, sizeof(EntityControlBlock)
   );
   pYb->pYbPrev = pYbCurr;                  // link it to the previous one
   pYbCurr = pYb;                           // new is current
   pYb->pEnt = pEnt;
   /*----------------------------------------+
   | Insert an Ee character at the beginning |
   | to avoid unreading before the entity    |
   +----------------------------------------*/
   iBufPos = 0;
   achBuffer[0] = EE_VALUE;             // placeholder: curr_char_kind is to use
   akocBuffer[0] = EE_KOC;
   ++iEntLvl;
   if (pEnt->is_external()) {
      iBufHead =  0;                    // Forces I/O
      pYb->ahlAccess = ahlAccess;
      pYb->lOffset = -1;
      pYb->pszApplInfo = dflt_pszApplInfo;
   }else {                              // It's a memory entity: not opened,
      iBufHead =  1;                    // the entity will be unread
      if (pEnt->is_bracketed()) {       // Bracketed
         const char *psz1, *psz2;       // for bracketed text inclusion
         unsigned int i1, i2;

         switch (pEnt->text_type()) {
      /* case PI_EFK:  has been handled above */
         case STARTTAG_EFK:
            psz1 = pDlmstr->value_of(STAGO_VDL_ix);
            psz2 = pDlmstr->value_of(TAGC_VDL_ix);
            break;
         case ENDTAG_EFK:
            psz1 = pDlmstr->value_of(ETAGO_VDL_ix);
            psz2 = pDlmstr->value_of(TAGC_VDL_ix);
            break;
         case MS_EFK:
            psz1 = pDlmstr->value_of(MDO_DSO_VDL_ix);
            psz2 = pDlmstr->value_of(MSC_MDC_VDL_ix);
            break;
         default:  // case MD_EFK:
            psz1 = pDlmstr->value_of(MDO_VDL_ix);
            psz2 = pDlmstr->value_of(MDC_VDL_ix);
            break;
         }
         i1 = strlen(psz1);
         i2 = strlen(psz2);
         memset(akocBuffer+iBufHead, 0, i1 + i2 + pEnt->value_length());
         memcpy(achBuffer+iBufHead, psz1, i1);  // first part
         iBufHead += i1;
         memcpy(achBuffer+iBufHead, pEnt->value(), pEnt->value_length());
         iBufHead += pEnt->value_length();
         memcpy(achBuffer+iBufHead, psz2, i2);  // second part
         iBufHead += i2;
         pYb->lOffset = iBufHead - i1;          // do not count 1st part
         pYb->lVdiff = 0 - i1;
      }else {
         memset(akocBuffer+iBufHead, 0, pEnt->value_length());
         memcpy(achBuffer+iBufHead, pEnt->value(), pEnt->value_length());
         iBufHead += pEnt->value_length();
         pYb->lOffset = iBufHead;
      }
      achBuffer[iBufHead] = EE_VALUE;
      akocBuffer[iBufHead] = EE_KOC;
   }
   read();                                   // get first real char
   return TRUE;
}

/*----------------------------------------------------------------------unread-+
| Unread i consecutive chars from the circular buffer.                         |
|                                                                              |
| To improve the efficiency, no validity check is done: the caller is          |
| advised to keep a record of the number of chunks he is reading, and          |
| to unread only those, and not to unread more than BUFSIZE, and not to read   |
| past the Ee signal...                                                        |
+-----------------------------------------------------------------------------*/
int EntityHandler::unread(unsigned int i)
{
   iBufPos = (iBufPos-i) & (BUFSIZE-1);
   return achBuffer[iBufPos];
}

/*PRIVATE-------------------------------------------------------store_new_char-+
| Store a new character in the circular buffer                                 |
+-----------------------------------------------------------------------------*/
inline void EntityHandler::store_new_char(int iNewChar, e_CharKind koc)
{
   iBufPos = (iBufPos + 1) & (BUFSIZE-1);
   iBufHead = iBufPos;
   achBuffer[iBufPos] = (char)iNewChar;
   akocBuffer[iBufPos] = koc;
}

/*PRIVATE---------------------------------------------------retrieve_next_char-+
| Retrieve an existing character from the circular buffer                      |
+-----------------------------------------------------------------------------*/
inline int EntityHandler::retrieve_next_char()
{
   iBufPos = (iBufPos + 1) & (BUFSIZE-1);
   return achBuffer[iBufPos];
}

/*------------------------------------------------------------------------read-+
| Read next chunk.                                                             |
|                                                                              |
| The last BUFSIZE chunks read are stored in a buffer exactly as they appear   |
| at the time they are fetched (either from the outside world, via a call      |
| to the client, or from the internal entity substitution algorithm).          |
| This allows a quick global unread.                                           |
+-----------------------------------------------------------------------------*/
int EntityHandler::read()
{
   int iNewChar;

   if (iBufPos < iBufHead) {               // still chars to re-read
      return retrieve_next_char();
   }
   // no more chunks to unread: look into the file buffer
   if (pYbCurr->iBufMax == -1) {           // iBufMax == -1 means EOF
      return achBuffer[iBufPos];           // no more reads!
   }
   ++pYbCurr->lOffset;
   if (++pYbCurr->iBufPos <= pYbCurr->iBufMax) {
      /*
      | if file buffer not empty, get next char from file buffer
      */
      iNewChar = pYbCurr->achBuffer[pYbCurr->iBufPos - 1];
   }else {
      /* Buffer empty: do real I/O
      |
      |  sr_xenrd returns:
      |  0:    Transparent IO
      |  1:    Record: insert RE/RS
      |  2:    only RS to insert
      |  3:    only RE to insert
      |  4:    End-of-File
      |  6:    File is not open
      |  8:    Bad handle
      |  >100: OS return code + 100
      */
      char *achBuffer;
      int iLength;

      switch (sr_xenrd(pYbCurr->ahlAccess, &achBuffer, &iLength)) {
      case 0:                              // Transparent IO
         pYbCurr->achBuffer = achBuffer;
         pYbCurr->iBufMax = iLength;
         pYbCurr->iBufPos = 1;
         iNewChar = achBuffer[0];
         break;
      case 1:                              // Record, RS/RE to insert
         achBuffer[iLength] = (char)pSyntax->get_REchar();
         pYbCurr->achBuffer = achBuffer;
         pYbCurr->iBufMax = iLength + 1;
         pYbCurr->iBufPos = 0;               // stay at first character
         iNewChar = (char)pSyntax->get_RSchar();
         break;
      case 2:                              // RS to insert
         pYbCurr->achBuffer = achBuffer;
         pYbCurr->iBufMax = iLength;
         pYbCurr->iBufPos = 0;             // stay at first character
         iNewChar = (char)pSyntax->get_RSchar();
         break;
      case 3:                              // RE to insert
         achBuffer[iLength] = (char)pSyntax->get_REchar();
         pYbCurr->achBuffer = achBuffer;
         pYbCurr->iBufMax = iLength + 1;
         pYbCurr->iBufPos = 1;
         iNewChar = achBuffer[0];
         break;
      case 4:                              // cannot read (EOF reached)
         pYbCurr->iBufMax = -1;            // prevent futher reads!
         store_new_char(EE_VALUE, EE_KOC); // Set Ee signal
         return EE_VALUE;
      default:                             // IO error
         if (iEntLvl > 1) {                // no name for the primary entity
            errorh(BADREAD, ERH_T_SEV, pYbCurr->pEnt->name());
         }else {
            errorh(BADRDPRIM, ERH_T_SEV);
         }
         break;
      }
   }
   if (pSyntax->is_RSchar(iNewChar)) {
      store_new_char(iNewChar, RS_KOC);
      ++pYbCurr->iRow;
   }else {
      store_new_char(iNewChar, REGULAR_KOC);
   }
   if (pSyntax->is_nonsgml(iNewChar)) {
      /*
      | Because errorh() reenters read(), the non-sgml
      | character is put in the buffer, before to call errorh
      | This insures this code will run only once
      */
      errorh(INVCHAR, this, ERH_E_SEV, iNewChar);
   }
   return iNewChar;
}

/*-----------------------------------------------------------------------close-+
| Close an entity - return FALSE if troubles..                                 |
+-----------------------------------------------------------------------------*/
boolean EntityHandler::close()
{
   boolean rc = TRUE;

   if (!iEntLvl) {                               // no more to close!
      errorh(INVEOF, this, ERH_T_SEV);           // can return from this
   }else {
      EntityDef *pEnt = pYbCurr->pEnt;     // save current entity,
      ApplicationHandle ahlAccess = pYbCurr->ahlAccess; // current access handle,
      EntityControlBlock *pYbPrev = pYbCurr->pYbPrev;   // pointer to previous
      --iEntLvl;

      if (iEntLvl) {                             // Reset prev entity state
         if ((iBufHead = pYbPrev->iForeItemCount) > 1) {
            memcpy(
               achBuffer+1, pYbPrev->manyFore.ach,
               pYbPrev->iForeItemCount * sizeof pYbPrev->manyFore.ach[0]
            );
            memcpy(
               akocBuffer+1, pYbPrev->manyFore.akoc,
               pYbPrev->iForeItemCount * sizeof pYbPrev->manyFore.akoc[0]
            );
            free(pYbPrev->manyFore.ach);
         }else {
            achBuffer[1] = pYbPrev->oneFore.ch;
            akocBuffer[1] = pYbPrev->oneFore.koc;
         }
         iBufPos = 1;
         /*----------------------------------------+
         | Insert an Ee character at the beginning |
         | to avoid unreading before the entity    |
         +----------------------------------------*/
         achBuffer[0] = EE_VALUE;
         akocBuffer[0] = EE_KOC;
      }
      free(pYbCurr);

      pYbCurr = pYbPrev;                             // Pretend it's closed
      if (pEnt->is_external()) {
         switch (sr_xencl(ahlAccess)) {
         /*-----------------------------+
         |  6:  File is not open        |
         |  8:  Bad handle              |
         +-----------------------------*/
         case 0:
            break;
         default:
            if (iEntLvl > 1) {           // no name for the primary entity
               errorh(BADCLPRIM, ERH_T_SEV);
            }else {
               errorh(BADCLOSE, ERH_T_SEV, pEnt->name());
            }
            rc = FALSE;
            break;
         }
      }
   }
   return rc;
}

/*---------------------------------------------------------------------get_pos-+
| Get the position of the current external entity                              |
+-----------------------------------------------------------------------------*/
void * EntityHandler::get_pos()
{
   XPOSBLK *pXposblk;
   void *pSysPosBlock;
   int  iSizeSysPosBlock;

   switch (
      sr_xengp(pYbCurr->ahlAccess, &pSysPosBlock, &iSizeSysPosBlock)
   ) {
   case 0:
      pXposblk = (XPOSBLK *)malloc(sizeof(XPOSBLK) + iSizeSysPosBlock);
      pXposblk->yb = *pYbCurr;   // Save current YB,
      pXposblk->yb.oneFore.ch = achBuffer[iBufPos];
      pXposblk->yb.oneFore.koc = akocBuffer[iBufPos];
      memcpy(                  // and the chunk returned by sr_xengp
         (char *)pXposblk + sizeof(XPOSBLK), pSysPosBlock, iSizeSysPosBlock
      );
      break;
   default:
      pXposblk = 0;
      if (iEntLvl > 1) {             // no name for the primary entity */
         errorh(BADREAD, ERH_T_SEV, pYbCurr->pEnt->name());
      }else {
         errorh(BADRDPRIM, ERH_T_SEV);
      }
      break;                         // this won't be reached: error T
   }
   return (void *)pXposblk;
}

/*---------------------------------------------------------------------set_pos-+
| Reset the position of the current external entity                            |
| Warning: no check done (current entity), and no unread are allowed!          |
+-----------------------------------------------------------------------------*/
void EntityHandler::set_pos(void *pXposblk)
{
   void *pSysPosBlock = (char *)pXposblk + sizeof(XPOSBLK);

   switch (sr_xensp(pYbCurr->ahlAccess, pSysPosBlock)) {
   case 0:
      pYbCurr->iBufMax = 0;
      iBufPos = iBufHead = 0;    // To force I/O
      /*----------------------------------------+
      | Insert an Ee character at the beginning |
      | to avoid unreading before the entity    |
      +----------------------------------------*/
      achBuffer[0] = EE_VALUE;
      akocBuffer[0] = EE_KOC;
      read();
      *pYbCurr = ((XPOSBLK *)pXposblk)->yb;  // restore EntityControlBlock
      achBuffer[iBufPos] = pYbCurr->oneFore.ch;
      akocBuffer[iBufPos] = pYbCurr->oneFore.koc;
      break;
   default:
      if (iEntLvl > 1) {                   // no name for the primary entity
         errorh(BADREAD, ERH_T_SEV, pYbCurr->pEnt->name());
      }else {
         errorh(BADRDPRIM, ERH_T_SEV);
      }
      break;                               // this wont be reached: error T
   }
}

/*---------------------------------------------------------------------replace-+
| Replace (in the read BUFFER) a string of character that is "count" bytes     |
| long by "new_char" (character reference handling).                           |
|                                                                              |
| On ENTRY:                                                                    |
|   The character currently read (curr_char) is the one that                   |
|   follows immediately the string to be replaced.                             |
| When RETURN:                                                                 |
|   The string has been replaced, and curr_char is new_char (followed          |
|   by the character that was currently read).                                 |
|                                                                              |
| Ex:                                                                          |
|                                                                              |
|    curr_char = 'A'                               curr_char = '{'             |
|                           replace('{',5)                                     |
|  |   X|&|1|2|3|;|A|                                |   X|}|A|                |
|  +-..-+-+-+-+-+-+-+         =====>                 +-..-+-+-+                |
|                                                                            |
|                  pos                                     pos                 |
|                                                                              |
|  Count must normally be > 0;  as a feature, we implemented:                  |
|                                                                              |
|  Count == 0 is used to replace the current_char by new_char:                 |
|                                                                              |
|    curr_char = 'A'                               curr_char = '{'             |
|                           replace('{',0)                                     |
|  |   X|&|1|2|3|;|A|                              |   X|&|1|2|3|;|{|          |
|  +-..-+-+-+-+-+-+-+         =====>               +-..-+-+-+-+-+-+-+          |
|                                                                            |
|                  pos                                             pos         |
+-----------------------------------------------------------------------------*/
void EntityHandler::replace(
   char new_char, unsigned int count, e_CharKind koc
)
{
   int iPosSource = iBufPos;           // remember where we are

   iBufPos = (iBufPos-count) & (BUFSIZE-1);
   achBuffer[iBufPos] = new_char;
   akocBuffer[iBufPos] = koc;
   if (count) {
      int iPosTarget = iBufPos;

      pYbCurr->lVdiff += (count-1);
      for (;;) {
         iPosTarget = (iPosTarget+1) & (BUFSIZE-1);
         achBuffer[iPosTarget] = achBuffer[iPosSource];
         akocBuffer[iPosTarget] = akocBuffer[iPosSource];
         if (iPosSource == iBufHead) break;
         iPosSource = (iPosSource+1) & (BUFSIZE-1);
      }
      iBufHead = iPosTarget;
   }
}

/*----------------------------------------------------------------------locate-+
| Get the identification/location of the nth nested entity.                    |
| Note: Yb.row always reflects the row of iBufHead.                            |
|       Some computation is required to obtain the row at iBufPos.             |
|       This is done here -- instead of doing it in the unread routine --      |
|       because locate is called only for messages.  It would be a waste of    |
|       time to compute the row at iBufPos in unread.                          |
+-----------------------------------------------------------------------------*/
void EntityHandler::locate(
   int  iRequestedEntLevel,
   char **ppszName,
   char **ppszApplInfo,
   int *piRow,
   long *plOffset
)
{
   int iFore;
   char *akoc;
   EntityControlBlock *pYb = pYbCurr;

   if ((iRequestedEntLevel -= iEntLvl) < 0) {
      do pYb = pYb->pYbPrev; while (++iRequestedEntLevel < 0);
      iFore = pYb->iForeItemCount;
      akoc = pYb->manyFore.akoc;    // Invalid if iFore == 1
   }else {
      iFore = diff_head_pos();
      akoc = akocBuffer;
   }
   *plOffset = pYb->lOffset - iFore;
   if (pYb->pEnt->is_external()) {     // compute row
      *piRow = pYb->iRow;
      while (--iFore) {
         if (akoc[iFore + 1] == RS_KOC) --(*piRow);
      }
      *ppszApplInfo = pYb->pszApplInfo;
   }else {
      *piRow = 0;
      *ppszApplInfo = 0;
   }
   // This is a NULL pointer if the entity is an internal literal
   *ppszName = (char *)pYb->pEnt->name();
}

/*----------------------------------------------------------------------offset-+
| Return the offset within current entity of curr_char                         |
| This is the offset as "seen" by the user, without taking into account        |
| replacement (char references) or insertion (bracketed text).                 |
+-----------------------------------------------------------------------------*/
long EntityHandler::offset()
{
   return pYbCurr->lOffset - diff_head_pos();
}

/*--------------------------------------------------------------virtual_offset-+
| Return the virtual offset within current entity of curr_char                 |
| The virtual offset takes into account replacement and insertion              |
| done after the entity has been read.  It's main purpose is to unread         |
| properly in an edited buffer.                                                |
+-----------------------------------------------------------------------------*/
long EntityHandler::virtual_offset()
{
   return pYbCurr->lOffset - pYbCurr->lVdiff - diff_head_pos();
}

//***>Note: This reflects the entity manager is also used for Utility Files.
//***>Note: The implementation is not finished. I plan to use more of the
//***>Note: standard em_functions.  No time today.

/*------------------------------------------------------------------------open-+
|Open a utility file                                                           |
|                                                                              |
|where:                                                                        |
|  exid.asgn cls is one of:                                                    |
|     UTSYNL_XFC, UTSYNR_XFC, UTSYNU_XFC,                                      |
|     UTODTR_XFC, UTODTW_XFC, UTOLTRD, UTOLTW                                  |
|  exid.name is defined only for UTO*_XFC (and is the document/link name)      |
|  for_write is TRUE, if the file is open in write mode                        |
|            is FALSE, if the file is open in read mode                        |
|  pAhlAccess is the address where to put the Access-AHL of the opened file    |
|                                                                              |
|Notes:                                                                        |
| UTSYNL_XFC (Local OCS) and UTSYNR_XFC (Reference OCS) cannot be open         |
| in write mode.  UTSYNU_XFC (User-owned OCS) is always written first.         |
+-----------------------------------------------------------------------------*/
boolean EntityHandler::open(
   ExternId *pExid,
   ApplicationHandle *pahlAccess,    // output (returned)
   boolean isWriteMode
)
{
   if (!pExid->find(this)) return FALSE;

   switch (sr_utopen(pExid->poem_id_handle(), pahlAccess, isWriteMode)) {
   case 0:
      return TRUE;
   default:
      return FALSE;
   }
}

/*------------------------------------------------------------------------read-+
| Read an Utility file                                                         |
+-----------------------------------------------------------------------------*/
boolean EntityHandler::read(
   ApplicationHandle ahlAccess,
   void *pchBuffer,
   int iLength
)
{
   switch (sr_utrd(ahlAccess, pchBuffer, iLength)) {
   case 0:
      return TRUE;
   default:
      errorh(BADREADUT, ERH_T_SEV);
      return FALSE;
   }
}

/*-----------------------------------------------------------------read_record-+
| Read a variable length record into a mallocated buffer                       |
+-----------------------------------------------------------------------------*/
int EntityHandler::read_record(
   ApplicationHandle ahlAccess, void **ppchBuffer
)
{
   unsigned int iLength;

   read(ahlAccess, &iLength, sizeof(iLength));
   if (iLength) {
      *ppchBuffer = malloc(iLength);
      read(ahlAccess, *ppchBuffer, iLength);
      return iLength;
   }else {
      *ppchBuffer = 0;
      return 0;
   }
}

/*-----------------------------------------------------------------------write-+
| Write an Utility file                                                        |
+-----------------------------------------------------------------------------*/
boolean EntityHandler::write(
   ApplicationHandle ahlAccess,
   const void *pchBuffer,
   unsigned int iLength
)
{
   switch (sr_utwrt(ahlAccess, pchBuffer, iLength)) {
   case 0:
      return TRUE;
   default:
      errorh(BADWRITUT, ERH_T_SEV);
      return FALSE;
   }
}

/*----------------------------------------------------------------write_record-+
| Write a variable length record from a mallocated buffer                      |
+-----------------------------------------------------------------------------*/
boolean EntityHandler::write_record(
   ApplicationHandle ahlAccess, const void *pchBuffer, unsigned int iLength
)
{
   write(ahlAccess, &iLength, sizeof iLength);
   if (iLength) write(ahlAccess, pchBuffer, iLength);
   return TRUE;
}

/*-----------------------------------------------------------------------close-+
| Close an Utility file                                                        |
+-----------------------------------------------------------------------------*/
boolean EntityHandler::close(
   ApplicationHandle ahlAccess
)
{
   switch (sr_utcl(ahlAccess)) {
   case 0:
      return TRUE;
   default:
      return FALSE;
   }
}
/******************************************************************************/
