/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// event.c
///  Report (stack) primary SGML events
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
/*
|  Primary and Decomposed events
*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include "gendef.h"
#include "event.h"
#include "element.h"
#include "entity.h"
#include "notation.h"
#include "prsstate.h"
#include "quantity.h"
#include "syntax.h"
#include "dtddcl.h"
#include "enthdlr.h"

static void display_tag_notif(
   const boolean isStartTag,
   const ElementType *pElem,
   const e_InferBy inferby,
   const e_Minimize minias
);
static void decompose_entity(EventList &evlst, EntityDef *pEnt);

/*------------------------------------------------------------------operator[]-+
| Random access to an indexed event.  ix <= 0, 0: last event                   |
+-----------------------------------------------------------------------------*/
Event * EventList::operator[](int ix)
{
   if (ix > 0) return 0;
   Event *pEv = pEvLast;

   while ((pEv) && (ix++ < 0)) {
      pEv = pEv->previous;
   }
   return pEv;
}

/*-----------------------------------------------------------------------first-+
| Set the event list to the 1st event                                          |
+-----------------------------------------------------------------------------*/
void EventList::first()
{
   if (pEvFirst) {
      pEvPrev = 0;
      pEvCurrent = pEvFirst;
      pEvNext = pEvCurrent->next;
   }
}

/*------------------------------------------------------------------------last-+
| Set the event list to the last event                                         |
+-----------------------------------------------------------------------------*/
void EventList::last()
{
   if (pEvLast) {
      pEvNext = 0;
      pEvCurrent = pEvLast;
      pEvPrev = pEvCurrent->previous;
   }
}

/*------------------------------------------------------------------------prev-+
| Get the previous event                                                       |
+-----------------------------------------------------------------------------*/
void EventList::prev()
{
   if (pEvPrev) {
      pEvCurrent = pEvPrev;
      pEvNext = pEvCurrent->next;
      pEvPrev = pEvCurrent->previous;
   }else if (pEvCurrent) {
      pEvNext = pEvCurrent;
      pEvCurrent = 0;
   }
}

/*------------------------------------------------------------------------next-+
| Get the next event                                                           |
| Never rely on the value of pEvCurrent on entry: it can have been removed.    |
+-----------------------------------------------------------------------------*/
void EventList::next()
{
   if (pEvNext) {
      pEvCurrent = pEvNext;
      pEvPrev = pEvCurrent->previous;
      pEvNext = pEvCurrent->next;
   }else if (pEvCurrent) {
      pEvPrev = pEvCurrent;
      pEvCurrent = 0;
   }
}

/*----------------------------------------------------------------------append-+
| Append an event                                                              |
+-----------------------------------------------------------------------------*/
void EventList::append(const Event& eventNew)
{
   Event *pEv = &(eventNew.make());
   pEv->previous = pEvLast;
   pEv->next = 0;
   if (pEvLast) {
      pEvLast->next = pEv;
      if (pEvCurrent == pEvLast) pEvNext = pEv;
   }
   pEvLast = pEv;
   if (!iCount) pEvFirst = pEv;
   ++iCount;
}

/*----------------------------------------------------------------------remove-+
| Remove the current event from the list.                                      |
+-----------------------------------------------------------------------------*/
boolean EventList::remove()
{
   if (!pEvCurrent) return FALSE;
   if (!pEvPrev) {
      pEvFirst = pEvNext;
      if (pEvFirst) pEvFirst->previous = 0;
   }else {
      pEvPrev->next = pEvNext;
   }
   if (!pEvNext) {
      pEvLast = pEvPrev;
      if (pEvLast) pEvLast->next = 0;
   }else {
      pEvNext->previous = pEvPrev;
   }
   --iCount;
   delete pEvCurrent;
   pEvCurrent = 0;
   return TRUE;
}

/*--------------------------------------------------------------------position-+
| Set current stack position                                                   |
| The position 0 is the first, the position (count() - 1) is the last.         |
+-----------------------------------------------------------------------------*/
void EventList::position(int iCountUp)
{
   if (!iCount) return;

   int iCountDown = iCount - iCountUp;
   if (iCountDown < iCountUp) {
      if (iCountDown <= 0) {
         pEvPrev = pEvLast;
         pEvCurrent = 0;
         pEvNext = 0;
      }else {
         pEvCurrent = pEvLast;
         while (--iCountDown > 0) {
            pEvCurrent = pEvCurrent->previous;
         }
         pEvNext = pEvCurrent->next;
         pEvPrev = pEvCurrent->previous;
      }
   }else {
      if (iCountUp < 0) {
         pEvNext = pEvFirst;
         pEvCurrent = 0;
         pEvPrev = 0;
      }else {
         pEvCurrent = pEvFirst;
         while (--iCountUp >= 0) {
            pEvCurrent = pEvCurrent->next;
         }
         pEvNext = pEvCurrent->next;
         pEvPrev = pEvCurrent->previous;
      }
   }
}

/*******************************************************************************
*                                                                              *
*                            StartElement                                      *
*                                                                              *
*******************************************************************************/

/*----------------------------------------------------------------StartElement-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
StartElement::StartElement(
   const ElementType *pElemArg,
   const AttributeList& attlistArg,
   const e_InferBy inferbyArg,
   const e_Minimize miniasArg,
   boolean isInclusionArg,
   boolean isExplConrefArg
) :
   PrimEvent(ELEMENT_START_EV), attlist(attlistArg)
{
   pElem = pElemArg;
   inferby = inferbyArg;
   minias = miniasArg;
   isInclusion = isInclusionArg;
   isExplConref = isExplConrefArg;
}

/*----------------------------------------------------------------StartElement-+
| Copy constructor                                                             |
+-----------------------------------------------------------------------------*/
StartElement::StartElement(
   const StartElement& selemSource
) :
   PrimEvent(ELEMENT_START_EV), attlist(selemSource.attlist)
{
   pElem = selemSource.pElem;
   inferby = selemSource.inferby;
   minias = selemSource.minias;
   isInclusion = selemSource.isInclusion;
   isExplConref = selemSource.isExplConref;
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& StartElement::make() const
{
   return * new StartElement(*this);
}

/*--------------------------------------------------------------------validate-+
| Validate this event, given the current parser state                          |
+-----------------------------------------------------------------------------*/
void StartElement::validate(ParseState *pPrsstat) const
{
   pPrsstat->validate_start_element(
       pElem, attlist, is_inferred(), is_explicit_conref()
   );
}

/*--------------------------------------------------------------init_decompose-+
| Initialize the decomposition of this primary event                           |
+-----------------------------------------------------------------------------*/
void StartElement::init_decompose()
{
   if (attlist.count()) {
      evlst.append(AttName(ATT_NAME_EV, attlist));
      evlst.first();
   }
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void StartElement::display(ParseState *pPrsstat) const
{
   display_tag_notif(TRUE, pElem, inferby, minias);
   attlist.display(pPrsstat, 10);
}

/*STATIC-----------------------------------------------------display_tag_notif-+
|Display a line indicating a tag notification was received                     |
+-----------------------------------------------------------------------------*/
static void display_tag_notif(
   const boolean isStartTag,
   const ElementType *pElem,
   const e_InferBy inferby,
   const e_Minimize minias
)
{
   char *pszStartOrEnd, *pszMinimi, *pszInfer;

   if (isStartTag) {
      pszStartOrEnd = "Start";
   }else {
      pszStartOrEnd = "End";
   }
   switch (minias) {
   case NET_ST_MNF:    pszMinimi = "NET-enabling Start-tag";  break;
   case EMPTY_ST_MNF:  pszMinimi = "Empty Start-tag";         break;
   case UNCL_ST_MNF:   pszMinimi = "Unclosed Start-tag";      break;
   case NULL_ET_MNF:   pszMinimi = "NULL End-tag";            break;
   case EMPTY_ET_MNF:  pszMinimi = "Empty End-tag";           break;
   case UNCL_ET_MNF:   pszMinimi = "Unclosed End-tag";        break;
   case NO_MNF:        pszMinimi = NULL;                      break;
   case DATA_ET_MNF:   pszMinimi = "Data End-tag";            break;
   default:            pszMinimi = "????";                    break;
   }
   switch (inferby) {
   case NO_IMF:
      pszInfer = NULL;
      break;
   case DOC_IMF:
      pszInfer = "start of document instance set";
      break;
   case DATA_IMF:
      pszInfer = "data";
      break;
   case STG_IMF:
      pszInfer = "start-tag";
      break;
   case ETG_IMF:
      pszInfer = "end-tag";
      break;
   case EDTG_IMF:
      pszInfer = "end-datatag: \"";
      break;
   case EOF_IMF:
      pszInfer = "end of entity";
      break;
   default:
      pszInfer = "????";
      break;
   }
   fprintf(stderr, "=> %s %s ", pszStartOrEnd, pElem->name());
   if (!pszInfer) {
      if (!pszMinimi) {
         fprintf(stderr, "\n");
      }else {
         fprintf(stderr, "(%s)\n", pszMinimi);
      }
   }else {
      if (!pszMinimi) {
         fprintf(stderr, "(implied by %s)\n", pszInfer);
      }else {
         fprintf(stderr, "(%s, implied by %s)\n", pszMinimi, pszInfer);
      }
   }
}

/*******************************************************************************
*                                                                              *
*                             EndElement                                       *
*                                                                              *
*******************************************************************************/

/*------------------------------------------------------------------EndElement-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
EndElement::EndElement(
   const ElementType *pElemArg,
   const e_InferBy inferbyArg,
   const e_Minimize miniasArg,
   boolean isPrematureArg,
   boolean isBadEmptyArg
) : PrimEvent(ELEMENT_END_EV)
{
   pElem = pElemArg;
   inferby = inferbyArg;
   minias = miniasArg;
   isPremature = isPrematureArg;
   isBadEmpty = isBadEmptyArg;
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& EndElement::make() const
{
   return * new EndElement(pElem, inferby, minias, isPremature, isBadEmpty);
}

/*--------------------------------------------------------------------validate-+
| Validate this event, given the current parser state                          |
+-----------------------------------------------------------------------------*/
void EndElement::validate(ParseState *pPrsstat) const
{
   pPrsstat->validate_end_element(
       pElem, is_inferred(), is_premature(), is_bad_empty()
   );
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void EndElement::display(ParseState *) const
{
   display_tag_notif(FALSE, pElem, inferby, minias);
}

/*******************************************************************************
*                                                                              *
*                                Data                                          *
*                                                                              *
*******************************************************************************/

/*------------------------------------------------------------------------Data-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
Data::Data() : PrimEvent(DATA_EV)
{
   pRep = (Rep *)malloc(sizeof(Rep));
   pRep->iLength = 0;
   pRep->iMaxLength = 0;
   pRep->iReferenceCount = 1;
}

/*------------------------------------------------------------------------Data-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
Data::Data(int iSize) : PrimEvent(DATA_EV)
{
   pRep = (Rep *)malloc(sizeof(Rep) + iSize);
   pRep->iLength = 0;
   pRep->iMaxLength = iSize;
   pRep->iReferenceCount = 1;
}

/*------------------------------------------------------------------------Data-+
| Copy constructor                                                             |
+-----------------------------------------------------------------------------*/
Data::Data(const Data& dataSource) : PrimEvent(DATA_EV)
{
   pRep = dataSource.pRep;
   ++pRep->iReferenceCount;
}

/*-----------------------------------------------------------------------~Data-+
| Destructor                                                                   |
+-----------------------------------------------------------------------------*/
Data::~Data()
{
   if (--pRep->iReferenceCount <= 0) free(pRep);
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& Data::make() const
{
    return * new Data(*this);
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void Data::display(ParseState *) const
{
   fprintf(stderr, "=> Data |%s|\n", zstring());
}

/*--------------------------------------------------------------------put_char-+
| Enter more characters                                                        |
| Return:                                                                      |
|  OK_RCPC if the character was sucessfully entered                            |
|  FULL_RCPC if the character entered filled the buffer up                     |
|  OVFLW_RCPC if the character was not entered because the buffer is now full  |
+-----------------------------------------------------------------------------*/
Data::e_RcPutChar Data::put_char(char c)
{
   if (pRep->iLength >= pRep->iMaxLength) {
      return OVFLW_RCPC;
   }
   pRep->data[pRep->iLength++] = c;
   if (pRep->iLength == pRep->iMaxLength) {
      end();
      return FULL_RCPC;
   }
   return OK_RCPC;
}

/*-------------------------------------------------------------------------end-+
| End the data buffer                                                          |
+-----------------------------------------------------------------------------*/
void Data::end()
{
   if ((pRep->iMaxLength > pRep->iLength) && (pRep->iReferenceCount == 1)) {
      pRep = (Rep *)realloc(pRep, sizeof(Rep) + pRep->iLength);
      pRep->iMaxLength = pRep->iLength;
   }
   pRep->data[pRep->iLength] = '\0';
}


/*******************************************************************************
*                                                                              *
*                         ProcessingInstruction                                *
*                                                                              *
*******************************************************************************/

/*-------------------------------------------------------ProcessingInstruction-+
| Default constructor                                                          |
+-----------------------------------------------------------------------------*/
ProcessingInstruction::ProcessingInstruction() : PrimEvent(PI_EV)
{
   pRep = (Rep *)malloc(sizeof(Rep));
   pRep->iLength = 0;
   pRep->iReferenceCount = 1;
}

/*-------------------------------------------------------ProcessingInstruction-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
ProcessingInstruction::ProcessingInstruction(
   EntityHandler *pEh,
   Syntax *pSyntax
) : PrimEvent(PI_EV)
{
   pRep = (Rep *)malloc(
      sizeof(Rep) + pSyntax->pQuantity->value_of(PILEN_QTY_ix)
   );
   pRep->iLength = pSyntax->read_to_end_delimiter(pEh, pRep->data);
   pRep = (Rep *)realloc(pRep, sizeof(Rep) + pRep->iLength);
   pRep->data[pRep->iLength] = '\0';
   pRep->iReferenceCount = 1;
}

/*-------------------------------------------------------ProcessingInstruction-+
| Copy constructor                                                             |
+-----------------------------------------------------------------------------*/
ProcessingInstruction::ProcessingInstruction(
   const ProcessingInstruction& procinstSource
) : PrimEvent(PI_EV)
{
   pRep = procinstSource.pRep;
   ++pRep->iReferenceCount;
}

/*------------------------------------------------------~ProcessingInstruction-+
| Destructor                                                                   |
+-----------------------------------------------------------------------------*/
ProcessingInstruction::~ProcessingInstruction()
{
   if (--pRep->iReferenceCount <= 0) free(pRep);
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& ProcessingInstruction::make() const
{
    return * new ProcessingInstruction(*this);
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void ProcessingInstruction::display(ParseState *) const
{
   fprintf(stderr, "=> PI |%s|\n", pRep->data);
}

/*******************************************************************************
*                                                                              *
*                               Comment                                        *
*                                                                              *
*******************************************************************************/

/*---------------------------------------------------------------------Comment-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
Comment::Comment() : PrimEvent(COMMENT_EV)
{
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& Comment::make() const
{
   return * new Comment();
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void Comment::display(ParseState *) const
{
   fprintf(stderr, "=> Comment declaration\n");
}

/*******************************************************************************
*                                                                              *
*                             RecordEnd                                        *
*                                                                              *
*******************************************************************************/

/*-------------------------------------------------------------------RecordEnd-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
RecordEnd::RecordEnd() : PrimEvent(RE_EV)
{
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& RecordEnd::make() const
{
   return * new RecordEnd();
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void RecordEnd::display(ParseState *) const
{
   fprintf(stderr, "=> Record End\n");
}

/*******************************************************************************
*                                                                              *
*                             EntityStart                                      *
*                                                                              *
*******************************************************************************/

/*-----------------------------------------------------------------EntityStart-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
EntityStart::EntityStart(
   EntityDef *pEntArg
) : PrimEvent(ENTITY_START_EV), pEnt(pEntArg)
{
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& EntityStart::make() const
{
   return * new EntityStart(pEnt);
}

/*STATIC------------------------------------------------------decompose_entity-+
| Decompose an entity into elementary events                                   |
+-----------------------------------------------------------------------------*/
static void decompose_entity(EventList &evlst, EntityDef *pEnt)
{
   switch (pEnt->text_type()) {
   case INVALID_EFK:
      break;
   case MD_EFK:
      evlst.append(DcmpLambda(ENT_INT_MD_EV, pEnt->value()));
      break;
   case PI_EFK:
      evlst.append(DcmpLambda(ENT_INT_PI_EV, pEnt->value()));
      break;
   case STARTTAG_EFK:
      evlst.append(DcmpLambda(ENT_INT_STARTTAG_EV, pEnt->value()));
      break;
   case ENDTAG_EFK:
      evlst.append(DcmpLambda(ENT_INT_ENDTAG_EV, pEnt->value()));
      break;
   case MS_EFK:
      evlst.append(DcmpLambda(ENT_INT_MS_EV, pEnt->value()));
      break;
   case TEXT_EFK:
      evlst.append(DcmpLambda(ENT_INT_SGML_EV, pEnt->value()));
      break;
   case CDATA_EFK:
      if (pEnt->is_internal()) {
         evlst.append(DcmpLambda(ENT_INT_CDATA_EV, pEnt->value()));
      }else {
         evlst.append(EntExt(ENT_EXT_CDATA_EV, pEnt));
      }
      break;
   case SDATA_EFK:
      if (pEnt->is_internal()) {
         evlst.append(DcmpLambda(ENT_INT_SDATA_EV, pEnt->value()));
      }else {
         evlst.append(EntExt(ENT_EXT_SDATA_EV, pEnt));
      }
      break;
   case NDATA_EFK:
      evlst.append(EntExt(ENT_EXT_NDATA_EV, pEnt));
      break;
   case SGML_EFK:
      evlst.append(EntExt(ENT_EXT_SGML_EV, pEnt));
      break;
   case SUBDOC_EFK:
      evlst.append(EntExt(ENT_EXT_SUBDOC_EV, pEnt));
      break;
   }
   evlst.first();
}

/*--------------------------------------------------------------init_decompose-+
| Initialize the decomposition of this primary event                           |
+-----------------------------------------------------------------------------*/
void EntityStart::init_decompose()
{
   decompose_entity(evlst, pEnt);
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void EntityStart::display(ParseState *pPrsstat) const
{
   fprintf(stderr, "-- Entity %s started\n", pEnt->name());
   pEnt->display(pPrsstat, 10);
}

/*******************************************************************************
*                                                                              *
*                              EntityEnd                                       *
*                                                                              *
*******************************************************************************/

/*-------------------------------------------------------------------EntityEnd-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
EntityEnd::EntityEnd() : PrimEvent(ENTITY_END_EV)
{
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& EntityEnd::make() const
{
   return * new EntityEnd();
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void EntityEnd::display(ParseState *) const
{
   fprintf(stderr, "-- Entity ended\n");
}

/*******************************************************************************
*                                                                              *
*                             SectionStart                                     *
*                                                                              *
*******************************************************************************/

/*----------------------------------------------------------------SectionStart-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
SectionStart::SectionStart(e_StatusSection ssArg) : PrimEvent(SECTION_START_EV)
{
   ss = ssArg;
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& SectionStart::make() const
{
   return * new SectionStart(ss);
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void SectionStart::display(ParseState *) const
{
   char *pchStatus;

   switch (ss) {
      case CDATA:          pchStatus = "CDATA";    break;
      case RCDATA:         pchStatus = "RCDATA";   break;
      case INCLUDE:        pchStatus = "INCLUDE";  break;
      default: /*IGNORE*/  pchStatus = "IGNORE";   break;
   }
   fprintf(stderr, "=> Marked section: %s started\n", pchStatus);
}

/*******************************************************************************
*                                                                              *
*                              SectionEnd                                      *
*                                                                              *
*******************************************************************************/

/*------------------------------------------------------------------SectionEnd-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
SectionEnd::SectionEnd() : PrimEvent(SECTION_END_EV)
{
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& SectionEnd::make() const
{
   return * new SectionEnd();
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void SectionEnd::display(ParseState *) const
{
   fprintf(stderr, "=> Marked section ended\n");
}

/*******************************************************************************
*                                                                              *
*                                Error                                         *
*                                                                              *
*******************************************************************************/

/*-----------------------------------------------------------------------Error-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
Error::Error() : PrimEvent(ERROR_EV)
{
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& Error::make() const
{
   return * new Error();
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void Error::display(ParseState *) const
{
   fprintf(stderr, "=> Error found; data skipped...\n");
}

/*******************************************************************************
*                                                                              *
*                             SgmlStart                                        *
*                                                                              *
*******************************************************************************/

/*-------------------------------------------------------------------SgmlStart-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
SgmlStart::SgmlStart() : PrimEvent(SGML_START_EV)
{
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& SgmlStart::make() const
{
   return * new SgmlStart();
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void SgmlStart::display(ParseState *) const
{
   fprintf(stderr, "=> Sgml declaration started\n");
}

/*******************************************************************************
*                                                                              *
*                              SgmlEnd                                         *
*                                                                              *
*******************************************************************************/

/*---------------------------------------------------------------------SgmlEnd-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
SgmlEnd::SgmlEnd() : PrimEvent(SGML_END_EV)
{
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& SgmlEnd::make() const
{
   return * new SgmlEnd();
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void SgmlEnd::display(ParseState *) const
{
   fprintf(stderr, "=> Sgml declaration ended\n");
}

/*******************************************************************************
*                                                                              *
*                           DoctypeStart                                       *
*                                                                              *
*******************************************************************************/

/*----------------------------------------------------------------DoctypeStart-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
DoctypeStart::DoctypeStart() : PrimEvent(DOCTYPE_START_EV)
{
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& DoctypeStart::make() const
{
   return * new DoctypeStart();
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void DoctypeStart::display(ParseState *) const
{
   fprintf(stderr, "=> Doctype declaration started\n");
}


/*******************************************************************************
*                                                                              *
*                             DoctypeEnd                                       *
*                                                                              *
*******************************************************************************/

/*------------------------------------------------------------------DoctypeEnd-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
DoctypeEnd::DoctypeEnd() : PrimEvent(DOCTYPE_END_EV)
{
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& DoctypeEnd::make() const
{
   return * new DoctypeEnd();
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void DoctypeEnd::display(ParseState *) const
{
   fprintf(stderr, "=> Doctype declaration ended\n");
}

/*******************************************************************************
*                                                                              *
*                           LinktypeStart                                       *
*                                                                              *
*******************************************************************************/

/*---------------------------------------------------------------LinktypeStart-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
LinktypeStart::LinktypeStart() : PrimEvent(LINKTYPE_START_EV)
{
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& LinktypeStart::make() const
{
   return * new LinktypeStart();
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void LinktypeStart::display(ParseState *) const
{
   fprintf(stderr, "=> Linktype declaration started\n");
}


/*******************************************************************************
*                                                                              *
*                             LinktypeEnd                                      *
*                                                                              *
*******************************************************************************/

/*-----------------------------------------------------------------LinktypeEnd-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
LinktypeEnd::LinktypeEnd() : PrimEvent(DOCTYPE_END_EV)
{
}

/*------------------------------------------------------------------------make-+
| Copy this to a new object of the same class                                  |
+-----------------------------------------------------------------------------*/
Event& LinktypeEnd::make() const
{
   return * new LinktypeEnd();
}

/*---------------------------------------------------------------------display-+
| Display the contents of this object                                          |
+-----------------------------------------------------------------------------*/
void LinktypeEnd::display(ParseState *) const
{
   fprintf(stderr, "=> Linktype declaration ended\n");
}

/*******************************************************************************
*                                                                              *
*                           DecomposedEvent                                    *
*                                                                              *
*******************************************************************************/

/*--------------------------------------------------DecomposedEvent::decompose-+
| Decompose a primary event - by default, end the decomposition chain          |
+-----------------------------------------------------------------------------*/
e_Event DecomposedEvent::decompose(
   ParseState *, EventList& evlstDecomposed
) {
   e_Event idReturned;

   if (++iRepeatCount <= iRepeatMax) {
      evlstDecomposed.append(TerminalDecomposedEvent());
      idReturned = id;
   }else {
      evlstDecomposed.remove();
      idReturned = EOP_EV;
   }
   evlstDecomposed.last();
   return idReturned;
}

/*-------------------------------------------------------------------decompose-+
| Decompose AttName in elementary events                                       |
+-----------------------------------------------------------------------------*/
e_Event AttName::decompose(ParseState *pPrsstat, EventList& evlst)
{
   e_Event idReturned;

   if (++iRepeatCount <= iRepeatMax) {
      if ((iRepeatCount == 1) && (pPrsstat->pEh->is_sort_attributes())) {
         attlist.sort();
      }
      Attribute *pAtt = attlist[iRepeatCount-1];
      const char *psz = pAtt->value();

      if (pAtt->is_implied()) {
         evlst.append(DcmpLambda(ATT_VAL_IMPLIED_EV, psz));
      }else {
         switch (pAtt->declared_value()) {
         case NOTATION_ADV:
            evlst.append(AttValNotation(psz));
            break;
         case CDATA_ADV:
            if (pAtt->is_composite()) {
//***>NOTE: NOT DONE: must be composite, or composite_sdata
               evlst.append(
                  AttValComposite(ATT_VAL_COMPOSITE_EV, pAtt->component())
               );
            }else {
               evlst.append(DcmpLambda(ATT_VAL_CDATA_EV, psz));
            }
            break;
         case GROUP_ADV:
            evlst.append(DcmpLambda(ATT_VAL_NMTOK_IN_GROUP_EV, psz));
            break;
         case ID_ADV:
            evlst.append(DcmpLambda(ATT_VAL_ID_EV, psz));
            break;
         case ENTITY_ADV:
            evlst.append(AttValEntity(ATT_VAL_ENTITY_EV, psz));
            break;
         case IDREF_ADV:
            evlst.append(DcmpLambda(ATT_VAL_IDREF_EV, psz));
            break;
         case NAME_ADV:
            evlst.append(DcmpLambda(ATT_VAL_NAME_EV, psz));
            break;
         case NMTOKEN_ADV:
            evlst.append(DcmpLambda(ATT_VAL_NMTOKEN_EV, psz));
            break;
         case NUMBER_ADV:
            evlst.append(DcmpLambda(ATT_VAL_NUMBER_EV, psz));
            break;
         case NUTOKEN_ADV:
            evlst.append(DcmpLambda(ATT_VAL_NUTOKEN_EV, psz));
            break;
         case ENTITIES_ADV:
            {
               AttValTokensArg arg(
                  ATT_VAL_TOKEN_ENTITY_EV,
                  psz, pPrsstat->pSyntax->get_SPACEchar()
               );
               evlst.append(AttValTokens(arg));
            }
            break;
         case IDREFS_ADV:
            {
               AttValTokensArg arg(
                  ATT_VAL_TOKEN_IDREF_EV,
                  psz, pPrsstat->pSyntax->get_SPACEchar()
               );
               evlst.append(AttValTokens(arg));
            }
            break;
         case NAMES_ADV:
            {
               AttValTokensArg arg(
                  ATT_VAL_TOKEN_NAME_EV,
                  psz, pPrsstat->pSyntax->get_SPACEchar()
               );
               evlst.append(AttValTokens(arg));
            }
            break;
         case NMTOKENS_ADV:
            {
               AttValTokensArg arg(
                  ATT_VAL_TOKEN_NMTOKEN_EV,
                  psz, pPrsstat->pSyntax->get_SPACEchar()
               );
               evlst.append(AttValTokens(arg));
            }
            break;
         case NUMBERS_ADV:
            {
               AttValTokensArg arg(
                  ATT_VAL_TOKEN_NUMBER_EV,
                  psz, pPrsstat->pSyntax->get_SPACEchar()
               );
               evlst.append(AttValTokens(arg));
            }
            break;
         case NUTOKENS_ADV:
            {
               AttValTokensArg arg(
                  ATT_VAL_TOKEN_NUTOKEN_EV,
                  psz, pPrsstat->pSyntax->get_SPACEchar()
               );
               evlst.append(AttValTokens(arg));
            }
            break;
         default: /* case INVALID_ADV: */
            evlst.append(DcmpLambda(ATT_VAL_INVALID_EV, psz));
            break;
         }
      }
      idReturned = id;
   }else {
      evlst.remove();
      idReturned = EOP_EV;
   }
   evlst.last();
   return idReturned;
}

/*-------------------------------------------------------------------decompose-+
| Decompose AttValNotation in elementary events                                |
+-----------------------------------------------------------------------------*/
e_Event AttValNotation::decompose(ParseState *pPrsstat, EventList& evlst)
{
   e_Event idReturned;

   switch (++iRepeatCount) {
   case 1:
      pNcb = (NotationType *)pPrsstat->pDtddcl->pHshNotation->search_value(
         pszValue
      );
      evlst.append(ExternIdEv(pNcb->external_id()));
      idReturned = id;
      break;
// Actually we don't want to return the notation attributes
// case 2:
//    pAttlistNotation = new AttributeList(pNcb);
//    evlst.append(AttName(DATA_ATT_NAME_EV, * pAttlistNotation));
//    idReturned = EOP_EV;
//    break;
   default:
//    delete pAttlistNotation;
      evlst.remove();
      idReturned = EOP_EV;
      break;
   }
   evlst.last();
   return idReturned;
}

/*-------------------------------------------------------------------decompose-+
| Decompose AttValEntity in elementary events                                  |
+-----------------------------------------------------------------------------*/
e_Event AttValEntity::decompose(ParseState *pPrsstat, EventList& evlst)
{
   e_Event idReturned;

   if (++iRepeatCount <= iRepeatMax) {
      EntityDef *pEnt = pPrsstat->find_general_entity((char *)pszValue);
      if (pEnt) {
         decompose_entity(evlst, pEnt);
      }else {
         evlst.append(TerminalDecomposedEvent());
      }
      idReturned = id;
   }else {
      evlst.remove();
      idReturned = EOP_EV;
   }
   evlst.last();
   return idReturned;
}

/*-------------------------------------------------------------------decompose-+
| Decompose AttValComposite in elementary events                               |
+-----------------------------------------------------------------------------*/
e_Event AttValComposite::decompose(ParseState *, EventList& evlst)
{
   e_Event idReturned;

   if (++iRepeatCount <= iRepeatMax) {
      const EntityDef *pEnt = aCompo->entity_def(iRepeatCount - 1);
      e_Event idCompo;
      if (!pEnt) {
         idCompo = ATT_VAL_COMP_DATA_EV;
      }else if (pEnt->text_type() == SDATA_EFK) {
         idCompo = ATT_VAL_COMP_SDATA_EV;
      }else {
         idCompo = ATT_VAL_COMP_CDATA_EV;
      }
      evlst.append(
         AttValComponent(
            idCompo, pEnt,
            aCompo->part_value(iRepeatCount - 1),
            aCompo->part_length(iRepeatCount - 1)
         )
      );
      idReturned = (iRepeatCount > 1)? EOP_EV : id;
   }else {
      evlst.remove();
      idReturned = EOP_EV;
   }
   evlst.last();
   return idReturned;
}

/*-------------------------------------------------------------------decompose-+
| Decompose AttValTokens in elementary events                                  |
+-----------------------------------------------------------------------------*/
e_Event AttValTokens::decompose(ParseState *, EventList& evlst)
{
   e_Event idReturned;

   if (++iRepeatCount <= iRepeatMax) {
      if (rep.idTokens == ATT_VAL_TOKEN_ENTITY_EV) {
         evlst.append(AttValEntity(rep.idTokens, rep.sl[iRepeatCount - 1]));
      }else {
         evlst.append(DcmpLambda(rep.idTokens, rep.sl[iRepeatCount - 1]));
      }
      idReturned = (iRepeatCount > 1)? EOP_EV : id;
   }else {
      evlst.remove();
      idReturned = EOP_EV;
   }
   evlst.last();
   return idReturned;
}

/*-------------------------------------------------------------------decompose-+
| Decompose EntNotation in elementary events                                   |
+-----------------------------------------------------------------------------*/
e_Event EntNotation::decompose(ParseState *, EventList& evlst)
{
   e_Event idReturned;

   switch (++iRepeatCount) {
   case 1:
      evlst.append(ExternIdEv(pNad->external_id()));
      idReturned = id;
      break;
   case 2:
      pAttlistNotation = new AttributeList(pNad);
      evlst.append(AttName(DATA_ATT_NAME_EV, * pAttlistNotation));
      idReturned = EOP_EV;
      break;
   default:
      delete pAttlistNotation;
      evlst.remove();
      idReturned = EOP_EV;
      break;
   }
   evlst.last();
   return idReturned;
}


/*-------------------------------------------------------------------decompose-+
| Decompose EntExt in elementary events                                        |
| Chain identifiers_events (and ENT_NOTATION if one of ENT_EXT_xDATA_EV)       |
+-----------------------------------------------------------------------------*/
e_Event EntExt::decompose(ParseState *, EventList& evlst)
{
   e_Event idReturned;

   switch (++iRepeatCount) {
   case 1:
      evlst.append(ExternIdEv(pEnt->external_id()));
      idReturned = id;
      break;
   case 2:
      if (
         (id == ENT_EXT_CDATA_EV) ||
         (id == ENT_EXT_SDATA_EV) ||
         (id == ENT_EXT_NDATA_EV)
      ) {
         evlst.append(EntNotation(pEnt->entity_notation()));
         idReturned = EOP_EV;
         break;
      }
      /* else, fall thru */
   default:
      evlst.remove();
      idReturned = EOP_EV;
      break;
   }
   evlst.last();
   return idReturned;
}

/*-------------------------------------------------------------------decompose-+
| Decompose ExternIdEv in elementary events                                    |
| [4] identifiers_events =                                                     |
|   (PUBLIC_ID, SYSTEM_ID?) | SYSTEM_ID | SYSTEM_ID_OMITTED                    |
+-----------------------------------------------------------------------------*/
e_Event ExternIdEv::decompose(ParseState *, EventList& evlst)
{
   switch (++iRepeatCount) {
   case 1:
      if (pExid->is_public()) {
         evlst.append(DcmpLambda(PUBLIC_ID_EV, pExid->public_id()));
      }else if (pExid->is_system()) {
         evlst.append(DcmpLambda(SYSTEM_ID_EV, pExid->system_id()));
      }else {
         evlst.append(DcmpLambda(SYSTEM_ID_OMITTED_EV, ""));
      }
      break;
   case 2:
      if ((pExid->is_public()) && (pExid->is_system())) {
         evlst.append(DcmpLambda(SYSTEM_ID_EV, pExid->system_id()));
         break;
      }
      /* else, fall thru */
   default:
      evlst.remove();
      break;
   }
   evlst.last();
   return EOP_EV;      // hidden node
}
/******************************************************************************/
