/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// event.h
///  Declarations for event
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _EVENT_H
#define _EVENT_H 1
//******************************************************************************

#include <stdlib.h>
#include <stdio.h>
#include "attrib.h"     // part of StartElement class

#include "element.eh"   // enums
#include "event.eh"
#include "strlst.h"

class Event;
class ElementType;
class Syntax;
class EntityHandler;
class EntityDef;
class SGMLparser;
class ParseState;

/*-----------------------------------------------------------------------------+
| EventList - events seen as a read-only or remove list.                       |
| "smart pointers" are used, making:                                           |
| ((++pEvlst, --pEvlst) == (--pEvlst, ++pEvlst)) always TRUE.                  |
| Also accessing the next or previous pEv still works after an event has been  |
| removed.                                                                     |
+-----------------------------------------------------------------------------*/
class EventList {
public:
   EventList();
   boolean is_no_more();       // Check if at one end of the list
   Event * operator->();       // Access current event
   operator Event*();          // Access current event (conversion)
   Event * operator++();       // Make current next event, access (prefix)
   Event * operator++(int);    // Make current next event, access (postfix)
   Event * operator--();       // Make current previous event, access (prefix)
   Event * operator--(int);    // Make current previous event, access (postfix)
   Event * operator[](int ix); // Random access: ix <= 0, 0: last event
   int  count();               // How many events in the list?
   void first();               // Make current first event
   void last();                // Make current last event
   void position(int);         // Make current event [0 ... count()-1]
   void next();                // Make current next event
   void prev();                // Make current previous event
   void append(const Event& eventNew); // Append a new last event
   boolean remove();           // Remove the current event
private:
   Event * pEvCurrent;
   Event * pEvNext;
   Event * pEvPrev;
   Event * pEvLast;
   Event * pEvFirst;
   int iCount;
};

/*--- INLINES ---*/
inline EventList::EventList() {          // Constructor
   iCount = 0;    // 0: no real event have been enrolled in the list so far.
   pEvCurrent = pEvNext = pEvPrev = pEvFirst = pEvLast = 0;
}
inline boolean EventList::is_no_more() {        // TRUE if there are no more events
   return (!pEvCurrent)? TRUE : FALSE;
}
inline Event * EventList::operator->() {        // Get current event
   return pEvCurrent;
}
inline EventList::operator Event *() {          // Get current event (conversion)
   return pEvCurrent;
}
inline Event * EventList::operator++() {        // Prefix increment
   next();
   return pEvCurrent;
}
inline Event * EventList::operator++(int) {     // Postfix increment
   Event *pEv = pEvCurrent;
   next();
   return pEv;
}
inline Event * EventList::operator--() {        // Prefix decrement
   prev();
   return pEvCurrent;
}
inline Event * EventList::operator--(int) {     // Postfix decrement
   Event * pEv = pEvCurrent;
   prev();
   return pEv;
}
inline int EventList::count() {                 // count of events stacked
   return iCount;
}

/*-----------------------------------------------------------------------------+
| Event abstract class                                                         |
+-----------------------------------------------------------------------------*/
class Event {                          // Stack Entry
   friend EventList;
public:
   const e_Event id;                   // identify the event
   Event(e_Event idArg);               // construct an event
   virtual ~Event();                   // delete an event
   virtual const char * zstring() const; // get the value
private:
   Event *previous;                    // link to previous, 0 if none
   Event *next;                        // link to next, 0 if none
   virtual Event& make() const = 0;    // copy `this' to a new one
};

/*--- INLINES ---*/
inline Event::Event(e_Event idArg) : id(idArg) {
   previous = next = 0;                // brand new: this event is unlinked
}
inline Event::~Event() { /*...*/ }
inline const char * Event::zstring() const { return ""; }

/*-----------------------------------------------------------------------------+
| Primary Event class                                                          |
+-----------------------------------------------------------------------------*/
class PrimEvent : public Event {
public:
   PrimEvent(e_Event idArg);
   boolean isDecomposed;
   EventList evlst;                    // decomposition list
   virtual void display(ParseState *pPrsstate) const = 0;
   virtual void validate(ParseState *pPrsstate) const;  // do a final check
   virtual void init_decompose();
};

inline PrimEvent::PrimEvent(e_Event idArg) : Event(idArg) {
   isDecomposed = FALSE;
}
inline void PrimEvent::init_decompose() { /*...*/ };
inline void PrimEvent::validate(ParseState *) const { /* ... */ }

/*-----------------------------------------------------------------------------+
| StartElement Event                                                           |
+-----------------------------------------------------------------------------*/
class StartElement : public PrimEvent {     // ELEMENT_START event
public:
   StartElement(
      const ElementType *pElem,
      const AttributeList& attlist,
      const e_InferBy inferby,
      const e_Minimize minias,
      boolean isInclusion,
      boolean isExplConref
   );
   const char * zstring() const;            // get the GI of the started-element
   const ElementType *element_type() const; // get the Element Type

   // are TRUE or FALSE
   boolean is_inferred() const;
   boolean is_inclusion() const;
   boolean is_explicit_conref() const;

   void init_decompose();
   void display(ParseState *) const;

private:
   const ElementType *pElem;       // How this element was declared
   AttributeList attlist;          // Attribute List
   e_InferBy inferby;              // If and what inferred this start tag
   e_Minimize minias;              // If and what minimization rule was applied
   boolean isInclusion  : 1;       // Allowed because of an inclusion
   boolean isExplConref : 1;       // Has an explicit content reference (empty)

   StartElement(const StartElement& selemSource); // copy operator
   void validate(ParseState *pPrsstate) const;
   Event& make() const;
};

/*-- INLINES --*/
inline const char * StartElement::zstring() const {
   return pElem->name();
}
inline const ElementType * StartElement::element_type() const {
   return pElem;
}
inline boolean StartElement::is_inferred() const {
   return (inferby != NO_IMF)? TRUE : FALSE;
}
inline boolean StartElement::is_inclusion() const {
   return (isInclusion)? TRUE : FALSE;
}
inline boolean StartElement::is_explicit_conref() const {
   return (isExplConref)? TRUE : FALSE;
}

/*-----------------------------------------------------------------------------+
| EndElement event                                                             |
+-----------------------------------------------------------------------------*/
class EndElement : public PrimEvent {      // ELEMENT_END event
public:
   /*
   |  Constructor
   */
   EndElement(
      const ElementType *pElem,
      const e_InferBy inferby,
      const e_Minimize minias,
      boolean isPremature,
      boolean isBadEmpty
   );

   const char * zstring() const;            // get the GI of the ended-element
   const ElementType *element_type() const; // get the Element Type

   // are TRUE or FALSE
   boolean is_inferred() const;
   boolean is_premature() const;
   boolean is_bad_empty() const;

   void display(ParseState *) const;

private:
   const ElementType *pElem;       // How this element was declared
   e_InferBy inferby;              // If and what inferred this end tag
   e_Minimize minias;              // If and what minimization rule was applied?
   boolean isPremature : 1;        // The element was prematurely ended
   boolean isBadEmpty  : 1;        // The element has an invalid empty content
   void validate(ParseState *pPrsstat) const;
   Event& make() const;
};

/*-- INLINES --*/
inline const char * EndElement::zstring() const {
   return pElem->name();
}
inline const ElementType * EndElement::element_type() const {
   return pElem;
}
inline boolean EndElement::is_premature() const {
   return (isPremature)? TRUE : FALSE;
}
inline boolean EndElement::is_bad_empty() const {
   return (isBadEmpty)? TRUE : FALSE;
}
inline boolean EndElement::is_inferred() const {
   return (inferby != NO_IMF)? TRUE : FALSE;
}

/*-----------------------------------------------------------------------------+
| Data event                                                                   |
+-----------------------------------------------------------------------------*/
class Data : public PrimEvent {
   friend SGMLparser;            // process_data has access to data
public:
   enum e_RcPutChar {            // Return code from put_char
      FULL_RCPC,                 // character entered filled buffer up
      OK_RCPC,                   // character sucessfully entered
      OVFLW_RCPC                 // character not entered because buffer full
   };

   const char * zstring() const; // get the data as an ASCIIZ string
   void display(ParseState *) const;

private:
   struct Rep {
      int iReferenceCount;
      int iMaxLength;
      int iLength;
      char data[1];
   };
   Rep *pRep;

   Data();
   Data(int iSize);
   Data(const Data& dataSource);
   #if (__BCPLUSPLUS__ == 0x320)
      // Borland C++ 4.0 does not understand friendship for destructors!
      public: ~Data(); private:
   #else
      ~Data();
   #endif
   Event& make() const;
   e_RcPutChar put_char(char c);
   void end();
};

/*-- INLINES --*/
inline const char *Data::zstring() const {
   pRep->data[pRep->iLength] = '\0';
   return pRep->data;
}

/*-----------------------------------------------------------------------------+
| Processing Instruction event                                                 |
+-----------------------------------------------------------------------------*/
class ProcessingInstruction : public PrimEvent { // PI event
public:
   ProcessingInstruction(EntityHandler *pEh, Syntax *pSyntax);
   ~ProcessingInstruction();

   const char *zstring() const;  // get the PI as an ASCIIZ string

   void display(ParseState *) const;

private:
   struct Rep {
      int iReferenceCount;
      int iLength;
      char data[1];
   };
   Rep *pRep;
   ProcessingInstruction();
   ProcessingInstruction(const ProcessingInstruction& procinstSource);
   Event& make() const;
};

/*-- INLINES --*/
inline const char * ProcessingInstruction::zstring() const {
   return pRep->data;
}

/*-----------------------------------------------------------------------------+
| Comment event                                                                |
+-----------------------------------------------------------------------------*/
class Comment : public PrimEvent {
public:
   Comment();
   const char *zstring() const; // get the comment as an ASCIIZ string
   void display(ParseState *) const;
private:
   Event& make() const;
};

/*-- INLINES --*/
inline const char * Comment::zstring() const {
   return "NOT YET IMPLEMENTED";
}

/*-----------------------------------------------------------------------------+
| RecordEnd event                                                              |
+-----------------------------------------------------------------------------*/
class RecordEnd : public PrimEvent {
public:
   RecordEnd();
   void display(ParseState *) const;

private:
   Event& make() const;
};

/*-----------------------------------------------------------------------------+
| EntityStart event                                                            |
+-----------------------------------------------------------------------------*/
class EntityStart : public PrimEvent {    // ENTITY_START event
public:
   EntityStart(EntityDef *pEnt);

   const char * zstring() const;          // get the name of the started entity
   const EntityDef * entity_def() const;  // get the EntityDef (from the DTD)

   void init_decompose();
   void display(ParseState *) const;

private:
   EntityDef *pEnt;
   Event& make() const;
};

/*-- INLINES --*/
inline const char * EntityStart::zstring() const {
   const char *pszName = pEnt->name();
   return pszName? pszName : "#PRIMARY";
}
inline const EntityDef * EntityStart::entity_def() const { return pEnt; }

/*-----------------------------------------------------------------------------+
| EntityEnd event                                                              |
+-----------------------------------------------------------------------------*/
class EntityEnd : public PrimEvent {   // ENTITY_END event
public:
   EntityEnd();
   void display(ParseState *) const;

private:
   Event& make() const;
};

/*-----------------------------------------------------------------------------+
| Section Start event                                                          |
+-----------------------------------------------------------------------------*/
class SectionStart : public PrimEvent { // SECTION_START event
public:
   enum e_StatusSection {
      CDATA,
      RCDATA,
      IGNORE,
      INCLUDE
   };
   SectionStart(e_StatusSection);

   //***>NOTE: NOT YET IMPLEMENTED - will return the status variant? name
   const char * zstring() const;
   e_StatusSection get_section_status() const; // status of the section
   void display(ParseState *) const;

private:
   e_StatusSection ss;
   Event& make() const;
};

/*-- INLINES --*/
inline SectionStart::e_StatusSection SectionStart::get_section_status() const {
   return ss;
}
inline const char * SectionStart::zstring() const {
   return "NOT YET IMPLEMENTED";
}

/*-----------------------------------------------------------------------------+
| Section End event                                                            |
+-----------------------------------------------------------------------------*/
class SectionEnd : public PrimEvent {   // SECTION_END event
public:
   SectionEnd();
   void display(ParseState *) const;

private:
   Event& make() const;
};

/*-----------------------------------------------------------------------------+
| Skipped data, because it was found invalid                                   |
+-----------------------------------------------------------------------------*/
class Error : public PrimEvent {        // ERROR event
public:
   Error();
   void display(ParseState *) const;

private:
   Event& make() const;
};

/*-----------------------------------------------------------------------------+
| SgmlStart event                                                              |
+-----------------------------------------------------------------------------*/
class SgmlStart : public PrimEvent {
public:
   SgmlStart();

   void display(ParseState *) const;

private:
   Event& make() const;
};

/*-----------------------------------------------------------------------------+
| SgmlEnd event                                                                |
+-----------------------------------------------------------------------------*/
class SgmlEnd : public PrimEvent {
public:
   SgmlEnd();

   void display(ParseState *) const;

private:
   Event& make() const;
};

/*-----------------------------------------------------------------------------+
| DoctypeStart event                                                           |
+-----------------------------------------------------------------------------*/
class DoctypeStart : public PrimEvent {
public:
   DoctypeStart();

   //***>NOTE: NOT YET IMPLEMENTED - will return the doctype name
   const char * zstring() const;

   void display(ParseState *) const;

private:
   Event& make() const;
};

/*-- INLINES --*/
inline const char * DoctypeStart::zstring() const {
   return "NOT YET IMPLEMENTED";
}

/*-----------------------------------------------------------------------------+
| DoctypeEnd event                                                             |
+-----------------------------------------------------------------------------*/
class DoctypeEnd : public PrimEvent {
public:
   DoctypeEnd();

   void display(ParseState *) const;

private:
   Event& make() const;
};

/*-----------------------------------------------------------------------------+
| LinktypeStart event                                                          |
+-----------------------------------------------------------------------------*/
class LinktypeStart : public PrimEvent {
public:
   LinktypeStart();

   //***>NOTE: NOT YET IMPLEMENTED - will return the linktype name
   const char * zstring() const;

   void display(ParseState *) const;

private:
   Event& make() const;
};

/*-- INLINES --*/
inline const char * LinktypeStart::zstring() const {
   return "NOT YET IMPLEMENTED";
}

/*-----------------------------------------------------------------------------+
| LinktypeEnd event                                                            |
+-----------------------------------------------------------------------------*/
class LinktypeEnd : public PrimEvent {
public:

   /*
   |  Constructor
   */
   LinktypeEnd();

   /*
   |  Debugging...
   */
   void display(ParseState *) const;

private:
   Event& make() const;
};

/*-----------------------------------------------------------------------------+
| Decomposed Event class                                                       |
+-----------------------------------------------------------------------------*/
class DecomposedEvent : public Event {
public:
   const int iRepeatMax;
   int iRepeatCount;

   DecomposedEvent(e_Event idArg, int iRepeatMaxArg);
   virtual e_Event decompose(ParseState *pPrsstat, EventList& evlstDecomposed);
   e_Event chain(EventList& evlst, const DecomposedEvent& ev);
};

inline DecomposedEvent::DecomposedEvent(
   e_Event idArg, int iRepeatMaxArg
) : Event(idArg), iRepeatMax(iRepeatMaxArg), iRepeatCount(0) {
   /*...*/
}

class TerminalDecomposedEvent : public DecomposedEvent {
public:
   TerminalDecomposedEvent() : DecomposedEvent(EOP_EV, 0) {}
   e_Event decompose(ParseState *pPrsstat, EventList& evlstDecomposed);
private:
   Event& make() const;
};

inline e_Event TerminalDecomposedEvent::decompose(
   ParseState *, EventList& evlstDecomposed
) {
   evlstDecomposed.remove();
   evlstDecomposed.last();
   return EOP_EV;
}
inline Event& TerminalDecomposedEvent::make() const {
   return * new TerminalDecomposedEvent();
}
/*-----------------------------------------------------------------------------+
| AttName Event: for ATT_NAME_EV as well as DATA_ATT_NAME_EV                   |
+-----------------------------------------------------------------------------*/
class AttName : public DecomposedEvent {
public:
   AttName(e_Event id, AttributeList &attlist);
   const char * zstring() const;
   e_Event decompose(ParseState *pPrsstat, EventList& evlstDecomposed);

private:
   AttributeList& attlist;
   Event& make() const;
};

/*-- INLINES --*/
inline AttName::AttName(
   e_Event id, AttributeList &attlistArg
) : DecomposedEvent(id, attlistArg.count()), attlist(attlistArg) {
}
inline Event& AttName::make() const {
   return * new AttName(id, attlist);
}
inline const char * AttName::zstring() const {
   Attribute *pAtt = attlist[iRepeatCount-1];
   return pAtt? pAtt->name() : "";
}

/*-----------------------------------------------------------------------------+
| DcmpLambda Event  - Regroup the following values:                            |
|    ATT_VAL_INVALID_EV                                                        |
|    ATT_VAL_IMPLIED_EV                                                        |
|    ATT_VAL_ID                                                                |
|    ATT_VAL_IDREF_EV            ATT_VAL_TOKEN_IDREF_EV                        |
|    ATT_VAL_NAME_EV             ATT_VAL_TOKEN_NAME_EV                         |
|    ATT_VAL_NMTOKEN_EV          ATT_VAL_TOKEN_NMTOKEN_EV                      |
|    ATT_VAL_NUMBER_EV,          ATT_VAL_TOKEN_NUMBER_EV                       |
|    ATT_VAL_NUTOKEN_EV          ATT_VAL_TOKEN_NUTOKEN_EV                      |
|    ATT_VAL_NMTOKEN_IN_GROUP_EV                                               |
|    PUBLIC_ID_EV                                                              |
|    SYSTEM_ID_EV                SYSTEM_ID_OMITTED_EV                          |
|    ENT_INT_MD_EV                                                             |
|    ENT_INT_PI_EV                                                             |
|    ENT_INT_STARTTAG_EV                                                       |
|    ENT_INT_ENDTAG_EV                                                         |
|    ENT_INT_MS_EV                                                             |
|    ENT_INT_SGML_EV                                                           |
|    ENT_INT_CDATA_EV                                                          |
|    ENT_EXT_CDATA_EV                                                          |
|    ENT_INT_SDATA_EV                                                          |
+-----------------------------------------------------------------------------*/
class DcmpLambda : public DecomposedEvent {
public:
   DcmpLambda(e_Event id, const char *pszValueArg, int iRepeatMaxArg = 1);
   const char * zstring() const;

private:
   Event& make() const;
   const char * pszValue;
};

/*-- INLINES --*/
inline DcmpLambda::DcmpLambda(
   e_Event id, const char *pszValueArg, int iRepeatMaxArg
) : DecomposedEvent(id, iRepeatMaxArg) {
   pszValue = pszValueArg;
}
inline const char * DcmpLambda::zstring() const {
   return pszValue;
}
inline Event& DcmpLambda::make() const {
   return * new DcmpLambda(id, pszValue, iRepeatMax);
}

/*-----------------------------------------------------------------------------+
| AttValNotation Event (ATT_VAL_NOTATION)                                      |
+-----------------------------------------------------------------------------*/
class AttValNotation : public DecomposedEvent {
public:
   AttValNotation(const char *pszValueArg, int iRepeatMaxArg = 1);
   const char * zstring() const;
   e_Event decompose(ParseState *pPrsstat, EventList& evlstDecomposed);

private:
   const char * pszValue;
   AttributeList * pAttlistNotation;
   const NotationType *pNcb;
   Event& make() const;
};

/*-- INLINES --*/
inline AttValNotation::AttValNotation(
   const char *pszValueArg, int iRepeatMaxArg
) : DecomposedEvent(ATT_VAL_NOTATION_EV, iRepeatMaxArg) {
   pszValue = pszValueArg;
   pAttlistNotation = 0;
}
inline Event& AttValNotation::make() const {
   return * new AttValNotation(pszValue, iRepeatMax);
}
inline const char * AttValNotation::zstring() const {
   return pszValue;
}

/*-----------------------------------------------------------------------------+
| AttValEntity Event (ATT_VAL_ENTITY, ATT_VAL_TOKEN_ENTITY)                    |
+-----------------------------------------------------------------------------*/
class AttValEntity : public DecomposedEvent {
public:
   AttValEntity(e_Event idArg, const char *pszValueArg, int iRepeatMaxArg = 1);
   const char * zstring() const;
   e_Event decompose(ParseState *pPrsstat, EventList& evlstDecomposed);

private:
   Event& make() const;
   const char *pszValue;
};

/*-- INLINES --*/
inline AttValEntity::AttValEntity(
   e_Event idArg, const char * pszValueArg, int iRepeatMaxArg
) : DecomposedEvent(idArg, iRepeatMaxArg) {
   pszValue = pszValueArg;
}
inline Event& AttValEntity::make() const {
   return * new AttValEntity(id, pszValue, iRepeatMax);
}
inline const char * AttValEntity::zstring() const {
   return pszValue;
}

/*-----------------------------------------------------------------------------+
| AttValComposite Event                                                        |
| - regroup following values:                                                  |
|  ATT_VAL_COMPOSITE_EV,                                                       |
|  ATT_VAL_COMPOSITE_SDATA_EV,                                                 |
| - Chain the following values:                                                |
|  ATT_VAL_COMP_DATA_EV,                                                       |
|  ATT_VAL_COMP_CDATA_EV,                                                      |
|  ATT_VAL_COMP_SDATA_EV                                                       |
+-----------------------------------------------------------------------------*/
class AttValComposite : public DecomposedEvent {
public:
   AttValComposite(e_Event idArg, Component *aCompo);
   const char * zstring() const;
   e_Event decompose(ParseState *pPrsstat, EventList& evlstDecomposed);

private:
   Event& make() const;
   Component *aCompo;
};

/*-- INLINES --*/
inline AttValComposite::AttValComposite(
   e_Event idArg, Component * aCompoArg
) : DecomposedEvent(idArg, aCompoArg->count() - 1) {
   aCompo = aCompoArg;
}
inline Event& AttValComposite::make() const {
   return * new AttValComposite(id, aCompo);
}
inline const char * AttValComposite::zstring() const {
   return aCompo->value();
}

/*-----------------------------------------------------------------------------+
| AttValComponent Event                                                        |
| - regroup following values:                                                  |
|  ATT_VAL_COMP_DATA_EV,                                                       |
|  ATT_VAL_COMP_CDATA_EV,                                                      |
|  ATT_VAL_COMP_SDATA_EV                                                       |
+-----------------------------------------------------------------------------*/
class AttValComponent : public DecomposedEvent {
public:
   AttValComponent(
      e_Event idArg, const EntityDef *pEnt,
      const char *pchValueArg, const int iLengthArg,
      int iRepeatMaxArg = 1
   );
   ~AttValComponent();
   const char * zstring() const;

private:
   Event& make() const;
   const EntityDef * const pEnt;
   char * pszValue;
   const int iLength;
};

/*-- INLINES --*/
inline AttValComponent::AttValComponent(
   e_Event idArg, const EntityDef *pEntArg,
   const char *pchValueArg, const int iLengthArg, int iRepeatMaxArg
) : DecomposedEvent(idArg, iRepeatMaxArg), iLength(iLengthArg), pEnt(pEntArg) {
   pszValue = new char [1 + iLength];
   pszValue[iLength] = '\0';
   memcpy(pszValue, pchValueArg, iLength);
}
inline AttValComponent::~AttValComponent() {
   delete [] (char *)pszValue;
}
inline Event& AttValComponent::make() const {
   return * new AttValComponent(id, pEnt, pszValue, iLength, iRepeatMax);
}
inline const char * AttValComponent::zstring() const {
   return pszValue;
}

/*-----------------------------------------------------------------------------+
| AttValTokens Event  (ATT_VAL_TOKENS) - Chain the following values:           |
|    ATT_VAL_TOKEN_ENTITY_EV                                                   |
|    ATT_VAL_TOKEN_IDREF_EV                                                    |
|    ATT_VAL_TOKEN_NAME_EV                                                     |
|    ATT_VAL_TOKEN_NMTOKEN_EV                                                  |
|    ATT_VAL_TOKEN_NUMBER_EV                                                   |
|    ATT_VAL_TOKEN_NUTOKEN_EV                                                  |
+-----------------------------------------------------------------------------*/
class AttValTokensRep {
   friend class AttValTokensArg;
   friend class AttValTokens;
private:
   AttValTokensRep(
      e_Event idTokensArg, const char *pszValueArg, const int iSepChar
   );
   int iReferenceCount;
   const char *pszValue;
   e_Event idTokens;
   StringList sl;
};
/*-- INLINES --*/
inline AttValTokensRep::AttValTokensRep(
   e_Event idTokensArg, const char * pszValueArg, const int iSepChar
) {
   iReferenceCount = 1;
   sl.add_tokens(pszValueArg, iSepChar);
   idTokens = idTokensArg;
   pszValue = pszValueArg;
}

/*----------------------------------------------------------------------------*/
class AttValTokensArg {
   friend class AttValTokens;
public:
   AttValTokensArg(
      e_Event idTokensArg, const char * pszValueArg, const int iSepChar
   );
   ~AttValTokensArg();
private:
   AttValTokensRep& rep;
};

/*-- INLINES --*/
inline AttValTokensArg::AttValTokensArg(
   e_Event idTokensArg, const char * pszValueArg, const int iSepChar
) : rep(* new AttValTokensRep(idTokensArg, pszValueArg, iSepChar)) {
   /*...*/
}
inline AttValTokensArg::~AttValTokensArg() {
   if (--rep.iReferenceCount <= 0) delete &rep;
}

/*----------------------------------------------------------------------------*/
class AttValTokens : public DecomposedEvent {
public:
   AttValTokens(AttValTokensArg& arg);
   AttValTokens(const AttValTokens& attvaltoksSource);
   ~AttValTokens();
   const char * zstring() const;
   e_Event decompose(ParseState *pPrsstat, EventList& evlstDecomposed);

private:
   Event& make() const;
   AttValTokensRep& rep;
};

/*-- INLINES --*/
inline AttValTokens::AttValTokens(
   AttValTokensArg& arg
) : DecomposedEvent(ATT_VAL_TOKENS_EV, arg.rep.sl.count()), rep(arg.rep) {
   ++(rep.iReferenceCount);
}
inline AttValTokens::AttValTokens(
   const AttValTokens& attvaltoksSource
) :
   DecomposedEvent(ATT_VAL_TOKENS_EV, attvaltoksSource.iRepeatMax),
   rep(attvaltoksSource.rep)
{
   ++(rep.iReferenceCount);
}
inline AttValTokens::~AttValTokens() {
   if (--(rep.iReferenceCount) <= 0) delete &rep;
}
inline Event& AttValTokens::make() const {
   return * new AttValTokens(* this);
}
inline const char * AttValTokens::zstring() const {
   return rep.pszValue;
}

/*-----------------------------------------------------------------------------+
| EntNotation Event  (ENT_NOTATION_EV)                                         |
+-----------------------------------------------------------------------------*/
class EntNotation : public DecomposedEvent {
public:
   EntNotation(const EntityNotation *pNadArg, int iRepeatMaxArg = 1);
   const char * zstring() const;
   e_Event decompose(ParseState *pPrsstat, EventList& evlstDecomposed);
private:
   const EntityNotation *pNad;
   AttributeList * pAttlistNotation;
   Event& make() const;
};

/*-- INLINES --*/
inline EntNotation::EntNotation(
   const EntityNotation *pNadArg, int iRepeatMaxArg
) : DecomposedEvent(ENT_NOTATION_EV, iRepeatMaxArg) {
   pNad = pNadArg;
   pAttlistNotation = 0;
}
inline const char * EntNotation::zstring() const {
   return pNad->name();
}
inline Event& EntNotation::make() const {
   return * new EntNotation(pNad, iRepeatMax);
}

/*-----------------------------------------------------------------------------+
| EntExt Event  - Regroup the following values:                                |
|    ENT_EXT_SGML_EV                                                           |
|    ENT_EXT_SUBDOC_EV                                                         |
|    ENT_EXT_CDATA_EV                                                          |
|    ENT_EXT_SDATA_EV                                                          |
|    ENT_EXT_NDATA_EV                                                          |
| Chain PUBLIC_ID, SYSTEM_ID (and ENT_NOTATION if one of ENT_EXT_xDATA_EV)     |
+-----------------------------------------------------------------------------*/
class EntExt : public DecomposedEvent {
public:
   EntExt(e_Event id, const EntityDef *pEntArg, int iRepeatMaxArg = 1);
   e_Event decompose(ParseState *pPrsstat, EventList& evlstDecomposed);

private:
   Event& make() const;
   const EntityDef *pEnt;
};

/*-- INLINES --*/
inline EntExt::EntExt(
   e_Event idArg, const EntityDef *pEntArg, int iRepeatMaxArg
) : DecomposedEvent(idArg, iRepeatMaxArg) {
   pEnt = pEntArg;
}
inline Event& EntExt::make() const {
   return * new EntExt(id, pEnt, iRepeatMax);
}

/*-----------------------------------------------------------------------------+
| ExternId event - chain PUBLIC_ID_EV, SYSTEM_ID_EV, and SYSTEM_ID_OMITTED_EV  |
+-----------------------------------------------------------------------------*/
class ExternIdEv : public DecomposedEvent {
public:
   ExternIdEv(const ExternId *pExidArg, int iRepeatMaxArg = 1);
   e_Event decompose(ParseState *pPrsstat, EventList& evlstDecomposed);

private:
   Event& make() const;
   const ExternId * const pExid;
};

/*-- INLINES --*/
inline ExternIdEv::ExternIdEv(
   const ExternId *pExidArg, int iRepeatMaxArg
) : DecomposedEvent(EOP_EV, iRepeatMaxArg), pExid(pExidArg) {
}
inline Event& ExternIdEv::make() const {
   return * new ExternIdEv(pExid, iRepeatMax);
}
#endif /* _EVENT_H ************************************************************/
