/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// externid.c
///  Class associated to External Identifiers
//   Pierre G. Richard 94/02/02 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <memory.h>
#include "gendef.h"
#include "syntax.h"
#include "quantity.h"
#include "prsstate.h"
#include "externid.h"
#include "enthdlr.h"
#include "errorh.h"

/*--------------------------------------------------------------------ExternId-+
| Constructor (minimal)                                                        |
+-----------------------------------------------------------------------------*/
ExternId::ExternId(const char *pszNameArg, e_TextClass asgn_clsArg)
{
   memset(this, 0, sizeof *this);          // zero the XID
   //***>NOTE: following cast required to violate constness
   pszName = (char *)pszNameArg;           // set the name
   asgn_cls = asgn_clsArg;                 // and assigned class
}

/*--------------------------------------------------------------------ExternId-+
| Constructor from the parse of an external identifier.                        |
| If the parser assigned class `asgn_cls' is CHARSET, this routine understands |
| that the PUBLIC keyword is implied.  Except for CHARSET, all ending "ps" are |
| are skipped.                                                                 |
|                                                                              |
| If (asgn_cls != INVALID_XFC), then `assign_class()' is called.               |
|                                                                              |
| [73] external identifier = ("SYSTEM" | ("PUBLIC", ps+, public identifier)),  |
|                      (ps+, system identifier)?                               |
|                                                                              |
|                                                                              |
| [74] public identifier = minimum literal                                     |
|                                                                              |
| [75] system identifier = ((lit, system data, lit)|(lita, system data, lita)) |
|                                                                              |
| [45] system data = character data  (no markup recognized)                    |
|                                                                              |
| A System Identifier is optional for PUBLIC                                   |
| A Public identifier doesn't exist for SYSTEM                                 |
|                                                                              |
| IMPLEMENTATION NOTE: required data areas are mallocated by this routine      |
| in one chunk, pointed by exid.pchBuffer.                                     |
+-----------------------------------------------------------------------------*/
ExternId::ExternId(
   ParseState *pPrsstat,
   const char *pszNameArg, e_TextClass asgn_clsArg
)
{
   EntityHandler * pEh = pPrsstat->pEh;
   Syntax * pSyntax = pPrsstat->pSyntax;

   static const VarntChoiceEntry ChoicesKindOfExid[] = {
      { PUBLIC_VRN_ix,    PUBLIC_TK },
      { SYSTEM_VRN_ix,    SYSTEM_TK },
      { NOT_FOUND_VRN_ix, NOT_FOUND_TK }
   };
   const int iMaxLitLen = pSyntax->pQuantity->value_of(LITLEN_QTY_ix);
   int iLength = 0;

   memset(this, 0, sizeof *this);                // zero the XID
   //***>NOTE: following cast required to violate constness
   pszName = (char *)pszNameArg;                 // set the name
   isIdentified = TRUE;                          // Even if empty identifier
   if (!*pszName) isDeflt = TRUE;                // unnamed == default entity

   int tk;
   if (asgn_clsArg != CHARSET_XFC) {
      tk = pSyntax->match_rsvname_in_list(pEh, ChoicesKindOfExid);
   }else {
      tk = PUBLIC_TK;         // assume PUBLIC was found!
   }
   switch (tk) {
   case PUBLIC_TK:
      pPrsstat->skip_ps();
      pchBuffer = (char *)malloc(1 + iMaxLitLen);
      isPublic = TRUE;
      pid.len = pPrsstat->read_minimum_literal(pEh, pchBuffer);

      if (
         (pPrsstat->pSgmldcl->features.isFormal) ||  // FORMAL YES
         (pPrsstat->pst0 < SGMLDCL_DONE_PST0)    // or still parsing <!SGML
      ) {
         EntityDef entPseudoLit(pchBuffer, pid.len);
         pEh->open(&entPseudoLit);               // re-read it
         formalize_public_id(
            pEh, pSyntax, (pPrsstat->pst0 >= SGMLDCL_DONE_PST0)? TRUE : FALSE
         );
         if (asgn_clsArg != INVALID_XFC) {       // if user asked for..
            assign_class(                        // check class
               pEh, asgn_clsArg, pPrsstat->pSgmldcl->features.isFormal
            );
         }
         pEh->close();
      }else {
         asgn_cls = asgn_clsArg;
      }
      iLength = 1 + pid.len;                     // > 0, even if ""
      break;
   case SYSTEM_TK:
      // isSystem is TRUE only if a literal follows
      asgn_cls = asgn_clsArg;
      break;
   default:
      errorh(INVKWD, pEh, ERH_T_SEV);
      break;
   }
   if (asgn_clsArg != CHARSET_XFC) {
      pPrsstat->skip_ps();
      if (
         (pPrsstat->get_curr_token() == LIT_TK) ||
         (pPrsstat->get_curr_token() == LITA_TK)
      ) {
         if (iLength) {
            pchBuffer = (char *)realloc(pchBuffer, iLength + 1 + iMaxLitLen);
         }else {
            pchBuffer = (char *)malloc(1 + iMaxLitLen);
         }
         isSystem = TRUE;
         sid.rel = iLength;
         sid.len = pPrsstat->read_cdata_literal(pEh, pchBuffer + iLength);
         iLength += 1 + sid.len;
         pPrsstat->skip_ps();
      }
      if (asgn_clsArg == INVALID_XFC) asgn_cls = TEXT_XFC;  // vanilia
   }
   if (iLength) pchBuffer = (char *)realloc(pchBuffer, iLength);
}

/*--------------------------------------------------------------------ExternId-+
| True Copy Constructor: for compiled exid's                                   |
+-----------------------------------------------------------------------------*/
ExternId::ExternId(
   const ExternId& exidSource, const char *pszNameArg, e_TextClass asgn_clsArg
)
{
   *this = exidSource;            // copy the exid
   //***>NOTE: following cast required to violate constness
   pszName = (char *)pszNameArg;
   if (exidSource.pchBuffer) {    // save the exid value
      int iLen = exidSource.pid.len + exidSource.sid.len + 2;

      pchBuffer = (char *)malloc(iLen);
      memcpy(pchBuffer, exidSource.pchBuffer, iLen);
   }
   isChecked = FALSE;             // FIND_PFC not issued
   isInvalid = FALSE;
   asgn_cls = asgn_clsArg;        // set the class
}

/*--------------------------------------------------------------------ExternId-+
| Copy Constructor to create an external id for a defaulted entity             |
+-----------------------------------------------------------------------------*/
ExternId::ExternId(const ExternId& exidSource, const char *pszNameArg)
{
   *this = exidSource;            // copy the exid
   //***>NOTE: following cast required to violate constness
   pszName = (char *)pszNameArg;
}

/*-------------------------------------------------------------------~ExternId-+
| Destructor                                                                   |
+-----------------------------------------------------------------------------*/
ExternId::~ExternId()
{
   free(pchBuffer);
}

/*-------------------------------------------------------------------dislocate-+
| Dislocate external id of notations                                           |
+-----------------------------------------------------------------------------*/
void ExternId::dislocate(
   StringKeeper& sk, const ExternId& exidSource, const char *pszNameArg
)
{
   *this = exidSource;                       // copy the exid
   ofsName = sk.park(pszNameArg, 1 + strlen(pszNameArg));
   ofsBuffer = sk.park(
      exidSource.pchBuffer, exidSource.sid.len + exidSource.pid.len + 2
   );
}

/*--------------------------------------------------------------------relocate-+
| Relocate external id of notations                                            |
+-----------------------------------------------------------------------------*/
void ExternId::relocate(char *strbuf)
{
   pszName = strbuf + ofsName;
   pchBuffer = strbuf + ofsBuffer;
}

/*---------------------------------------------------------------------pile_up-+
| Pile up external id's of entities in the external id array                   |
|                                                                              |
| IMPLEMENTATION NOTE: it's assumed the default entity (if any) will come      |
| first -- see implementation note in EntityDcl::parse()                       |
+-----------------------------------------------------------------------------*/
int ExternId::pile_up(
   StringKeeper &sk,
   int ofsNameArg,
   ExternId*& pExidMP, const ExternId* const aExidTable,
   unsigned int &ofsBufferExidEntDefault
)
{
   *pExidMP = *this;                   // copy
   pExidMP->isInvalid = FALSE;         // reset session bits
   pExidMP->isChecked = FALSE;
   pExidMP->ofsName = ofsNameArg;
   if ((!isDeflt) || (!ofsNameArg)) {  // not defaulted, or the default itself
      pExidMP->ofsBuffer = sk.park(pchBuffer, sid.len + pid.len + 2);
      if (!ofsNameArg) ofsBufferExidEntDefault = pExidMP->ofsBuffer; // default
   }else {                             // defaulted, but not the default itself
      pExidMP->ofsBuffer = ofsBufferExidEntDefault;
      this->pchBuffer = 0;             // already deleted!
   }
   int ix = diffptr(pExidMP, aExidTable);
   ++pExidMP;
   return ix;
}

/*---------------------------------------------------------------relocate_pile-+
| Relocate iMaxEntries ExternId objects piled_up in an array                   |
+-----------------------------------------------------------------------------*/
void ExternId::relocate_pile(unsigned int iMaxEntries, char * strbuf)
{
   ExternId *pExid = this;

   for (int i=0; i < iMaxEntries; ++i, ++pExid) {
      pExid->pszName = strbuf + pExid->ofsName;
      pExid->pchBuffer = strbuf + pExid->ofsBuffer;
   }
}

/*------------------------------------------------------------------------find-+
| External Entities/Notations: obtain the FileId ApplicationHandle in EXID.    |
| return FALSE if not found, or TRUE if OK                                     |
+-----------------------------------------------------------------------------*/
boolean ExternId::find(EntityHandler *pEh)
{
   if (!isChecked) {
      switch (pEh->find(this, ahlIdent)) {
      case 0:
         isInvalid = FALSE;
         isChecked = TRUE;
         break;
      default:
         isInvalid = TRUE;
         break;
      }
   }
   if (isInvalid) {
      switch (asgn_cls) {
      case PRIMARY_XFC:
         errorh(BADOPPRIM, ERH_T_SEV);
         isChecked = TRUE;         // Don't ask twice!
         break;
      case UTSYNL_XFC:   // Local OCS,
      case UTSYNR_XFC:   // Reference OCS,
      case UTSYNU_XFC:   // User-owned OCS,
      case UTODTR_XFC:   // ODT Read
      case UTODTW_XFC:   // ODT Write
      case UTOLTR_XFC:   // OLT Read
      case UTOLTW_XFC:   // OLT Write
         break;          // All utility files: be quiet
      default:
         errorh(INVEXID, pEh, ERH_I_SEV, pszName);
         isChecked = TRUE;
         break;
      }
      return FALSE;          // check_code = 0
   }else {
      return TRUE;
   }
}

/*--------------------------------------------------check_duplicate_formal_pid-+
| Check that FPI's are different each other (10.2.2, last paragraph).          |
+-----------------------------------------------------------------------------*/
void ExternId::check_duplicate(StringKeeper& sk, const int iMaxEntries)
{
   ExternId *pExid1 = this;
   for (int i1=1; i1 < iMaxEntries; ++pExid1, ++i1) {
      if (pExid1->isValidfpi) {
         int len = pExid1->pid.len;
         char *psz = sk.point(pExid1->ofsBuffer + pExid1->pid.rel);
         ExternId *pExid2 = pExid1 + 1;
         for (int i2=i1+1; i2 <= iMaxEntries; ++pExid2, ++i2) {
            if (
               (pExid2->isValidfpi) && (len == pExid2->pid.len) &&
               !strcmp(psz, sk.point(pExid2->ofsBuffer + pExid2->pid.rel))
            ) {
               errorh(                         // YES: identical
                  DUPFPI, ERH_E_SEV,
                  sk.point(pExid1->ofsName), sk.point(pExid2->ofsName)
               );
            }
         }
      }
   }
}

/*---------------------------------------------------------formalize_public_id-+
| Build a public identifier or a formal public identifier                      |
|                                                                              |
| The FPI structure is filled "the best we can": unrecognized field are set    |
| to zero.  While parsing the <!SGML, just issue informative messages.         |
|                                                                              |
| This code must be enhanced: the following "DEFINE"(s) will permit (some day) |
| implementing a CHARSET_independent version.                                  |
+-----------------------------------------------------------------------------*/
#define SOLIDUS       '/'
#define SOLIDI_LEN    2
#define PLUS_PREFIX   '+'
#define MINUS_PREFIX  '-'
void ExternId::formalize_public_id(
   EntityHandler *pEh, Syntax *pSyntax, boolean isTrueWarning
)
{
   static const VarntChoiceEntry ChoicesPublicTextClass[] = {
      { CAPACITY_VRN_ix,  CAPACITY_XFC }, // capacity set
      { CHARSET_VRN_ix,   CHARSET_XFC  }, // character data
      { DOCUMENT_VRN_ix,  DOCUMENT_XFC }, // SGML document
      { DTD_VRN_ix,       DTD_XFC      }, // doc type decl subset
      { ELEMENTS_VRN_ix,  ELEMENTS_XFC }, // element set
      { ENTITIES_VRN_ix,  ENTITIES_XFC }, // entity set
      { LPD_VRN_ix,       LPD_XFC      }, // link type decl subset
      { NONSGML_VRN_ix,   NONSGML_XFC  }, // non-SGML data entity
      { NOTATION_VRN_ix,  NOTATION_XFC }, // character data
      { SHORTREF_VRN_ix,  SHORTREF_XFC }, // short reference set
      { SUBDOC_VRN_ix,    SUBDOC_XFC   }, // SGML subdoc entity
      { SYNTAX_VRN_ix,    SYNTAX_XFC   }, // concrete syntax
      { TEXT_VRN_ix,      TEXT_XFC     }, // SGML text entity
      { NOT_FOUND_VRN_ix, INVALID_XFC  }  // INVALID
   };
   int i;
   unsigned int rel = 0;
   e_ErrorSev sev;
   boolean check_charset;

   if (isTrueWarning) {
      sev = ERH_E_SEV;
   }else {
      sev = ERH_I_SEV;
   }

   /*------------------------------------------+
   | (minimum data)//  => ISO Owner            |
   | +//(minimum data)// => Registered Owner   |
   | -//(minimum data)// => Unregistered Owner |
   +------------------------------------------*/
   switch (pEh->curr_char_val()) {
   default:                 // ISO
      isRegistered = TRUE;
      isIsoreg = TRUE;
      check_charset = TRUE;
      break;
   case PLUS_PREFIX:
      isRegistered = TRUE;
      // fall thru
   case MINUS_PREFIX:       // unregistered
      pEh->read();
      i = skip_past_solidi(pEh);
      if (i != SOLIDI_LEN) {
         errorh(INVOWNKN, pEh, sev);  // Invalid Owner Kind
         return;
      }
      rel = 1 + SOLIDI_LEN;
      check_charset = FALSE;
      break;
   }
   ownid.rel = rel;
   i = skip_past_solidi(pEh);
   if (i <= 0) {
      errorh(INVOWNID, pEh, sev);     // Invalid Owner Ident
      return;
   }
   ownid.len = i-SOLIDI_LEN;
   if (check_charset) {
      if (
         strncmp("ISO Registration Number", pchBuffer + ownid.rel, ownid.len)
      ) {
        check_charset = FALSE;     // if not "ISO Reg..."
      }
   }

   /*-------------------------+
   | Public_text_class, SPACE |
   | Entered in upper-case    |
   +-------------------------*/
   {
      boolean iCaseGeneralSave = pSyntax->tv.isCaseGeneral; // save CASE option

      pSyntax->tv.isCaseGeneral = FALSE;                    // turn it off
      spec_cls = (e_TextClass)pSyntax->match_rsvname_in_list(
         pEh, ChoicesPublicTextClass
      );
      pSyntax->tv.isCaseGeneral = iCaseGeneralSave;         // and restore it
   }
   switch(spec_cls) {
   case INVALID_XFC:
      errorh(INVTXTCL, pEh, sev); // Invalid Text Class
      return;
   case CHARSET_XFC:
      isPtds = TRUE;
      isProvptdv = TRUE;
      break;
   case NOTATION_XFC:
   case CAPACITY_XFC:
   case SYNTAX_XFC:
      isProvptdv = TRUE;                 // PTDV must be omitted
   default:
      if (check_charset) {               // ISO Reg Nmbr w/o CHARSET
         errorh(INVPTC, pEh, sev);
      }
      break;
   }
   if (!pEh->is_SPACEchar()) {
      errorh(NOSPTXCL, pEh, sev);
      return;
   }
   pEh->read();

   /*------------------------------------------+
   | "-//"?,(public text description),"//"     |
   | -// => unavailable public text (optional) |
   +------------------------------------------*/
   rel = txtdesc.rel = (unsigned int)pEh->offset();
   if (pEh->curr_char_val() == MINUS_PREFIX) {
      i = skip_past_solidi(pEh);              // unavailable public text ?
      if (i == 1+SOLIDI_LEN) {                // check better
         rel += 1+SOLIDI_LEN;                 // Yep: update rel
         txtdesc.rel = rel;                   // point text description
         i = skip_past_solidi(pEh);           // skip over
      }else {                                 // No:
         isAvailptext = TRUE;                 // Text is available
      }
   }else {
      isAvailptext = TRUE;
      i = skip_past_solidi(pEh);
   }
   if (i > 0) {
      txtdesc.len = i-SOLIDI_LEN;
      rel += i;
   }else {
      isAvailptext = FALSE;
      errorh(NOPTXDS, pEh, sev);
      return;
   }

   /*--------------------------------------+
   | (public text language                 |
   |  | public text designating sequence), |
   |  ("//",public text display version)?  |
   +--------------------------------------*/
   lang.rel = rel;                            // Lang or Pub txt design seq
   i = skip_past_solidi(pEh);
   if (i <= 0 ) {                             // Ee reached
      i = -i;
      rel += i;
   }else {
      rel += i;
      i -= SOLIDI_LEN;
   }
   lang.len = i;
   if (isPtds) {                              // Public Text Design Seq:
      if (!i) {
         errorh(NOPTDS, pEh, sev);
         return;
      }
   }else {                                    // Language:
      /*-------------------------------+
      | Check "a posteriori" the       |
      | language name is OK (10.2.2.3) |
      +-------------------------------*/
      {
         char *ach = pchBuffer + lang.rel;
         int j = 0;
         boolean first_time = TRUE;

         if ((!i) || (!pSyntax->is_namestrt(ach[j]))) {
            errorh(INVPTL, pEh, sev);
            return;
         }
         do {
            if (j >= pSyntax->pQuantity->value_of(NAMELEN_QTY_ix)) {
               errorh(INVPTL, pEh, sev); // Too big
               return;
            }
            if (!pSyntax->is_name(ach[j])) {
               errorh(INVPTL, pEh, sev);
               return;
            }
            if (ach[j] != pSyntax->uppercase(ach[j])) {
               ach[j] = (char)pSyntax->uppercase(ach[j]);
               if (first_time) {
                  errorh(NOLANGUP, pEh, sev);
                  first_time = FALSE;
               }
            }
         }while (++j < i);
      }
   }
   if (pEh->is_Ee()) {                        // PTDV omitted
      isProvptdv = TRUE;                      // system may provide it
   }else {
      if (isProvptdv) {                       // PDTV given, but not allowed
         errorh(OOCPTDV, pEh, sev);
         return;
      }
      ptdv.rel = rel;
      i=skip_past_solidi(pEh);
      if (i <= 0) {                            // try to reach Ee
         ptdv.len = -i;
      }else {                                  // Ee not reached
         ptdv.len = i;
         errorh(INVPTDV, pEh, sev);
         return;
      }
   }
   isValidfpi = TRUE;
}

/*----------------------------------------------------------------assign_class-+
| Check the public text class of an external identifier                        |
| and set the parser assigned class exid.asgn_cls.                             |
|                                                                              |
| This routine is called systematically each time an external identifier       |
| has been found.  It checks that the spec_cls is compatible with the          |
| spec_cls (class specified in the formal public identifier).                  |
|                                                                              |
|   Asgn_cls:       Comes from:           If incompatible class,               |
|                                         Err Lvl:  VALIDFPI is:               |
|                                                                              |
|   CHARSET_XFC     <!SGML... BASESET        I      Off                        |
|   SYNTAX_XFC      <!SGML... CAPACITY       I      Off                        |
|   CAPACITY_XFC    <!SGML... SYNTAX         I      Off                        |
|   NOTATION_XFC    <!NOTATION               W      Off                        |
|   NONSGML_XFC     <!ENTITY... N/C/SDATA    I      Unchanged                  |
|   SUBDOC_XFC      <!ENTITY... SUBDOC       I      Unchanged                  |
|   DTD_XFC         <!DOCTYPE                I      Unchanged                  |
|   LPD_XFC         <!LINKTYPE               I      Unchanged                  |
|   0               <!ENTITY... (no type)    I      Unchanged                  |
|                                                                              |
|                                                                              |
| A Public Text Class is said compatible:                                      |
|                                                                              |
|   - When asgn_cls is one of CHARSET_XFC, SYNTAX_XFC, CAPACITY_XFC            |
|     and (asgn_cls == spec_cls).  If not, only an Informational               |
|     message is issued: if later FORMAL YES is found, then a final            |
|     Warning message will be issued.                                          |
|                                                                              |
|   - When FORMAL is YES and                                                   |
|      - asgn_cls is one of NOTATION_XFC, NONSGML_XFC,                         |
|        SUBDOC_XFC, DTD_XFC, LPD_XFC and (asgn_cls == spec_cls).              |
|        If not, Warning or Informational messages will be issued depending    |
|        whether this case was flagged as an error in ISO8879, or not.         |
|      - asgn_cls is 0 and spec_cls is one of DOCUMENT_XFC, DTD_XFC,           |
|        ELEMENTS_XFC, ENTITIES_XFC, LPD_XFC, SHORTREF_XFC, TEXT_XFC.          |
|        If not, -- i.e in an ENTITY declaration, the spec_cls is one          |
|        of: CAPACITY_XFC, CHARSET_XFC, NONSGML_XFC, NOTATION_XFC,             |
|        SUBDOC_XFC, SYNTAX_XFC -- an informational message will be issued     |
|        and the asgn_cls will be TEXT_XFC.                                    |
|                                                                              |
+-----------------------------------------------------------------------------*/
void ExternId::assign_class(
   EntityHandler *pEh, e_TextClass asgn_clsArg, boolean isFormal
)
{
   asgn_cls = asgn_clsArg;          // set the assigned class anyway
   if ((!isValidfpi) || (!pEh)) {   // return: message already issued
      return;                       // or no message, but FORMAL NO
   }
   switch (asgn_cls) {
   case CAPACITY_XFC:
   case CHARSET_XFC:
   case SYNTAX_XFC:
      if (asgn_cls != spec_cls) {
         isValidfpi = FALSE;
         errorh(BADCLS, pEh, ERH_I_SEV);
      }
      break;
   case NOTATION_XFC:
      if ((isFormal) && (asgn_cls != spec_cls)) {
         isValidfpi = FALSE;
         errorh(BADCLS, pEh, ERH_E_SEV);
      }
      break;
   case NONSGML_XFC:
   case SUBDOC_XFC:
   case DTD_XFC:
   case LPD_XFC:
      if ((isFormal) && (asgn_cls != spec_cls)) {
         errorh(BADCLS, pEh, ERH_I_SEV);
      }
      break;
   default:            // case TEXT_XFC
      if (isFormal) {
         switch (spec_cls) {
         case DOCUMENT_XFC:
         case DTD_XFC:
         case ELEMENTS_XFC:
         case ENTITIES_XFC:
         case LPD_XFC:
         case SHORTREF_XFC:
         case TEXT_XFC:
            break;
         default:
            errorh(BADCLS, pEh, ERH_I_SEV);
            break;
         }
      }
      break;
   }
}

/*STATIC------------------------------------------------------skip_past_solidi-+
| Skip after next solidi ("//").                                               |
| Return number of bytes skipped, negated if Ee was reached.                   |
+-----------------------------------------------------------------------------*/
int ExternId::skip_past_solidi(EntityHandler *pEh)
{
   boolean isFirstSolidusFound = FALSE;

   for (int count=0; !pEh->is_Ee(); pEh->read(), ++count) {
      if (pEh->curr_char_val() == SOLIDUS) {
         if (isFirstSolidusFound) {
            pEh->read();
            return count+1;
         }
         isFirstSolidusFound = TRUE;
      }else {
         isFirstSolidusFound = FALSE;
      }
   }
   return -count;
}
/*============================================================================*/
