/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// modele.c
///  Check validity of elements against the content models
//   Pierre G. Richard 88/02/23 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
/*
|
| Check that, at a given point, a proposal to start or end an element,
| or to enter #PCDATA matches the "ContentToken" tree -- a binary tree ,built
| when decoding the DTD's ELEMENT declaration.
| To achieve this goal, the history of already entered elements is kept in
| a OpenElement stack, that is up to the point reached in the document
| structure.
|
| Given a OpenElement stack, this code is able to tell if a proposed change
| in the document structure is valid or not, and if not, why: for instance,
| if a known required element is missing, if some required element is missing
| but unable to find out what element, or if an element is out of context.
|
| The "ContentToken" tree is built in the following manner.  To each "token"
| (either a model group, or a primitive content token) is associated a node.
| If the token is a model group, then the node is a true parent node and its
| child link points the first content token.  its content token.  If the
| content token is a primitive content token, the node is a leaf, and the
| child link is reused to point at the ElementType structure.
|
| For a more detailled description, see comments in the code.
|
| IMPLEMENTATION NOTES =
|
| This code is a tricky one: don't fiddle with it !
|
| DEPENDENCIES =  (see Include list)
|
| IMPORT=
|   (none)
|
| EXPORT=
|   "mdl_goto" requires to reach a given element in the model tree
|   "mdl_fini" requires to find a way out from the model.
|   "mdl_edtg" close a datatag when the template matches the data.
|   "mdl_qaelem" asks for a list of allowed elements
|
*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <memory.h>
#include "gendef.h"
#include "dtddcl.h"
#include "modele.h"
#include "element.h"
#include "oestack.h"

/*----------------------------------------------------------------------Modele-+
| Regular constructor                                                          |
+-----------------------------------------------------------------------------*/
Modele::Modele(const ElementType *pElem)
{
   ixGrpCurrent = -1;
   aGrpStack[0].pCnttok = pElem->content_token();
}

/*PRIVATE---------------------------------------------------open_content_token-+
| Register a new OpenModelGroup entry on the current stack of OpenModelGroup's.|
| ixToken indicates the index of the ContentToken of the new entry.            |
|                                                                              |
| Note: to access a token given the "father" ContentToken and the index of the |
|       child: father + father.mdl.rx points at the first "child" ContentToken,|
|       the add ix to points the ix-th child (stringized binary tree).         |
+-----------------------------------------------------------------------------*/
inline void Modele::open_content_token(OpenModelGroup * pGrp, int ixToken) {
   pGrp[0].ix = ixToken;
   pGrp[1].pCnttok = pGrp[0].pCnttok + pGrp[0].pCnttok->mdl.rx + ixToken;
}

/*PRIVATE-----------------------------------------------------------next_unhit-+
| Returns the index of the first unset bit in the group pointed by "pGrpMdl",  |
| after and including the caller's "first" bit.                                |
| If all bits are set, it returns -1;                                          |
+-----------------------------------------------------------------------------*/
int Modele::next_unhit(
   OpenModelGroup * pGrpMdl,                  // group
   unsigned int first                         // Index of first bit to be tested
)
{
   unsigned int last = pGrpMdl->pCnttok->mdl.num;  // Last bit to be tested

   while (first < last) {
      if (!pGrpMdl->hits[first]) return first;
      ++first;
   }
   return -1;
}

/*PRIVATE--------------------------------------------------all_tokens_were_hit-+
| Returns TRUE if all hit bits for the specified group are turned on,          |
| (except for optional tokens if "isOptionalExcluded" is TRUE).                |
| Otherwise, returns FALSE.                                                    |
| pGrpMdl points the group header to be checked against.                       |
+-----------------------------------------------------------------------------*/
boolean Modele::all_tokens_were_hit(
   OpenModelGroup * pGrpMdl,          // model group pointer
   boolean isOptionalExcluded         // TRUE: optional tokens must been hit
)
{
   const ContentToken *pCnttok = pGrpMdl->pCnttok; // point model node
   int last;
   int i;

   last = pCnttok->mdl.num;           // Ending index (number of bits to test)

   for (pCnttok += pCnttok->mdl.rx, i=-1; ++i < last; ++pCnttok) {
      if (
         !pGrpMdl->hits[i] &&
         (!isOptionalExcluded || !(pCnttok->occur & OMISSIBLE))
      ) {
         return FALSE;
      }
   }
   return TRUE;
}

/*PRIVATE--------------------------------------------------------is_last_token-+
| Return TRUE if current token is the last token in this model group           |
+-----------------------------------------------------------------------------*/
inline boolean Modele::is_last_token(const OpenModelGroup * const pGrpMdl)
{
   return (pGrpMdl->ix == (pGrpMdl->pCnttok->mdl.num  - 1)) ? TRUE : FALSE;
}

/*PRIVATE----------------------------------------------------is_optional_token-+
| Return TRUE if current token is contextually optional, FALSE otherwise.      |
|                                                                              |
| see ISO8879-1986(E), page 8:                                                 |
| 4.59 "contextually optional element"                                         |
| 4.60: "contextually optional token"                                          |
|                                                                              |
| Contextually optional element:                                               |
|   1) a plus inclusion; or                                                    |
|   2) currently applicable model token is a "contextually optional token":    |
|      i.e. a content token that                                               |
|      a) has an inherently optional token; or                                 |
|      b) has a PLUS occurence indicator and has been satisfied; or            |
|      c) is in a model group that is itself a contextually optional token,    |
|         no tokens of which has been statisfied.                              |
|                                                                              |
| Note: the "inherently optional" rule has been already applied at DTD compile |
|       time.                                                                  |
+-----------------------------------------------------------------------------*/
boolean Modele::is_optional_token(
   const boolean mhit                // Hit bit of the element model
)
{
   int ix = ixGrpCurrent + 1;
   OpenModelGroup * pGrp = aGrpStack + ix;                 // Points token

   while (!(pGrp->pCnttok->occur & OMISSIBLE)) {           // Rule a)
      if (--pGrp, --ix < 0)               return mhit;     // Rule b)
      if (pGrp->hits[pGrp->ix])           return TRUE;     // Rule b)
      if (pGrp->hits != 0)                return FALSE;    // Rule c)
   }
   return TRUE;
}

/*PRIVATE----------------------------------------------------is_required_token-+
| Return TRUE if current elmnt is contextually required, FALSE otherwise.      |
|                                                                              |
| see ISO8879-1986(E), page 8:                                                 |
| 4.61 "contextually required element"                                         |
| 4.62: "contextually required token"                                          |
|                                                                              |
| Not a contextually optional element; and                                     |
|   1) gi of the Document Type Name; or                                        |
|   2) currently applicable model token is a "contextually required token":    |
|      i.e. a content token that                                               |
|      a) is the only one in its model group; or                               |
|      b) is a member of a "seq" group                                         |
|         i) that                                                              |
|            - is itself a contextually required token; or                     |
|            - contains a token which has been satisfied;                      |
|            and                                                               |
|         ii) all preceding tokens of which                                    |
|            - have been satisfied; or                                         |
|            - are contextually optional.                                      |
|                                                                              |
| Note: "b) ii" will be achieved by next_token().                              |
+-----------------------------------------------------------------------------*/
boolean Modele::is_required_token(
   const boolean mhit              // Hit bit of the element model
)
{
   if (is_optional_token(mhit)) return FALSE;

   int ix = ixGrpCurrent + 1;
   OpenModelGroup * pGrp = aGrpStack + ix;

   while (--pGrp, --ix >= 0) {
      if (pGrp->pCnttok->conn != SEQ_CONN) return FALSE;
      if (pGrp->hits != 0)                 return TRUE;
   }
   return TRUE;
}

/*PRIVATE-----------------------------------------------------check_exceptions-+
| See if an element is in a plus or minus exception group on the stack.        |
| This function must be called for real tags (neither PCDATA, or datatag).     |
|                                                                              |
| Returns: > 0 if in a minus group                                             |
|              --this is also "ex_depth", (stack depth level of minus group)   |
|           -1 if in a plus group, otherwise                                   |
|            0 if in neither                                                   |
+-----------------------------------------------------------------------------*/
int Modele::check_exceptions(
   const ElementType * const pElem,        // Element to look for
   const OpenElementStack & oelemstk
)
{
   /*--------------------------------------------+
   | First check in the minus groups of all TCBs |
   | Return i > 0, meaning: exclusion found at   |
   | level i (0 < i <= tcb_stack_ix)             |
   +--------------------------------------------*/
   {
      OpenElementStackIterator iterate(oelemstk);
      const OpenElement *pOelem;
      for (int i=1; pOelem = iterate.prev(), pOelem; ++i) {
         ExclusionListIterator iterate(pOelem->pElem);
         const ElementType *pElemExclusion;
         while ((pElemExclusion = iterate.next()) != 0) {
            if (pElemExclusion == pElem) return i;
         }
      }
   }
   /*------------------------------------------+
   | Then check in the plus groups of all TCBs |
   +------------------------------------------*/
   {
      OpenElementStackIterator iterate(oelemstk);
      const OpenElement *pOelem;
      while (pOelem = iterate.prev(), pOelem) {
         InclusionListIterator iterate(pOelem->pElem);
         const ElementType *pElemInclusion;
         while ((pElemInclusion = iterate.next()) != 0) {
            if (pElemInclusion == pElem) return -1;
         }
      }
   }
   /*-----------------------+
   |Nothing found: return 0 |
   +-----------------------*/
   return 0;
}

/*PRIVATE-----------------------------------------------------------next_token-+
| Find the next token to test, set new "ixGrpCurrent" in the tcb.              |
| "ntk" the new_token code is a one of:                                        |
|      - FLUSH: don't make stops at each possible token                        |
|      - AFTER_HIT: assume the current token was hit                           |
|      - NONE: None of these                                                   |
|                                                                              |
| Set "ms" in the OpenElement to:                                              |
|      FINISH_MS  if element has ended (no more tokens to test);               |
|      INVALID_MS if a new token cannot be found because current token         |
|                 (which was not hit) was neither required nor optional;       |
|                 (see ISO8879-1986(E) 4.61, NOTE)                             |
|                                                                              |
| In addition if FLUSH is off:                                                 |
|      MATCH_MS  if a new token was found, or a previous token                 |
|                could be repeated;                                            |
|                                                                              |
| The current token will either be a group header,                             |
| or some token that could not end its group.                                  |
+-----------------------------------------------------------------------------*/
void Modele::next_token(
   OpenElement * const pOelemCurrent,
   const e_NextContext ntk      // new_token_code: see comment in the header
)
{
   OpenModelGroup *pGrpMdl = aGrpStack + ixGrpCurrent;

   /*----------------------------------------------------+
   | If the GI was a hit, turn on the hit bit.           |
   | If the current token is repeatable, exit so it will |
   | stand as the next token to test (note at 11.2.4).   |
   +----------------------------------------------------*/
   if (ntk == AFTER_HIT_NTK) {
      if (ixGrpCurrent < 0) {
         pOelemCurrent->isMhit = TRUE;
      }else {
         pGrpMdl->hits += pGrpMdl->ix;      // turn on hit bit;
      }
      if ((pGrpMdl+1)->pCnttok->occur & REPEATABLE) {
         pOelemCurrent->msStatus = MATCH_MS;
         return;
      }
   }
   /*-------------------------------------------------------------------+
   | To find the next token, we will try first to end the current group |
   | and any we can that contain it.                                    |
   +-------------------------------------------------------------------*/
   while (ixGrpCurrent >= 0) {
      if (
         (pGrpMdl->hits == 0) &&              // none were hit
         (
            (is_last_token(pGrpMdl)) ||
            (
               (!((pGrpMdl+1)->pCnttok->occur & OMISSIBLE)) &&
               (pGrpMdl->pCnttok->conn == SEQ_CONN)
            )
         )
      ) {
         /*-------------------------------------------------------+
         | It's worthless to further scan in the current group if |
         | nothing was hit and we reached the last token, or, for |
         | a SEQ group, we reached the first non-optional token.  |
         | Leave this group to try to pursue in the model.        |
         +-------------------------------------------------------*/
         --pGrpMdl, --ixGrpCurrent;
         continue;
      }
      /*-----------------------------------------------------------+
      | The current group can be safely closed                     |
      | (i.e. no more token in it could be a candidate) if:        |
      | - in an OR group, the current token was hit                |
      | - in a SEQ group, the current token is the last and was    |
      |   either hit, or is omittable                              |
      | - in an AND group, all tokens were hit (excluding optional |
      |   tokens if FLUSH)                                         |
      +-----------------------------------------------------------*/
      {
         boolean isAllHit;       // TRUE if all non-optionals were hit

         switch (pGrpMdl->pCnttok->conn) {
         case OR_CONN:
            isAllHit = pGrpMdl->hits[pGrpMdl->ix]; // current token has been hit
            break;
         case SEQ_CONN:
            if (
               (is_last_token(pGrpMdl)) && (    // last token and either
                  ((pGrpMdl+1)->pCnttok->occur & OMISSIBLE) || // omissible,
                  (pGrpMdl->hits[pGrpMdl->ix])     // or this one was hit
               )
            ) {
              isAllHit = TRUE;
            }else {
              isAllHit = FALSE;
            }
            break;
         default:   // case AND_CONN
            isAllHit = all_tokens_were_hit(
               pGrpMdl, (ntk==FLUSH_NTK)? TRUE : FALSE
            );
            break;
         }
         if (isAllHit) {
            /*----------------------------------------------------------+
            | If all non-optional tokens in the group were hit, leave   |
            | the group (successfully).  Set the hit bit.               |
            | If we were called from "mdl_goto", and if the group is    |
            | repeateable, return: the same group is again a candidate. |
            | Otherwise, "mdl_fini" is trying to find a way out: loop.  |
            +----------------------------------------------------------*/
            if (--pGrpMdl, --ixGrpCurrent < 0) {
               pOelemCurrent->isMhit = TRUE;
            }else {
               pGrpMdl->hits += pGrpMdl->ix;    // turn on hit bit;
            }
            if (
               (ntk != FLUSH_NTK) && ((pGrpMdl+1)->pCnttok->occur & REPEATABLE)
            ) {
               pOelemCurrent->msStatus = MATCH_MS;
               return;
            }
            continue;
         }
      }
      /*-----------------------------------------------------------------+
      | At this point, the "current token" is either the original one,   |
      | or the token of the highest level unhit group that it ended.     |
      | If we were called from "mdl_fini", and the element is optional   |
      | we can jump over.  Otherwise, it's INVALID.                      |
      | For an AND group, it's immediately invalid as we already skipped |
      | non-optional tokens.                                             |
      +-----------------------------------------------------------------*/
      if (ntk == FLUSH_NTK) {
         if (
            (pGrpMdl->pCnttok->conn == AND_CONN) ||
            (!is_optional_token(pOelemCurrent->isMhit))
         ) {
            pOelemCurrent->msStatus = INVALID_MS; // No more group to end, return
            return;
         }
         open_content_token(pGrpMdl, pGrpMdl->ix + 1);
         continue;                           // and loop
      }
      /*---------------------------------------------+
      | Called from "mdl_goto".  Get the next token. |
      +---------------------------------------------*/
      if (pGrpMdl->pCnttok->conn == AND_CONN) {
         if (
            (ntk == AFTER_HIT_NTK) ||
            (pGrpMdl->hits[pGrpMdl->ix])     // current token has been hit
         ) {
            /*-------------------------------------+
            | Back to an AND group after a hit:    |
            | either because if was just hit and   |
            | is not repeatable, or it was a       |
            | repeatable token hit before.         |
            |                                      |
            | Restart from the first unhit token.  |
            | Note: next_unhit will not returns -1,|
            | or group would have ended            |
            +-------------------------------------*/
            open_content_token(pGrpMdl, next_unhit(pGrpMdl, 0));
            pOelemCurrent->msStatus = MATCH_MS;
            return;
         }
         /*-------------------------------------+
         | And group, but no hit:               |
         | the next HIT token is the next unhit.|
         +-------------------------------------*/
         {
            int iNext = next_unhit(pGrpMdl,pGrpMdl->ix+1);  // next sibling

            if (iNext != -1) {
               open_content_token(pGrpMdl, iNext);
               pOelemCurrent->msStatus = MATCH_MS;
               return;
            }
         }
         /*--------------------------------------------+
         | If no next_unhit (AND group exhausted)      |
         | and if some non-optional token were not hit,|
         | it's definitively INVALID.                  |
         +--------------------------------------------*/
         if (!all_tokens_were_hit(pGrpMdl, TRUE)) { // some non optional remains...
            pOelemCurrent->msStatus = INVALID_MS;     // definitively INVALID
            return;
         }
         /*-------------------------------------------------+
         | End the group if all non-optional token were hit.|
         | If the group is repeateable, return:             |                           |
         | the same group is again a candidate.             |
         | Otherwise, pursue the quest.                     |
         +-------------------------------------------------*/
         if (--pGrpMdl, --ixGrpCurrent < 0) {
            pOelemCurrent->isMhit = TRUE;
         }else {
            pGrpMdl->hits += pGrpMdl->ix;    // turn on hit bit;
         }
         if ((pGrpMdl+1)->pCnttok->occur & REPEATABLE) {
            pOelemCurrent->msStatus = MATCH_MS;
            return;
         }
         continue;
      }
      /*------------------------------------------------------------------+
      | The next valid token is the next sequential one if current token: |
      | - is omittable,                                                   |
      | - or was hit (i.e., because it is repeatable and was hit by a     |
      |   previous GI or because it is a hit group that just ended),      |
      | - or is in an OR group (it is not the last testable token,        |
      |   because group would have ended)                                 |
      +------------------------------------------------------------------*/
      if (
         ((pGrpMdl+1)->pCnttok->occur & OMISSIBLE) ||
         (pGrpMdl->pCnttok->conn == OR_CONN) ||
         (pGrpMdl->hits[pGrpMdl->ix])
      ) {
         open_content_token(pGrpMdl, pGrpMdl->ix+1);
         pOelemCurrent->msStatus = MATCH_MS;
      }else {
         pOelemCurrent->msStatus = INVALID_MS;
      }
      return;
   }
   /*-----------------------------------------------------------+
   | The model is now exhausted: return FINISH_MS or INVALID_MS |
   +-----------------------------------------------------------*/
   if ((pOelemCurrent->isMhit) || ((pGrpMdl+1)->pCnttok->occur & OMISSIBLE)) {
      pOelemCurrent->msStatus = FINISH_MS;   // hit, or not hit but allowed to
   }else {
      pOelemCurrent->msStatus = INVALID_MS;
   }
   return;
}

/*--------------------------------------------------------------------mdl_goto-+
| Determine whether a GI is valid within the current position reached          |
| in the element model.                                                        |
|                                                                              |
| -- INVARIANTS --                                                             |
| ixGrpCurrent is the index of the most recently opened model group on the     |
|    stack of OpenModelGroup                                                   |
| next entry describes the current token in this group                         |
|    This is the sole entry in the stack of OpenModelGroup that can            |
|    be associated to a primiive content token, rather than a model group,     |
|    i.e.:                                                                     |
|      aGrpStack[ixGrpCurrent+1].pCnttok is a primitive content token          |
|      aGrpStack[ixGrpCurrent+1].hits and ix are irrelevant                    |
|                                                                              |
| -- INPUT --                                                                  |
| "pElemGiven" points the ElementType structure associated to the               |
|     expected primitive content, or a NULL pointer if caller wants to         |
|     check #PCDATA                                                            |
| "oelemstk" is the stack of opened elements                                    |
| "piMexDepth" (optional) is a pointer to an integer                           |
|                                                                              |
| -- PRECONDITION --                                                           |
| - The currently opened element is the pElem member of the last entry in       |
|   the OpenElement stack.                                                     |
| - This element can never be EMPTY, nor #PCDATA, nor DCDATA:                  |
|   - for EMPTY, the element is immediately closed, so it does not stay        |
|     on the stack of OpenElement.                                             |
|                                                                              |
|   - #PCDATA is never put on the stack,                                       |
|   - in case of CDATA or RCDATA, this procedure is not called:                |
|      - it cannot be a start-tag, as it is not recognized,                    |
|      - DATA within DCDATA is required to be trapped by the caller            |
|                                                                              |
| -- OUTPUT --                                                                 |
|                                                                              |
| Return codes are:                                                            |
|    MATCH_MS    if valid,                                                     |
|    MATCHX_MS   if valid, but appears invalidly in the model exclusion list,  |
|    MATCHT_MS   if the current PCDATA is also a TOSL in the model,            |
|    INCLUS_MS   is valid as an inclusion (plus exception),                    |
|    FINISH_MS   if element model is finished,                                 |
|    REQRD_MS    if a different element is required,                           |
|    REQRDX_MS   as above, but the required element was illegally excluded,    |
|    INVALID_MS  if it is totally invalid,                                     |
|    EXCLUS_MS   if invalid as an exclusion (minus exception),                 |
|                *piMexDepth (if not a NULL pointer) has the MEX depth.        |
|                                                                              |
| *piMexDepth (if specified) has the value of the depth level of the           |
|             ancestor element which excluded "pGivenElem": 1 for               |
|             the current element, 2 for its parent, etc...                    |
|             This value is only valid for a return code of EXCLUS_MS.         |
|                                                                              |
| If the return code is MATCH_MS, INCLUS_MS, REQRD_MS, REQRDX_MS:              |
|    *ppElemReturned points the effectively valid ElementType                   |
|    (either the proposed pElemGiven, or a required element if REQRD?_MS)       |
|    If this is a "target elem" (4.315)                                         |
|      *ppDtpReturned points its data_tag pattern.                             |
|                                                                              |
| -- POSTCONDITION --                                                          |
|                                                                              |
|    aGrpStack[ixGrpCurrent+1] describes the next token                        |
|    oelemstk[0]->msStatus reflects the status reached by next token:           |
|        MATCH_MS, FINISH_MS, or INVALID_MS                                    |
|                                                                              |
| -- NOTE --                                                                   |
|    The EXCLUDE validation service is provided thru the MATCHX and REQRDX     |
|    return codes (see ISO 8879-1986(E) p.42, 11.2.5.2, b)                     |
|    MATCHX_MS cannot be returned for #PCDATA  (pElem == 0)                     |
+-----------------------------------------------------------------------------*/
e_StatusModele Modele::mdl_goto(
   const OpenElementStack& oelemstk,     // Stack of opened element
   const ElementType * const pElemGiven, // ElementType of new GI (0: #PCDATA)
   const ElementType ** const ppElemReturned,
   const DataTagPattern ** const ppDtpReturned,
   int *piMexDepth                      // 0, or where to output mex_level
)
{
   OpenElement * const pOelemCurrent = oelemstk[0];
   int iDepthEx;                    // >0 => MEX; -1 => PEX; 0 => none

   if (!piMexDepth) piMexDepth = &iDepthEx;
   *ppElemReturned = pElemGiven;      // Assume it's OK
   *ppDtpReturned = 0;              // and no target found
   if (pElemGiven) {
      *piMexDepth = check_exceptions(pElemGiven, oelemstk);
   }else {
      *piMexDepth = 0;
   }
   if (pOelemCurrent->pElem->is_any()) {
      if (*piMexDepth > 0) {
         return EXCLUS_MS;
      }else {
         return MATCH_MS;           // ANY has no model -> not a target
      }
   }else {
      if (pOelemCurrent->msStatus == FINISH_MS) {
         if (*piMexDepth < 0) {
            return INCLUS_MS;
         }else {
            return FINISH_MS;
         }
      }
   }
   do {
      OpenModelGroup *pGrpTok = aGrpStack + (ixGrpCurrent + 1); // Points token
      /*---------------------------------------------------+
      | Open enough model groups until we meet a primitive |
      | content token.  Have "pCnttok" point to it.        |
      | Note that pCntTok is rather a pCntMdl, until the   |
      | loop exits.                                        |
      +---------------------------------------------------*/
      while (pGrpTok->pCnttok->conn) {      // if still a pGrpMdl...
         pGrpTok->hits = 0;                 // reset all hits
         open_content_token(pGrpTok, 0);    // start at token 0
         ++pGrpTok, ++ixGrpCurrent;         // make it the current group
      }
      const ContentToken * const pCnttok = pGrpTok->pCnttok;
      /*----------------------------------------------------+
      | If "pElemGiven" matches the primitive content token: |
      | - If it's a valid exclusion, return EXCLUS_MS       |
      | - If it's not excluded, return MATCH_MS             |
      | - If its exclusion is invalid, return MATCHX_MS     |
      +----------------------------------------------------*/
      if (
         (pElemGiven == pCnttok->pElem) ||         // Hit in model
         ((!pElemGiven) && (pCnttok->isDatatag))  // or PCDATA && datatag
      ) {
         if (
            (*piMexDepth > 0) &&                 // if in minus group
            (
               (pCnttok->occur & OMISSIBLE) ||
               (pGrpTok-1)->pCnttok->conn == OR_CONN // Model has OR
            )
         ) {
            next_token(pOelemCurrent, AFTER_HIT_NTK);
            return EXCLUS_MS;                    // Valid exclusion
         }                                       // else Invalid exclusion
         if (pCnttok->isTarget) {
            *ppDtpReturned = (pCnttok+1)->pDtp;
         }
         next_token(pOelemCurrent, AFTER_HIT_NTK);
         if (*piMexDepth > 0) {
            return MATCHX_MS;                    // report invalid exclusion
         }else {
            return MATCH_MS;                     // vanilla match
         }
      }
      /*-------------------------------------------------------+
      | If "pElemGiven" doesn't match the current primitive     |
      | content token, see if "pElemGiven" could be an          |
      | inclusion.  Returns INCLUS_MS if so.                   |
      |                                                        |
      | Otherwise, see if the current primitive content token  |
      | is required: returns REQRD_MS if so.                   |
      |                                                        |
      | Notes: - if #PCDATA is found and the primitive content |
      |          token is a target elem, it is required         |
      |          (not said in ISO)                             |
      |        - a #PCDATA primitive content token is never    |
      |          required (see ISO 8879-1986, 11.2.4.2)        |
      +-------------------------------------------------------*/
      if (*piMexDepth < 0) return INCLUS_MS;          // hit in plus group
      if (
         (is_required_token(pOelemCurrent->isMhit)) ||
         ((!pElemGiven) && (pCnttok->isTarget))
      ) {
         *ppElemReturned = pCnttok->pElem;              // let's caller know
         *piMexDepth = check_exceptions(pCnttok->pElem, oelemstk);
         if (pCnttok->isTarget) {
            *ppDtpReturned = (pCnttok+1)->pDtp;
         }
         next_token(pOelemCurrent, AFTER_HIT_NTK);
         if (*piMexDepth > 0) {
            return REQRDX_MS;
         }else {
            return REQRD_MS;
         }
      }
      /*--------------------------------------------------------+
      | The current primitive content token, that doesn't match |
      | "pElemGiven", is however not required: try jumping over, |
      | and loop if succesful.  But, first at all, check if     |
      | "pElemGiven" is on the red list (excluded).              |
      +--------------------------------------------------------*/
      if (*piMexDepth > 0) return EXCLUS_MS;         // Hit in minus group
      next_token(pOelemCurrent, NONE_NTK);

   }while (pOelemCurrent->msStatus == MATCH_MS);      // MATCH: loop
   return (e_StatusModele)pOelemCurrent->msStatus; // INVALID or FINISH: leave
}

/*--------------------------------------------------------------------mdl_fini-+
| Determine whether the current element can be ended, or whether               |
| non-optional tokens remain at the current level or higher.                   |
|                                                                              |
| Returns:                                                                     |
|    FINISH_MS  if element can be and has ended,                               |
|    REQRD_MS   if an element is required before to end this one,              |
|    INVALID_MS if element cannot be ended, due to a non-optional non-required |
|                                                                              |
| If the return code is REQRD_MS:                                              |
|    *ppElemReturned points the required ElementType                            |
|    If this is a "target elem" (4.315)                                         |
|      *ppDtpReturned points the data_tag pattern.                             |
|                                                                              |
+-----------------------------------------------------------------------------*/
e_StatusModele Modele::mdl_fini(
   const OpenElementStack & oelemstk,  // Stack of opened element
   const ElementType ** const ppElemReturned,
   const DataTagPattern ** const ppDtpReturned
)
{
   OpenElement * const pOelemCurrent = oelemstk[0];
   /*----------------------------------------------+
   | If the element is one of RCDATA, CDATA or ANY,|
   | or if the tcb rc shows the element is ended,  |
   | returns FINISH_MS.                            |
   | Note: the tag cannot be #PCDATA, nor EMPTY:   |
   |       a tcb is not created for those          |
   +----------------------------------------------*/
   if (
      (pOelemCurrent->pElem->declared_data_content() != DENY_DCDATA) ||
      (pOelemCurrent->pElem->is_any()) ||
      (pOelemCurrent->pElem->is_empty()) ||
      (pOelemCurrent->msStatus == FINISH_MS)
   ) {
      return FINISH_MS;
   }
   next_token(pOelemCurrent, FLUSH_NTK);
   if (pOelemCurrent->msStatus == FINISH_MS) return FINISH_MS;

   /*-----------------------------------------------------------+
   | To check if a required element is missing, call mdl_goto,  |
   | pretending a dummy element started, and check if REQ is on.|
   | Note that our dummy element must successfully pass the     |
   | "check_exception" test, or no check for REQ would be done: |
   | that's why the dummy element role is played by #PCDATA.    |
   +-----------------------------------------------------------*/
   pOelemCurrent->msStatus = MATCH_MS;
   switch (mdl_goto(oelemstk, 0, ppElemReturned, ppDtpReturned)) { // 0 is PCDATA
   case REQRD_MS:          // required element found
      return REQRD_MS;
   case REQRDX_MS:
      return REQRDX_MS;    // illegal attempt to exclude a required element
   default:
      return INVALID_MS;
   }
}

/*--------------------------------------------------------------------mdl_edtg-+
| End a datatag: asynchronous entry to next_token, after the content satisfied |
| a template.                                                                  |
| Trick: don't tell next_token "PCDATA" was hit: it was so, anyway,            |
| at the time the first character was entered (after the target) -             |
| this allows us to jump over the PCDATA (repeatable) token                    |
+-----------------------------------------------------------------------------*/
void Modele::mdl_edtg(OpenElement * const pOelemCurrent)
{
   next_token(pOelemCurrent, NONE_NTK);
}

/*PRIVATE----------------------------------------------include_plus_exceptions-+
| Turn on boolean of the list for elements that are plus exceptions            |
+-----------------------------------------------------------------------------*/
void Modele::include_plus_exceptions(
   AllowedElems *pAelem,
   const OpenElementStack & oelemstk   // Stack of opened element
)
{
   for (int i=0; i < pAelem->iCountElem; ++i) {
      if (pAelem->aisAllowed[i+1] == FALSE) {
         OpenElementStackIterator iterate(oelemstk);
         const OpenElement *pOelem;
         while (pOelem=iterate.prev(), pOelem) {
            InclusionListIterator iterate(pOelem->pElem);
            const ElementType *pElemInclusion;
            while ((pElemInclusion = iterate.next()) != 0) {
               if (pElemInclusion == pAelem->pElem + i) {
                  pAelem->aisAllowed[i+1] = TRUE;
               }
            }
         }
      }
   }
}

/*PRIVATE---------------------------------------------exclude_minus_exceptions-+
| Turn off boolean of the list for elements that are minis exceptions          |
+-----------------------------------------------------------------------------*/
void Modele::exclude_minus_exceptions(
   AllowedElems *pAelem,
   const OpenElementStack & oelemstk   // Stack of opened element
)
{
   for (int i=0; i < pAelem->iCountElem; ++i) {
      if (pAelem->aisAllowed[i+1] == TRUE) {
         OpenElementStackIterator iterate(oelemstk);
         const OpenElement *pOelem;
         while (pOelem=iterate.prev(), pOelem) {
            ExclusionListIterator iterate(pOelem->pElem);
            const ElementType *pElemExclusion;
            while ((pElemExclusion = iterate.next()) != 0) {
               if (pElemExclusion == pAelem->pElem + i) {
                  pAelem->aisAllowed[i+1] = FALSE;
               }
            }
         }
      }
   }
}

/*--------------------------------------------------------------------mdl_aelem-+
| Check what tags are allowed at some point                                    |
+-----------------------------------------------------------------------------*/
void Modele::mdl_aelem(
   const OpenElementStack & oelemstk,  // Stack of opened element
   AllowedElems *pAelem
)
{
   OpenElement * const pOelemCurrent = oelemstk[0];
   if (pOelemCurrent->pElem->is_any()) {
      memset(
         pAelem->aisAllowed, TRUE,
         (1+pAelem->iCountElem) * sizeof pAelem->aisAllowed[0]
      );
      exclude_minus_exceptions(pAelem, oelemstk);
      return;
   }else {
      memset(
         pAelem->aisAllowed, FALSE,
         (1+pAelem->iCountElem) * sizeof pAelem->aisAllowed[0]
      );
   }
   if (pOelemCurrent->pElem->is_empty()) return; // don't go further if EMPTY
   if (pOelemCurrent->msStatus != FINISH_MS) {
      do {
         OpenModelGroup *pGrpTok = aGrpStack + (ixGrpCurrent + 1);
         /*---------------------------------------------------+
         | Open enough model groups until we meet a primitive |
         | content token.  Have "pCnttok" point to it.        |
         +---------------------------------------------------*/
         while (pGrpTok->pCnttok->conn) {     // if current token is a group
            pGrpTok->hits = 0;                // reset all hits
            open_content_token(pGrpTok, 0);   // start at token 0
            ++pGrpTok, ++ixGrpCurrent;        // make it the current group
         }
         const ContentToken * const pCnttok = pGrpTok->pCnttok;
         /*----------------------------------------------+
         | Turn on the corresponding bit in boolean list |
         +----------------------------------------------*/
         if ((pCnttok->isDatatag) || (!pCnttok->pElem)) {
            pAelem->aisAllowed[0] = TRUE;      // PCDATA is allowed
         }else {
            for (int i=0; i < pAelem->iCountElem; ++i) {
               if (pAelem->pElem + i == pCnttok->pElem) {
                  pAelem->aisAllowed[i+1] = TRUE;
                  break;
               }
            }
         }
         /*-----------------------------------+
         | Stop at the first required element |
         +-----------------------------------*/
         if (is_required_token(pOelemCurrent->isMhit)) {
            next_token(pOelemCurrent, AFTER_HIT_NTK);
            break;
         }
         next_token(pOelemCurrent, NONE_NTK);
      }while (pOelemCurrent->msStatus == MATCH_MS);    // MATCH: loop
   }
   include_plus_exceptions(pAelem, oelemstk);
   exclude_minus_exceptions(pAelem, oelemstk);
   return;
}
/******************************************************************************/
