/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// oestack.c
///  Manages internal stack of open elements
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include "gendef.h"
#include "oestack.h"
#include "element.h"

/*-----------------------------------------------------------------OpenElement-+
| Regular constructor for OpenElement                                          |
+-----------------------------------------------------------------------------*/
OpenElement::OpenElement(
   const ElementType *pElemArg,
   const DataTagPattern *pDtpArg,
   const OpenElement *pOelemPrevious
) : mdlCurrent(pElemArg)
{
   pElem = pElemArg;
   pDtp = pDtpArg;
   pSrm = pElem->shortref_map();            // make it active
   if ((!pSrm) && (pOelemPrevious)) {       // If new element has no own MAP
      pSrm = pOelemPrevious->pSrm;          // inheritate from parent
   }
   msStatus = MATCH_MS;
   isNetEnabling = FALSE;
   if (pOelemPrevious) {
      isNetEnabled = pOelemPrevious->isNetEnabled;
      isZero = FALSE;
   }else {
      isNetEnabled = FALSE;
      isZero = TRUE;
   }
   isPex = FALSE;
   isMhit = FALSE;
   if (pDtp) {                             // Data Tag (set by mdl_goto)
      isDataTagEnabled = TRUE;             // Activate...
   }else if (pOelemPrevious) {              // else
      isDataTagEnabled = pOelemPrevious->isDataTagEnabled; // inheritate
   }else {
      isDataTagEnabled = FALSE;            // FALSE if no parent
   }
   isEmptyOrBad = FALSE;
}

/*------------------------------------------------------------OpenElementStack-+
| Default constructor.                                                         |
+-----------------------------------------------------------------------------*/
OpenElementStack::OpenElementStack() :
   iStkentMax(1),
   ixStkentTop(-1),
   pStkentTop(0),
   elemZero(&cnttokZero)
{
   push(&elemZero, 0);
}

/*------------------------------------------------------------OpenElementStack-+
| Copy constructor, to save a stack of opened elements                         |
+-----------------------------------------------------------------------------*/
OpenElementStack::OpenElementStack(const OpenElementStack& source) :
   iStkentMax(source.iStkentMax),
   ixStkentTop(-1),
   pStkentTop(0),
   elemZero(&cnttokZero)
{
   int i = -source.ixStkentTop;
   while (i <= 0) push(source[i++]);
}

/*-----------------------------------------------------------~OpenElementStack-+
| Destructor                                                                   |
+-----------------------------------------------------------------------------*/
OpenElementStack::~OpenElementStack()
{
   while (pop());
}

/*--------------------------------------------------------------------activate-+
| Initialize the stack for a given DTD and TAGLVL quantity                     |
+-----------------------------------------------------------------------------*/
void OpenElementStack::activate(ElementType *pElemBase, int iMaxEntries)
{
   iStkentMax = 1 + iMaxEntries;
   cnttokZero.pElem = pElemBase;
}

/*-------------------------------------------------------------------operator=-+
| Assignment by copy, to restore a stack of opened elements                    |
+-----------------------------------------------------------------------------*/
OpenElementStack& OpenElementStack::operator=(
   const OpenElementStack& source
)
{
   if (this != &source) {
      while (pop());
      int i = -source.ixStkentTop;
      while (i <= 0) push(source[i++]);
   }
   return *this;
}

/*------------------------------------------------------------------------push-+
| Push a new entry onto the stack.  Return FALSE if overflow.                  |
+-----------------------------------------------------------------------------*/
OpenElement * OpenElementStack::push(
   const ElementType *pElem, const DataTagPattern *pDtp
)
{
   if ((ixStkentTop + 1) >= iStkentMax)  return 0;  // Overflow!

   StackEntry *pStkent = new StackEntry(
      pElem, pDtp, pStkentTop? &pStkentTop->oelem : 0
   );

   ++ixStkentTop;
   pStkent->previous = pStkentTop;
   pStkentTop = pStkent;
   return &pStkentTop->oelem;
}

/*------------------------------------------------------------------------push-+
| Push an OpenElement                                                          |
+-----------------------------------------------------------------------------*/
OpenElement * OpenElementStack::push(const OpenElement *pSource)
{
   if ((ixStkentTop + 1) >= iStkentMax)  return 0;       // Overflow!

   StackEntry *pStkent = new StackEntry(pSource);

   ++ixStkentTop;
   pStkent->previous = pStkentTop;
   pStkentTop = pStkent;
   return &pStkentTop->oelem;
}

/*-------------------------------------------------------------------------pop-+
| Pop an entry from the stack.  Return FALSE if underflow.                     |
+-----------------------------------------------------------------------------*/
boolean OpenElementStack::pop()
{
   if (ixStkentTop < 0) return FALSE;  // Underflow!

   StackEntry *pStkent = pStkentTop;

   pStkentTop = pStkent->previous;
   --ixStkentTop;
   delete pStkent;
   return TRUE;
}

/*------------------------------------------------------------------operator[]-+
| Get a pointer to the (last-in - ix) stack entry                              |
+-----------------------------------------------------------------------------*/
OpenElement * OpenElementStack::operator[](int ix) const
{
   StackEntry *pStkent = pStkentTop;

   while ((pStkent) && (ix++ < 0)) {
      pStkent = pStkent->previous;
   }
   if (pStkent) return &(pStkent->oelem);
   return 0;
}

/*----------------------------------------------------OpenElementStackIterator-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
OpenElementStackIterator::OpenElementStackIterator(
   const OpenElementStack& oestk
) :
   pStkent(oestk.pStkentTop)
{
}

/*-----------------------------------------------------------------------first-+
| Get a pointer to the first stack entry                                       |
+-----------------------------------------------------------------------------*/
OpenElement * OpenElementStack::first() const  // get first value
{
   return (*this)[-ixStkentTop];
}

/*------------------------------------------------------------------------prev-+
| Iterator to the previous stack entry                                         |
+-----------------------------------------------------------------------------*/
const OpenElement * OpenElementStackIterator::prev() {
   const OpenElementStack::StackEntry& stkentReturn = *pStkent;

   if (pStkent) {
      pStkent = pStkent->previous;
      return &stkentReturn.oelem;
   }
   return 0;
}
/******************************************************************************/
