/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// parser.c
///  Parse an SGML document and report events (part 2 of 2)
//   Pierre G. Richard 93/10/05 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "gendef.h"
#include "rsvname.h"
#include "parser.h"
#include "event.h"
#include "element.h"
#include "quantity.h"
#include "capacity.h"
#include "hash.h"
#include "enthdlr.h"
#include "prsmode.h"
#include "syntax.h"
#include "sgmldcl.h"
#include "dtddcl.h"
#include "procpil.h"
#include "utilctl.h"

static boolean tvrs_check_idref(void *, const char *pszName, void *piElemId);

/*------------------------------------------------------------------SGMLparser-+
| This constructor is used to create a new parser session.                     |
+-----------------------------------------------------------------------------*/
SGMLparser::SGMLparser(const Document& doc) : ParseState(doc)
{
}

/*------------------------------------------------------------------SGMLparser-+
| This constructor is used to create another parser session from a given       |
| ParseState (to resume from)                                                  |
+-----------------------------------------------------------------------------*/
SGMLparser::SGMLparser(const ParseState& prsctxt) : ParseState(prsctxt)
{
}

/*PRIVATE----------------------------------------------------copy_context_into-+
| This function copies the current ParseState of a Parser                      |
+-----------------------------------------------------------------------------*/
void SGMLparser::copy_context_into(ParseState&) const
{
   //***>NOTE: Not yet implemented
}

/*----------------------------------------------------------------check_events-+
| Do a final check of the validity of the events.                              |
+-----------------------------------------------------------------------------*/
void SGMLparser::check_events(const int iCountStart)
{
   evlst.position(iCountStart);
   for (Event *pPrmev = (Event *)evlst; pPrmev; pPrmev = ++evlst) {
      ((PrimEvent *)pPrmev)->validate(this);
   }
}

/*-----------------------------------------------------------record_elem_start-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void SGMLparser::record_elem_start(
   const ElementType *pElem,
   const AttributeList& attlist,
   const DataTagPattern *pDtp,
   const e_InferBy inferby,
   const e_Minimize minias,
   boolean isInclusion,
   boolean isExplConref
)
{
   oelemstk[0]->isEmptyOrBad = FALSE;       // previous element wasn't empty

   OpenElement *pOelem = oelemstk.push(pElem, pDtp);

   if (!pOelem) {                           // no stack space left!
      errorh(BIGTAGLVL, pEh, ERH_T_SEV, pElem->name());
   }
   if (isInclusion) {                      // Included element
      pOelem->isPex = TRUE;
      rcdstate.declare_start_included_elem();
   }else {                                 // Proper Element
      if (rcdstate.declare_start_proper_elem()) {
         evlst.append(RecordEnd());
      }
   }
   evlst.append(StartElement(
         pElem, attlist, inferby, minias, isInclusion, isExplConref
      )
   );
   if ((pElem->is_empty()) || (isExplConref)) {
      /*
      | If the element is empty, end it now,
      | so it won't appear on the stack of open elements.
      */
      const ElementType *pElemReturned;
      const DataTagPattern *pDtpReturned;

      record_elem_end(&pElemReturned, &pDtpReturned, inferby);
   }else {
      if (inferby != NO_IMF) {             // if inferred,
         pOelem->isEmptyOrBad = TRUE;       // it's bad if empty
      }
   }
}

/*-------------------------------------------------------------record_elem_end-+
| Returns FALSE if unable to, a required element being needed.                 |
+-----------------------------------------------------------------------------*/
boolean SGMLparser::record_elem_end(
   const ElementType **ppElemReturned,
   const DataTagPattern **ppDtpReturned,
   const e_InferBy inferby,
   const e_Minimize minias
)
{
   boolean isPremature = FALSE;
   boolean isBadEmpty = FALSE;
   OpenElement *pOelem = oelemstk[0];

   switch (pOelem->mdlCurrent.mdl_fini(oelemstk, ppElemReturned, ppDtpReturned)) {
   case REQRDX_MS:                         // invalid exclusion found
      errorh(INVMEX, pEh, ERH_E_SEV, (*ppElemReturned)->name());
      /* fall thru */
   case REQRD_MS:
      return FALSE;
   case INVALID_MS:
      isPremature = TRUE;
      break;                               // end it anyway... is that wrong?
   default: /* case FINISH_MS: */
      break;
   }
   if (pOelem->isPex) {
      rcdstate.declare_end_included_elem();
   }else {
      rcdstate.declare_end_proper_elem();
      isBadEmpty = pOelem->isEmptyOrBad;    // omitted and empty
   }
   evlst.append(EndElement(
         pOelem->pElem, inferby, minias, isPremature, isBadEmpty
      )
   );
   pElemLastEnded = pOelem->pElem;
   oelemstk.pop();
   return TRUE;
}

/*----------------------------------------------------validate_element_context-+
| Validate the element pElemFound within the current model.                    |
| Return INVALID_MS if the element is out of context with this model,          |
| Return INCLUS_MS or MATCH_MS if it's OK, pDtpReturned points the             |
| data tag pattern in case ElemFound is a target element.                      |
+-----------------------------------------------------------------------------*/
e_StatusModele SGMLparser::validate_element_context(
    const ElementType *pElemFound,
    const DataTagPattern **ppDtpReturned,
    const e_InferBy inferby
)
{
   for (;;) {
      int mex_depth;                    // to catch the mex depth
      const ElementType *pElemReturned;

      switch (
         oelemstk[0]->mdlCurrent.mdl_goto(
           oelemstk, pElemFound, &pElemReturned, ppDtpReturned, &mex_depth
        )
      ) {
      case REQRDX_MS:                   // New required is invalid exclusion
         errorh(INVMEX, ERH_E_SEV, pElemReturned->name());
         /* fall thru */
      case REQRD_MS:
         /*-----------------------------------------------------------+
         | Another element is required.  If this element is DCDATA,   |
         | this must be CDATA or RDDATA (#PCDATA is never required).  |
         | In this case start and end it, so the current (invalid)    |
         | ElemFound is recognized.  This is more a recovery strategy,|
         | since start tag minimization of DCDATA is prohibited       |
         +-----------------------------------------------------------*/
         switch (pElemReturned->declared_data_content()) {
         case NONR_DCDATA:
         case RPLC_DCDATA:                    // see 7.3.1.1
            record_elem_start(
               pElemReturned, AttributeList(pElemReturned),
               *ppDtpReturned, inferby
            );
            record_elem_end(&pElemReturned, ppDtpReturned, NO_IMF);
            break;
         default:                             // i.e DENY_DCATA
            record_elem_start(
                pElemReturned, AttributeList(pElemReturned),
                *ppDtpReturned, inferby
            );
            break;
         }
         break;                               // and try again
      case FINISH_MS:
         record_elem_end(&pElemReturned, ppDtpReturned, inferby);
         if (oelemstk[0]->isZero) {            // Are we still alive?
            return INVALID_MS;                // Nope: invalid!
         }
         break;                               // else, try again
      case EXCLUS_MS:                         // cannot be for #PCDATA!
         while (--mex_depth >= 0) {           // imply end for depth mex elems
            if (!record_elem_end(&pElemReturned, ppDtpReturned, inferby)) {
               record_elem_start(
                  pElemReturned, AttributeList(pElemReturned),
                  *ppDtpReturned, inferby
               );
               break;
            }
         }
         if (oelemstk[0]->isZero) {            // Are we still alive?
            return INVALID_MS;                // Nope: invalid!
         }
         break;                               // else, try again
      case MATCHX_MS:                         // OK, but invalid exclusion
         errorh(INVMEX, ERH_E_SEV, pElemFound->name());
         return MATCH_MS;
      case INCLUS_MS:
         return MATCH_MS;
      case MATCH_MS:
         return INCLUS_MS;
      case INVALID_MS:
         return INVALID_MS;
      }
   }
}

/*----------------------------------------------------------start_base_element-+
| Imply the start-tag for the document element.                                |
|                                                                              |
| Given [10], [11], [12] and [13], if the document instance set does not       |
| start with a STAGO or a STAGO_TAGC, the start-tag for the document element   |
| must be implied.                                                             |
+-----------------------------------------------------------------------------*/
void SGMLparser::start_base_element()
{
   const ElementType *pElemReturned;
   const DataTagPattern *pDtpReturned;

   OpenElement *pOelemZero = oelemstk.first();
   pOelemZero->mdlCurrent.mdl_goto(               // Goto base element
       oelemstk, pDtddcl->pElemBase,              // always returns MATCH_MS
       &pElemReturned, &pDtpReturned
   );

   const int iCountStart = evlst.count();         // next? event to stack
   record_elem_start(
      pElemReturned, AttributeList(pElemReturned),
      pDtpReturned, DOC_IMF
   );
   check_events(iCountStart);
   adjust_recognition_mode(pElemReturned->declared_data_content());
}

/*---------------------------------------------------------------start_element-+
| Register a change in the tag structure, when opening new elements.           |
|                                                                              |
| If something's wrong, restore the element-context and return FALSE.          |
| If everything is fine, return TRUE.                                          |
+-----------------------------------------------------------------------------*/
boolean SGMLparser::start_element()
{
   const ElementType *pElemFound;
   unsigned long ulOfsStart;

   /*----------------------------------------+
   | Have pElemFound describing the start-tag |
   +----------------------------------------*/
   switch (get_curr_token()) {
   case STAGO_TK:
      ulOfsStart = pEh->offset();             // Remember start offset
      {
         char pszName[NAMELEN_MAX+1];         // Get the generic identifier

         pSyntax->read_general_name(pEh, pszName);
         pElemFound = pDtddcl->search_element_type(pszName);

         if (!pElemFound) {
            errorh(INVELEMGI, pEh, ERH_E_SEV, pszName);
            finish_parse_tag(0);              // Bad: get rid of any attrib
            rcdstate.reset();                 // discard RE if any
            evlst.append(Error());            // Tag skipped
            return FALSE;                     // Tag was invalid
         }
      }
      break;
   default:                                   // case STAGO_TAGC_TK "<>"
      /*-------------------------------------------------+
      | By 7.4.1.1, a GI specification is implied for an |
      | empty start tag before to determine whether      |
      | any tags were omitted before it                  |
      | a) if (OMITTAG==YES) <> == most recently started |
      | b) if (OMITTAG==NO)  <> == most recently ended   |
      | c) or the base document element                  |
      +-------------------------------------------------*/
      if (pSgmldcl->features.isOmittag) {     // 7.4.1.1 a)
         OpenElement *pOelem = oelemstk[0];
         if (!pOelem->isZero) {
            pElemFound = pOelem->pElem;
            break;
         }
      }else {
         if ((pElemFound = pElemLastEnded) != 0) { // 7.4.1.1 b)
            break;
         }
      }
      pElemFound = pDtddcl->pElemBase;        // 7.4.1.1 c)
      break;
   }

   boolean isInclusion = FALSE;
   const int iCountStart = evlst.count();     // next? event to stack
   const DataTagPattern *pDtpReturned;
   {
      /*----------------------------------------+
      | Before to check if the element is valid |
      | within the current model, save the      |
      | element-context to allow backup.        |
      +----------------------------------------*/
      OpenElementStack oelemstkSave(oelemstk);  // save element stack,
      RecordState rcdstateSave(rcdstate);       // record state,
      const ElementType *pElemLastEndedSave = pElemLastEnded; //Last ended element

      e_StatusModele ms;
      int iTimes = 0;

      while (
         ms = validate_element_context(pElemFound, &pDtpReturned, STG_IMF),
         ms == INVALID_MS
      ) {
         /*--------------------------------------------------+
         | This start-tag is definitively invalid:           |
         | Restore the open element stack, the record state, |
         | last ended element.  Destroy all events pushed    |
         | since we started                                  |
         +--------------------------------------------------*/
         oelemstk = oelemstkSave;
         rcdstate = rcdstateSave;
         pElemLastEnded = pElemLastEndedSave;
         for (evlst.position(iCountStart); evlst.remove(); evlst.next());
         /*----------------------------------------------------+
         | Last chance to recover before to definitively get   |
         | rid of the tag: If any RE was pending, they could   |
         | have inferred bad tags: suppress the pending RE and |
         | RE and restart the loop.  If invalid, RecordState   |
         | reflects the state reached after the start of a     |
         | proper element, which has been prooved to be the    |
         | less harmfull condition after such an error.        |
         +----------------------------------------------------*/
         if (
            (!rcdstate.declare_start_proper_elem()) || //no RE pending
            (++iTimes >= 2)                   // or already done
         ) {
            errorh(OOCSTG, pEh, ERH_E_SEV);   // unrecoverable...
            evlst.append(Error());
            if (get_curr_token() != STAGO_TAGC_TK) {
               finish_parse_tag(0);           // Get rid of any attributes
            }
            return FALSE;
         }
      }
      if (ms == INCLUS_MS) isInclusion = TRUE;
   }
   /*-----------------------------------------------------+
   | Element is valid: either MATCH_MS or INCLUS_MS.      |
   | For a non-empty start_tag, parse the attributes.     |
   | Note that "finish_parse_tag" changes curr_token.     |
   +-----------------------------------------------------*/
   if (get_curr_token() == STAGO_TAGC_TK) {
      record_elem_start(
         pElemFound, AttributeList(pElemFound),
         pDtpReturned, NO_IMF, EMPTY_ST_MNF, isInclusion, FALSE
      );
   }else {
      const AttributeList attlist(pElemFound);
      boolean isExplConref;

      if (finish_parse_tag(&attlist)) {     // CONREF specified
         isExplConref = TRUE;
      }else {
         isExplConref = FALSE;
      }

      int iTagLength = (int)(pEh->offset() - ulOfsStart);
      e_Minimize minias;

      switch (get_curr_token()) {
      case NET_TK:
         minias = NET_ST_MNF;
         if (!pSgmldcl->features.isShorttag) { // rule 7.4.1
            errorh(NOTAGC, pEh, ERH_E_SEV);
         }
         iTagLength -= get_curr_count();
         break;
      case TAGC_TK:
         minias = NO_MNF;
         iTagLength -= get_curr_count();
         break;
      default:
         minias = UNCL_ST_MNF;
         break;
      }
      if (iTagLength > pSyntax->pQuantity->value_of(
            TAGLEN_QTY_ix
         )
      ) {
         errorh(BIGTAGLEN, pEh, ERH_E_SEV, iTagLength);
      }
      record_elem_start(
         pElemFound, attlist,
         pDtpReturned, NO_IMF, minias, isInclusion, isExplConref
      );
      if (minias == NET_ST_MNF) {
         if ((pElemFound->is_empty()) || (isExplConref))  {   // if  empty
            errorh(INVMTYNET, pEh, ERH_E_SEV);
         }else {
            oelemstk[0]->isNetEnabling = TRUE;
            oelemstk[0]->isNetEnabled = TRUE;
         }
      }
   }
   check_events(iCountStart);
   adjust_recognition_mode(pElemFound->declared_data_content());
   isNetEnabled = oelemstk[0]->isNetEnabled;
   return TRUE;
}

/*------------------------------------------------------------prepare_for_data-+
| Set up a convenient OpenElement structure before to enter DATA               |
| (including DATA RE)                                                          |
|                                                                              |
|  1) If DATA is OOC, do not unstack the "tells" to Application                |
|     and restore the environment (struct t_COMMON) as it was on entry.        |
|  2) Otherwise if DATA is OK (MATCH_MS)                                       |
|     unstack the "tells", and return TRUE                                     |
|                                                                              |
|  NOTE: Within this function the CON mode is always set.                      |
|        If on entry the current active element is CDATA or RCDATA,            |
|        return TRUE immediately.                                              |
|        If a DCDATA tag is required, it is opened and closed immediately,     |
|        in conformance with 7.3.1.1 a).                                       |
|        Therefore, there is no need to "adjust_rcgn_mode".                    |
+-----------------------------------------------------------------------------*/
boolean SGMLparser::prepare_for_data()
{
   if (!oelemstk[0]->pElem->declared_data_content()) {
      /*-----------------------------------+
      | Before to check if DATA is valid   |
      | within the current model, save the |
      | element-context to allow backup.   |
      +-----------------------------------*/
      OpenElementStack oelemstkSave(oelemstk);  // save element stack,
      RecordState rcdstateSave(rcdstate);       // record state,
      const ElementType *pElemLastEndedSave = pElemLastEnded; // last ended element
      int iCountStart = evlst.count();          // events currently stacked
      const DataTagPattern *pDtpReturned;

      if (validate_element_context(0, &pDtpReturned, DATA_IMF) == INVALID_MS) {
         /*-------------------------------------------------+
         | This end-tag is definitively invalid:            |
         | Restore the open elemnt stack, the record state, |
         | last ended element.  Destroy all events pushed   |
         | since we started and return FALSE.               |
         +-------------------------------------------------*/
         oelemstk = oelemstkSave;
         rcdstate = rcdstateSave;
         pElemLastEnded = pElemLastEndedSave;
         for (evlst.position(iCountStart); evlst.remove(); evlst.next());
         return FALSE;
      }
      check_events(iCountStart);
      isNetEnabled = oelemstk[0]->isNetEnabled;
   }
   return TRUE;                               // Data is now OK
}

/*-----------------------------------------------------------------end_element-+
| Issue iCountElementToEnd times "record_elem_end".                            |
| The iCountElementToEnd-1 first times, tye element is inferred (by inferby).  |
| iCountElementToEnd==0 is special for ETAGO_TAGC                              |
|                                                                              |
| Note: inferred enum is one of ETG_IMF, EDTG_IMF, EOF_IMF                     |
+-----------------------------------------------------------------------------*/
void SGMLparser::end_element(
   unsigned int iCountElementToEnd,
   e_InferBy inferby,
   e_Minimize minias
)
{
   const ElementType *pElemReturned;
   const DataTagPattern *pDtpReturned;
   const int iCountStart = evlst.count();      // events currently stacked

   if (!iCountElementToEnd) {                  // ETAGO_TAGC
      /*
      | end elements, until one (current, or inferred) element-end worked
      */
      while (!record_elem_end(
         &pElemReturned, &pDtpReturned, NO_IMF, minias)
      ) {
         record_elem_start(
            pElemReturned, AttributeList(pElemReturned), pDtpReturned, inferby
         );
      }
   }else do {
      boolean isSuccessfulEnd;

      if (iCountElementToEnd > 1) {
         isSuccessfulEnd = record_elem_end(
            &pElemReturned, &pDtpReturned, inferby
         );
      }else {
         isSuccessfulEnd = record_elem_end(
            &pElemReturned, &pDtpReturned, NO_IMF, minias
         );
      }
      if (!isSuccessfulEnd) {
         record_elem_start(
            pElemReturned, AttributeList(pElemReturned), pDtpReturned, inferby
         );
         ++iCountElementToEnd;               // count one more
      }else {
         --iCountElementToEnd;
      }
   }while (iCountElementToEnd && !oelemstk[0]->isZero);
   adjust_recognition_mode(oelemstk[0]->pElem->declared_data_content());
   check_events(iCountStart);
   isNetEnabled = oelemstk[0]->isNetEnabled;
}

/*------------------------------------------------------------how_many_NET_end-+
| Return the count of element a valid trapped NET would end.                   |
+-----------------------------------------------------------------------------*/
int SGMLparser::how_many_NET_end()
{
   OpenElementStackIterator iterate(oelemstk);
   const OpenElement *pOelem;

   for (int i=1; pOelem = iterate.prev(), pOelem->isNetEnabling; ++i);
   return i;
}

/*--------------------------------------------------------how_many_element_end-+
| Check the validity of an end tag - returns 0 if invalid or the               |
| count of element the end tag is ending.                                      |
+-----------------------------------------------------------------------------*/
int SGMLparser::how_many_element_end()
{
   const ElementType *pElemFound;
   int ecnt;

   /*--------------------------------------------------------+
   | Set ecnt to the # of times "end_element" must be issued |
   +--------------------------------------------------------*/
   {
      char pszName[NAMELEN_MAX+1];          // to get the generic identifier

      pSyntax->read_general_name(pEh, pszName);
      pElemFound = pDtddcl->search_element_type(pszName);
   }
   if (!pElemFound) {                       // Unknown element
      ecnt = 0;
   }else {
      /*-------------------------------------------------+
      | Find out if End-tag matches any tag on the stack |
      +-------------------------------------------------*/
      OpenElementStackIterator iterate(oelemstk);
      const OpenElement *pOelem;

      for (ecnt=1; pOelem=iterate.prev(), pOelem->pElem != pElemFound; ++ecnt) {
         if (pOelem->isZero) {              // No match found
            ecnt = 0;                       // invalid end
            break;
         }
      }
   }
   if (!ecnt) {                             // invalid end tag
      if (oelemstk[0]->pElem->declared_data_content()) {  // RCDATA or CDATA
         errorh(FAKEEND, pEh, ERH_E_SEV);   // 7.6 p.24, last paragraph
         ecnt = 1;                          // so it ends ones
      }else {
         errorh(IGNOREEND, pEh, ERH_E_SEV);
      }
      evlst.append(Error());
   }
   finish_parse_tag(0);                     // jump to the end of the end-tag
   if (get_curr_token() == NET_TK) errorh(NOTAGC, pEh, ERH_E_SEV);
   return ecnt;
}

/*--------------------------------------------------------------parse_instance-+
| Main loop when parsing the document instance.                                |
| Return TRUE if parsing can pursue.                                           |
+-----------------------------------------------------------------------------*/
boolean SGMLparser::parse_instance()
{
   switch (recognize_delimiter()) {

   case NOT_FOUND_TK:                            // PCDATA/DCDATA (C or RC)
      if (pEh->is_Ee()) {                        // not a char: system signal
         return process_Ee();
      }else {
         return process_data();
      }

   case ERO_TK:                                  // &
      if ((oelemstk[0]->isZero) && (oelemstk[0]->isMhit)) {
         errorh(OUTDOCDAT, pEh, ERH_E_SEV);
         evlst.append(Error());
         return FALSE;
      }
      process_entity(open_general_entity_reference());
      return TRUE;

   case PIO_TK:                                  // <?
      rcdstate.declare_other_markup_dcl();       // Register new state
      evlst.append(ProcessingInstruction(pEh, pSyntax));
      return TRUE;

   case CRO_TK:                                  // &# (character reference)
      if ((oelemstk[0]->isZero) && (oelemstk[0]->isMhit)) {
         errorh(OUTDOCDAT, pEh, ERH_E_SEV);
         evlst.append(Error());
         return FALSE;
      }
      pSyntax->translate_character_ref(pEh, get_curr_count());
      return TRUE;                               // (be back soon: no event!)

   case STAGO_TK:                                // STAGO      "<"
   case STAGO_TAGC_TK:                           // STAGO_TAGC "<>"
      if ((oelemstk[0]->isZero) && (oelemstk[0]->isMhit)) {
         errorh(OUTDOCSTG, pEh, ERH_E_SEV);
         evlst.append(Error());                  // tell application
         return FALSE;
      }
      pst1 = TAG_PST1;
      start_element();
      return TRUE;

   case NET_TK:                                  // NET "/"
   case ETAGO_TK:                                // ETAGO      "</"
   case ETAGO_TAGC_TK:                           // ETAGO_TAGC "</>"
      if ((oelemstk[0]->isZero) && (oelemstk[0]->isMhit)) {
         errorh(OUTDOCETG, pEh, ERH_E_SEV);
         evlst.append(Error());                  // tell application
         return FALSE;
      }
      pst1 = TAG_PST1;
      switch (get_curr_token()) {
      case ETAGO_TAGC_TK:
         end_element(0, ETG_IMF, EMPTY_ET_MNF);
         break;
      case NET_TK:
         end_element(how_many_NET_end(), ETG_IMF, NULL_ET_MNF);
         break;
      case ETAGO_TK:
         {
            int ecnt = how_many_element_end();

            if (ecnt > 0) {
               e_Minimize minias;
               if (get_curr_token() == TAGC_TK) {
                  minias = NO_MNF;
               }else {
                  minias = UNCL_ET_MNF;
               }
               end_element(ecnt, ETG_IMF, minias);
            }
         }
         break;
      }
      return TRUE;

   case MDO_COM_TK:                              // <!--
      rcdstate.declare_other_markup_dcl();
      evlst.append(Comment());                   // Comment skipped
      skip_cdc();                                // does read
      return TRUE;

   case MDO_MDC_TK:                              // <!>
      rcdstate.declare_other_markup_dcl();
      evlst.append(Comment());
      return TRUE;

   case MDO_DSO_TK:                              // <![
      if ((oelemstk[0]->isZero) && (oelemstk[0]->isMhit)) {
         errorh(OUTDOCDAT, pEh, ERH_E_SEV);
         evlst.append(Error());
         return FALSE;
      }
      if (!marked_section_open()) {
         evlst.append(Error());
         marked_section_ignore();                // MSC_MDC reached
      }else {
         switch (pPrsmode->get_mode()) {
         case FCDMS_MODE:
            evlst.append(SectionStart(SectionStart::CDATA));
            break;
         case RCDMS_MODE:
            evlst.append(SectionStart(SectionStart::RCDATA));
            break;
         case IGNMS_MODE:
            evlst.append(SectionStart(SectionStart::IGNORE));
            marked_section_ignore();             // MSC_MDC reached
            evlst.append(SectionEnd());
            marked_section_close();
            break;
         default:                                // INCLUDE or TEMP
            evlst.append(SectionStart(SectionStart::INCLUDE));
            break;
         }
         rcdstate.declare_other_markup_dcl();    // Register new state
      }
      return TRUE;

   case MSC_MDC_TK:                              // ]]>
      if ((oelemstk[0]->isZero) && (oelemstk[0]->isMhit)) {
         errorh(OUTDOCDAT, pEh, ERH_E_SEV);
         evlst.append(Error());
         return FALSE;
      }
      rcdstate.declare_other_markup_dcl();
      if (!marked_section_close()) {             // Too many ]]>  ...
         evlst.append(Error());                  // Skipped...
      }else {
         evlst.append(SectionEnd());
      }
      return TRUE;

   case MDO_TK:                                  // <!USEMAP
      if ((oelemstk[0]->isZero) && (oelemstk[0]->isMhit)) {
         errorh(OUTDOCDAT, pEh, ERH_E_SEV);
         evlst.append(Error());
         return FALSE;
      }
      rcdstate.declare_other_markup_dcl();
      process_usedcl();
      return TRUE;                               // (be back soon: no event!)

   case ERO_GRPO_TK:
      errorh(NIMPL, pEh, ERH_T_SEV,1001);        // CONCUR/LINK
      return FALSE;

   case ETAGO_GRPO_TK:
   case STAGO_GRPO_TK:
      errorh(NIMPL, pEh, ERH_T_SEV,1004);        // CONCUR
      return FALSE;

   default:                                      // SHORTREF
      {
         /*----------------------------------------+
         | SHORTREFS: If a current map is activated|
         | AND the shortref number is not bigger   |
         | than the one the DTD was compiled with  |
         | AND the shortref is mapped...           |
         +----------------------------------------*/
         EntityDef *pEnt;
         int  iIdSr;
         if (
            (oelemstk[0]->pSrm) &&                // If a current map exists
            (
               iIdSr = (int)get_curr_token()-(int)SR_first,
               iIdSr < pDtddcl->iCountShortref   // and a valid iIdSr
            ) &&
            (
              pEnt = oelemstk[0]->pSrm->entity_def(iIdSr),
              pEnt != 0                          // and mapped
            )
         ) {
            if (!pEh->open(pEnt)) pEnt = 0;
            process_entity(pEnt);
            return TRUE;
         }
      }
      iCountForcedData = get_curr_count();
      pEh->unread(iCountForcedData);
      return process_data();                     // see 9.4.6
   }
}

/*----------------------------------------------------------------process_data-+
| Process data: either because it's not a delimiter (NOT_FOUND_TK) or          |
| this is an unmapped SHORTREF (9.4.6)                                         |
| Return TRUE if parsing can pursue.                                           |
+-----------------------------------------------------------------------------*/
boolean SGMLparser::process_data()
{
   Data data(255);

   do {
      switch (pst1) {

      case TAG_PST1:
         if (pEh->is_Ee()) return TRUE;         // not a char: system signal
         if ((pEh->is_separator()) && (!pEh->is_numchar())) {
            /*-------------------------------+
            |An RS that is not interpreted as|
            |markup is ignored  (7.6.1)      |
            +-------------------------------*/
            if (pEh->is_RSchar()) {              // never generates RE
               rcdstate.declare_RS();
               break;
            }
            /*-----------------------------------------+
            | Even an RE is recognized as DATA, before |
            | to consider if it can be ignored.        |
            +-----------------------------------------*/
            if (
               (!oelemstk[0]->pElem->is_pcdata()) && // Neither mixed content
               (!oelemstk[0]->pElem->declared_data_content()) // nor declared
            ) {
               break;                            // separator ignored
            }
         }else {                                 // Not a separator
            if ((oelemstk[0]->isZero) && (oelemstk[0]->isMhit)) {
               errorh(OUTDOCDAT, pEh, ERH_E_SEV);
               evlst.append(Error());
               pst1 = INVDATA_PST1;              // Data out of document
               return FALSE;
            }
         }
         if (!prepare_for_data()) {              // cannot enter data here
            if ((pEh->is_separator()) && (!pEh->is_numchar())) {
               errorh(OOCSEP, pEh, ERH_E_SEV);
            }else {
               errorh(OOCDATA, pEh, ERH_E_SEV);
            }
            evlst.append(Error());
            pst1 = INVDATA_PST1;
            return TRUE;
         }
         pst1 = DATA_PST1;                       // Valid character
         /* Fall thru */

      case DATA_PST1:                            // Mother state
      case DATA_STOWED_PST1:
         if (pEh->is_Ee()) {                     // not a char: system signal
            if (pst1 == DATA_STOWED_PST1) {
               evlst.append(data);
               pst1 = DATA_PST1;
            }
            return TRUE;
         }
         if ((pEh->is_nonsgml()) && (!pEh->is_numchar())) {
            if (pst1 == DATA_STOWED_PST1) {
               evlst.append(data);
            }
            evlst.append(Error());               // tell application
            pst1 = DATA_DISCARD_CHAR_PST1;
            return TRUE;
         }
         if (!pEh->is_numchar()) {               // if *not* a numeric char ref
            if (pEh->is_RSchar()) {              // never generates RE
               rcdstate.declare_RS();            // Change for RS state
               break;                            // Discard data
            }
            if (pEh->is_REchar()) {
               if (!rcdstate.declare_RE()) {     // RE state
                  break;                         // RE ignored
               }
               if (pst1 == DATA_STOWED_PST1) {
                  evlst.append(data);
               }else {
                  oelemstk[0]->isEmptyOrBad = FALSE; // Element content not empty
               }
               evlst.append(RecordEnd());        // Generate RE
               pst1 = DATA_DISCARD_CHAR_PST1;    // This one is discarded
               return TRUE;
            }
            if ((pEh->is_ms_char()) && (!iCountForcedData)) {
               if (pEh->is_msochar()) {          // MSO:
                  pPrsmode->disable_markup_recognition();
               }else if (pEh->is_msichar()) {    // MSI, orphan or not
                  pPrsmode->enable_markup_recognition();   // restore anyway
               }else {                           // MSS: process directly
                  iCountForcedData = 2;          // i.e MSS + next char
               }
            }
         }
         if (rcdstate.declare_data()) {          // register new state
            if (pst1 == DATA_STOWED_PST1) {
               evlst.append(data);
            }else {
               oelemstk[0]->isEmptyOrBad = FALSE; // Element content not empty
            }
            evlst.append(RecordEnd());           // RE inserted
            pst1 = DATA_PENDING_CHAR_PST1;
            return TRUE;
         }
         if (pst1 != DATA_STOWED_PST1) {
            oelemstk[0]->isEmptyOrBad = FALSE;   // Element content not empty
         }
         /* Fall thru */

      case DATA_PENDING_CHAR_PST1:
         switch (data.put_char(                  // enter curr char
            (char)pEh->curr_char_val())
         ) {
         case Data::OK_RCPC:                     // sucessfully entered
            pst1 = DATA_STOWED_PST1;
            break;
         default:                                // buffer full up
            pst1 = DATA_DISCARD_CHAR_PST1;
            evlst.append(data);
            return TRUE;
         }
         break;

      case DATA_DISCARD_CHAR_PST1:
         pst1 = DATA_PST1;
         break;

      default: /* INVDATA_PST1: */               // Invalid data (mother state)
         if (pEh->is_Ee()) return TRUE;          // not a char: system signal
         break;
      }
      pEh->read();

   }while (
      (
         (iCountForcedData) && (--iCountForcedData > 0)
      ) || (
         !may_be_delimiter()
      )
   );
   if (pst1 == DATA_STOWED_PST1) {
      evlst.append(data);
      pst1 = DATA_PST1;
   }
   return TRUE;
}

/*--------------------------------------------------------------process_entity-+
| Process an entity found in contents                                          |
| Note: open is smart and does not open non-parseable entities                 |
+-----------------------------------------------------------------------------*/
void SGMLparser::process_entity(EntityDef *pEnt)
{
   if (!pEnt) {                                  // EntityDef not found! Error.
      evlst.append(Error());                     // Skipped...
   }else if (pEnt->is_rcdata()) {                // parseable data
      evlst.append(EntityStart(pEnt));           // use regular algorithm
   }else if (pEnt->text_type() == PI_EFK) {      // PI entity
      rcdstate.declare_other_markup_dcl();       // Register new state
      evlst.append(EntityStart(pEnt));
      evlst.append(EntityEnd());
   }else {
      /*
      | CDATA, SDATA, NDATA, SUBDOC: check if data is allowed
      */
      switch (pst1) {
      case TAG_PST1:
         if (!prepare_for_data()) {              // cannot have data here
            errorh(OOCDATA, pEh, ERH_E_SEV);
            evlst.append(Error());
            pst1 = INVDATA_PST1;
            break;
         }
         pst1 = DATA_PST1;
         /* Fall thru */

      case DATA_PST1:
         oelemstk[0]->isEmptyOrBad = FALSE;       // Not empty
         if (rcdstate.declare_data()) {          // register new state
           evlst.append(RecordEnd());
         }
         switch (pEnt->text_type()) {
         case SUBDOC_EFK:                        // SUBDOC entity
            //***>NOTE: What do we do with SUBDOC entities?
            break;
         default:                                // C/S/N Data
            //***>NOTE: How do we get the data back from the outside?
            break;
         }
         evlst.append(EntityStart(pEnt));
         evlst.append(EntityEnd());
         break;

      default: /* INVDATA_PST1 */                // Invalid data entity
         break;
      }
   }
}

/*STATIC------------------------------------------------------tvrs_check_idref-+
|Check if all idref have been set (defined)                                    |
+-----------------------------------------------------------------------------*/
static boolean tvrs_check_idref(
   void *, const char *pszName, void *piElemId
)
{
   if (*(int *)piElemId == 0) {
      errorh(INVIDREF, ERH_E_SEV, pszName);
   }
   return TRUE;
}

/*------------------------------------------------------------------process_Ee-+
| Process end of entity                                                        |
+-----------------------------------------------------------------------------*/
boolean SGMLparser::process_Ee()
{
   if (pst2 == EE_BEFORE_CLOSE_PST2) {
      pst2 = EE_AFTER_CLOSE_PST2;
      if ((pPrsmode->get_mode() == RCD_MODE) &&
         (pEh->curr_entity_level() <= cn_ent_level)) {
         errorh(INVLNEST, pEh, ERH_E_SEV);
      }
      pPrsmode->enable_markup_recognition(); // recognize MARKUP,see 9.7
      evlst.append(EntityEnd());
      return TRUE;
   }
   pst2 = EE_BEFORE_CLOSE_PST2;
   if (pEh->curr_entity_level() > 1) {  // Not #PRIMARY
      pEh->close();                  // get next char in prev. file
      return TRUE;                      // and continue
   }
   /*-----------------------------------------+
   | If the document element at least started,|
   | end the maximum of elements              |
   | Check if all references were solved      |
   +-----------------------------------------*/
   if (!(oelemstk[0]->isZero)) {
      end_element(
         1 + pSyntax->pQuantity->value_of(TAGLVL_QTY_ix), EOF_IMF, NO_MNF
      );
      ExpnhashId.traverse(tvrs_check_idref, 0);
   }
   if (ms_level) errorh(UBLNMSO, pEh, ERH_E_SEV);
   return FALSE;
}

/*--------------------------------------------------------------process_usedcl-+
| Enter here to process either USELINK or USEMAP, when an MDO has been found   |
| in content.                                                                  |
+-----------------------------------------------------------------------------*/
void SGMLparser::process_usedcl()
{
   enum use_keyword {            // Value doesn't belong to the list
      INVALID_VL,                // To find if it's USEMAP or USELINK
      USEMAP_VL,
      USELINK_VL
   };
   static VarntChoiceEntry const usedcl_choices[] = {
      { USELINK_VRN_ix,    USELINK_VL },
      { USEMAP_VRN_ix,     USEMAP_VL  },
      { NOT_FOUND_VRN_ix,  INVALID_VL }
   };
   boolean isErrorFound = FALSE;

   pPrsmode->push_mode(MD_MODE);
   switch (pSyntax->match_rsvname_in_list(pEh, usedcl_choices)) {

   case INVALID_VL:
      errorh(INVMDIGN, pEh, ERH_S_SEV);
      isErrorFound = TRUE;
      break;

   case USELINK_VL:
      errorh(NIMPL, pEh, ERH_T_SEV,1002);
      isErrorFound = TRUE;
      skip_ps();
      break;

   case USEMAP_VL:
      {
         char pszName[NAMELEN_MAX+1];

         skip_ps();
         if (get_curr_token() == RNI_TK) {
            if (!pSyntax->match_rsvname(pEh, EMPTY_VRN_ix)) {
               errorh(
                  NOKYWD, pEh, ERH_S_SEV,
                  pSyntax->pRsvname->name_of(EMPTY_VRN_ix)
               );
               isErrorFound = TRUE;
               break;
            }
            pszName[0] = '\0';
         }else {
            if (get_curr_token() != NOT_FOUND_TK) {
               errorh(NONAME, pEh, ERH_S_SEV);
               isErrorFound = TRUE;
               break;
            }
            pSyntax->read_general_name(pEh, pszName);  // get the name
         }
         ShortrefMap *pSrm = (ShortrefMap *)pDtddcl->pHshSrMap->search_value(
            pszName
         );
         if (!pSrm) {
            errorh(NOSRM, pEh, ERH_E_SEV, pszName);
            isErrorFound = TRUE;
            skip_ps();
         }else {
            /*
            | at pSrm + 0 is a dummy entry that has the mapname
            */
            oelemstk[0]->pSrm = pSrm + 1;          // new short map
            skip_ps();
         }
      }
      break;
   }
   pPrsmode->pop_mode();
   if (get_curr_token() != MDC_TK) {
      if (!isErrorFound) {
         errorh(NOMDC, pEh, ERH_S_SEV);
         isErrorFound = TRUE;
      }
      pSyntax->skip_to_delimiter(pEh, MDC_VDL_ix);     // Try to get the MDC...
   }
   if (isErrorFound) evlst.append(Error());
}
/******************************************************************************/
