/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// prsstate.c
///  Parser state class and methods
//   Pierre G. Richard 87/07/28 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <assert.h>
#include <new.h>

#include "gendef.h"
#include "externid.h"
#include "entity.h"
#include "dlmstr.h"
#include "rsvname.h"
#include "prsstate.h"
#include "dlmvect.h"
#include "sgmldcl.h"
#include "dtddcl.h"
#include "enthdlr.h"
#include "utilctl.h"
#include "quantity.h"
#include "capacity.h"
#include "hash.h"
#include "prsmode.h"
#include "parser.h"

enum { ID_HASHSIZE = 128 };  // For about 200 entries

/*PROTECTED---------------------------------------------------------ParseState-+
| Default constructor                                                          |
| This constructor cannot be called.  It is required because Instance has      |
| access to the unique and already created ParseState.                         |
+-----------------------------------------------------------------------------*/
ParseState::ParseState() : exidPrimary(0, INVALID_XFC), entPrimary(0)
{
   assert("ParseState NULL constructor cannot be called" == FALSE);
}

/*PROTECTED---------------------------------------------------------ParseState-+
| Regular constructor to create a ParseState for a new parser session          |
+-----------------------------------------------------------------------------*/
ParseState::ParseState(
   const Document& doc
) :
   ExpnhashId(ID_HASHSIZE),
   exidPrimary(NULL, PRIMARY_XFC),  // NULL name denotes a primary entity
   entPrimary(&exidPrimary),
   isEnabledEvent(EV_MAX_SIZE)
{
   ms_ent_level = 0;
   cn_ent_level = 0;
   floor_level = 0;
   ms_level = 0;
   iCountOfExid = 0;
   isNetEnabled = FALSE;
   isQuietAtFloor = FALSE;
   isEnabledEvent = 1;                       // all events enabled
   pPrsmode = new ParseMode;                 // need a ParseMode
   pEh = new EntityHandler(doc, pPrsmode);   // start the entity manager
   pSgmldcl = new SGMLdcl;                   // set a minimal SGMLdcl
   pDtddcl = 0;                              // no DTD so far
   shadow_syntax(pSgmldcl->pSyntax);         // private copy of new pointers
   curr_token = NOT_FOUND_TK;                // current delimiter token found
   curr_count = 0;                           // length of current delimiter
   curr_elem_id = 1;

   pElemLastEnded = 0;
   iCountForcedData = 0;
   memset(aulCapacityCount, 0, sizeof aulCapacityCount); // capacity reached
   pst0 = INITIALIZED_PST0;
}

/*------------------------------------------------------------------ParseState-+
| Constructor to take a resume point.                                          |
| This constructor is used to get a ParseState at a given point,               |
| so it's then possible to resume from this point.                             |
+-----------------------------------------------------------------------------*/
ParseState::ParseState(
   const SGMLparser *pParserSource
) :
   oelemstk(pParserSource->oelemstk),
   exidPrimary(pParserSource->exidPrimary),
   entPrimary(pParserSource->entPrimary, &exidPrimary),
   isEnabledEvent(pParserSource->isEnabledEvent)
{
   pParserSource->copy_context_into(*this);
}

/*PROTECTED---------------------------------------------------------ParseState-+
| Copy constructor                                                             |
| This constructor is used to resume a Parser session form a given             |
| ParseState.  Only called by SGMLparser(const ParseState& prsctxt),           |
| it is therefore protected.                                                   |
+-----------------------------------------------------------------------------*/
ParseState::ParseState(
   const ParseState& prsstatSource
) :
   oelemstk(prsstatSource.oelemstk),
   exidPrimary(prsstatSource.exidPrimary),
   entPrimary(prsstatSource.entPrimary, &exidPrimary),
   isEnabledEvent(prsstatSource.isEnabledEvent)
{
   //***>Note: not yet implemented
}

#include <iostream.h>

/*------------------------------------------------------------------Destructor-+
|                                                                              |
+-----------------------------------------------------------------------------*/
ParseState::~ParseState()
{
   delete pDtddcl;               //***> TEMPORARY (?)
   delete pSgmldcl;
   delete pEh;
   delete pPrsmode;
}

/*---------------------------------------------------------------shadow_syntax-+
| Copy a few pointers in private area when those have been changed             |
+-----------------------------------------------------------------------------*/
void ParseState::shadow_syntax(Syntax *pSyntaxArg)
{
   pSyntax = pSyntaxArg;               // Shadow from SGMLdcl member
   pDv = pSyntax->pDv;                 // Shadow for recognize delimiter
   pEh->shadow_syntax(pSyntax);        // also, tell the EntityHandler
}

/*------------------------------------------------------------activate_SGMLdcl-+
| This must be called after the SGML dcl has been parsed (UTSYNUser_XFC)       |
| or when the default SGML dcl needs to be loaded (UTSYNLocal_XFX)             |
|                                                                              |
| clsOcs must be one of:                                                       |
|    UTSYNU_XFC:   utility user-modified ocs                                   |
|    UTSYNL_XFC:   utility local-system ocs                                    |
|    UTSYNR_XFC:   utility refcon-system ocs                                   |
+-----------------------------------------------------------------------------*/
void ParseState::activate_SGMLdcl(e_TextClass clsOcs)
{
   delete pSgmldcl;

   ApplicationHandle ahlAccess = open_OCS(clsOcs, FALSE);
   pSgmldcl = new SGMLdcl(ahlAccess);
   pEh->close(ahlAccess);
   shadow_syntax(pSgmldcl->pSyntax);  // private copy of new pointers
}

/*-------------------------------------------------------------activate_syntax-+
| This must be called when another syntax must be activated (scope instance)   |
|                                                                              |
| clsOdt must be one of:                                                       |
|    UTSYNU_XFC:   utility user-modified ocs                                   |
|    UTSYNL_XFC:   utility local-system ocs                                    |
|    UTSYNR_XFC:   utility refcon-system ocs                                   |
+-----------------------------------------------------------------------------*/
void ParseState::activate_syntax(e_TextClass clsOcs)
{
   delete pSgmldcl->pSyntax;

   ApplicationHandle ahlAccess = open_OCS(clsOcs, FALSE);
   pSgmldcl->pSyntax = new Syntax(ahlAccess);
   pEh->close(ahlAccess);
   shadow_syntax(pSgmldcl->pSyntax);  // private copy of new pointers
}

/*---------------------------------------------------------find_general_entity-+
| Find the EntityDef structure given the name of a general entity              |
+-----------------------------------------------------------------------------*/
EntityDef * ParseState::find_general_entity(char *pszName)
{
   EntityDef *pEnt;

   pEnt = (EntityDef *)pHshGenEnt->search_value(pszName);
   if (pEnt) return pEnt;                        // found: return

   pEnt = (EntityDef *)pHshGenEnt->search_value(""); // else try DEFAULT
   if (pEnt) {                                   // if it exists...
      errorh(NOGERDEF, pEh, ERH_I_SEV, pszName); // gen ent defaulted
      /*-------------------------------------------------------+
      | Make a copy of the Default Entity.                     |
      | If the DEFAULT entity is external, also copy the EXID: |
      | The system's strategy is allowed to use this name      |
      | (see ISO8879 + Amendment 1, 10.1.6)                    |
      +-------------------------------------------------------*/

      //***>NOTE: when the following is executed during the parse
      //***>NOTE  of the instance:
      //***>NOTE  1) it violates the constness of DTDdcl
      //***>NOTE  2) the new (copied) EXID is never deleted
      //***>NOTE     Warning: if you plan to implement this
      //***>NOTE     remember *not* to delete the ExternId::pchBuffer!

      char *pszNameSave = pszName;
      EntityDef *pEntNew = (EntityDef *)pHshGenEnt->add_value(
         &pszNameSave, sizeof(EntityDef)
      );
      ExternId *pExidNew;
      if (pEnt->is_external()) {
         pExidNew = new ExternId(*(pEnt->external_id()), pszNameSave);
         ++iCountOfExid;
      }else {
         pExidNew = 0;
      }
      new (pEntNew) EntityDef(*pEnt, pszNameSave, pExidNew);
      ++aulCapacityCount[ENTCAP_ix];             // see 10.5.1.2
      return pEntNew;
   }
   errorh(NOGERIGN, pEh, ERH_E_SEV, pszName);    // gen ent ignored
   return 0;
}

/*-----------------------------------------------open_general_entity_reference-+
| Try to access a general entity reference.                                    |
| Return a pointer to the entity, (NULL pointer if unable to access).          |
| The entity can be opened, if the recognition mode allows it (see open).      |
|                                                                              |
| At compile time, we can only be called by att_val_spec() parsing             |
| a default AVS.                                                               |
+-----------------------------------------------------------------------------*/
EntityDef * ParseState::open_general_entity_reference()
{
   char pszName[NAMELEN_MAX+1];    /* see 9.3.1 */

   pSyntax->read_refd_entity_name(pEh, pszName);
   EntityDef *pEnt = find_general_entity(pszName);
   if ((pEnt) && (pEh->open(pEnt))) return pEnt;
   return 0;
}

/*PRIVATE--------------------------------------open_parameter_entity_reference-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void ParseState::open_parameter_entity_reference(EntityDef **ppEntPI)
{
   char pszName[NAMELEN_MAX + 1];
   EntityDef *pEntPI;

   if (!ppEntPI) ppEntPI = &pEntPI;  // temporary area
   pSyntax->read_refd_entity_name(pEh, pszName);
   *ppEntPI = (EntityDef *)pHshPrmEnt->search_value(pszName);
   if (!*ppEntPI) {
      errorh(NOPER, pEh, ERH_E_SEV); // entname not found
   }else {
      /*------------------------------------------+
      | Only PI (no C/S/N Data nor SUBDOC) can be |
      | found by open and not opened.  If this |
      | is the case, then *ppEnt points the ent.  |
      | If not, zero *ppEnt: nothing to signal.   |
      +------------------------------------------*/
      if (!(pEh->open(*ppEntPI)) || ((*ppEntPI)->text_type() != PI_EFK)) {
         *ppEntPI = 0;
      }
   }
}

/*------------------------------------------------------read_parameter_literal-+
| Read a parameter literal into pchBuffer, returns its length                  |
|                                                                              |
| -- INVARIANTS --                                                             |
| This function requires access to the Syntax, ParseMode, DelimString,         |
| and Quantity classes.  This function is called only while in the prolog.     |
|                                                                              |
| The mode is LITP_MODE, so CRO_TK, LIT_TK, LITA_TK, PERO_TK, PERO_GRPO_TK     |
| are recognized.  Since this function is only called while parsing the        |
| prolog, PERO_GRPO is an error (see 9.4.4).                                   |
|                                                                              |
| -- INPUT --                                                                  |
| `for_datatag' indicates if the param literal is within a datatag template    |
| `pchBuffer' points a target buffer                                           |
|                                                                              |
| -- PRECONDITION --                                                           |
| `pchBuffer is LITLEN+1 long, or DTEMPLEN+1 if `for_datatag'                  |
| `curr_token' has the last delimiter token as found by                        |
|              `recognize_delimiter'.                                          |
|                                                                              |
| -- OUTPUT --                                                                 |
| If the maximum length was exceeded, this function exits -- which is not      |
| too nasty in prologs.                                                        |
|                                                                              |
| Otherwise, `pchBuffer' contains the parameter literal, of course             |
| not including the ending delimiter.                                          |
| The length of the literal is returned.                                       |
|                                                                              |
| -- POSTCONDITION --                                                          |
| `curr_token' has the new delimiter token found by `recognize_delimiter'      |
| (it is expected to be the same as on entry).                                 |
+-----------------------------------------------------------------------------*/
unsigned int ParseState::read_parameter_literal(
   char *pchBuffer, boolean for_datatag
)
{
   const int iMaxLength = pSyntax->pQuantity->value_of(
      for_datatag ? DTEMPLEN_QTY_ix : LITLEN_QTY_ix
   );
   const e_LexToken delim_token = curr_token;
   const int init_ent_level = pEh->curr_entity_level();
   e_LexToken new_token;
   int iLength = 0;

   if ((delim_token != LIT_TK) && (delim_token != LITA_TK)) {
      errorh(NOLIT, pEh, ERH_T_SEV);
   }
   pPrsmode->push_mode(LITP_MODE);
   while (                      // until same delimiter, at same nesting level
      new_token = recognize_delimiter(),
      (new_token != delim_token) || (pEh->curr_entity_level() != init_ent_level)
   ) {
      switch (new_token) {
      case LIT_TK:
      case LITA_TK:       /* LIT or LITA in a nested entity */
         pEh->unread(curr_count);
         break;
      case NOT_FOUND_TK:
         if (pEh->is_Ee()) {
            pPrsmode->enable_markup_recognition();
            if (pEh->curr_entity_level() == init_ent_level) {
               errorh(
                  UBLNLIT, pEh, ERH_T_SEV,
                  pSyntax->pDlmstr->value_of(
                     delim_token==LIT_TK ? LIT_VDL_ix : LITA_VDL_ix
                  )
               );
            }else {
               pEh->close();   // read a new character in previous file
               continue;
            }
         }else if (pEh->is_ms_char()) { // mso, msi, mss
            if (pEh->is_msochar()) {
               pPrsmode->disable_markup_recognition();
            }else if (pEh->is_msichar()) {
               pPrsmode->enable_markup_recognition();
            }else {             /* mss_char: treat it directly */
               pchBuffer[iLength] = (char)pEh->curr_char_val();
               pEh->read();
               if (++iLength > iMaxLength) {
                  errorh(BIGLIT, pEh, ERH_T_SEV);
               }
               if (pEh->is_Ee()) continue;
            }
         }
         break;
      case CRO_TK:
         switch (pSyntax->translate_character_ref(pEh, curr_count)) {
         case INVALID_CHRF:
            continue;
         case NUMERIC_CHRF:
            if ((for_datatag) && ((pEh->is_nonsgml()) || (pEh->is_fctchar()))) {
               errorh(INVDTPLIT, pEh, ERH_E_SEV);
               pEh->read();
               continue;
            }
            break;
         default: // case NAMED_CHRF:
            break;
         }
         /*----------------------------------------+
         |As of now we don't care numeric vs. named|
         |so a numeric could be interpreted as     |
         |markup?                                  |
         |Hmmmm... Better keep numeric unparsed ?  |
         +----------------------------------------*/
         break;
      case PERO_TK:
#ifndef KEEP_PER_UNPARSED
         open_parameter_entity_reference();
         continue;
#else
         pEh->unread(curr_count);    /* to keep PER unparsed */
         break;
#endif
      default:
      case PERO_GRPO_TK:
         errorh(INVDLM, pEh, ERH_T_SEV);
         break;
      }
      pchBuffer[iLength] = (char)pEh->curr_char_val();
      pEh->read();
      if (++iLength > iMaxLength) {
         if (for_datatag) {
            errorh(BIGDLIT, pEh, ERH_T_SEV);
         }else {
            errorh(BIGLIT, pEh, ERH_T_SEV);
         }
      }
   }
   pPrsmode->pop_mode();
   pchBuffer[iLength] = '\0';
   return iLength;
}

/*------------------------------------------------------------parse_name_group-+
| Parse a name group, call an user-defined install function for each name      |
|                                                                              |
| [69] name_group = GRPO, ts*,name,ts*,   , GRPC                               |
|                           (connector, ts*,name,ts* )*                        |
+-----------------------------------------------------------------------------*/
void ParseState::parse_name_group(
   void (*inst_function)(void *pvBase),
   void *pvBase
)
{
   int curr_floor_level = floor_level;        // save current floor_level
   int init_ent_level  = pEh->curr_entity_level(); // remember where we started
   unsigned int grpcnt = 0;

   pPrsmode->push_mode(GRP_MODE);
   for (;;) {
      floor_level = pEh->curr_entity_level(); // No Ee allowed
      skip_ts();                              // opt ts after GRPO
      if (curr_token != NOT_FOUND_TK) {
         errorh(INVNAMGRP, pEh, ERH_T_SEV);
      }
      if (++grpcnt > pSyntax->pQuantity->value_of(GRPCNT_QTY_ix)) {
         errorh(BIGGRPCNT, pEh, ERH_T_SEV);
      }
      (*inst_function)(pvBase);               // call the install function
      floor_level = init_ent_level;
      skip_ts();                              // skip opt ts
      switch (curr_token) {
      case AND_TK:
      case SEQ_TK:
      case OR_TK:
         continue;                            // more to do
      case GRPC_TK:
         pPrsmode->pop_mode();
         if (pEh->curr_entity_level() != init_ent_level) {
            errorh(INVHNEST, pEh, ERH_T_SEV);
         }
         floor_level = curr_floor_level;      // restore floor_level
         return;
      default:
         errorh(INVCONN, pEh, ERH_T_SEV);
      }
   }
}

/*------------------------------------------------------------may_be_delimiter-+
| Examine if the current character has some chances to be a delimiter.         |
|                                                                              |
| Return FALSE if it will never be recognized as such.                         |
| Return TRUE if it could be: this doesn't mean that it *is* a delimiter.      |
|                                                                              |
| IMPLEMENTATION NOTE:                                                         |
| This function is used to interrupt the process_data loop, while avoiding     |
| the look-ahead required by the full `recognize delimiter' function.          |
| This allows the Parser to report events with more synchronism.               |
+-----------------------------------------------------------------------------*/
boolean ParseState::may_be_delimiter()
{
   curr_token = NOT_FOUND_TK;
   curr_count = 0;
   if (
      (!(pDv->dltab[pEh->curr_char_val()] & (1 << pPrsmode->get_mode()))) ||
      (pPrsmode->is_markup_recognition_disabled()) ||
      (pEh->is_numchar())
   ) {
      return FALSE;
   }
   return TRUE;
}

/*---------------------------------------------------------recognize_delimiter-+
| Try to decode a delimiter on input.                                          |
| Set up curr_count, curr_token,                                               |
| Return curr_token (could be NOT_FOUND_TK).                                   |
|                                                                              |
| -- INVARIANTS --                                                             |
| This function requires access to the                                         |
| DelimiterVector, Syntax, ParseMode, and Quantity classes                     |
|                                                                              |
| `dvtbent' and `dltab' of the DelimiterVector class contains specific data    |
| in order to process to delimiter recognition.   Only `recognize_delimiter'   |
| has access to this data -- this function is a friend of the DelimiterVector  |
| class.                                                                       |
|                                                                              |
| ParseMode::get_mode() returns the current recognition mode.                  |
|                                                                              |
| -- INPUT --                                                                  |
| (none)                                                                       |
|                                                                              |
| -- PRECONDITION --                                                           |
| The recognition mode must have been properly set (ParseMode::set_mode)       |
|                                                                              |
| -- OUTPUT --                                                                 |
| `curr_token' contains the identifier of the delimiter_token found            |
|              (one of the e_LexToken values, including NOT_FOUND_TK)          |
|                                                                              |
| -- POSTCONDITION --                                                          |
| `curr_count' has the length of the number of characters (read on input)      |
|              forming the delimiter -- 0 if no delimiter found.               |
|                                                                              |
| -- NOTES --                                                                  |
|                                                                              |
| IMPLEMENTATION NOTE:                                                         |
|      Due to the complexity of the Delimiters Recognition rules, with         |
|      B-sequence imbedded, the following code is a very tough one.            |
|      Change it with care!  Sloppy goto's are used to optimize                |
|      execution and (paradoxically) for readibility !                         |
|                                                                              |
| CODE ANNOTATIONS:                                                            |
|                                                                              |
| (1) Re: "A BSEQ cannot be preceded/followed by a blank"                      |
|     if pViAlt was set, this means that more than one alternate was           |
|     found, in violation of the rule stated above.                            |
|                                                                              |
| (2) Re: "only one BSEQ is allowed"                                           |
|     The BSEQ can be descended only one time, so we have no scruples          |
|     to chain the previous VI; this also acts as a protection for (3);        |
|                                                                              |
| (3) At this point, "in_bseq_recur" is always FALSE, so we cannot decode      |
|     a B-seq more than 2 times.  Ad absurbo:                                  |
|     - if we were flunked in the search of the first BSEQ PCL,                |
|       the scanned char is reputed to be non-blank                            |
|       (all blanks have been exhausted  during the B-seq descent).            |
|       Then: "is_blank(curr_char)" is FALSE.                                  |
|     - All other suceeding PCLs are reputed *not* to contain any B-seq,       |
|       since only one single B-seq is allowed.                                |
|       Then: "pVi->isBseq" is FALSE.                                          |
|     Note: if the single B-seq Rule was not existing, this would have         |
|           implied the stacking of prev_count, prev_tk, and pViPrev values.   |
|                                                                              |
+-----------------------------------------------------------------------------*/
e_LexToken ParseState::recognize_delimiter()
{
   if (!may_be_delimiter()) return NOT_FOUND_TK;

   VectorItem  *pViCur = pDv->dvtbent[pPrsmode->get_mode()].pVi;  // Current
   VectorItem  *pViPrv = 0;                     // Previous
   VectorItem  *pViAlt = 0;                     // Alternate
   unsigned int iPegAlt;
   VectorItem  *pViFnd = 0;                     // Found
   unsigned int iPegFnd = 0;
   VectorItem  *pViFndBefore;                   // Previously Found
   unsigned int iPegFndBefore = 0;
   unsigned int iPegAfterBseq = 0;
   char        *pch;
   unsigned int i;
   unsigned int iMaxBseq;
   unsigned int iCount = 0;
   boolean      in_bseq_recur = FALSE;

main_loop: //----------------------------------------------------------------
   /*-------------------------------------------------+
   |Search thru the DV structure the "best" delimiter |
   +-------------------------------------------------*/
   if (pEh->is_numchar()) {
     pch = 0;
   }else {
     pch = (char *)memchr(
        pViCur->str, pEh->curr_char_val(), pViCur->h.count
     );
   }
   if (pch) {
      /*----------------------------------------------------+
      |Found in the PCL: set up any alternate, and continue |
      +----------------------------------------------------*/
      if (
         (pEh->is_blank()) &&          // Blank conflict? (alternate found)
         (pViCur->h.isBseq) &&         // see note (3)
         (!pViAlt)                     // PROTECTION: see Note (1)
      ) {
         pViAlt = pViCur;              // yes, set alternate path
         iPegAlt = iCount;             // remember where
      }
      pEh->read();                  // new curr_char
      ++iCount;                        // increase unread count
      pViCur += 1 + diffptr(pch, pViCur->str); // get 2nd node

examine_cont_string: //------------------------------------------------------
      /*------------------------------------+
      |Check the continuation string matches|
      +------------------------------------*/
      pch = pViCur->str;                      // point continuation string
      while (
         (*pch) &&
         (*pch == (char)pEh->curr_char_val()) &&
         (!pEh->is_numchar())
      ) {                                     // leave at eos
         ++pch;
         ++iCount;                            // increase unread count
         pEh->read();                      // look for a match
      }
      if (*pch) {                             // eos not found (it's bad):
         if (in_bseq_recur) goto try_another_B_dv;
         goto break_it;                       // no recursion: leave
      }                                       // eos found (it's ok):
      if (pViCur->tk) {                       // largest token gotten
         if (check_contextual_constraint(pViCur->tk)) {
            pViFnd = pViCur;                  // get its VectorItem
            iPegFnd = iCount;
         }                                    // else, as not found...
         if (in_bseq_recur) goto try_another_B_dv;
         goto break_it;                       // and leave...
      }
      if (
         ((pViCur=pViCur->next)->tk != NOT_FOUND_TK) &&
         check_contextual_constraint(pViCur->tk)
      ) {                                     // may be found...
         pViFnd = pViCur;                     // set may_be VectorItem
         iPegFnd = iCount;                    // peg unread here
      }
      goto main_loop;
   }
   if (in_bseq_recur) goto try_another_B_dv;  // see note (2)
   /*---------------------------------+
   |Not found in the PCL, first time: |
   |try a blank seq recognition       |
   +---------------------------------*/
   if (
      (pEh->is_blank()) &&                    // need a Bseq,
      (pViCur->h.isBseq) &&                   // and an alternate path
      (!pEh->is_numchar())
   ) {
      goto decode_Bseq;                       // see note (3)
   }
   goto break_it;

try_another_B_dv: //---------------------------------------------------------
   /*----------------------------------------------------------------+
   |Not found: while doing blank sequence nodes examinition,         |
   |go to another previous B-dv if needed and possible (see note (2) |
   |For a same count, prev_F is always more specific..               |
   +----------------------------------------------------------------*/
   if (iPegFnd > iPegFndBefore) {             // between both...
      iPegFndBefore = iPegFnd;                // choose the longest
      pViFndBefore = pViFnd;
   }
   pViCur = pViPrv = (pViPrv+1)->previous;
   if (!pViCur) {                             // not possible
      iPegFnd = iPegFndBefore;                // set final answer
      pViFnd = pViFndBefore;
      goto break_it;
   }
   pEh->unread(iCount - iPegAfterBseq);       // back where we were
   iCount = iPegAfterBseq;
   goto examine_cont_string;

decode_Bseq: //--------------------------------------------------------------
   for (;;) {
      /*-------------------------------------------+
      |descend thru the maximum of B-sequence node |
      +-------------------------------------------*/
      pViCur += pViCur->h.count + 1;          // point BSEQ node
      i = (pViCur+1)->bseqlen;                // get minimum BSEQ length
      iMaxBseq = iCount + pSyntax->pQuantity->value_of(BSEQLEN_QTY_ix);
      do {                                    // skip minimum blanks:
         ++iCount;
         pEh->read();
      }while ((pEh->is_blank()) && (!pEh->is_numchar()) && (--i));
      switch(i) {
      case 0:                                 // more than min blanks found
         if ((pViCur+1)->must_recur) {        // more B specificity:
            pViCur = pViCur->next;            // chain next node
            continue;                         // try it...
         }
         do {                                 // no more B-sequence node:
            pEh->read();                      // skip all others blanks
         }while (
            (++iCount < iMaxBseq) &&
            (pEh->is_blank()) && (!pEh->is_numchar())
         );
         break;
      case 1:                                 // exactly min blanks found
         break;
      default:                                // less than min blanks found
         pViCur = (pViCur+1)->previous;           // back to previous
         break;                               // we'll check pViCur after
      }
      break;
   }
   if ((pViPrv = pViCur) != 0) {
      in_bseq_recur = TRUE;
      iPegFndBefore = iPegFnd;                // save F_peg
      pViFndBefore = pViFnd;                  // and pViFnd
      iPegAfterBseq = iCount;                 // peg unread(s) here
      goto examine_cont_string;
   }

break_it: //-----------------------------------------------------------------
   if (pViAlt) {                              // Alternate Bseq co-exists
      pEh->unread(iCount - iPegAlt);          // unread to it
      iCount = iPegAlt;
      pViCur = pViAlt;                        // go try it
      pViAlt = 0;
      goto decode_Bseq;                       // one time only....
   }
   if (iCount != iPegFnd) {
      if (iCount < iPegFnd) {
         while (++iCount < iPegFnd) pEh->read(); // read one less
         pEh->read();                         // set curr char for the last
      }else {
         pEh->unread(iCount - iPegFnd);       // unread
      }
   }
   if (pViFnd) {
      curr_count = iPegFnd;   // iPegFnd has the length of the decoded string
      curr_token = pViFnd->tk;
   }
   return curr_token;
}

/*PRIVATE------------------------------------------check_contextual_constraint-+
| Check contextual constraints for a delimiter to be recognized.               |
|                                                                              |
| -- INVARIANTS --                                                             |
| This function requires access to the Syntax, and ParseMode classes.          |
|                                                                              |
| When the current element has a declared content of RCDATA, the mode          |
| is RCD_MO and `cn_ent_level' records the entity nesting level                |
| reached at the time the current element started.                             |
|                                                                              |
| Within a Marked Section containing RCDATA, the mode is RCDMS_MO.             |
| `ms_ent_level' records the entity nesting level reached at the time          |
| the Marked Section started.                                                  |
|                                                                              |
| -- INPUT --                                                                  |
| `tk' has the token of the delimiter the contextual constraints of which      |
| need to be checked.                                                          |
|                                                                              |
| -- PRECONDITION --                                                           |
| Following values of ParseMode class must have been properly set:             |
|   - recognition mode                                                         |
|   - level of nested entities (ent_level, cn_ent_level, ms_ent_level)         |
|   - net_enabled                                                              |
|                                                                              |
| -- OUTPUT --                                                                 |
| if (the delimiter is recognizable) return TRUE  else return FALSE;           |
|                                                                              |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
boolean ParseState::check_contextual_constraint(e_LexToken tk)
{
   e_RecognitionMode current_mode = pPrsmode->get_mode();

   switch(tk) {

   case MSC_MDC_TK:
      /*---------------------------------------------+
      |MSC_MDC cannot be recognized as such if it    |
      |occurs in an RCDATA marked section at a higher|
      |entity level than the one the marked section  |
      |started with (see marked_section_open)        |
      +---------------------------------------------*/
      if (
         (current_mode == RCDMS_MODE) &&
         (pEh->curr_entity_level() > ms_ent_level)
      ) {
         return FALSE;
      }
      return TRUE;

   case CRO_TK:                                 // CREF: digit or namestart
      if (pEh->is_cref()) return TRUE;
      return FALSE;

   case ERO_TK:                                 // NMS: namestart
   case MDO_TK:                                 // DCL: namestart
   case PERO_TK:                                // NMS: namestart
      if (!pEh->is_namestrt()) {
         return FALSE;
      }
      return TRUE;

   case STAGO_TK:                               // GI: namestart, or 7.4.1.2
   case ETAGO_TK:                               // GI: namestart, or 7.4.5.2
      /*---------------------------------------------+
      | if in TAG_MODE                               |
      |    always TRUE (no constraints)              |
      |    (see ISO 8879, 7.4.1.2 and 7.4.5.2)       |
      | else (CON_MO, RCD_MO, FCD_MO)                |
      |    TRUE iff followed by a name_start char    |
      |    except for RCDATA-content within a nested |
      |    entity.                                   |
      |                                              |
      | Note that STAGO is dl_found only for CON_MO  |
      |      (not for RCD_MO, neither FCD_MO)        |
      +---------------------------------------------*/
      if (current_mode == TAG_MODE) return TRUE;
      if (
         (!(pEh->is_namestrt())) ||
         (
            (current_mode == RCD_MODE) &&
            (pEh->curr_entity_level() > cn_ent_level)
         )
      ) {
         return FALSE;
      }
      return TRUE;

   case ETAGO_TAGC_TK:
   case ETAGO_GRPO_TK:
      /*---------------------------------------------+
      | TRUE except for RCDATA-content, within a     |
      | nested entity.                               |
      +---------------------------------------------*/
      if (current_mode == TAG_MODE) return TRUE;
      if (
         (current_mode == RCD_MODE) &&
         (pEh->curr_entity_level() > cn_ent_level)
      ) {
         return FALSE;
      }
      return TRUE;

   case NET_TK:                                 // ELEM (special)
      /*---------------------------------------------+
      | if in TAG_MODE                               |
      |    always TRUE (no constraints)              |
      |    (see ISO 8879, 7.4.1.2 and 7.4.5.2)       |
      | if CON_MODE || RCD_MODE || FCD_MODE          |
      |    TRUE iff within an element whose          |
      |    start-tag was a net-enabling start tag    |
      |    except for RCD_MODE (RCDATA elem content) |
      |    within a nested entity.                   |
      +---------------------------------------------*/
      if (current_mode == TAG_MODE) return TRUE;
      if (
         (current_mode == RCD_MODE) &&
         (pEh->curr_entity_level() > cn_ent_level)
      ) {
         return FALSE;
      }
      if (!isNetEnabled) return FALSE;
      return TRUE;

   default:
      return TRUE;                              /* ---  any character type */
   }
}

/*-------------------------------------------------read_processing_instruction-+
| Read a processing instruction                                                |
| See comments in `read_to_end_delimiter'                                      |
+-----------------------------------------------------------------------------*/
unsigned int ParseState::read_processing_instr(
   EntityHandler *pEh, char *pchBuffer
)
{
   return pSyntax->read_to_end_delimiter(pEh, pchBuffer);
}

/*--------------------------------------------------------read_minimum_literal-+
| Read a minimum literal.                                                      |
| See comments in `read_to_end_delimiter'                                      |
+-----------------------------------------------------------------------------*/
unsigned int ParseState::read_minimum_literal(
   EntityHandler *pEh, char *pchBuffer
)
{
   return pSyntax->read_to_end_delimiter(
      pEh, pchBuffer, curr_token, MINLIT_GLC
   );
}

/*----------------------------------------------------------read_cdata_literal-+
| Read a character data literal                                                |
| See comments in `read_to_end_delimiter'                                      |
+-----------------------------------------------------------------------------*/
unsigned int ParseState::read_cdata_literal(
   EntityHandler *pEh, char *pchBuffer
)
{
   return pSyntax->read_to_end_delimiter(
      pEh, pchBuffer, curr_token, LIT_GLC
   );
}

/*---------------------------------------------------------marked_section_open-+
| Enter here when a Marked Section Start (MDO_DSO_TK) has been decoded         |
| Returns with the right mode pushed, if not INCLUDE or TEMP.                  |
|                                                                              |
| [93] marked section declaration =                                            |
|              marked section start, status keyword specification, dso,        |
|              marked section, marked section end                              |
|                                                                              |
| [94] marked section start = mdo, dso                                         |
|                                                                              |
| [95] marked section end = msc, mdc                                           |
|                                                                              |
| [96] marked section = SGML character*                                        |
|                                                                              |
| [97] status keyword specification =                                          |
|                (ps+, (status keyword | "TEMP"))*, ps*                        |
|                                                                              |
| [100] status keyword = "CDATA" | "IGNORE" | "INCLUDE" | "RCDATA"             |
|                                                                              |
| As stated in 9.1 and note of 4.263, a "marked section end" in an RCDATA      |
| marked section is not recognized as such if it occurs in an entity           |
| imbedded in the marked section.                                              |
| If RCDATA is specified, remember in ms_ent_level the current                 |
| entity level, to avoid false "]]>" recognition.                              |
| "check_contextual_constraint" in dlfind.c will then deny a delimiter to be   |
| an MSC_MDC if it occurs at a higher entity level than the entity_level       |
| the RCDATA marked section started with.                                      |
+-----------------------------------------------------------------------------*/
boolean ParseState::marked_section_open()
{
   enum marked_section_value {
      INVALID_MSVL =  0x00,  // Value doesn't belong to the list
      TEMP_MSVL    =  0x01,  // Keywords code, weight shows precedence order
      INCLUDE_MSVL =  0x02,
      RCDATA_MSVL  =  0x04,
      CDATA_MSVL   =  0x08,
      IGNORE_MSVL  =  0x10
   };
   static const VarntChoiceEntry ChoicesMarkedSection[] = {
      { TEMP_VRN_ix     , TEMP_MSVL       },
      { INCLUDE_VRN_ix  , INCLUDE_MSVL    },
      { RCDATA_VRN_ix   , RCDATA_MSVL     },
      { CDATA_VRN_ix    , CDATA_MSVL      },
      { IGNORE_VRN_ix   , IGNORE_MSVL     },
      { NOT_FOUND_VRN_ix, INVALID_MSVL    }
   };
   int value = 0;
   int isErrorFound = FALSE;

   pPrsmode->push_mode(MD_MODE);
   while (!isErrorFound) {
      skip_ps();
      /*-----------------------+
      |(ps+, keywd)*, ps*, DSO |
      +-----------------------*/
      switch(get_curr_token()) {
      case NOT_FOUND_TK:
         {
            int new_value = pSyntax->match_rsvname_in_list(
               pEh, ChoicesMarkedSection
            );

            switch (new_value) {
            case INVALID_MSVL:
               errorh(INVKWD, pEh, ERH_S_SEV);
               isErrorFound = TRUE;
               break;
            case TEMP_MSVL:
            case INCLUDE_MSVL:
               break;
            default:         // RCDATA, CDATA, IGNORE
               value |= new_value;
               break;
            }
         }
         continue;
      case DSO_TK:
         break;           // OK...
      default:
         /*
         | PLUS_GRPO, MINUS_GRPO, RNI, LIT, LITA, MDC, DSC, GRPO
         | may be we could try to forgive the user for the DSO omission...
         */
         errorh(INVDLM, pEh, ERH_S_SEV);
         isErrorFound = TRUE;
         break;
      }
      break;   // break the for loop, if no "continue" bypassed it
   }
   pPrsmode->pop_mode();
   if (isErrorFound) return FALSE;

   if (++ms_level > pSyntax->pQuantity->value_of(TAGLVL_QTY_ix)) {
      errorh(BIGMSLVL, pEh, ERH_E_SEV);
      value = IGNORE_MSVL;
   }
   if (
      (pst0 < PROLOG_DONE_PST0) &&                       // In prolog, and
      ((value < IGNORE_MSVL) && (value >= RCDATA_MSVL))  // CDATA or RCDATA
   ) {
      errorh(INVMSPRO, pEh, ERH_E_SEV);                  // is an error
      value = IGNORE_MSVL;                               // pretend IGNORE
   }
   if (value >= IGNORE_MSVL) {
      pPrsmode->push_mode(IGNMS_MODE);
   }else if (value >= CDATA_MSVL) {
      pPrsmode->push_mode(FCDMS_MODE);
   }else if (value >= RCDATA_MSVL) {
      ms_ent_level = pEh->curr_entity_level(); // MSC_MDC won't be recognized
      pPrsmode->push_mode(RCDMS_MODE);
   }
   return TRUE;
}

/*-------------------------------------------------------marked_section_ignore-+
| When a marked section is being ignored, it's faster to skip all nested       |
| marked sections in a separate procedure, while matching MDSO/MSDCs           |
+-----------------------------------------------------------------------------*/
void ParseState::marked_section_ignore()
{
   int ms_ignore_depth = 1;

   for (;;) {
      recognize_delimiter();
      switch (get_curr_token()) {
      case MDO_DSO_TK:
         ++ms_ignore_depth;
         continue;
      case MSC_MDC_TK:
         if (--ms_ignore_depth == 0) break;
         continue;
      case NOT_FOUND_TK:
      default:
         if (pEh->is_Ee()) {
            errorh(UBLNMSO, pEh, ERH_T_SEV);
         }
         pEh->read();
         continue;
      }
      break;
   }
}

/*--------------------------------------------------------marked_section_close-+
| Enter here when a Marked Section End (MSC_DSC_TK) has been decoded           |
| Pop previous mode if not INCLUDE or TEMP.                                    |
| Return FALSE if unable to pop.                                               |
+-----------------------------------------------------------------------------*/
boolean ParseState::marked_section_close()
{
   if (--ms_level < 0) {
      ++ms_level;
      errorh(INVMSE, pEh, ERH_E_SEV);
      return FALSE;
   }else {
      /*
      | Pop regular mode if not INCLUDE (one of IGNORE/RCDATA/CDATA)
      */
      if (pPrsmode->get_mode() >= RCDMS_MODE) pPrsmode->pop_mode();
      return TRUE;
   }
}

/*----------------------------------------------------------------------skip_S-+
|Skip S separator and issue recognize_delimiter                                |
+-----------------------------------------------------------------------------*/
boolean ParseState::skip_S()
{
   if (!pEh->is_separator()) {
      recognize_delimiter();
      return FALSE;
   }
   while (recognize_delimiter() == NOT_FOUND_TK) {
      pEh->read();
      if (!pEh->is_separator()) {
         recognize_delimiter();
         break;
      }
   }
   return TRUE;
}

/*------------------------------------------------------------CheckActiveParam-+
| Parameter to check active param                                              |
+-----------------------------------------------------------------------------*/
struct CheckActiveParam {
   EntityHandler *pEh;
   UtilityControl& utilctl;
   Syntax *pSyntax;
   boolean isActive;
   CheckActiveParam(
      EntityHandler *pEhArg,
      UtilityControl& utilctlArg,
      Syntax *pSyntaxArg,
      boolean isActiveArg
   ) : utilctl(utilctlArg) {
     pEh = pEhArg, pSyntax = pSyntaxArg, isActive = isActiveArg;
   }
};

/*STATIC------------------------------------------------inst_check_active_type-+
| Check if a document type or a link type is active                            |
+-----------------------------------------------------------------------------*/
static void inst_check_active_type(void *pvBase)
{
   CheckActiveParam *pArg = (CheckActiveParam *)pvBase;
   char pszName[NAMELEN_MAX + 1];

   pArg->pSyntax->read_general_name(pArg->pEh, pszName);
   if (pArg->isActive == TRUE) {    // already activated group
      return;                       // no needs to check...
   }
   pArg->isActive = pArg->utilctl.is_active(pszName);
}

/*---------------------------------------------------------------------skip_ts-+
| Skip token separator                                                         |
|                                                                              |
| [70] ts = s | Ee | parameter entity reference                                |
|  [5]  s = SPACE | RE | RS | SEPCHAR                                          |
|                                                                              |
| IMPLEMENTATION NOTE:                                                         |
| - In a group, parameter entities reference must consist at least of one      |
|   token (skip_ts has been called directly, not from skip_ps or skip_ds).     |
| - When skip_ts is called to `skip ds', a parameter entity referencing        |
|   a PI is valid.  If so, return the entity address into ppEntPI.             |
|                                                                              |
| RETURN:                                                                      |
|    TRUE if at least one ts was found                                         |
|    FALSE if no ts were found                                                 |
+-----------------------------------------------------------------------------*/
boolean ParseState::skip_ts(EntityDef **ppEntPI)
{
   boolean isFound = FALSE;
   boolean check_empty_per;

   if (
      (pPrsmode->get_mode() == GRP_MODE) &&
      (pEh->curr_entity_level() == floor_level)
   ) {
      check_empty_per = TRUE;
   }else {
      check_empty_per = FALSE;
   }
   for (;;) {
      isFound = skip_S()? TRUE : isFound;
      switch (get_curr_token()) {
      case PERO_GRPO_TK:                     // Specific Parameter Entity
         isFound = TRUE;
         {
            CheckActiveParam arg(pEh, utilctl, pSyntax, FALSE); // 9.4.4:
            if (pst0 < PROLOG_DONE_PST0) {     // still in prolog
               errorh(INVDLM, pEh, ERH_E_SEV); // is an error
               arg.isActive = TRUE;            // recovery: make it active
            }
            parse_name_group(inst_check_active_type, &arg);
            if (!arg.isActive) {               // Not active:
               char pszName[NAMELEN_MAX+1];

               pSyntax->read_refd_entity_name( // get rid of it, by 9.4.4
                  pEh, pszName
               );
               continue;
            }
         }
         // else, fall thru
      case PERO_TK:                          // Parameter Entity
         isFound = TRUE;
         open_parameter_entity_reference(ppEntPI);
         if ((ppEntPI) && (*ppEntPI)) return TRUE;
         continue;
      case NOT_FOUND_TK:
         if (pEh->is_Ee()) {
            isFound = TRUE;
            if (pEh->curr_entity_level() > floor_level) {
               pEh->close();
               /*
               | Except if the PER is kept unparsed
               | ent_level is always equal here to floor_level.
               */
               if (
                  (check_empty_per) &&
                  (pEh->curr_entity_level() == floor_level)
               ) {
                  errorh(INVEMPPER, pEh, ERH_E_SEV);
               }
               continue;
            }else {
               if (!isQuietAtFloor) {
                  errorh(INVLNEST, pEh, ERH_T_SEV); // can't return from this
               }
            }
         }
         // fall thru
      default:
         return isFound;
      }
   }
}

/*---------------------------------------------------------------------skip_ps-+
| Skip parameter separator                                                     |
|                                                                              |
| ASSUME: mode is MD_MODE                                                      |
|                                                                              |
| [65] ps = s | Ee | parameter entity reference | comment                      |
|                                                                              |
| IMPLEMENTATION NOTE:                                                         |
| [70] ts = s | Ee | parameter entity reference                                |
|     ==> ps = ts | comment                                                    |
|                                                                              |
| RETURN:                                                                      |
|    TRUE if at least one ps was found                                         |
|    FALSE if no ps were found                                                 |
|                                                                              |
+-----------------------------------------------------------------------------*/
boolean ParseState::skip_ps()
{
   boolean isFound = FALSE;

   for (;;) {
      isFound = skip_ts()? TRUE : isFound;
      if (get_curr_token() == COM_TK) {        // COM (start) found
         if (!pSyntax->skip_to_delimiter(pEh, COM_VDL_ix)) {
            errorh(UBLNCOM, pEh, ERH_T_SEV);
         }
         isFound = TRUE;
         continue;
      }
      return isFound;
   }
}

/*------------------------------------------------------------skip_specific_ps-+
| Skip specific parameter separator                                            |
|                                                                              |
| "ps" in production [110] and [154] are not *really* ps, as stated by last    |
| paragraph of 11.1: they cannot contain Ee, nor parameter entity reference.   |
+-----------------------------------------------------------------------------*/
void ParseState::skip_specific_ps()
{
   for (;;) {
      skip_S();
      if (get_curr_token() != COM_TK) break;
      if (!pSyntax->skip_to_delimiter(pEh, COM_VDL_ix)) {
         errorh(UBLNCOM, pEh, ERH_T_SEV);
      }
   }
}

/*--------------------------------------------------skip_inferable_required_ps-+
| Required by ISO 8879, when para 3 of 10.1.1 doesn't apply.                   |
| However YAO can pursue (severity W).                                         |
+-----------------------------------------------------------------------------*/
void ParseState::skip_inferable_required_ps()
{
   e_LexToken prev_token = get_curr_token();

   if (
      (!skip_ps()) &&
      (prev_token == NOT_FOUND_TK) &&
      (get_curr_token() == NOT_FOUND_TK)
   ) {
      errorh(NOPS, pEh, ERH_E_SEV);
   }
}

/*--------------------------------------------------------------------skip_cdc-+
| Skip a comment declaration                                                   |
|                                                                              |
| [91] comment declaration = mdo, (comment, (s | comment)*)?, mdc              |
+-----------------------------------------------------------------------------*/
void ParseState::skip_cdc()
{
   pPrsmode->push_mode(MD_MODE);
   for (;;) {
      if (!pSyntax->skip_to_delimiter(pEh, COM_VDL_ix)) {
         errorh(UBLNCOM, pEh, ERH_T_SEV);
      }
      skip_S();
      switch (get_curr_token()) {
      case MDC_TK:
           break;
      case COM_TK:
           continue;
      default:
           errorh(NOMDC, pEh, ERH_T_SEV);
           break;
      }
      break;
   }
   pPrsmode->pop_mode();
}

/*-----------------------------------------------------adjust_recognition_mode-+
| As stated in 9.1 and note of 4.263, an "end-tag" in RCDATA-content is not    |
| recognized as such if it occurs in an imbedded entity.                       |
| If RCDATA is specified, remember in cn_ent_level the current entity level to |
| avoid false "/", "</", or "</>" recognition.  "check_contextual_constraint"  |
| will then deny a delimiter to be an end-tag if it occurs at a higher entity  |
| level than the entity_level the RCDATA-content started with.                 |
+-----------------------------------------------------------------------------*/
void ParseState::adjust_recognition_mode(e_Dcdata dcdata)
{
   switch (dcdata) {                      // ISO 8879 - Note at 7.6
   case NONR_DCDATA:
      pPrsmode->set_mode(FCD_MODE);
      break;
   case RPLC_DCDATA:
      pPrsmode->set_mode(RCD_MODE);
      cn_ent_level = pEh->curr_entity_level(); // avoid false end-tag recognition
      break;
   default:                               // not dcdata
      pPrsmode->set_mode(CON_MODE);
      break;
   }
}

/*--------------------------------------------------------------------open_OCS-+
| Open an OCS, and ensure its validity...                                      |
|                                                                              |
| clsOcs must be one of:                                                       |
|    UTSYNU_XFC:   utility user-modified ocs                                   |
|    UTSYNL_XFC:   utility local-system ocs                                    |
|    UTSYNR_XFC:   utility refcon-system ocs                                   |
|                                                                              |
| isWriteMode can be TRUE only for UTSYNU_XFC                                  |
+-----------------------------------------------------------------------------*/
ApplicationHandle ParseState::open_OCS(e_TextClass clsOcs, boolean isWriteMode)
{
   ApplicationHandle ahlAccess;
   char  const * const ocs_banner = OCS_BANNER;

   /*
   | Open the OCS file, and check
   */
   if (!pEh->open(
         utilctl.build_compiled_exid(clsOcs), &ahlAccess, isWriteMode
      )
   ) {
      if (isWriteMode) {
         errorh(BADOPSWR, ERH_T_SEV);
      }else {
         errorh(BADOPSRD, ERH_T_SEV, 1);
      }
   }
   if (isWriteMode) {
      pEh->write(ahlAccess, ocs_banner+1, 1 + *ocs_banner);
   }else {
      char  banner[81];
      unsigned int i;

      if (
         (i=1, !pEh->read(ahlAccess, banner, 1 + *ocs_banner)) ||
         (i=2, strlen(banner) != *ocs_banner) ||
         (i=3, strcmp(banner, ocs_banner+1))
      ) {
         errorh(BADOPSRD, ERH_T_SEV, i);
      }
   }
   return ahlAccess;
}

/*--------------------------------------------------------------------open_ODT-+
| Open an ODT, or an OLT and ensure its validity...                            |
+-----------------------------------------------------------------------------*/
ApplicationHandle ParseState::open_ODT(ExternId *pExid, boolean isWriteMode)
{
   char  const * const odt_banner = ODT_BANNER;
   ApplicationHandle ahlAccess;

   /*
   | Open the ODT file, and check
   */
   if (!pEh->open(
         pExid, &ahlAccess, isWriteMode
      )
   ) {
      if (isWriteMode) {
         errorh(BADOPDWR, ERH_T_SEV);
      }else {
         errorh(BADOPDRD, ERH_T_SEV, 1);
      }
   }
   if (isWriteMode) {
      pEh->write(ahlAccess, odt_banner+1, 1 + *odt_banner);
   }else {
      char  banner[81];
      unsigned int i;

      if (
         (i=1, !pEh->read(ahlAccess, banner, 1 + *odt_banner)) ||
         (i=2, strlen(banner) != *odt_banner) ||
         (i=3, strcmp(banner, odt_banner+1))
      ) {
         errorh(BADOPDRD, ERH_T_SEV, i);
      }
   }
   return ahlAccess;
}
/******************************************************************************/
