/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// rsvname.c
///  Class for SGML reserved names
//   Pierre G. Richard 93/10/25 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include "gendef.h"
#include "enthdlr.h"
#include "rsvname.h"
#include "refstr.h"
#include "skmgr.h"

/*----------------------------------------------------------------ReservedName-+
| Constructor from scratch: this is the minimal CORE Concrete Syntax           |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
ReservedName::ReservedName()
{
   memcpy(
      aRsvnameVarnt,
      ReferenceString::get_aKywdVarntNameValues(),
      sizeof aRsvnameVarnt
   );
}

/*----------------------------------------------------------------ReservedName-+
| Constructor from OCS                                                         |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
ReservedName::ReservedName(
   ApplicationHandle ahlAccess,       // OCS file handle
   char *strbuf                       // string area of this OCS
)
{
   EntityHandler::read(ahlAccess, aRsvnameVarnt, sizeof aRsvnameVarnt);
   /*
   | Relocate aRsvnameVarnt
   */
   for (int i=0; i < sizeof aRsvnameVarnt / sizeof aRsvnameVarnt[0]; ++i) {
      aRsvnameVarnt[i].p = strbuf + aRsvnameVarnt[i].ofs;
   }
}

/*PROTECTED-------------------------------------------------------ReservedName-+
| Constructor for building reserved names from the SGML declaration            |
+-----------------------------------------------------------------------------*/
ReservedName::ReservedName(int)
{
   memset(aRsvnameVarnt, 0, sizeof aRsvnameVarnt);
}

/*-------------------------------------------------------------NewReservedName-+
| NewReservedName constructor                                                  |
+-----------------------------------------------------------------------------*/
NewReservedName::NewReservedName(
   StringKeeper& skArg
) : ReservedName(0), sk(skArg)
{
}

/*---------------------------------------------------------------------name_of-+
| Derived: must be able to return 0 in case the name has not yet been replaced.|
| Also, use a string keeper to point the replacement name.                     |
+-----------------------------------------------------------------------------*/
const char * NewReservedName::name_of(const e_IxRsvnameVarnt ix) const
{
   if (aRsvnameVarnt[ix].ofs) {
      return sk.point(aRsvnameVarnt[ix].ofs);
   }
   return 0;    // Not yet replaced!
}

/*--------------------------------------------------------clearance_to_replace-+
| Check if a reserved name is allowed to be replaced.                          |
+-----------------------------------------------------------------------------*/
e_ErrorCode NewReservedName::clearance_to_replace(
   const char *pszRsvnameReplaced, e_IxRsvnameVarnt * const pixRsvnameReplaced
) const
{
   e_ErrorCode msgcode = ReferenceString::clearance_to_replace(
      pszRsvnameReplaced, pixRsvnameReplaced
   );

   if (((msgcode == ER_NULL) && (aRsvnameVarnt[*pixRsvnameReplaced].ofs))) {
      msgcode = DUPVANAM;                 // already replaced!
      *pixRsvnameReplaced = NOT_FOUND_VRN_ix;
   }
   return msgcode;
}

/*---------------------------------------------------------------------replace-+
| Replace a reserved name or tell why you can't.                               |
| Return ER_NULL if OK, of INVRPTNAM, or DUPNAMEREP                            |
|                                                                              |
|   on INPUT,                                                                  |
|      *pixVarnt = the index of the reserved name to be replaced               |
|      pszRsvnameReplacement points the replacement name,                      |
|   on OUPUT,                                                                  |
|      if DUPNAMEREP is returned:                                              |
|         *pixVarnt = index of the replaced this replacement replaces          |
|      else                                                                    |
|         unchanged                                                            |
+-----------------------------------------------------------------------------*/
e_ErrorCode NewReservedName::replace(
   const char *pszRsvnameReplacement,
   e_IxRsvnameVarnt * const pixRsvnameReplaced
)
{
   /*
   |  First check the replacement is not already a keyword in the CS
   */
   e_IxRsvnameRefer ixRefer;
   ReferenceString::get_index(pszRsvnameReplacement, &ixRefer);
   if (ixRefer != NOT_FOUND_RRN_ix) {        // already a keyword
      return INVRPTNAM;
   }
   /*
   |  Now, check that this replacement name is not already in use
   */
   for (int i=0; i < sizeof aRsvnameVarnt / sizeof aRsvnameVarnt[0]; ++i) {
      if (
         (aRsvnameVarnt[i].ofs) &&
         (!strcmp(pszRsvnameReplacement, sk.point(aRsvnameVarnt[i].ofs)))
      ) {
         *pixRsvnameReplaced = (e_IxRsvnameVarnt)i;
         return DUPNAMREP;
      }
   }
   /*
   |  It's OK: replace
   */
   aRsvnameVarnt[*pixRsvnameReplaced].ofs = sk.park(
      pszRsvnameReplacement, 1+strlen(pszRsvnameReplacement)
   );
   return ER_NULL;
}

/*--------------------------------------------------------default_non_replaced-+
| Default the non-replaced name with the ones of the given (core) syntax       |
+-----------------------------------------------------------------------------*/
void NewReservedName::default_non_replaced(const Syntax *pSyntaxCore)
{
   ReservedName *pRsvnameCore = pSyntaxCore->pRsvname;

   for (int i=0; i < sizeof aRsvnameVarnt / sizeof aRsvnameVarnt[0]; ++i) {
      if (!aRsvnameVarnt[i].ofs) {
         const char *pszNameCore = pRsvnameCore->name_of((e_IxRsvnameVarnt)i);
         aRsvnameVarnt[i].ofs = sk.park(pszNameCore, 1+strlen(pszNameCore));
      }
   }
}

/*----------------------------------------------------------------------unload-+
| Write the reserved names to disk                                             |
+-----------------------------------------------------------------------------*/
void NewReservedName::unload(ApplicationHandle ahlAccess)
{
   EntityHandler::write(ahlAccess, aRsvnameVarnt, sizeof aRsvnameVarnt);
}
/******************************************************************************/
