/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// samgr.c
///  Sorted Array class
//   Pierre G. Richard 87/05/27 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
/*
| These procedures are designed to gather in a sorted array elements
| of a same kind
|
| Typical calls are:
|
| At construction:
|     +---------------------------------------+
|     | SortArray(                            |
|     |    const unsigned int iItemSize,      |
|     |    int (*compare)(                    |
|     |       const void *pItem1,             |
|     |       const void *pItem2,             |
|     |       const void * const pBase        |
|     |    ),                                 |
|     |    unsigned int * const piTotalCount, |
|     |    void ** const ppArray,             |
|     |    const void *pBase                  |
|     | ) :                                   |
|     +---------------------------------------+
|     request the creation of a new array,
|     each element having a length of `iItemSize',
|     the user-supplied compare function being `compare'.
|
|     When the SortArray instance is destructed, the sorted array of
|     items is returned at address pointed by `ppArray', while the
|     count of entered items will be copied at address pointed by
|     `piTotalCount'.
|
| When a new element is encountered,
|     +------------------------------------+
|     | void enter(void *pItem)            |
|     +------------------------------------+
|     stores this new element in the array.
|
*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <memory.h>
#include "samgr.h"
#include "sort.h"

/*-------------------+
|Local data and types|
+-------------------*/
enum { INCRT = 50 };

/*-------------------------------------------------------------------SortArray-+
| SortArray constructor                                                        |
+-----------------------------------------------------------------------------*/
SortArray::SortArray(
   const unsigned int iItemSizeArg,
   int (*compareArg)(
      const void *pItem1, const void *pItem2, const void * const pBase
   ),
   unsigned int * const piTotalCountArg,
   void ** const ppArrayArg,
   const void *pBaseArg
) :
   pBase(pBaseArg),
   ppArray(ppArrayArg),               // where to return the built array
   piTotalCount(piTotalCountArg),     // and the count of items
   iItemSize(iItemSizeArg)
{
   compare = compareArg;

   count = -1;
   max = INCRT;
   ppItem = (void **)malloc(max * sizeof(void **));
}

/*-----------------------------------------------------------------------enter-+
| Install a new item.                                                          |
+-----------------------------------------------------------------------------*/
void SortArray::enter(void *pItem)
{
   if (++count >= max) {
      ppItem = (void **)realloc(ppItem, sizeof(void **) * (max += INCRT));
   }
   memcpy((ppItem[count] = malloc(iItemSize)), pItem, iItemSize);
}

/*------------------------------------------------------------------sa_compare-+
| Compare                                                                      |
+-----------------------------------------------------------------------------*/
inline int SortArray::sa_compare(int ix1, int ix2)
{
   return (*compare)(ppItem[ix1], ppItem[ix2], pBase);
}

/*-----------------------------------------------------------------sa_exchange-+
| Exchange                                                                     |
+-----------------------------------------------------------------------------*/
inline void SortArray::sa_exchange(int ix1, int ix2)
{
   void *pItem;

   pItem = ppItem[ix1];
   ppItem[ix1] = ppItem[ix2];
   ppItem[ix2] = pItem;
}

/*STATIC----------------------------------------------------------compare_item-+
| Compare (for wqsort)                                                         |
+-----------------------------------------------------------------------------*/
static int compare_item(int ix1, int ix2, void *pvBase)
{
   return ((SortArray *)pvBase)->sa_compare(ix1, ix2);
}

/*STATIC---------------------------------------------------------exchange_item-+
| Exchange (for wqsort)                                                        |
+-----------------------------------------------------------------------------*/
static void exchange_item(int ix1, int ix2, void *pvBase)
{
   ((SortArray *)pvBase)->sa_exchange(ix1, ix2);
}

/*------------------------------------------------------------------~SortArray-+
|Terminate: create and return the array                                        |
+-----------------------------------------------------------------------------*/
SortArray::~SortArray()
{
   *piTotalCount = ++count;
   if (compare) {
      Sort::quick(count, compare_item, exchange_item, this);
   }
   void *pArray = *ppArray = malloc(count * iItemSize);
   for (int i=0; i < count; ++i) {
      memcpy(pArray, ppItem[i], iItemSize);
      pArray = (char *)pArray + iItemSize;
      free(ppItem[i]);
   }
   free(ppItem);
}
/******************************************************************************/
