/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// server.c
///  A few Entity Manager functions...
//   Pierre G. Richard 89/08/15 - Petrus Services and Consulting, Ltd.
//   Added PUBLIC ID resolver - Pekka Kataja 1995 #ifdef _PKHACKED ... #endig
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include "gendef.h"            // general definitions
#include "server.h"            // Application services: I/O, memory
#include "externid.h"          // External identifier

#include <stdio.h>             // Standard I/O
#include <stdlib.h>            // Standard Library
#include <string.h>            // String manipulation
#include <errno.h>
#include <ctype.h>

/*------------------+
| Local Definitions |
+------------------*/
#define FILE_BUFSIZE   BUFSIZ  // Buffer size: deflt value from stdio.h
#define TERM_BUFSIZE   140     // see CMS RDTERM

const char pszMsgRemoveFail[] = "Remove of file \"%s\" failed\n";
const char pszMsgTempNameFail[] = "Cannot make a temporary name\n";
const char pszMsgTermInput[] = "Terminal Input:\n";
const char pszMsgOpenFileFail[] = "Can't open file %s\n";
const char pszMsgAllocateFail[] = "Out of heap space!\n";

/*--------------+
|Local Functions|
+--------------*/
static unsigned int pubid_to_fspec(const char *, char *);

/*--------------------------------------------------------------------Document-+
|                                                                              |
+-----------------------------------------------------------------------------*/
Document::Document(
   const char *pszPrimFileSpecArg,        // Primary Entity File Spec
   const char *pszOcsFileNameArg,         // OCS File to write / read
   const char *pszTerminalFileSpecArg,    // Reserved name for terminal
   boolean isProductionArg,               // OK to use a precompiled DTD
   boolean isSortAttributesArg            // Sort attribute lists
)
{
   pszPrimFileSpec = pszPrimFileSpecArg;
   pszOcsFileName = pszOcsFileNameArg;
   pszTerminalFileSpec = pszTerminalFileSpecArg;
   isProduction = isProductionArg;
   isSortAttributes = isSortAttributesArg;
}

/*----------------------------------------------------------------------Server-+
|                                                                              |
+-----------------------------------------------------------------------------*/
Server::Server(const Document& docArg)
{
   doc = docArg;
   pFyleTop = 0;
   pFyleCur = 0;
   pFyleTerm = 0;
   err_fd = stderr;
   pchFnChain = 0;
}

/*---------------------------------------------------------------------~Server-+
| Functional Description: Do cleanup after finished with system services.      |
|                         Release storage for all filenames.  Delete           |
|                         temporary files and release storage for all FYLES.   |
|                                                                              |
| Return: none                                                                 |
+-----------------------------------------------------------------------------*/
Server::~Server()
{
   char *pszFn, *pszFnNext;

   /*-------------------------+
   |delete temporary files and|
   |release the filename areas|
   +-------------------------*/
   pszFn = pchFnChain;                         // First filename area
   while (pszFn) {                             // Loop on the filenames
      pszFnNext = *((char **)pszFn);           // Next one in pszFnNext
      if ((*(pszFn + sizeof(char *))=='T') &&  // temporary files are marked T
         (-1 == remove(pszFn + 1 + sizeof(char *)))) {  // delete them
         printf(pszMsgRemoveFail, pszFn + 1 + sizeof(char *));
      }
      free(pszFn);                             // free this filename area
      pszFn = pszFnNext;                       // loop to the next one
   }
}

/*PRIVATE----------------------------------------------------------set_current-+
| Function Name: set_current                                                   |
|                                                                              |
| Functional Description: Check the fyle pointer pFyleToFind belongs to the    |
|                         list of pointers to opened FYLE structures           |
|                                                                              |
| Parameters: pFyleToFind fyle pointer                                         |
|                                                                              |
| Return: TRUE  pFyleToFind points one of the opened FYLE structures           |
|         FALSE pFyleToFind is not a known FYLE pointer                        |
+-----------------------------------------------------------------------------*/
boolean Server::set_current(FYLE *pFyleToFind)
{
   FYLE *pFyle;

   for (pFyle=pFyleTop; pFyle; pFyle=pFyle->next) { // for all FYLEs
      if (pFyle == pFyleToFind) {                   // Is it the correct one ?
         pFyleCur = pFyle;                          // Yes: update current FYLE
         return TRUE;                               // Exit: found!
      }
   }
   return FALSE;                                    // Exit: not found...
}

/*STATIC-------------------------------------------------------pubid_to_fspec-+
| Map a PUBLIC Identifier to a true filespec.                                 |
|                                                                             |
| A static table is currently used.  A real Application could instead use     |
| an external file.                                                           |
+----------------------------------------------------------------------------*/
#ifdef _PKHACKED
#include <pubid.h>
#endif

static unsigned int pubid_to_fspec(const char *id, char *buffer)
{

#ifdef _PKHACKED
	return ( ResolvePublicId::resolve ( id , buffer ) ) ;
#endif


   static char const * const pubdef[] = {
      "ISO 8879:1986//CAPACITY Reference//EN"                   ,"standard.CAP",
      "ISO 8879:1986//SYNTAX Reference//EN"                     ,"standard.CSD",
      "ISO 8879:1986//ENTITIES Added Latin 1//EN"               ,"iso$lat1.ent",
      "ISO 8879:1986//ENTITIES Added Latin 2//EN"               ,"iso$lat2.ent",
      "ISO 8879:1986//ENTITIES Greek Letters//EN"               ,"iso$grk1.ent",
      "ISO 8879:1986//ENTITIES Monotoniko Greek//EN"            ,"iso$grk2.ent",
      "ISO 8879:1986//ENTITIES Greek Symbols//EN"               ,"iso$grk3.ent",
      "ISO 8879:1986//ENTITIES Alternative Greek Symbols//EN"   ,"iso$grk4.ent",
      "ISO 8879:1986//ENTITIES Russian Cyrillic//EN"            ,"iso$cyr1.ent",
      "ISO 8879:1986//ENTITIES Non-Russian Cyrillic//EN"        ,"iso$cyr2.ent",
      "ISO 8879:1986//ENTITIES Numeric and Special Graphic//EN" ,"iso$num.ent",
      "ISO 8879:1986//ENTITIES Diacritical Marks//EN"           ,"iso$dia.ent",
      "ISO 8879:1986//ENTITIES Publishing//EN"                  ,"iso$pub.ent",
      "ISO 8879:1986//ENTITIES Box and Line Drawing//EN"        ,"iso$box.ent",
      "ISO 8879:1986//ENTITIES General Technical//EN"           ,"iso$tech.ent",
      "ISO 8879:1986//ENTITIES Added Math Symbols: Ordinary//EN","iso$amso.ent",
      "ISO 8879:1986//ENTITIES Added Math Symbols: Binary Operators//EN",
                                                                 "iso$amsb.ent",
      "ISO 8879:1986//ENTITIES Added Math Symbols: Relations//EN",
                                                                 "iso$amsr.ent",
      "ISO 8879:1986//ENTITIES Added Math Symbols: Negated Relations//EN",
                                                                 "iso$amsn.ent",
      "ISO 8879:1986//ENTITIES Added Math Symbols: Arrow Relations//EN",
                                                                 "iso$amsa.ent",
      "ISO 8879:1986//ENTITIES Added Math Symbols: Delimiters//EN",
                                                                 "iso$amsc.ent",
      "-//IBM/PGM 5684-025//ENTITIES Added Math Symbols//EN",    "ibmmath.ent"
   };

   for (int i=0; i < sizeof pubdef /(2 * sizeof pubdef[0]); i += 2) {
      if (!strcmp(pubdef[i], id)) {
         strcpy(buffer, pubdef[i+1]);
         return strlen(buffer);
      }
   }
   return 0;    // Not found
}

/*---------------------------------------------------------------------sr_find-+
| Functional Description: Given a named resource identifier descriptor,        |
|                         find an associated ID_handle.                        |
|                         The class of the external identifier will            |
|                         determine how to get the system ID handle.           |
|                         UTODTW and UTSYNU requests are to create a           |
|                         temporary internal use only file ID for a            |
|                         compiled DTD file (ODT), and compiled SGML           |
|                         declarations file (OCS) respectively.                |
|                         When not in compilation mode, don't save the         |
|                         compiled .ODT and .OCS files.                        |
|                                                                              |
|                         Allocate space for remembering each filename.        |
|                         This filename is prefixed by a letter to describe    |
|                         the kind of file (see sr_xenop).                     |
|                                                                              |
|  FIND is called:                                                             |
|  - before to open an external resource, or                                   |
|  - before to pass an external resource to the *application*:                 |
|       - not parseable or SUBDOC ENTITY found in content or                   |
|         in an Attribute Value Spec of an Element                             |
|       - NOTATION (either directly in an AVS or ENTITY associated)            |
|  It is guaranteed this routine will not be called more than once for the     |
|  same resource.                                                              |
|                                                                              |
|  This routine must put in the field ApplicationHandle an unique identifier   |
|  that will be used to:                                                       |
|  - open the file; or,                                                        |
|  - update the ApplicationHandle field in the                                 |
|    EntityDef block / EntityNotation block                                    |
|                                                                              |
|  When there is not a SYSTEM of PUBLIC external id the application should     |
|  have a system default.  For the purposes of this application there is no    |
|  system default.                                                             |
|                                                                              |
| NOTES: NOTATIONs are generally mapped to an executable module                |
|       while ENTITIES are just Non-SGML files, or SUBDOC.                     |
|                                                                              |
|       It's up to the *application* to map the external ressource to          |
|       whatever kind of "system address" is suitable.  Some prefer a          |
|       handle, some prefer a pointer to whatever they want.  For this         |
|       reason, the ApplicationHandle is an union that can handle both.        |
|                                                                              |
|       There is a conceptual difference between Id_AHL (ID Handles            |
|       updated by FIND) and Acc_AHL (ACCESS handles returned by OPEN).        |
|       (even if these values are identical for some *applications*)           |
|                                                                              |
|       Here, the ID Handles are pointers to the fully qualified name,         |
|       and the ACCESS handles are pointers to the FYLE structure associated   |
|       to opened entities.                                                    |
|                                                                              |
| Parameters:                                                                  |
|                                                                              |
| Return: 0:  Normal                                                           |
|         2:  Warning, external entity is already opened                       |
|         4:  Could not associate a Fileid-AHL                                 |
+-----------------------------------------------------------------------------*/
int Server::sr_find(
   const ExternId *pExid, ApplicationHandle& ahlIdent
)
{
#if (INT_MAX == SHRT_MAX)
   static char const LOCS_NAME_STR[] = "LOCAL16";    // System OCS
#else
   static char const LOCS_NAME_STR[] = "LOCAL32";    // System OCS
#endif
   static char const OCS_EXT_STR[] = ".OCS";         // OCS extension
   static char const ODT_EXT_STR[] = ".ODT";         // ODT extension
   static char const DTD_EXT_STR[] = ".DTD";         // DTD extension

   boolean isInputFile = TRUE;
   char const *pszFileSpec;
   char buffer[FILENAME_MAX+1];               // buffer to build filespec
   boolean isTempfile = FALSE;

   switch(pExid->assigned_class()) { // Which class ?
   case PRIMARY_XFC:         // (primary entity)    --,  --, No Name
      pszFileSpec = doc.pszPrimFileSpec;      // set the file spec
      break;
   case UTODTR_XFC:          // ODT file read      Sid, Pid, doctype name
   case UTOLTR_XFC:          // OLT file read       --,  --, linktype name
      if (!doc.isProduction) {                // standard mode
         return 4;                            // pretend it's not found
      }else {                                 // else, production mode
         strcpy(buffer, pExid->name());
         strcat(buffer, ODT_EXT_STR);
      }
      pszFileSpec = buffer;
      break;
   case UTODTW_XFC:          // ODT file write     Sid, Pid, doctype name
   case UTOLTW_XFC:          // OLT file write      --,  --, linktype name
      if (!doc.isProduction) {                // standard mode
         if (!tmpnam(buffer)) {               // write to a temp file
            printf(pszMsgTempNameFail);
            return 4;                         // problem with tempnam
         }
         isTempfile = TRUE;                   // to tell it's a tempfile
      }else {                                 // production mode: save ODT
         strcpy(buffer, pExid->name());
         strcat(buffer, ODT_EXT_STR);
      }
      isInputFile = FALSE;
      pszFileSpec = buffer;
      break;
   case UTSYNL_XFC:          // No SGML dcl found: default syntax
   case UTSYNR_XFC:          // SGML dcl, but SCOPE INSTANCE and not in INSTANCE
      if (doc.pszOcsFileName) {               // take user preference
         strcpy(buffer, doc.pszOcsFileName);
      }else {                                 // if none, take LOCAL
         strcpy(buffer, LOCS_NAME_STR);
      }
      strcat(buffer, OCS_EXT_STR);
      pszFileSpec = buffer;
      break;
   case UTSYNU_XFC:          // SGML decl found: OCS file to write
      if (doc.pszOcsFileName) {               // take user preference
         strcpy(buffer, doc.pszOcsFileName);
         strcat(buffer, OCS_EXT_STR);
      }else {                                 // if none, tempname
         if (!tmpnam(buffer)) {
            printf(pszMsgTempNameFail);
            return 4;                         // problem with tempnam
         }
         isTempfile = TRUE;
      }
      isInputFile = FALSE;
      pszFileSpec = buffer;
      break;
   case DTD_XFC:             // <!DOCTYPE          Sid, Pid, doctype name
      if ((pExid) && (pExid->is_system())) {  // at least, SYSTEM ""
         strcpy(buffer, pExid->system_id());
      }else {
#ifdef _PKHACKED 
	 pubid_to_fspec ( pExid->public_id () , buffer ) ;
#else
         strcpy(buffer, pExid->name());
         strcat(buffer, DTD_EXT_STR);
#endif // end _PKHACKED
      }
      pszFileSpec = buffer;
      break;
   default:
      // CHARSET_XFC:           <!SGML.. CHARSET   Sid, Pid, no name
      // NOTATION_XFC:          <!NOTATION         Sid, Pid, notation name
      // NONSGML_XFC:           <!ENTITY C/N/SDATA Sid, Pid, entity name
      // CAPACITY_XFC:          <!SGML.. CAPACITY  Sid, PId, no name
      // DOCUMENT_XFC:          <!ENTITY (no type) Sid, Pid, entity name
      // ELEMENTS_XFC:          <!ENTITY (no type) Sid, Pid, entity name
      // ENTITIES_XFC:          <!ENTITY (no type) Sid, Pid, entity name
      // LPD_XFC:               <!LINKTYPE         Sid, Pid, linktype name
      // SHORTREF_XFC:          <!ENTITY (no type) Sid, Pid, entity name
      // SUBDOC_XFC:            <!ENTITY SUBDOC    Sid, Pid, entity name
      // SYNTAX_XFC:            <!SGML.. SYNTAX    Sid, Pid, No Name
      // TEXT_XFC:              <!ENTITY (no type) Sid, Pid, entity name
      if (
         (pExid->is_public()) && (pubid_to_fspec(pExid->public_id(), buffer))
      ) {
         pszFileSpec = buffer;          // known public identifier given
      }else if (pExid->is_system()) {   // or system identifier
         pszFileSpec = pExid->system_id();
      }else {
         return 0;                      // if none, don't tell parser
      }
      break;
   }

   /*----------------------------------+
   |Copy the fspec into a mallocated   |
   |structure, linked with the previous|
   |one to be able to release it later |
   |The filename is preceded by letter |
   |'T' if it is temporary, else 'X'.  |
   +----------------------------------*/
   {
   //***>NOTE: not implemented
   // if (!isTempfile) {
   //    char *filename_save = duplicate_string(pszFileSpec);
   //    findfile(filename_save, isInputFile, buffer);
   //    free(filename_save);
   //    pszFileSpec = buffer;
   // }
      isInputFile;          // to avoid the warning until implemented
      char **ppch, *pch;
      unsigned int iLenFileSpec = strlen(pszFileSpec);
      ppch = (char **)malloc(sizeof(char *) + iLenFileSpec + 2);
      if (!ppch) {
         printf(pszMsgAllocateFail);
         return 4;
      }
      *ppch = pchFnChain;             // chain to previous filenames
      pchFnChain=(char *)ppch;        // remember current filename
      pch = (char *)(ppch+1) + 1;
      memcpy(pch, pszFileSpec, iLenFileSpec);
      pch[iLenFileSpec] = '\0';
      if (isTempfile) {
         *(pch-1) = 'T';
      }else {
         *(pch-1) = 'X';
      }
      ahlIdent.p = pch;                // ahlIdent.p is the name
   }
   return 0;
}

/*--------------------------------------------------------------------sr_xenop-+
| Functional Description: Open a text file (external entity) using the         |
|                         ahlIdent found by sr_find.                           |
|                         Open the file.                                       |
|                         Allocate space for a FYLE struct to remember things  |
|                         about the file.                                      |
|                         Place the FYLE pointer in pahlAccess                 |
|                                                                              |
| Note: some "applications" can consider identifying and accessing returns     |
|       the same handle (ahlIdent == ahlAccess), some don't.                   |
|                                                                              |
| Parameters:                                                                  |
|                                                                              |
| Return: 0  Normal                                                            |
|         4  File is already opened                                            |
|         6  Cannot open (ex: UNIX exclusive access)                           |
|         8  Invalid FileId-AHL                                                |
+-----------------------------------------------------------------------------*/
int Server::sr_xenop(
   const ApplicationHandle ahlIdent,
   ApplicationHandle& ahlAccess
)
{
   FYLE *pFyle;
   char *filespec = (char *)ahlIdent.p;

   if (!strcmp(filespec, doc.pszTerminalFileSpec)) {        // TERMINAL:
      if (pFyleTerm) return 4;                  // already opened
      pFyle = (FYLE *)malloc(sizeof(FYLE) + TERM_BUFSIZE + 1);
      if (!pFyle) {
         printf(pszMsgAllocateFail);
         return 6;                              // or cannot be "opened"
      }
      pFyle->buffer = (char *)(pFyle + 1);
      pFyle->iBufSize = TERM_BUFSIZE;
      pFyleTerm = pFyle;                        // remember fyle pointer
      printf(pszMsgTermInput);
   }else {
      FILE *fd = fopen(filespec, "r");          // Try to open the file

      if (!fd) {
         printf(pszMsgOpenFileFail, filespec);
         return 8;                               // cannot open: return
      }
      /*--------------------------------------------+
      | Associate a FYLE structure for bufferizing, |
      | RS/RE handling, a.s.o                       |
      +--------------------------------------------*/
      pFyle = (FYLE *)malloc(sizeof(FYLE) + FILE_BUFSIZE);
      if (!pFyle) {
         printf(pszMsgAllocateFail);
         return 6;
      }
      pFyle->buffer = (char *)(pFyle + 1);     // set buffer pointer
      pFyle->iBufSize = FILE_BUFSIZE;          // and buffer size
      pFyle->fd = fd;                          // remember handle
   }
   pFyle->pszName = filespec;                  // mallocated string in FnChain
   pFyle->isTruncated = FALSE;
   pFyle->isFirst = TRUE;
   /*--------------------+
   | Enter the new FYLE  |
   | into the FYLE queue |
   +--------------------*/
   {
      FYLE **ppFyle = &(pFyleTop);

      while (*ppFyle) ppFyle = &(*ppFyle)->next;
      ahlAccess.p = pFyleCur = *ppFyle = pFyle;
      pFyle->next = 0;                         // tells it's the last file

   }
   return 0;
}

/*--------------------------------------------------------------------sr_xencl-+
| Functional Description: Close a text file (external entity).                 |
|                                                                              |
| Parameters:                                                                  |
|                                                                              |
| Return: 0     Normal                                                         |
|         8     Cannot close                                                   |
+-----------------------------------------------------------------------------*/
int Server::sr_xencl(ApplicationHandle ahlAccess)
{
   FYLE *pFyle = (FYLE *)ahlAccess.p;   // pointer part of ApplicationHandle

   if ((!pFyle) || ((pFyle != pFyleCur) && (!set_current(pFyle)))) {
      return 8;                         // Bad Handle: return
   }
   if (pFyle == pFyleTerm) {            // Do not fclose the TERMINAL
      pFyleTerm = 0;                    // just nullify the pointer
   }else {                              // for regular files
      if (fclose(pFyle->fd)) return 8;  // if fclose fails, return
   }
   {                                    // Unlink the FYLE chain
      FYLE **pFylep = &(pFyleTop);
      while (*pFylep != pFyle) pFylep = &(*pFylep)->next; // find file to close
      *pFylep = pFyle->next;            // Fix link chain
   }
   free(pFyle);                         // remove FYLE structure
   pFyleCur = 0;                        // current file no more valid
   return 0;
}

/*--------------------------------------------------------------------sr_xenrd-+
| Functional Description: Read a text file (external entity).                  |
|                                                                              |
| There is no function to perform basic IO in the IBM C SAA/CPI set:           |
| for text files, we use fgets and assume there is no such things as           |
| extra-control characters or '\0' within these files.                         |
| - if there is a '\0' within the record read, then strlen does not return     |
|   the number of bytes read                                                   |
| - if the record is longer than the receiving buffer but contains             |
|   a '\n' at offset BUFSIZE, then record truncation is not detected           |
|                                                                              |
| Note 1: When a document is reparsed (RESUME_PSF), only the Application knows |
|         what the row, col and offset the primary entity is resumed from.     |
|         These informations need to be put in some place...                   |
|         The code must be inserted where it's indicated "** See Note 1 **".   |
|                                                                              |
| Parameters:                                                                  |
|                                                                              |
| Return: 1:    RE/RS to be inserted                                           |
|         4:    End-of-File                                                    |
|         8:    Bad handle                                                     |
|         >100: system code + 100                    -- Bad IO --              |
+-----------------------------------------------------------------------------*/
int Server::sr_xenrd(
    ApplicationHandle ahlAccess,
    char **ppchBuffer,
    int *iLength
)
{
   FYLE *pFyle = (FYLE *)ahlAccess.p;    // pointer part of ApplicationHandle
   int len_read, rc;
   char *pch;

   if ((!pFyle) || ((pFyle != pFyleCur) && (!set_current(pFyle)))) {
      return 8;                          // Bad Handle
   }
   if (pFyle->iBufSize == -1) {          // EOF was in progress
      pFyle->posblk.isAtEof = TRUE;
      return 4;                          // EOF is now reached
   }
   if (pFyle != pFyleTerm) {             // not for Terminal!
      fgetpos(pFyle->fd, &(pFyle->posblk.fpos));
      pFyle->posblk.isFirst = pFyle->isFirst;
      pFyle->posblk.isTruncated = pFyle->isTruncated;
      pFyle->posblk.isAtEof = FALSE;
   }
   if (pFyle->isFirst) {                // for first read
      pFyle->isFirst = FALSE;
      // oeb.appl_info = pFyle->pszName; -- Tell the file name
      // ** See Note 1 **
   }
   if (pFyle == pFyleTerm) {             // Read TERMINAL: no RE/RS inserted
      /*--------------+
      |Terminal entry |
      +--------------*/
      int ch;
      pch = pFyle->buffer - 1;
      len_read = -1;
      while ((++len_read < TERM_BUFSIZE) &&
         ((ch=getc(stdin)) != '\n') && (ch != EOF)) {
         *(++pch) = (char)ch;
      }
      if ((len_read >= 6) &&             // check for $(EOF)
         (*(pch - 5) == '$') && (*(pch - 4) == '(') && (*(pch - 3) == 'E') &&
         (*(pch - 2) == 'O') && (*(pch - 1) == 'F') && (*pch       == ')')) {
         len_read -= 6;
         ch = EOF;
      }
      if (ch == EOF) {                   // EOF in progress:
         pFyle->iBufSize = -1;           // no more further reads!
         if (len_read) {
            rc = 1;                      // RE/RS need to be inserted
         }else {
            rc = 4;                      // signal EOF reached now
         }
      }else {
         rc = 1;                         // RE/RS need to be inserted
      }
   }else {
      /*-----------------------------+
      |fill the buffer up to bufsize |
      +-----------------------------*/
      pch = fgets(pFyle->buffer, pFyle->iBufSize, pFyle->fd);
      if (!pch) {                           // EOF or ERROR
         len_read = 0;
         pFyle->iBufSize = -1;              // avoid further reads
         if (feof(pFyle->fd)) {             // just EOF
            rc = 4;                         // set rc
         }else {                            // ERROR
            rc = 100+errno;                 // add 100 to system code
         }
      }else if (
         len_read = strlen(pch) - 1,
         pch[len_read] == '\n'              // '\n' ended
      ) {
         if (pFyle->isTruncated) {         // truncated record
            rc = 3;                         // only RE to insert
            pFyle->isTruncated = FALSE;
         }else {
            rc = 1;                         // RE/RS to insert
         }
      }else {                               // last char not '\n'
         ++len_read;                        // take it into account
         if (feof(pFyle->fd)) {             // EOF reached
            pFyle->iBufSize = -1;           // remember this
            if (pFyle->isTruncated)  {     // truncated record
               rc = 3;                      // only RE to insert
               pFyle->isTruncated = FALSE; // for next calls...
            }else {
               rc = 1;                      // RE/RS to insert
            }
         }else {                            // not at EOF
            if (pFyle->isTruncated)  {     // truncated record
               rc = 0;                      // nothing to insert
            }else {
               rc = 2;                      // only RS to insert
               pFyle->isTruncated = TRUE;  // cos' it's truncated
            }
         }
      }
   }
   if (rc < 4) {                                   // EOF not reached
      *ppchBuffer = pFyle->buffer;
      *iLength = len_read;
   }
   return rc;
}

/*--------------------------------------------------------------------sr_xengp-+
| Functional Description: Get the position of an opened entity                 |
|                                                                              |
| Parameters:                                                                  |
|                                                                              |
| Return: 0:    OK                                                             |
|         4:    EOF                                                            |
|         8:    Bad handle                                                     |
+-----------------------------------------------------------------------------*/
int Server::sr_xengp(
   ApplicationHandle ahlAccess,
   void **ppSysPosBlock,
   int *iSizeSysPosBlock
)
{
   FYLE *pFyle = (FYLE *)ahlAccess.p;   // pointer part of ApplicationHandle

   if (
      (!pFyle) || ((pFyle != pFyleCur) && (!set_current(pFyle))) ||
      (pFyle == pFyleTerm)              // not for Terminal!
   ) {
      return 8;                         // Bad Handle
   }
   if (pFyle->posblk.isAtEof) {
      return 4;
   }
   *ppSysPosBlock = &(pFyle->posblk);
   *iSizeSysPosBlock = sizeof pFyle->posblk;
   return 0;
}

/*--------------------------------------------------------------------sr_xensp-+
| Functional Description: Reset the position of an opened entity               |
|                                                                              |
| Parameters:                                                                  |
|                                                                              |
| Return: 0:    OK                                                             |
|         8:    Bad handle                                                     |
|         >100: system code + 100                    -- Bad IO --              |
+-----------------------------------------------------------------------------*/
int Server::sr_xensp(ApplicationHandle ahlAccess, void *pSysPosBlock)
{
   FYLE *pFyle = (FYLE *)ahlAccess.p;    // pointer part of ApplicationHandle
   int rc;

   if (
      (!pFyle) || ((pFyle != pFyleCur) && (!set_current(pFyle))) ||
      (pFyle == pFyleTerm)               // not for Terminal!
   ) {
      return 8;                          // Bad Handle
   }
   if (((POSBLK *)pSysPosBlock)->isAtEof) {
      pFyle->iBufSize = -1;              // EOF reached
   }else {
      if ((rc = fsetpos(pFyle->fd, &((POSBLK *)pSysPosBlock)->fpos)) != 0) {
         pFyle->iBufSize = -1;           // avoid further reads
         return rc + 100;
      }
      pFyle->isFirst = ((POSBLK *)pSysPosBlock)->isFirst;
      pFyle->isTruncated = ((POSBLK *)pSysPosBlock)->isTruncated;
      pFyle->iBufSize = FILE_BUFSIZE;
   }
   return 0;
}

/*-------------------------------------------------------------------sr_utopen-+
|Open a utility file, to read or to write                                      |
|                                                                              |
|Return: 0  Normal                                                             |
|        6: Cannot open                                                        |
+-----------------------------------------------------------------------------*/
int Server::sr_utopen(
   const ApplicationHandle ahlIdent,
   ApplicationHandle *pahlAccess,
   const boolean isWriteMode
)
{
   FILE *fd;
   char *pchMode;

   if (isWriteMode) {
      pchMode = "wb";
   }else {
      pchMode = "rb";
   }
   fd = fopen((char *)ahlIdent.p, pchMode);
   if (!fd) {
      return 6;
   }else {
      pahlAccess->p = fd;
      return 0;
   }
}

/*---------------------------------------------------------------------sr_utrd-+
| Functional Description: Read iLength bytes from file ahlAccess.p to          |
|                         buffer pchBuffer.  This function is valid only for   |
|                         utility files.                                       |
|                                                                              |
| Parameters:                                                                  |
|                                                                              |
| Return: 0     Normal                                                         |
|         8     Invalid Access-AHL                                             |
+-----------------------------------------------------------------------------*/
int Server::sr_utrd(
    ApplicationHandle ahlAccess,
    void *pchBuffer,
    int iLength
)
{
   if (
      (iLength) &&
      (fread(pchBuffer, 1, iLength, (FILE *)(ahlAccess.p)) != iLength)
   ) {
      return 8;
   }else {
      return 0;
   }
}

/*--------------------------------------------------------------------sr_utwrt-+
| Functional Description: Write iLength bytes from pchBuffer to file           |
|                         ahlAccess.p.  This function is valid only for        |
|                         utility files.                                       |
|                                                                              |
| Parameters:                                                                  |
|                                                                              |
| Return: 0     Normal                                                         |
|         8     Invalid Access-AHL                                             |
+-----------------------------------------------------------------------------*/
int Server::sr_utwrt(
    ApplicationHandle ahlAccess,
    const void *pchBuffer,
    int iLength
)
{
   if (
      (iLength) &&
      (fwrite(pchBuffer, 1, iLength, (FILE *)(ahlAccess.p)) != iLength)
   ) {
      return 8;
   }else {
      return 0;
   }
}

/*---------------------------------------------------------------------sr_utcl-+
| Functional Description: Close a Utility file.                                |
|                                                                              |
| Parameters:                                                                  |
|                                                                              |
| Return: 0:    OK                                                             |
|         8:    Cannot close                                                   |
+-----------------------------------------------------------------------------*/
int Server::sr_utcl(ApplicationHandle ahlAccess)
{
   if (fclose((FILE *)ahlAccess.p)) {
      return 8;
   }else {
      return 0;
   }
}
/******************************************************************************/
