/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// sgmldcl.c
///  Agglomerate for components of the SGML declaration
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <memory.h>
#include <string.h>

#include "gendef.h"
#include "syntax.h"
#include "externid.h"
#include "sgmldcl.h"
#include "enthdlr.h"
#include "errorh.h"
#include "quantity.h"
#include "capacity.h"
#include "prsstate.h"
#include "prsmode.h"
#include "dlmstr.h"
#include "dlmvect.h"
#include "rsvname.h"
#include "samgr.h"
#include "baseset.h"
#include "shortref.h"

/*------------------------------------------------------------------NewSGMLdcl-+
|                                                                              |
+-----------------------------------------------------------------------------*/
NewSGMLdcl::NewSGMLdcl(
   ParseState *pPrsstatArg,
   EventList& evlstArg
) :
   pPrsstat(pPrsstatArg),
   pEh(pPrsstatArg->pEh),
   pPrsmode(pPrsstatArg->pPrsmode),
   pSyntax(pPrsstatArg->pSyntax),
   evlst(evlstArg)
{
   pSyntaxNew = new NewSyntax(sk);
   pCapacityNew = new NewCapacity();
   isScopeInstanceNew = FALSE;
   memset(&featuresNew, 0, sizeof featuresNew);
   pszAppinfoNew = 0;

   memset(aChdefTable, 0, sizeof aChdefTable);
   memset(parmlit, 0, sizeof parmlit);

   isSevereErrorFound = FALSE;
   isNotFormalPID = FALSE;
   isMdoChanged = FALSE;
   isPioChanged = FALSE;
   isComChanged = FALSE;
   isMdcChanged = FALSE;

   pchLitBuffer = new char [
      2 * (1 + pSyntax->pQuantity->value_of(LITLEN_QTY_ix))
   ];
   {
      for (int i=0; i < MAX_CHARSET; ++i) {
         achSwitchTable[i] = (char)i; // default switch table is identity
      }
   }
}

/*-----------------------------------------------------------------~NewSGMLdcl-+
|                                                                              |
+-----------------------------------------------------------------------------*/
NewSGMLdcl::~NewSGMLdcl()
{
   delete pSyntaxNew;
   delete pCapacityNew;
   delete [] pchLitBuffer;
}

/*---------------------------------------------------------------------compile-+
| Compile the SGML Declaration.                                                |
| Enter here when SGML has been decoded.                                       |
+-----------------------------------------------------------------------------*/
boolean NewSGMLdcl::compile()
{
   static ReferChoiceEntry const scope_choices[] = {
      { DOCUMENT_RRN_ix,  DOCUMENT_TK  },
      { INSTANCE_RRN_ix,  INSTANCE_TK  },
      { NOT_FOUND_RRN_ix, NOT_FOUND_TK }
   };
   pPrsmode->push_mode(MD_MODE);
   /*--------------------------------------+
   | Check we read: <!SGML "ISO 8879:1986" |
   +--------------------------------------*/
   pPrsstat->skip_ps();
   pPrsstat->read_minimum_literal(pEh, pchLitBuffer);
   if (strcmp(pchLitBuffer, ReferenceString::ISO8879_banner())) {
      errorh(INVSGMLID, pEh, ERH_E_SEV);
   }
   pPrsstat->skip_ps();

   /*-------------------------------------+
   | 1) CHARSET  (document character set) |
   |    Get the System Character Set      |
   |    Set up gtype and uctab tables     |
   +-------------------------------------*/
   pSyntax->match_rsvname_or_die(pEh, CHARSET_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   parse_basesets_and_descsets(TRUE); // build document character set tables

   /*---------------------------+
   | 2) CAPACITY (capacity set) |
   +---------------------------*/
   parse_capacity_parameter();

   /*---------------------------------+
   | 3) SCOPE (concrete syntax scope) |
   +---------------------------------*/
   pSyntax->match_rsvname_or_die(pEh, SCOPE_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   switch (pSyntax->match_rsvname_in_list(pEh, scope_choices)) {
   case DOCUMENT_TK:
      break;
   case INSTANCE_TK:
      isScopeInstanceNew = TRUE;
      break;
   default:
      errorh(ISORDOC, pEh, ERH_T_SEV);
      break;
   }
   pPrsstat->skip_inferable_required_ps();

   /*----------------------------+
   | 4) SYNTAX (concrete syntax) |
   +----------------------------*/
   pSyntax->match_rsvname_or_die(pEh, SYNTAX_RRN_ix);
   pPrsstat->skip_inferable_required_ps();

   EntityDef *pEntSyntax = 0;
   ExternId *pExidSyntax = 0;
   int floor_level_save;

   if (!pSyntax->match_rsvname(pEh, SHUNCHAR_RRN_ix)) {
      /*
      | SYNTAX not followed by SHUNCHAR: get the external identifier
      */
      pExidSyntax = new ExternId(
         pPrsstat, ReservedName::name_of(SYNTAX_RRN_ix), SYNTAX_XFC
      );
      if (!pExidSyntax->is_formal()) isNotFormalPID = TRUE;

      /*
      | Change the switch table if any switch are found
      */
      {
         unsigned int iChnb1, iChnb2;
         boolean isChnb1Ok, isChnb2Ok;

         pPrsstat->skip_ps();
         if (pSyntax->match_rsvname(pEh, SWITCHES_RRN_ix)) {
            pPrsstat->skip_inferable_required_ps();
            if (
               (pPrsstat->get_curr_token() != NOT_FOUND_TK) ||
               (!pEh->is_digit())
            ) {
               errorh(NODIG, pEh, ERH_T_SEV);
            }
            do {
               isChnb1Ok = pSyntax->read_character_number(pEh, &iChnb1);
               pPrsstat->skip_inferable_required_ps();
               isChnb2Ok = pSyntax->read_character_number(pEh, &iChnb2);
               pPrsstat->skip_inferable_required_ps();
               if ((isChnb1Ok) && (isChnb2Ok)) {
                  achSwitchTable[iChnb1] = (char)iChnb2;
                  achSwitchTable[iChnb2] = (char)iChnb1;
               }
            }while (
              (pPrsstat->get_curr_token()==NOT_FOUND_TK) && (pEh->is_digit())
            );
         }
      }
      /*
      | Now open the Public Syntax
      */
      pEntSyntax = new EntityDef(pExidSyntax, SYNTAX_RRN_ix);
      if (!pEh->open(pEntSyntax)) {
         errorh(ABORTED, ERH_T_SEV);
      }
      floor_level_save = pPrsstat->floor_level;            // save floor_level
      pPrsstat->floor_level = pEh->curr_entity_level();    // No Ee allowed
      pPrsstat->isQuietAtFloor = TRUE;
      pPrsstat->skip_inferable_required_ps();
      pSyntax->match_rsvname_or_die(pEh, SHUNCHAR_RRN_ix); // SHUNCHAR expected
   }

   parse_shunchar_parameter();            // SHUNCHAR
   parse_basesets_and_descsets(FALSE);    // (BASESET, DESCSET)+
   {                                      // Get rid of what parmlit points to.
      for (int i=0; (parmlit[i]) && (i < MAX_CHARSET); ++i) free(parmlit[i]);
   }

   /*
   | Check if SWITCHES were properly assigned (13.4.1)
   */
   {
      for (int iChOrig=0; iChOrig < MAX_CHARSET; ++iChOrig) {
         unsigned int iChSwch = achSwitchTable[iChOrig];
         if (iChOrig != iChSwch) {           // have been switched
            e_ErrorCode msgcode;
            const CharsetDescriptor& dsetOrig = pSyntaxNew->get_char_descriptor(
               iChOrig
            );
            const CharsetDescriptor& dsetSwch = pSyntaxNew->get_char_descriptor(
               iChSwch
            );
            if (
               (pSyntax->is_alphanum(dsetOrig.s_chnb)) ||
               (pSyntax->is_alphanum(dsetSwch.s_chnb))
            ) {
               msgcode = INVSWITCH;          // switch digit or letter: bad
            }else if (
               (pSyntax->is_nonsgml(dsetOrig.s_chnb)) ||
               (pSyntax->is_nonsgml(dsetSwch.s_chnb))
            ) {
               msgcode = INVSWTCH2;          // switch not in syntax: bad
            }else {
               continue;                     // else it's ok
            }
            errorh(msgcode, ERH_E_SEV, iChOrig, iChSwch);
            achSwitchTable[iChSwch] = (char)iChSwch;
            achSwitchTable[iChOrig] = (char)iChOrig;
         }
      }
   }
   pSyntaxNew->chBSEQ = (aChdefTable + 'B')->doc_d_chnb;

   parse_function_parameter();            // FUNCTION
   parse_naming_parameter();              // NAMING
   if (isSevereErrorFound) errorh(ABORTED, ERH_T_SEV);
   parse_delim_parameter();               // DELIM
   if (isSevereErrorFound) errorh(ABORTED, ERH_T_SEV);
   parse_names_parameter();               // NAMES
   parse_quantity_parameter();            // QUANTITY

   if (pEntSyntax) {
      if (!pEh->is_Ee()) errorh(INVKWD, pEh, ERH_E_SEV);
      pEh->close();
      delete pEntSyntax;
      delete pExidSyntax;
      pPrsstat->isQuietAtFloor = FALSE;
      pPrsstat->floor_level = floor_level_save; // restore current floor_level
   }

   /*--------------------------+
   | 5) FEATURES (feature use) |
   +--------------------------*/
   process_features();
   if ((featuresNew.isFormal) && (isNotFormalPID)) {
      errorh(INVSGFPI, ERH_E_SEV);
   }

   /*----------------------------------------------+
   | 6) APPINFO (application-specific information) |
   |    Store the value into pszAppinfo            |
   +----------------------------------------------*/
   pSyntax->match_rsvname_or_die(pEh, APPINFO_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   if (pPrsstat->get_curr_token() == NOT_FOUND_TK) {
      pSyntax->match_rsvname_or_die(pEh, NONE_RRN_ix);
      pszAppinfoNew = 0;
   }else {
      int iLen = 1 + pPrsstat->read_minimum_literal(pEh, pchLitBuffer);
      pszAppinfoNew = (char *)malloc(iLen);
      memcpy(pszAppinfoNew , pchLitBuffer, iLen);
   }

   /*------------------------------+
   | Get the new dvtab and dvtbent |
   +------------------------------*/
   pSyntaxNew->pDv->build_delimiter_vectors(featuresNew, *(pSyntaxNew->pDlmstr));

   if (isScopeInstanceNew) check_instance_chars();

   /*-------------------+
   | Write the OCS file |
   +-------------------*/
   ApplicationHandle ahlAccess = pPrsstat->open_OCS(UTSYNU_XFC, TRUE);
   unload(ahlAccess);
   pEh->close(ahlAccess);
   pPrsstat->skip_ps();
   if (pPrsstat->get_curr_token() != MDC_TK) errorh(NOMDC, pEh, ERH_E_SEV);
   pPrsmode->pop_mode();
   return FALSE;
}

/*PRIVATE------------------------------------------parse_basesets_and_descsets-+
| BASESET / DESCSET parameters                                                 |
| Build an UCTAB and a GTYPE table.                                            |
|                                                                              |
| IMPLEMENTATION NOTE:                                                         |
|                                                                              |
| The same `dsettab' is used both for the document character set and the       |
| syntax character set.  This doesn't matter since, it is a temporary area for |
| the document character set: it is used to build the `gtype' and `uctab'      |
| tables, which is ended before `parse_basesets_and_descsets' is called for    |
| the syntax.                                                                  |
+-----------------------------------------------------------------------------*/
void NewSGMLdcl::parse_basesets_and_descsets(boolean isDocument)
{
   memset(pSyntaxNew->dsettab, 0, sizeof pSyntaxNew->dsettab);
   pSyntax->match_rsvname_or_die(pEh, BASESET_RRN_ix);

   do {
      /*---------------------------------------------+
      | BASESET: get the corresponding SetDescriptor |
      +---------------------------------------------*/
      const SetDescriptor * aSetBase;
      {
         ExternId exid(
            pPrsstat, ReservedName::name_of(CHARSET_RRN_ix), CHARSET_XFC
         );
         if (
            (!exid.find(pEh)) ||
            (aSetBase = find_baseset(exid.public_id()), !aSetBase)
         ) {            // if Application or myself can't find it,
            errorh(INVBASET, pEh, ERH_T_SEV);
            return;     // avoid compiler warnings - can't return from ERH_T
         }
      }
      /*----------------------------------------------------------------+
      |DESCSET: Changes the meaning based on the BASESET                |
      +----------------------------------------------------------------*/
      pPrsstat->skip_ps();
      pSyntax->match_rsvname_or_die(pEh, DESCSET_RRN_ix);
      pPrsstat->skip_inferable_required_ps();
      if ((pPrsstat->get_curr_token() != NOT_FOUND_TK) || (!pEh->is_digit())) {
         errorh(NODIG, pEh, ERH_T_SEV);
         return;        // avoid compiler warnings - can't return from ERH_T
      }
      do {
         unsigned int iChnbDesc;

         boolean isChnbOk = pSyntax->read_character_number(pEh, &iChnbDesc);
         pPrsstat->skip_ps();
         unsigned long ulNbrOfChar = pSyntax->read_number(pEh);
         pPrsstat->skip_inferable_required_ps();

         unsigned int iChnbBase;
         int iLenParmLit;
         enum {
            UNUSED_MEANING, BASED_MEANING, LITERAL_MEANING
         }described_as;

         if (pPrsstat->get_curr_token() != NOT_FOUND_TK) { // assume LIT/LITA
            iLenParmLit = pPrsstat->read_minimum_literal(pEh, pchLitBuffer);
            described_as = LITERAL_MEANING;
            iChnbBase = 0;
         }else if (pEh->is_digit()) {
            isChnbOk = pSyntax->read_character_number(pEh, &iChnbBase);
            described_as = BASED_MEANING;
         }else if (pSyntax->match_rsvname(pEh, UNUSED_RRN_ix)) {
            described_as = UNUSED_MEANING;
            iChnbBase = 0;
         }else {
            errorh(NUMLITUNU, pEh, ERH_T_SEV);
            iChnbBase = 0;
            return;        // avoid compiler warnings - can't return from ERH_T
         }

         if ((isChnbOk) && (ulNbrOfChar != 0)) {
            unsigned short *pusGtype  = pSyntaxNew->gtype + iChnbDesc;
            CharsetDescriptor *pDset = pSyntaxNew->dsettab + iChnbDesc;

            do {
               if ((iChnbDesc >= MAX_CHARSET) || (iChnbBase >= MAX_CHARSET)) {
                  errorh(BIGCHDESC, pEh, ERH_I_SEV, iChnbDesc);
                  break;
               }
               if (!pDset->isDescribed) {
                  pDset->isDescribed = TRUE;
               }else {
                  errorh(DUPCHDESC, pEh, ERH_E_SEV, iChnbDesc);
                  continue;
               }
               CharDefinition *pChdef;
               switch (described_as) {
               case BASED_MEANING:
                  {
                     const SetDescriptor *pSet = aSetBase + iChnbBase++;
/*09.22.91*/         if (pSet->usType == S_NONSGML) {
                        pDset->isUnused = TRUE;
                        if (isDocument) *pusGtype = S_NONSGML;
                        continue;
                     }
                     pDset->isBased = TRUE;
                     pDset->s_chnb = pSet->chSysNb;
                     pChdef = aChdefTable + pDset->s_chnb;
                     if (isDocument) {
                        pChdef->upper = pSet->chUpper;
                        *pusGtype = pSet->usType;
                     }
                  }
                  break;
               case LITERAL_MEANING:
                  pDset->isParmLit = TRUE;
                  {
                     char **ppsz = parmlit;
                     int i = 0;

                     while ((*ppsz) && (strcmp(*ppsz, pchLitBuffer))) {
                        if (++i >= MAX_CHARSET) {
                           errorh(BIGDSCLIT, pEh, ERH_T_SEV);
                           return;
                        }
                        ++ppsz;
                     }
                     if (!*ppsz) {
                        *ppsz = (char *)malloc(iLenParmLit+1);
                        memcpy(*ppsz, pchLitBuffer, iLenParmLit+1);
                     }
                     pDset->s_chnb = (char)i;
                     pChdef = aChdefTable + i + MAX_CHARSET;
                  }
                  if (isDocument) *pusGtype = S_SGML;
                  break;
               default: // UNUSED_MEANING:
                  pDset->isUnused = TRUE;
                  if (isDocument) *pusGtype = S_NONSGML;
                  continue;
               }
               if (isDocument) {
                  if (!pChdef->isDescribedInDoc) {
                     pChdef->isDescribedInDoc = TRUE;
                     pChdef->doc_d_chnb = (char)iChnbDesc;
                  }else {
/*09.22.91*/         errorh(
                        DUPCHASGN, pEh, ERH_I_SEV, iChnbDesc,
                        pChdef->doc_d_chnb
                     );
                     pChdef->isMultipleInDoc = TRUE;
                  }
               }else {
                  if (!pChdef->isDescribedInSyn) {
                     pChdef->isDescribedInSyn = TRUE;
                     pChdef->syn_d_chnb = (char)iChnbDesc;
                  }else {
/*09.22.91*/         errorh(
                        DUPCHASGN, pEh, ERH_I_SEV, iChnbDesc,
                        pChdef->syn_d_chnb
                     );
                     pChdef->isMultipleInSyn = TRUE;
                  }
               }
            }while (++pusGtype, ++pDset, ++iChnbDesc, --ulNbrOfChar);
         }
         pPrsstat->skip_inferable_required_ps();
      }while (
         (pPrsstat->get_curr_token() == NOT_FOUND_TK) && (pEh->is_digit())
      );
   }while (pSyntax->match_rsvname(pEh, BASESET_RRN_ix));

   /*------------------------------------------+
   | Check all LETTERS, DIGIT and Minimum Data |
   | have been described properly              |
   +------------------------------------------*/
   {
      static char const pszMinset[] =
         "abcdefghijklmnopqrstuvwxyz"   // must start with lower case letters
         "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
         "0123456789'()+,-./:=?";

      boolean err_found = FALSE;
      for (char const *pchMin = pszMinset; *pchMin; ++pchMin) {
         CharDefinition *pChdef = aChdefTable + *pchMin;
         if (isDocument) {
            if (!pChdef->isDescribedInDoc) {
               err_found = TRUE;
               errorh(NODRDOC, ERH_E_SEV, *pchMin);
            }else if (pChdef->isMultipleInDoc) {
               err_found = TRUE;
               errorh(DUPDRDOC, ERH_E_SEV, *pchMin);
            }
         }else {
            if (!pChdef->isDescribedInSyn) {
               err_found = TRUE;
               errorh(NODRSYN, ERH_E_SEV, *pchMin);
            }else if (pChdef->isMultipleInSyn) {
               err_found = TRUE;
               errorh(DUPDRSYN, ERH_E_SEV, *pchMin);
            }
         }
      }
      if (err_found) errorh(ABORTED, ERH_T_SEV);
   }
}

/*PRIVATE---------------------------------------------parse_shunchar_parameter-+
| Enter here when SHUNNED has just been decoded.                               |
| Update gtype for shunned characters, finish gtype and uctab.                 |
+-----------------------------------------------------------------------------*/
void NewSGMLdcl::parse_shunchar_parameter()
{
   static ReferChoiceEntry const shun_choices[] = {
      { CONTROLS_RRN_ix,  CONTROLS_TK  },
      { NONE_RRN_ix,      NONE_TK      },
      { NOT_FOUND_RRN_ix, NOT_FOUND_TK }
   };
   unsigned int iChnbDesc;
   boolean isShunnedControls = FALSE;

   pPrsstat->skip_inferable_required_ps();
   switch (pSyntax->match_rsvname_in_list(pEh, shun_choices)) {
   case NONE_TK:
      pPrsstat->skip_inferable_required_ps();
      break;
   case CONTROLS_TK:
      isShunnedControls = TRUE;
      pPrsstat->skip_inferable_required_ps();
      /* fall thru */
   default:  /* NOT_FOUND_TK */
      if (!pEh->is_digit()) {
         if (isShunnedControls) {
            break;
         }else {
            errorh(NODIG, pEh, ERH_T_SEV);
         }
      }
      do {
         if (
            (pSyntax->read_character_number(pEh, &iChnbDesc)) &&
            (pSyntaxNew->gtype[iChnbDesc] & S_SHUNNABLE)
         ) {
            pSyntaxNew->gtype[iChnbDesc] = S_SHUNNED;
         }
         pPrsstat->skip_inferable_required_ps();
      }while (pEh->is_digit());
      break;
   }
   unsigned short *pusGtype = pSyntaxNew->gtype;
   char *pUctab = pSyntaxNew->uctab;
   CharsetDescriptor *pDset = pSyntaxNew->dsettab;
   int iCount;
   for (iCount=0; iCount < MAX_CHARSET; ++iCount,++pusGtype,++pDset,++pUctab) {
      if (!pDset->isDescribed) {
         *pusGtype = S_NONSGML;
      }else if (*pusGtype & S_SHUNNABLE) {
         if ((isShunnedControls) && (*pusGtype & S_CONTROLS)) {
            *pusGtype = S_SHUNNED;
         }else {
            *pusGtype = S_SGML;
         }
      }
      *pUctab = pDset->isBased ?
         aChdefTable[aChdefTable[pDset->s_chnb].upper].doc_d_chnb :
         (char)iCount
      ;
   }
}

/*PRIVATE---------------------------------------------parse_function_parameter-+
| Parse the function parameter of the SGML declaration                         |
| Create the pSyntaxNew->fncrtab                                               |
+-----------------------------------------------------------------------------*/
void NewSGMLdcl::parse_function_parameter()
{
   static ReferChoiceEntry const fcl_choices[] = {
      { FUNCHAR_RRN_ix,   FUNCHAR_FCTCLS },
      { MSICHAR_RRN_ix,   MSICHAR_FCTCLS },
      { MSOCHAR_RRN_ix,   MSOCHAR_FCTCLS },
      { MSSCHAR_RRN_ix,   MSSCHAR_FCTCLS },
      { SEPCHAR_RRN_ix,   SEPCHAR_FCTCLS },
      { NOT_FOUND_RRN_ix, NULL_FCTCLS    }
   };

   pSyntax->match_rsvname_or_die(pEh, FUNCTION_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   SortArray& saFncr = * new SortArray(
      sizeof(FunctionCharAssoc),       // size of each array item
      0,                               // comparison function
      &pSyntaxNew->fncr_max,           // where to return count of items entered
      (void **)&(pSyntaxNew->fncrtab), // where to return address of the array
      0
   );
   {
      for (int i=0; i<3; ++i) {
         int iChRef;
         int iChNew;
         FunctionCharAssoc fncr;
         e_IxRsvnameRefer ixFunctioncharName;
         char *pchNewSyntax;

         switch(i) {
         case 0:
            iChRef = pSyntax->get_REchar();
            pchNewSyntax = &pSyntaxNew->tv.chRE;
            fncr.fctcls = RE_FCTCLS;
            ixFunctioncharName = RE_RRN_ix;
            break;
         case 1:
            iChRef = pSyntax->get_RSchar();
            pchNewSyntax = &pSyntaxNew->tv.chRS;
            fncr.fctcls = RS_FCTCLS;
            ixFunctioncharName = RS_RRN_ix;
            break;
         default:
            iChRef = pSyntax->get_SPACEchar();
            pchNewSyntax = &pSyntaxNew->tv.chSPACE;
            fncr.fctcls = SPACE_FCTCLS;
            ixFunctioncharName = SPACE_RRN_ix;
            break;
         }
         pSyntax->match_rsvname_or_die(pEh, ixFunctioncharName);

         /*
         | RE, RS and SPACE can be changed by the NAMES parameter;
         | however, it's easier to "park" their value now, even if
         | they become useless later (at most, 12 bytes wasted.)
         */
         {
            const char * pszName = ReservedName::name_of(ixFunctioncharName);
            fncr.ofs = sk.park(pszName, 1+strlen(pszName));
         }
         pPrsstat->skip_inferable_required_ps();
         if (!pSyntax->read_character_number(pEh, (unsigned int *)&iChNew)) {
            iChNew = iChRef;
         }
         iChNew = map_to_doc_chnb((char)iChNew);
         if (iChNew != -1) {
            fncr.ch_nb = iChNew;
            *pchNewSyntax = (char)iChNew;
            pSyntaxNew->gtype[iChNew] |= S_SSPCHAR;
         }
         pPrsstat->skip_inferable_required_ps();
         saFncr.enter(&fncr);
      }

   }
   {
      unsigned int iCharClassAll, iCharClassAdd;
      iCharClassAll = 0;                            // to check MSO/MSI
      while (!(pSyntax->match_rsvname(pEh, NAMING_RRN_ix))) {
         pSyntax->read_general_name(pEh, pchLitBuffer);
         pPrsstat->skip_ps();                       // added function

         FunctionCharAssoc fncr;
         fncr.fctcls = (e_FunctionClass)pSyntax->match_rsvname_in_list(
            pEh, fcl_choices
         );
         switch (fncr.fctcls) {
         case FUNCHAR_FCTCLS: iCharClassAdd = S_FUNCHAR; break;
         case MSICHAR_FCTCLS: iCharClassAdd = S_MSICHAR; break;
         case MSOCHAR_FCTCLS: iCharClassAdd = S_MSOCHAR; break;
         case MSSCHAR_FCTCLS: iCharClassAdd = S_MSSCHAR; break;
         case SEPCHAR_FCTCLS: iCharClassAdd = S_SEPCHAR; break;
         default: errorh(INVFCTCL, pEh, ERH_T_SEV); break;
         }
         pPrsstat->skip_inferable_required_ps();

         int iChNew;
         if (
            (pSyntax->read_character_number(pEh, (unsigned int *)&iChNew)) &&
            (iChNew = map_to_doc_chnb((char)iChNew), iChNew != -1)
         ) {
            fncr.ch_nb = iChNew;
            pSyntaxNew->gtype[iChNew] |= iCharClassAdd;
            if (
               (pSyntaxNew->is_minldat(iChNew)) &&
               (iCharClassAdd & (S_MSICHAR | S_MSOCHAR | S_MSSCHAR))
            ) {
               errorh(INVSCASGN, pEh, ERH_E_SEV);
            }else {
               fncr.ofs = sk.park(pchLitBuffer, 1+strlen(pchLitBuffer));
               saFncr.enter(&fncr);
               iCharClassAll |= iCharClassAdd;
            }
         }
         pPrsstat->skip_inferable_required_ps();
      }
      if ((iCharClassAll & S_MSOCHAR) && (!(iCharClassAll & S_MSICHAR))) {
         errorh(NOMSICHAR, ERH_E_SEV);
      }
   }
   delete &saFncr;
   if (!pSyntaxNew->check_functionchar()) errorh(ABORTED, ERH_T_SEV);
}

/*PRIVATE-----------------------------------------------parse_naming_parameter-+
| Update the pSyntaxNew->gtype table                                           |
+-----------------------------------------------------------------------------*/
void NewSGMLdcl::parse_naming_parameter()
{
   pPrsstat->skip_inferable_required_ps();

   for (int i=0; i<4; ++i) {
      char *pch;
      e_IxRsvnameRefer ixNamingName;
      unsigned int iCharClassAdd;
      boolean isUpper;

      switch (i) {
      case 0:
         pch = pchLitBuffer;
         ixNamingName = LCNMSTRT_RRN_ix;
         iCharClassAdd = S_NMSTRT;
         isUpper = FALSE;
         break;
      case 1:
         ++pch;                      // pchLitBuffer is 2 * LITLEN
         ixNamingName = UCNMSTRT_RRN_ix;
         iCharClassAdd = S_NMSTRT;
         isUpper = TRUE;
         break;
      case 2:
         pch = pchLitBuffer;
         ixNamingName = LCNMCHAR_RRN_ix;
         iCharClassAdd = S_NMCHAR;
         isUpper = FALSE;
         break;
      default:
         ++pch;
         ixNamingName = UCNMCHAR_RRN_ix;
         iCharClassAdd = S_NMCHAR;
         isUpper = TRUE;
         break;
      }
      pSyntax->match_rsvname_or_die(pEh, ixNamingName);
      pPrsstat->skip_ps();
      read_parameter_literal(pch, FALSE);

      for (char *pchLower = pchLitBuffer; *pch; ++pch) {
         pSyntaxNew->gtype[*pch] |= iCharClassAdd;
         if (
            (pSyntaxNew->is_alphanum(*pch)) ||
            (pSyntaxNew->is_separator(*pch)) ||
            ((iCharClassAdd & S_NMCHAR) && (iCharClassAdd & S_NMSTRT))
         ) {
            errorh(INVNASGN, pEh, ERH_T_SEV, *pch);
         }
         if (isUpper) {
            if (!*pchLower) errorh(DIFUCLC, pEh, ERH_T_SEV);
            pSyntaxNew->uctab[*pchLower++] = *pch;
         }
      }
      pPrsstat->skip_ps();
      if ((isUpper) && (*pchLower)) errorh(DIFUCLC, pEh, ERH_T_SEV);
   }
   pSyntax->match_rsvname_or_die(pEh, NAMECASE_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   pSyntax->match_rsvname_or_die(pEh, GENERAL_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   if (answer_is_yes()) pSyntaxNew->tv.isCaseGeneral = TRUE;
   pPrsstat->skip_inferable_required_ps();
   pSyntax->match_rsvname_or_die(pEh,  ENTITY_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   if (answer_is_yes()) pSyntaxNew->tv.isCaseEntity = TRUE;
   pPrsstat->skip_inferable_required_ps();
}

/*PRIVATE------------------------------------------------parse_delim_parameter-+
| Decode the DELIM parameter of the SYNTAX of the SGML declaration.            |
|                                                                              |
| [190] delimiter set =                                                        |
|            "DELIM", ps+, general delimiters, short reference delimiters      |
|                                                                              |
| [191] general delimiters =                                                   |
|            "GENERAL", ps+, "SGMLREF", (ps+, name, ps+, parameter literal)*   |
|                                                                              |
| [192] short reference delimiters =                                           |
|            "SHORTREF", ps+, ("SGMLREF"|"NONE), (ps+, parameter literal)*     |
|                                                                              |
| At entry to this main routine, the next character on input                   |
| must be "DELIM".                                                             |
|                                                                              |
| It returns when all short references have been processed (next character on  |
| input should be "NAMES")                                                     |
|                                                                              |
| On INPUT:                                                                    |
| `gtype' is built (so pSyntaxNew->is_xxx calls are OK)                        |
|                                                                              |
| On OUTPUT:                                                                   |
| `aDlmstrVarnt' (offset of some delimiters) is created,                       |
|                but need to be updated later by dlm_dvtb (double-delim)       |
|                                                                              |
| `dltab' (256 byte "could_be_delimiter" table) is created,                    |
|                but need to be updated later by dlm_dvtb.                     |
|                                                                              |
| `srtab' is created.  The SHORTREF delimiter table is used to recognize       |
|         the parameter literal of the short_ref_mapping declaration.          |
|                                                                              |
+-----------------------------------------------------------------------------*/
void NewSGMLdcl::parse_delim_parameter()
{
   pSyntax->match_rsvname_or_die(pEh, DELIM_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   pSyntax->match_rsvname_or_die(pEh, GENERAL_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   pSyntax->match_rsvname_or_die(pEh, SGMLREF_RRN_ix);

   boolean isInstalledDelim[DLI_MAX_SIZE];
   memset(isInstalledDelim, 0, sizeof isInstalledDelim);

   /*-----------------------------------+
   | Decode and Install User Delimiters |
   +-----------------------------------*/
   for (;;) {
      e_IxDlmitmRefer ix;

      pPrsstat->skip_inferable_required_ps();
      pSyntax->read_general_name(pEh, pchLitBuffer);
      ReferenceString::get_index(pchLitBuffer, &ix);
      if (ix == NOT_FOUND_DLI_ix) {
         errorh(INVROL, pEh, ERH_T_SEV);
      }
      if (ReferenceString::token_of(ix) == SHORTREF_TK) { // just a generic name
         isInstalledDelim[ix] = TRUE;
         break;                                           // and breaks the loop
      }
      pPrsstat->skip_ps();
      if (!read_parameter_literal(pchLitBuffer, TRUE)) {  // if solely funchar
         if (*pchLitBuffer) {
            errorh(INVDLSTR1, pEh, ERH_E_SEV);
         }else {
            errorh(INVDLSTR0, pEh, ERH_E_SEV);
         }
      }else if (isInstalledDelim[ix]) {
         errorh(DUPNAMDLM, pEh, ERH_E_SEV);
      }else if (strcmp(ReferenceString::value_of(ix), pchLitBuffer)) { // differ
         pSyntaxNew->pDlmstr->set_delimiter_string_table(ix, pchLitBuffer);
         pSyntaxNew->pDv->define_delimiter(
            pchLitBuffer,
            ReferenceString::token_of(ix),
            ReferenceString::recognition_modes_of(ix),
            TRUE
         );
         /*
         | Later, if SCOPE INSTANCE, check the delimiter change validity
         */
         switch(ReferenceString::token_of(ix)) {
         case MDO_TK:  isMdoChanged = TRUE; break;
         case COM_TK:  isPioChanged = TRUE; break;
         case PIO_TK:  isComChanged = TRUE; break;
         case MDC_TK:  isMdcChanged = TRUE; break;
         default: break;
         }
      }
   }
   /*------------------------------------------+
   | Default uninstalled delimiters to SGMLREF |
   +------------------------------------------*/
   for (int i=0; i < DLI_MAX_SIZE; ++i) {
      if (!isInstalledDelim[i]) {        /* if not already installed */
         e_IxDlmitmRefer ix = (e_IxDlmitmRefer)i;
         char *pchTarget = pchLitBuffer;
         const char *pchSource = ReferenceString::value_of(ix);
         int chnb;
         while (*pchSource) {
            chnb = check_and_map_to_doc_chnb(*pchSource++);
            if (chnb != -1) *(pchTarget++) = (char)chnb;
         }
         *pchTarget = '\0';
         pSyntaxNew->pDlmstr->set_delimiter_string_table(ix, pchLitBuffer);
         pSyntaxNew->pDv->define_delimiter(
            pchLitBuffer,
            ReferenceString::token_of(ix),
            ReferenceString::recognition_modes_of(ix),
            TRUE
         );
      }
   }
   /*-------------------------------------------+
   | If SHORTREF SGMLREF, install the shortrefs |
   | from the reference concrete syntax         |
   +-------------------------------------------*/
   pPrsstat->skip_inferable_required_ps();
   int iToken = SR_first;

   if (pSyntax->match_rsvname(pEh, SGMLREF_RRN_ix)) {
      build_reference_shortref_literal(pchLitBuffer); // string of ref shortrefs
      char *pch = pchLitBuffer;
      while (*(pch += 1 + strlen(pch))) {
         pSyntaxNew->pShortref->define_shortref((e_LexToken)iToken, pch);
         pSyntaxNew->pDv->define_delimiter(
            pch, (e_LexToken)iToken, 1 << CON_MODE, FALSE
         );
         ++iToken;
      }
   }else if (!pSyntax->match_rsvname(pEh, NONE_RRN_ix)) {
      errorh(SGMLNONE, pEh, ERH_T_SEV);
   }
   /*----------------------------------+
   | Decode and Install User Shortrefs |
   +----------------------------------*/
   while (pPrsstat->skip_ps(), pPrsstat->get_curr_token() != NOT_FOUND_TK) {
      if (!(read_parameter_literal(pchLitBuffer, FALSE))) {
         errorh(INVDLSTR0, pEh, ERH_E_SEV);
         continue;
      }
      /*
      | Rule at 13.4.6.2:
      | no more than one BSEQ,
      | BSEQ cannot be followed/preceeded by SEPCHAR
      */
      {
         char *pch = pchLitBuffer;
         boolean isBseqFound = FALSE;
         e_ErrorCode msgcode = ER_NULL;
         while (*pch) {
            if (*pch != pSyntaxNew->chBSEQ) {
               ++pch;
            }else {
               if (isBseqFound) {
                  msgcode = INVSRBS1;
                  break;
               }else if (
                  (pch > pchLitBuffer) && (pSyntaxNew->is_blank(pch[-1]))
               ) {
                  msgcode = INVSRBS2;
                  break;
               }else {
                  isBseqFound = TRUE;
                  while (*++pch == pSyntaxNew->chBSEQ);
                  if (pSyntaxNew->is_blank(*pch)) {
                     msgcode = INVSRBS2;
                     break;
                  }
               }
            }
         }
         if (msgcode != ER_NULL) {
            errorh(msgcode, pEh, ERH_E_SEV);
            continue;
         }
      }
      pSyntaxNew->pShortref->define_shortref((e_LexToken)iToken, pchLitBuffer);
      pSyntaxNew->pDv->define_delimiter(
         pchLitBuffer, (e_LexToken)iToken, 1 << CON_MODE, FALSE
      );
      ++iToken;
   }
   pSyntaxNew->pShortref->build();
}

/*PRIVATE------------------------------------------------parse_names_parameter-+
|                                                                              |
+-----------------------------------------------------------------------------*/
void NewSGMLdcl::parse_names_parameter()
{
   /*--------------------+
   |Check where we are.. |
   +--------------------*/
   pSyntax->match_rsvname_or_die(pEh, NAMES_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   pSyntax->match_rsvname_or_die(pEh, SGMLREF_RRN_ix);
   pPrsstat->skip_inferable_required_ps();

   /*-----------------------+
   |Enter the variant names |
   +-----------------------*/
   char pszName[NAMELEN_RCS+1];
   while (                                         // until QUANTITY found,
      pSyntax->read_general_name(pEh, pszName),    // get replaced name
      strcmp(pszName, ReservedName::name_of(QUANTITY_RRN_ix))
   ) {
      e_ErrorCode msgcode;
      e_IxRsvnameVarnt ixVarnt;

      msgcode = pSyntaxNew->pRsvname->clearance_to_replace(pszName, &ixVarnt);
      switch (msgcode) {
      case ER_NULL:
         break;
      case DUPVANAM:                               // already replaced
         errorh(
            DUPVANAM, pEh, ERH_E_SEV, pszName,
            pSyntaxNew->pRsvname->name_of(ixVarnt)
         );
         break;
      default:
         errorh(msgcode, pEh, ERH_E_SEV, pszName);
         break;
      }
      pPrsstat->skip_ps();
      pSyntax->read_general_name(pEh, pszName);    // get replacement name
      if (msgcode == ER_NULL) {  // replaced name found and replaceable
         msgcode = pSyntaxNew->pRsvname->replace(pszName, &ixVarnt);
         switch (msgcode) {
         case ER_NULL:           // OK
            break;
         case DUPNAMREP:
            errorh(
               DUPNAMREP, pEh, ERH_E_SEV,
               pszName, pSyntax->pRsvname->name_of(ixVarnt)
            );
            break;
         default: // INVRPTNAM:
            errorh(msgcode, pEh, ERH_E_SEV);
            break;
         }
      }
      pPrsstat->skip_ps();
   }
   /*---------------------------------------+
   | Complete unchanged with default values |
   +---------------------------------------*/
   pSyntaxNew->pRsvname->default_non_replaced(pSyntax);

   /*----------------------------------------------+
   | RE, RS, SPACE can have changed: set and check |
   +----------------------------------------------*/
   pSyntaxNew->reset_functionchar_rsvname();
   if (!pSyntaxNew->check_functionchar()) errorh(ABORTED, ERH_T_SEV);
}

/*PRIVATE---------------------------------------------parse_capacity_parameter-+
| [180] capacity set = "CAPACITY", ps+,                                        |
|                  (("PUBLIC", ps+, public identifier) |                       |
|                   ("SGMLREF", (ps+, name, ps+, number)+ ))                   |
|                                                                              |
+-----------------------------------------------------------------------------*/
void NewSGMLdcl::parse_capacity_parameter()
{
   EntityDef *pEntCapacity = 0;
   ExternId *pExidCapacity = 0;
   int floor_level_save;
   boolean isCapacityDefined[CAP_MAX_SIZE] = { FALSE };
   boolean isEmptySet;

   /*---------------------+
   | Check where we are.. |
   +---------------------*/
   pSyntax->match_rsvname_or_die(pEh, CAPACITY_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   if (pSyntax->match_rsvname(pEh, SGMLREF_RRN_ix)) {
      isEmptySet = TRUE;                         // SGMLREF: check empty set
      pPrsstat->skip_inferable_required_ps();    // Required ps
   }else {
      isEmptySet = FALSE;                        // discard empty set
      /*-------------------------+
      | Public Capacity Found    |
      | Build an exid to open it |
      +-------------------------*/
      pExidCapacity = new ExternId(
         pPrsstat, ReservedName::name_of(CAPACITY_RRN_ix), CAPACITY_XFC
      );
      if (!pExidCapacity->is_formal()) isNotFormalPID = TRUE;
      if (pPrsstat->get_curr_token() != NOT_FOUND_TK) {   // SCOPE must follow
         errorh(INVDLM, pEh, ERH_T_SEV);
      }
      pEntCapacity = new EntityDef(pExidCapacity, CAPACITY_RRN_ix);
      if (!pEh->open(pEntCapacity)) {
         delete pEntCapacity;
         delete pExidCapacity;
         pEntCapacity = 0;
      }else {
         floor_level_save = pPrsstat->floor_level;         // save floor_level
         pPrsstat->floor_level = pEh->curr_entity_level(); // No Ee allowed
         pPrsstat->isQuietAtFloor = TRUE;
      }
      pPrsstat->skip_ps();
   }
   /*----------------------------------+
   | Enter the variant capacity values |
   +----------------------------------*/
   for (;;) {
      char pszName[NAMELEN_MAX+1];
      e_IxCapacity ix;

      if (pPrsstat->get_curr_token() != NOT_FOUND_TK) {
         errorh(INVDLM, pEh, ERH_T_SEV);
      }
      if (pEh->is_Ee()) break;
      pSyntax->read_general_name(pEh, pszName);  // get capacity name
      ReferenceString::get_index(pszName, &ix);
      if (ix == NOT_FOUND_CAP_ix) {              // not found
         pEh->unread(strlen(pszName));          // unread... SCOPE?
         break;
      }
      isEmptySet = FALSE;                        // At least one value found
      if (isCapacityDefined[ix]) {
         errorh(DUPDEF, pEh, ERH_E_SEV, pszName);
      }
      pPrsstat->skip_inferable_required_ps();
      unsigned long value = pSyntax->read_number(pEh);

      if (
         (!isCapacityDefined[ix]) &&
         (pCapacityNew->set_value(ix, value) < value)
      ) {
         errorh(INVCAP, pEh, ERH_E_SEV, pCapacityNew->value_of(ix));
      }
      pPrsstat->skip_inferable_required_ps();    // even if PUBLIC: Ee is a ps
   }
   pCapacityNew->check_consistency();

   /*------------------------------------+
   | Close the external capacity, if any |
   +------------------------------------*/
   if (pEntCapacity) {
      if (!pEh->is_Ee()) errorh(INVKWD, pEh, ERH_E_SEV);
      pEh->close();
      delete pEntCapacity;
      delete pExidCapacity;
      pPrsstat->isQuietAtFloor = FALSE;
      pPrsstat->floor_level = floor_level_save; // restore current floor_level
   }
   if (isEmptySet) {                      // SGMLREF can't have empty set
      errorh(NOCAPSET, pEh, ERH_E_SEV);
   }
}

/*PRIVATE---------------------------------------------parse_quantity_parameter-+
| [194] quantity set = "QUANTITY", ps+, "SGMLREF", (ps+, name, ps+, number)*   |
+-----------------------------------------------------------------------------*/
void NewSGMLdcl::parse_quantity_parameter()
{
   boolean isQuantityDefined[QTY_MAX_SIZE] = { FALSE };

   /*---------------------+
   | Check where we are.. |
   +---------------------*/
   pPrsstat->skip_inferable_required_ps();
   pSyntax->match_rsvname_or_die(pEh, SGMLREF_RRN_ix);

   /*----------------------------------+
   | Enter the variant quantity values |
   +----------------------------------*/
   for (;;) {
      char pszName[NAMELEN_MAX+1];
      e_IxQuantity ix;

      pPrsstat->skip_inferable_required_ps();
      if ((pEh->is_Ee() || (pPrsstat->get_curr_token() != NOT_FOUND_TK))) break;

      pSyntax->read_general_name(pEh, pszName);      // get quantity name
      ReferenceString::get_index(pszName, &ix);
      if (ix == NOT_FOUND_QTY_ix) {                  // not found
         pEh->unread(strlen(pszName));              // unread... FEATURES?
         break;
      }
      if (isQuantityDefined[ix]) {
         errorh(DUPDEF, pEh, ERH_E_SEV, pszName);
      }
      pPrsstat->skip_inferable_required_ps();
      unsigned long value = pSyntax->read_number(pEh);
      if (
         (!isQuantityDefined[ix]) &&
         (!pSyntaxNew->pQuantity->set_value(ix, value))
      ) {
         errorh(INVQNTY, pEh, ERH_E_SEV);
      }
   }
}

/*PRIVATE-------------------------------------build_reference_shortref_literal-+
| Get in pchBuf the Reference (default) Shortref Strings, separated with '\0'  |
| and ended with a doubled '\0'.                                               |
|                                                                              |
| We cannot have a static translation of the Character Reference,              |
| as these could have been changed by the FUNCTION parameter.                  |
| To translate the ShortrefLiteral, this string is "unread" in a pseudo entity |
| and processed by get_lit (that does the character reference translation).    |
| Due to the presence of LIT mixed with LITA, two literals are needed.         |
|                                                                              |
| Note that there is no problem to use the name SPACE or RE: these cannot      |
| have been changed so far !  (NAMES is after DELIM ... SHORTREF).             |
+-----------------------------------------------------------------------------*/
void NewSGMLdcl::build_reference_shortref_literal(char *pchBuf)
{
   static char const pchShortrefLiteral[] =
      "' &#TAB; &#RE; &#RS; &#RS;B &#RS;&#RE; "
      //       2,    2,    2,     3,         3,
      "&#RS;B&#RE; B&#RE; &#SPACE; BB \" # '"
      //          4,     3,       2, 3, 2,2,
      "\"% ' ( ) * + , - -- : ; = @ [ ] ^ _ { | } ~  \""
      //  2,2,2,2,2,2,2,2, 3,2,2,2,2,2,2,2,2,2,2,2,2,1
   ;
   static char const chIncrtShortrefLit[] = {
      29,                                 // string size expected first
      2,2,2,3,3,4,3,2,3,2,2,2,2,2,2,2,2,2,2,3,2,2,2,2,2,2,2,2,2,2,2,2,1
   };

   char const *pchIncrtShortrefLit = chIncrtShortrefLit; // array of increments

   EntityDef entPseudoLit(                     // to decode pchShortrefLiteral
      (char *)pchShortrefLiteral, sizeof(pchShortrefLiteral) - 1
   );
   pEh->open(&entPseudoLit);                   // re-read it
   /*
   | Read the literal with the LIT in it
   | If &#TAB; was not replaced by `translate_character_ref',
   | the length of the string is off by 1
   */
   pPrsstat->recognize_delimiter();            // get the LITA delimiter
   int iLen = read_parameter_literal(pchBuf, FALSE);
   if (iLen != pchIncrtShortrefLit[0]) {
      memmove(pchBuf, pchBuf+1, --iLen);
      ++pchIncrtShortrefLit;
   }
   /*
   | Read the literal with the LITA in it
   */
   pPrsstat->recognize_delimiter();            // get the LIT delimiter
   read_parameter_literal(pchBuf+iLen, FALSE); // catenate
   pEh->close();
   {
      int i = 0;
      int j;
      while (j = *++pchIncrtShortrefLit, j) {
         pchBuf[i] = '\0';                     // Make each SR an Z-string
         i += j;
      }
      pchBuf[i] = '\0';                        // double zero at end
   }
}

/*PRIVATE--------------------------------------------------------answer_is_yes-+
| Get YES or NO, or abort                                                      |
+-----------------------------------------------------------------------------*/
boolean NewSGMLdcl::answer_is_yes()
{
   static ReferChoiceEntry const yesno_choices[] = {
      { YES_RRN_ix,       YES_TK       },
      { NO_RRN_ix,        NO_TK        },
      { NOT_FOUND_RRN_ix, NOT_FOUND_TK }
   };
   switch (pSyntax->match_rsvname_in_list(pEh, yesno_choices)) {
   case YES_TK:  return TRUE;
   case NO_TK:   return FALSE;
   default:
      errorh(YESNO, pEh, ERH_T_SEV);
      return FALSE;
   }
}

/*PRIVATE--------------------------------------------check_and_map_to_doc_chnb-+
| Map a character entered directly to a document character number.             |
| Normally, this is not allowed when the character number is different in      |
| both the syntax-reference character set and the document character set.      |
|                                                                              |
| However, if the character can be mapped without ambiguity (not multiple)     |
| to a syntax-reference character number, it can be considered as if it was    |
| entered using a character reference.  Is this a YASP extension?              |
|                                                                              |
| See the comments in "map_to_doc_chnb" for other required conditions          |
| in doing such a mapping.                                                     |
+-----------------------------------------------------------------------------*/
int NewSGMLdcl::check_and_map_to_doc_chnb(char system_chnb)
{
   CharDefinition *pChdef = aChdefTable + system_chnb;

   if (!pChdef->isDescribedInSyn) {
      errorh(NODRSYN, pEh, ERH_E_SEV, system_chnb);
      isSevereErrorFound = TRUE;
      return -1;
   }
   if (pChdef->isMultipleInSyn) {
      errorh(DUPDRSYN, pEh, ERH_E_SEV, system_chnb);
      isSevereErrorFound = TRUE;
      return -1;
   }
   return map_to_doc_chnb(pChdef->syn_d_chnb);
}

/*PRIVATE------------------------------------------------------map_to_doc_chnb-+
| Map a syntax character number to a document character number.                |
| Returns the document character number, or -1 if in error.                    |
|                                                                              |
| There is a lot of reason not to be able to map a syntax character:           |
|                                                                              |
|  -- NODRSYN2 --                                                              |
|  1) The syntax character number has not be described                         |
|  2) The syntax character has been described as UNUSED                        |
|  3) The syntax character set is mapped on its BASESET to an UNUSED character |
|                                                                              |
|  -- NODRDOC2 --                                                              |
|  4) A corresponding document character number has not been described,        |
|  5) The corresponding document character number has been described as        |
|     UNUSED.                                                                  |
|  6) The corresponding document character number is mapped on its BASESET     |
|     to an UNUSED character                                                   |
|  In such cases, the CharDefinition table indicates: not described            |
|  7) More than one corresponding document character number has been found     |
|                                                                              |
+-----------------------------------------------------------------------------*/
int NewSGMLdcl::map_to_doc_chnb(char syn_chnb)
{
   CharDefinition *pChdef;
   {
      const CharsetDescriptor& dsetSyn = pSyntaxNew->get_char_descriptor(
         achSwitchTable[syn_chnb]
      );
      if ((!dsetSyn.isDescribed) || (dsetSyn.isUnused)) {
         errorh(NODRSYN2, pEh, ERH_E_SEV, syn_chnb);
         isSevereErrorFound = TRUE;
         return -1;
      }
      pChdef = aChdefTable + dsetSyn.s_chnb;
      if (dsetSyn.isParmLit) pChdef += MAX_CHARSET;
   }
   if (
      (!(pChdef->isDescribedInDoc)) || (pChdef->isMultipleInDoc) ||
      (pSyntaxNew->is_nonsgml(pChdef->doc_d_chnb))
   ) {
      errorh(NODRDOC2, pEh, ERH_E_SEV, syn_chnb);
      isSevereErrorFound = TRUE;
      return -1;
   }
   return pChdef->doc_d_chnb;
}

/*PRIVATE-----------------------------------------------read_parameter_literal-+
| Get a parameter literal into `pchBuffer', returns its length.                |
|                                                                              |
| `pchBuffer' must point to a buffer that is at least LITLEN + 1 long.         |
|                                                                              |
| When the boolean `is_solely_funchar' is TRUE, and if the parameter was made  |
| solely of function characters (introduced by named character reference),     |
| returns a length of 0.                                                       |
|                                                                              |
| IMPLEMENTATION NOTE:                                                         |
| This variant version of ParseState::read_parameter_literal is used           |
| when parsing the SYNTAX parameter of the SGMLdeclaration.                    |
|                                                                              |
| 1) Only CRO, LIT or LITA will be recognized.                                 |
| 2) The gotten parameter will be expressed using the document character set.  |
|    a) If a character reference number is found, it will stay unchanged.      |
|    b) If a named character reference is found, it is translated into         |
|       a syntax character number by translate_character_ref,                  |
|       then mapped to a document character (map_to_do_chnb)                   |
|    c) If a character is entered directly, it is transformed into             |
|       a syntax character number, then mapped to a document character number  |                                  |
|       (check_and_map_to_doc_chnb).                                           |
+-----------------------------------------------------------------------------*/
unsigned int NewSGMLdcl::read_parameter_literal(
   char *pchBuffer, boolean isSolelyFunchar
)
{
   const int iMaxLength = pSyntax->pQuantity->value_of(LITLEN_QTY_ix);
   const e_LexToken delim_token = pPrsstat->get_curr_token();
   int iLength = 0;
   e_LexToken new_token;

   if ((delim_token != LIT_TK) && (delim_token != LITA_TK)) {
      errorh(NOLIT, pEh, ERH_T_SEV);
   }
   pPrsstat->pPrsmode->push_mode(LITP_MODE);
   while (                               // until same delimiter
      new_token = pPrsstat->recognize_delimiter(),
      (new_token != delim_token)
   ) {
      int chnb;

      switch (new_token) {
      case NOT_FOUND_TK:
         if (pEh->is_Ee()) {
            errorh(
               UBLNLIT, pEh, ERH_T_SEV,
               pSyntax->pDlmstr->value_of(
                  delim_token==LIT_TK ? LIT_VDL_ix : LITA_VDL_ix
               )
            );
            chnb = -1;
         }else {
            chnb = check_and_map_to_doc_chnb((char)pEh->curr_char_val());
         }
         break;
      case CRO_TK:
         switch (pSyntax->translate_character_ref(
               pEh, pPrsstat->get_curr_count()
            )
         ) {
         case INVALID_CHRF:
            continue;                    // curr_char already at next char
         case NAMED_CHRF:                // Named: is already a doc_chnb
            chnb = map_to_doc_chnb((char)pEh->curr_char_val());
            break;
         case NUMERIC_CHRF:              // Numeric: user's translated
            chnb = pEh->curr_char_val(); // OK: enter this chnb
            break;
         }
         break;
      default:
         pEh->unread(pPrsstat->get_curr_count());
         chnb = check_and_map_to_doc_chnb((char)pEh->curr_char_val());
         break;
      }
      if (chnb != -1) {
         if ((isSolelyFunchar) && (!(pSyntaxNew->is_fctchar(chnb)))) {
            isSolelyFunchar = FALSE;      // release caller's bad assumption
         }
         pchBuffer[iLength] = (char)chnb;
         if (++iLength > iMaxLength) errorh(BIGLIT, pEh, ERH_T_SEV);
      }
      pEh->read();
   }
   pPrsstat->pPrsmode->pop_mode();
   pchBuffer[iLength] = '\0';
   return isSolelyFunchar ? 0 : iLength;
}


/*PRIVATE-------------------------------------------------check_instance_chars-+
| If INSTANCE, verify that no SGML character can lead to an ambiguity,         |
| as decribed by 13.3 b                                                        |
+-----------------------------------------------------------------------------*/
void NewSGMLdcl::check_instance_chars()
{
   if (pSyntaxNew->functionchar_value(0) != pSyntax->get_REchar()) {
      errorh(INVISTCH, ERH_E_SEV, ReservedName::name_of(RE_RRN_ix));
   }
   if (pSyntaxNew->functionchar_value(1) != pSyntax->get_RSchar()) {
      errorh(INVISTCH, ERH_E_SEV, ReservedName::name_of(RS_RRN_ix));
   }
   if (pSyntaxNew->functionchar_value(2) != pSyntax->get_SPACEchar()) {
      errorh(INVISTCH, ERH_E_SEV, ReservedName::name_of(SPACE_RRN_ix));
   }
   {                                              // look for a proper TAB
      for (int i=3; i < pSyntaxNew->max_functionchar(); ++i) {
         if (
           (SEPCHAR_FCTCLS == pSyntaxNew->functionchar_class(i)) &&
           ('\t' == (char)pSyntaxNew->functionchar_value(i)) &&
           (!strcmp(
                 ReservedName::name_of(TAB_RRN_ix),
                 pSyntaxNew->functionchar_name(i)
              )
           )
         ) {
           break;                                 // found!
         }
      }
      if (i == 0) {
         errorh(INVISTCH, ERH_E_SEV, ReservedName::name_of(TAB_RRN_ix));
      }
   }
   if (isMdoChanged) {
      errorh(INVISTDLM, ERH_E_SEV, ReservedName::name_of(MDO_RRN_ix));
   }
   if (isComChanged) {
      errorh(INVISTDLM, ERH_E_SEV, ReservedName::name_of(COM_RRN_ix));
   }
   if (isPioChanged) {
      errorh(INVISTDLM, ERH_E_SEV, ReservedName::name_of(PIO_RRN_ix));
   }
   if (isMdcChanged) {
      errorh(INVISTDLM, ERH_E_SEV, ReservedName::name_of(MDC_RRN_ix));
   }
   {
      boolean isSevErrSave = isSevereErrorFound;  // Must be restored

      for (char *pchList = "<!>-?"; *pchList; ++pchList) {
         int i = map_to_doc_chnb(*pchList);
         if ((i == -1) || (i != *pchList)) {
            char asz[2] = "";
            asz[1] = *pchList;
            errorh(INVISTCH, ERH_E_SEV, asz);
         }
      }
      isSevereErrorFound = isSevErrSave; // if map_to_doc_chnb changed it
   }
}

/*PRIVATE----------------------------------------------------------get_feature-+
| Get and check the values related to a feature                                |
|                                                                              |
| Input: Read pointer at the beginning of the feature name                     |
|        ixFeatureName indexes the name of the feature                         |
|        isSupported is TRUE if this version of YASP supports this feature     |
|                    is FALSE otherwise                                        |
|        piNumber points an area where to put the associated number, if any    |
|              or is a NULL pointer if no associated number are expected       |
|        iMaxNumber is the maximum value for the associated number             |
|                irrelevant when piNumber is NULL                              |
|                                                                              |
| Output: *piNumber has the associated number                                  |
|         Returns 1 if the feature is to be used, 0 otherwise;                 |
|         Read pointer after the (feature_name, answer, number?)               |
|         and any ps that follows                                              |
+-----------------------------------------------------------------------------*/
boolean NewSGMLdcl::get_feature(
   e_IxRsvnameRefer ixFeatureName,
   boolean isSupported,
   unsigned int *piNumber,
   unsigned int iMaxNumber
)
{
   boolean isFeatureOn = FALSE;

   pSyntax->match_rsvname_or_die(pEh, ixFeatureName);
   pPrsstat->skip_inferable_required_ps();
   if (answer_is_yes()) {
      if (!isSupported) {
         errorh(INVFEAT, pEh, ERH_E_SEV, ReservedName::name_of(ixFeatureName));
      }else {
         isFeatureOn = TRUE;
      }
      if (piNumber) {
         pPrsstat->skip_inferable_required_ps();
         unsigned long ul = pSyntax->read_number(pEh);
         if (isSupported) {
            if (ul > iMaxNumber) {
               errorh(BIGFEAT, pEh, ERH_E_SEV, iMaxNumber);
               *piNumber = iMaxNumber;
            }else if (!ul) {
               isFeatureOn = FALSE;   // number == 0 also means NO
            }else {
               *piNumber = (unsigned int)ul;
            }
         }
      }
   }
   pPrsstat->skip_inferable_required_ps();
   return isFeatureOn;
}

/*PRIVATE-----------------------------------------------------process_features-+
| Process the FEATURES entry of the SGML declaration                           |
|                                                                              |
| Input: Read pointer (supposedly) pointing the FEATURES keyword               |
|                                                                              |
| Output: featuresNew is set                                                   |
|         Read pointer at the end of [195], plus any ps that follows           |
|                                                                              |
+-----------------------------------------------------------------------------*/
void NewSGMLdcl::process_features()
{
   pSyntax->match_rsvname_or_die(pEh, FEATURES_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   pSyntax->match_rsvname_or_die(pEh, MINIMIZE_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   featuresNew.isDatatag = get_feature(DATATAG_RRN_ix, FALSE);
   featuresNew.isOmittag = get_feature(OMITTAG_RRN_ix, TRUE);
   featuresNew.isRank = get_feature(RANK_RRN_ix, TRUE);
   featuresNew.isShorttag = get_feature(SHORTTAG_RRN_ix, TRUE);
   pSyntax->match_rsvname_or_die(pEh, LINK_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   featuresNew.isSimple = get_feature(        /* working on it... */
      SIMPLE_RRN_ix, FALSE,
      &featuresNew.iMaxSimple, MAX_SIMPLE
   );
   featuresNew.isImplicit = get_feature(IMPLICIT_RRN_ix, FALSE);
   featuresNew.isExplicit = get_feature(
      EXPLICIT_RRN_ix, FALSE, &featuresNew.iMaxExplicit, MAX_EXPLICIT
   );
   pSyntax->match_rsvname_or_die(pEh, OTHER_RRN_ix);
   pPrsstat->skip_inferable_required_ps();
   featuresNew.isConcur = get_feature(
      CONCUR_RRN_ix, FALSE, &featuresNew.iMaxConcur, MAX_CONCUR
   );
   featuresNew.isSubdoc = get_feature(
      SUBDOC_RRN_ix, TRUE, &featuresNew.iMaxSubdoc, MAX_SUBDOC
   );
   featuresNew.isFormal = get_feature(FORMAL_RRN_ix, TRUE);
}

/*PRIVATE---------------------------------------------------------------unload-+
| Save the compiled Sgml declaration to disk                                   |
+-----------------------------------------------------------------------------*/
void NewSGMLdcl::unload(ApplicationHandle ahlAccess)
{
   pSyntaxNew->unload(ahlAccess);
   pCapacityNew->unload(
      ahlAccess, pSyntaxNew->pQuantity->value_of(NAMELEN_QTY_ix)
   );
   pEh->write(ahlAccess, &isScopeInstanceNew, sizeof isScopeInstanceNew);
   pEh->write(ahlAccess, &featuresNew, sizeof featuresNew);
   if (pszAppinfoNew) {
      pEh->write_record(ahlAccess, pszAppinfoNew, 1 + strlen(pszAppinfoNew));
      free(pszAppinfoNew);
   }else {
      pEh->write_record(ahlAccess, 0, 0);
   }
}

/*---------------------------------------------------------------------SGMLdcl-+
| SGMLdcl constructor from scratch -- Enough to read an SGML declaration       |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
SGMLdcl::SGMLdcl()
{
   pSyntax = new Syntax();
   pCapacity = new Capacity();
   isScopeInstance = FALSE;               // Unused
   memset(&features, 0, sizeof features); // Unused
   pszAppinfo = 0;                        // Unused
}

/*---------------------------------------------------------------------SGMLdcl-+
| SGMLdcl constructor from OCS                                                 |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
SGMLdcl::SGMLdcl(ApplicationHandle ahlAccess)
{
   pSyntax = new Syntax(ahlAccess);
   pCapacity = new Capacity(ahlAccess);
   EntityHandler::read(ahlAccess, &isScopeInstance, sizeof isScopeInstance);
   EntityHandler::read(ahlAccess, &features, sizeof features);
   EntityHandler::read_record(ahlAccess, (void **)&pszAppinfo); // NULL if INFO NONE
}

/*--------------------------------------------------------------------~SGMLdcl-+
| SGMLdcl destructor                                                           |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
SGMLdcl::~SGMLdcl()
{
   free(pszAppinfo);
   delete pCapacity;
   delete pSyntax;
}
/******************************************************************************/
