/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// skmgr.c
///  String Keeper class
//   Pierre G. Richard 87/06/23 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
/*
| This module stores characters string in contiguous large areas, to:
|  - avoid the overhead to store them individually,
|  - permit creation on an external medium of a unique record of all the
|    strings that have been declared here.
|
| For this purpose, when the caller needs nbytes of storage for a given
| string, it calls "sk_park" that stores the string and return an OFS value.
|
| The OFS value is unique to each string.  After all strings have been
| processed, and written in unique record on the external media, OFS values
| will be the offset of each string within the beginning of the record.
| Offset zero is *always* a null string.
|
| "sk_point" allow recovery of the address of a string, given its OFS value.
|
| Note that only string of chars can be handled this way, no boundary
| alignment being done by this module.
|
*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <memory.h>
#include "gendef.h"
#include "errorh.h"
#include "skmgr.h"

/*-------------------------------------------------------------------------new-+
| Create a Parking.  Use malloc to be able to do realloc.                      |
+-----------------------------------------------------------------------------*/
void *Parking::operator new(size_t size)
{
   return (Parking *)malloc(size);
}

/*----------------------------------------------------------------------delete-+
| Insure consistency of the alloc/free library routines.                       |
+-----------------------------------------------------------------------------*/
void Parking::operator delete(void *pVoid)
{
   free(pVoid);
}

/*---------------------------------------------------------------------Parking-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
Parking::Parking(StringKeeper *pSk)
{
   prev = pSk->pPark;
   room_left = PARKING_LOT_SIZE;
   ofs_start = pSk->ofs;
   if (prev) {
      prev = (Parking *)realloc(prev, sizeof (Parking) - prev->room_left);
      prev->room_taken = PARKING_LOT_SIZE - prev->room_left;
   }else {
      pSk->ofs = 1;
      lot[0] = '\0';      // start with an empty string
      --room_left;
   }
   pSk->pPark = this;
}

/*----------------------------------------------------------------StringKeeper-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
StringKeeper::StringKeeper()
{
   pPark = 0;
   ofs = 0;
}

/*---------------------------------------------------------------~StringKeeper-+
| Destructor                                                                   |
+-----------------------------------------------------------------------------*/
StringKeeper::~StringKeeper()
{
   while (pPark) {
      Parking *pParkWork = pPark->prev;
      delete pPark;
      pPark = pParkWork;
   }
}

/*------------------------------------------------------------------------park-+
| Park a new string - return its offset.                                       |
| pchSource is a NULL pointer if a reservation is just needed.                 |
+-----------------------------------------------------------------------------*/
unsigned int StringKeeper::park(
   const char * const pchSource, const unsigned int iLenSource
)
{
   if (iLenSource > (MAX_ALLOC - ofs)) errorh(BIGALLOC, ERH_T_SEV, SKMG_BGAL);
   if ((!pPark) || (iLenSource > pPark->room_left)) {
      pPark = new Parking(this);
      if (iLenSource > pPark->room_left) exit(104);
   }
   if (pchSource) {
      memcpy(pPark->lot + (ofs - pPark->ofs_start), pchSource, iLenSource);
   }
   ofs += iLenSource;
   pPark->room_left -= iLenSource;
   return ofs - iLenSource;
}

/*-----------------------------------------------------------------------point-+
|Given a parked string offset value, return a pointer to the string.           |
+-----------------------------------------------------------------------------*/
char * StringKeeper::point(const unsigned int ofsSource) const
{
   Parking *pParkSource;

   for (
      pParkSource = pPark;
      pParkSource->ofs_start > ofsSource;
      pParkSource = pParkSource->prev
   );
   return pParkSource->lot + (ofsSource - pParkSource->ofs_start);
}

/*----------------------------------------------------------------------output-+
|Output all strings in a whole unique record, preceeded by its length.         |
|output is a "write" kind of function                                          |
+-----------------------------------------------------------------------------*/
void StringKeeper::output(
   boolean (*out)(ApplicationHandle, void const *, unsigned int),
   ApplicationHandle ahl
)
{
   if (!pPark) return;

   pPark->room_taken = Parking::PARKING_LOT_SIZE - pPark->room_left;
   (*out)(ahl, &ofs, sizeof ofs);

   Parking *pParkWork1 = pPark;
   Parking *pParkWork2 = 0;
   do {                                 // Inverse the pointer chain
      Parking *pParkWork3 = pParkWork2; // w3 is the next park
      pParkWork2 = pParkWork1;          // w2 is the current park
      pParkWork1 = pParkWork2->prev;    // w1 is the previous park
      pParkWork2->prev = pParkWork3;    // so "prev" is changed to "next"
   }while (pParkWork1);
   do {
      (*out)(ahl, pParkWork2->lot, pParkWork2->room_taken);
      pParkWork1 = pParkWork2;
      pParkWork2 = pParkWork2->prev;   // rather be named "next"
      delete pParkWork1;
   }while (pParkWork2);

   pPark = 0;                          // Reset initial values
   ofs = 0;
}
/******************************************************************************/
