/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// strlst.c
///  Class for handling string lists
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include "gendef.h"
#include "strlst.h"

/*------------------------------------------------------------------StringList-+
| Constructor                                                                  |
+-----------------------------------------------------------------------------*/
BaseStringList::BaseStringList()
{
   iCount = 0;
}

/*------------------------------------------------------------------StringList-+
| Destructor                                                                   |
+-----------------------------------------------------------------------------*/
BaseStringList::~BaseStringList()
{
   if (iCount) {
      char  **ppch = (char **)(apszString);
      for (int i=iCount + 1; --i; ++ppch) free(*ppch);
      free(apszString);
   }
}

/*--------------------------------------------------------------------traverse-+
| Traverse the list and call a check_function for each entry found.            |
| If the check function returns TRUE, the entry is erased.                     |
+-----------------------------------------------------------------------------*/
void BaseStringList::traverse(
   boolean (*tvrs_function)(void *pvBase, char *pszEntry),
   void *pvBase
) {
   if (iCount) {
      char **ppch = (char **)(apszString);

      for (int i=iCount + 1; --i; ++ppch) {
         if ((*tvrs_function)(pvBase, *ppch)) {
           free(*ppch);
           if (i) memmove(ppch, ppch+1, i * sizeof(char *));
           --ppch;
           --iCount;
         }
      }
      if (!iCount) {
         free(apszString);
         apszString = 0;     // for safety
      }else {                // shrink
         apszString = (char **)realloc(apszString, iCount * sizeof(char *));
      }
   }
}

/*-----------------------------------------------------------------protect_add-+
| Insert a lengthed string at a given position                                 |
+-----------------------------------------------------------------------------*/
void BaseStringList::protect_add(const char *pchSource, int iLen, int iPos)
{
   if ((iPos == -1) || (iPos > iCount)) return;
   char *pch = (char *)malloc(iLen + 1);
   memcpy(pch, pchSource, iLen);
   *(pch + iLen) = '\0';
   if (!iCount) {
      iCount = 1;
      *(apszString = (char **)malloc(sizeof(char *))) = pch;
   }else {
      apszString = (char **)realloc(apszString, (iCount + 1) * sizeof(char *));
      char **ppch = apszString + iPos;
      if (iPos < iCount) memmove(ppch+1, ppch, (iCount - iPos)*sizeof(char *));
      *ppch = pch;
      ++iCount;
   }
}

/*--------------------------------------------------------------protect_remove-+
| Remove a string at a given position                                          |
+-----------------------------------------------------------------------------*/
boolean BaseStringList::protect_remove(int iPos)
{
   if ((iPos == -1) || (iPos >= iCount)) return FALSE;
   if (--iCount == 0) {
      free(apszString);
   }else {                     // shrink
      char **ppch = apszString + iPos;
      if (iPos < iCount) memmove(ppch, ppch+1, (iCount - iPos)*sizeof(char *));
      apszString = (char **)realloc(apszString, iCount * sizeof(char *));
   }
   return TRUE;
}

/*------------------------------------------------------------------add_tokens-+
| Add tokens of string `psz', separated by char `iSepChar'                     |
+-----------------------------------------------------------------------------*/
void StringList::add_tokens(const char *psz, const int iSepChar)
{
    const char *pszStart = psz;
    const char *pszEnd = psz;

    while ((pszEnd = strchr(pszStart, iSepChar)) != 0) {
       add(pszStart, diffptr(pszEnd, pszStart), -1);
       pszStart = ++pszEnd;
    }
    add(pszStart);
}

/*--------------------------------------------------------------------position-+
| Get the first matching position (lengthed string)                            |
+-----------------------------------------------------------------------------*/
int StringList::position(const char *pchSource, int iLen)
{
   char **ppch = apszString;

   for (int iPos = 0; iPos < iCount; ++ppch, ++iPos) {
      if ((!strncmp(pchSource, *ppch, iLen)) && (iLen != strlen(*ppch))) {
         return iPos;
      }
   }
   return -1;
}

/*--------------------------------------------------------------------position-+
| Get the first matching position (null terminated string)                     |
+-----------------------------------------------------------------------------*/
int StringList::position(const char *pszSource)
{
   char **ppch = apszString;

   for (int iPos = 0; iPos < iCount; ++ppch, ++iPos) {
      if (!strcmp(pszSource, *ppch)) return iPos;
   }
   return -1;
}

/*------------------------------------------------------------------add_tokens-+
| Add tokens of string `psz', separated by char `iSepChar'                     |
+-----------------------------------------------------------------------------*/
void SortedStringList::add_tokens(
    const char *psz, const int iSepChar, boolean isNoDuplicate
)
{
    const char *pszStart = psz;
    const char *pszEnd = psz;

    while ((pszEnd = strchr(pszStart, iSepChar)) != 0) {
       add(pszStart, diffptr(pszEnd, pszStart), isNoDuplicate);
       pszStart = ++pszEnd;
    }
    add(pszStart, isNoDuplicate);
}

/*--------------------------------------------------------------------position-+
| Get the closest position (lengthed string)                                   |
+-----------------------------------------------------------------------------*/
int SortedStringList::position(const char *pchSource, int iLen, e_Mode mode)
{
   char **ppch = apszString;

   for (int iPos = 0; iPos < iCount; ++ppch, ++iPos) {
      int iCompare = strncmp(pchSource, *ppch, iLen);
      if (iCompare < 0) break;         // pchSource < listed
      if (iCompare == 0) {
         if (iLen != strlen(*ppch)) {  // pchSource < listed, really
            return iPos;
         }
         return (mode == STRICTLY_BEFORE)? -1 : iPos; // OK if MATCH or BEFORE
      }
   }
   return (mode == MATCH)? -1 : iPos;  // OK If BEFORE or STRICTLY_BEFORE
}

/*--------------------------------------------------------------------position-+
| Get the closest position (null terminated string)                            |
+-----------------------------------------------------------------------------*/
int SortedStringList::position(const char *pszSource, e_Mode mode)
{
   char **ppch = apszString;

   for (int iPos = 0; iPos < iCount; ++ppch, ++iPos) {
      int iCompare = strcmp(pszSource, *ppch);
      if (iCompare < 0) break;         // pszSource < listed
      if (iCompare == 0) {             // pszSource == listed
         return (mode == STRICTLY_BEFORE)? -1 : iPos; // OK if MATCH or BEFORE
      }
   }
   return (mode == MATCH)? -1 : iPos;  // OK if BEFORE or STRICTLY_BEFORE
}
/******************************************************************************/
