/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// strlst.h
///  Declarations for strlst
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _STRLST_H
#define _STRLST_H 1
/*
| StringList and SortedStringList classes
|
| For some of these functions, the string value, passed by argument,
| does not need to be null-terminated.  In this case, the length
| must be given, and no '\0' must be found within this length.
|
| Strings in a StringList object are accessible by their position: 0 being
| the first one,  count() - 1 being the last.  -1, as a passed position
| parameter, means: end-of-list. -1, as a returned position, means: not found.
|
| There is no such concept of position for strings in a SortedStringList
| object.  SortedStringList are used to remember if a string value was
| defined at least once (isNoDuplicate == TRUE), or to check that each
| string in a list of strings was processed (check list role).
|
| An object of the BaseStringList class cannot be directly instanciated.
|
*/
//******************************************************************************

#include <string.h>     // this class is used by inlines

class BaseStringList {        // bstrl
public:
   /*-------------------------------------+
   | Functions that can be used either    |
   | for StringList or Sorted StringList  |
   +-------------------------------------*/
   /*
   | Traverse the list, call `tvrs_function(pvBase, pszEntry)'
   | for each string entry found.
   */
   void traverse(
      boolean (*tvrs_function)(void *pvBase, char *pszEntry), void *pvBase
   );

   /*
   | Return the number of entries in the list
   */
   int count();

protected:
   char **apszString;
   int iCount;
   BaseStringList();
   _export ~BaseStringList();
   void protect_add(const char *pch, int iLen, int iPos = -1);
   boolean protect_remove(int iPos = 0);
};

class StringList : public BaseStringList {  // strl
public:
   /*
   | Add a new string at the position specified in `iPos'
   | (-1: at the end of the list).
   | The value of the string to add is described by `pch[0] ... pch[iLen-1]'.
   | `iLen' has the length of the string `pch' which does not
   | need to be NULL-terminated.
   */
   void add(const char *pch, int iLen, int iPos);

   /*
   | Add a new string at the position specified by `iPos'
   | (default is -1: at the end of the list).
   | The value of the string to add is given by `psz'.
   | `psz' must be a NULL-terminated string.
   */
   void add(const char *psz, int iPos = -1);

   /*
   | Decompose a string into individual sub-strings,
   | each sub-string, except the last one, being delimited by the
   | character `chSepchar'.  Add each of these sub-strings in the list.
   | The value of the string to decompose is given by `psz'.
   | `psz' must be a NULL-terminated string.
   */
   void add_tokens(const char *psz, const int iSepChar = ' ');

   /*
   | Normalize the 'psz' string by applying the following rules:
   |  - ignore all characters that are in "pszIgnoreSet"
   |  - ignore all leading/trailing sequence of characters
   |    that are in pszReduceSet
   |  - reduce all other sequence of characters that are in pszReduceSet
   |    to a single `chSepChar' character
   |
   | Note: 1) this is a static function (call by `StringList::normalize')
   |       2) the input string is overwritten
   |       Hence, a typical call would be:
   |       pszNormalizedString = StringList::normalize(strdup(pszInputString));
   */
   //***>Note: NOT YET IMPLEMENTED!
   static char * normalize(
      char *psz,
      const int iSepChar = ' ',
      const char *pszReduceSet = " \t\r",
      const char *pszIgnoreSet = "\n"
   );

   /*
   | Remove the first string from the list
   | which matches the value described by `pch[0] ... pch[iLen-1]'.
   | `iLen' has the length of the string `pch' which does not
   | need to be NULL-terminated.
   | return TRUE if OK
   */
   boolean remove(const char *pch, int iLen);

   /*
   | Remove the first string from the list
   | which matches the value given in `psz'.
   | `psz' must be a NULL-terminated string.
   | Return TRUE if OK
   */
   boolean remove(const char *psz);

   /*
   | Remove the entry at the position specified by `iPos'
   | (default: at the end of the list).
   | Return TRUE if OK
   */
   boolean remove(int iPos = -1);

   /*
   | Get the position of the first string from the list
   | which matches the value described by `pch[0] ... pch[iLen-1]'.
   | `iLen' has the length of the string `pch' which does not
   | need to be NULL-terminated.
   | Return -1 if not found
   */
   int position(const char *pch, int iLen);

   /*
   | Get the position of the first string from the list
   | which matches the value given in `psz'.
   | `psz' must be a NULL-terminated string.
   | Return -1 if not found
   */
   int position(const char *psz);

   /*
   | Return a string from its position
   | Returns NULL if out of bound
   */
   char * operator[](int iPos);
};

class SortedStringList : public BaseStringList {      // sstrl
public:
   /*
   | Add a new string.
   | The value of the string to add is described by `pch[0] ... pch[iLen-1]'.
   | `iLen' has the length of the string `pch' which does not
   | need to be NULL-terminated.
   | if isNoDuplicate == TRUE: don't add if already there (default)
   | if isNoDuplicate == FALSE: add even if already there
   */
   void add(const char *pch, int iLen, boolean isNoDuplicate = TRUE);

   /*
   | Add a new string.
   | The value of the string to add is given by `psz'.
   | `psz' must be a NULL-terminated string.
   | if isNoDuplicate == TRUE: don't add if already there (default)
   | if isNoDuplicate == FALSE: add even if already there
   */
   void add(const char *psz, boolean isNoDuplicate = TRUE);

   /*
   | Decompose a string into individual sub-strings,
   | each sub-string, except the last one, being delimited by the
   | character `chSepchar'.  For each of these sub-strings:
   | if isNoDuplicate == TRUE: don't add to the list if already there (default)
   | if isNoDuplicate == FALSE: add to the list, even if already there
   | The value of the string to decompose is given by `psz'.
   | `psz' must be a NULL-terminated string.
   */
   void add_tokens(
      const char *psz, const int iSepChar = ' ', boolean isNoDuplicate = TRUE
   );

   /*
   | Remove the first string from the list
   | which matches the value described by `pch[0] ... pch[iLen-1]'.
   | `iLen' has the length of the string `pch' which does not
   | need to be NULL-terminated.
   | return TRUE if OK
   */
   boolean remove(const char *pch, int iLen);

   /*
   | Remove the first string from the list
   | which matches the value given in `psz'.
   | `psz' must be a NULL-terminated string.
   | Return TRUE if OK
   */
   boolean remove(const char *psz);

   /*
   | Find if a string from the list
   | matches the value described by `pch[0] ... pch[iLen-1]'.
   | `iLen' has the length of the string `pch' which does not
   | need to be NULL-terminated.
   | Return -1 if not found
   */
   boolean find(const char *pch, int iLen);

   /*
   | Find if a string from the list
   | matches the value given in `psz'.
   | `psz' must be a NULL-terminated string.
   | Return -1 if not found
   */
   boolean find(const char *psz);

private:
   enum e_Mode { BEFORE, STRICTLY_BEFORE, MATCH };
   int position(const char *pch, int iLen, e_Mode mode);
   int position(const char *psz, e_Mode mode);
};

/*-- INLINES --*/
inline int BaseStringList::count() {
   return iCount;
}

inline void StringList::add(const char *pch, int iLen, int iPos)
{
   protect_add(pch, iLen, (iPos == -1)? iCount : iPos);
}

inline void StringList::add(const char *psz, int iPos)
{
   protect_add(psz, strlen(psz), (iPos == -1)? iCount : iPos);
}

inline boolean StringList::remove(const char *pchSource, int iLen)
{
   return protect_remove(position(pchSource, iLen));
}

inline boolean StringList::remove(const char *pszSource)
{
   return protect_remove(position(pszSource));
}

inline boolean StringList::remove(int iPos)
{
   return protect_remove((iPos == -1)? iCount-1 : iPos);
}

inline char * StringList::operator[](int iPos)
{
   return (
      (iPos < -1) || (iPos >= iCount)
   )? 0 : apszString[(iPos == -1)? iCount-1 : iPos];
}

inline void SortedStringList::add(
   const char *pch, int iLen, boolean isNoDuplicate
) {
   protect_add(
      pch, iLen, position(pch, iLen, isNoDuplicate? STRICTLY_BEFORE : BEFORE)
   );
}
inline void SortedStringList::add(const char *psz, boolean isNoDuplicate)
{
   protect_add(
      psz, strlen(psz), position(psz, isNoDuplicate? STRICTLY_BEFORE : BEFORE)
   );
}

inline boolean SortedStringList::remove(const char *pchSource, int iLen)
{
   return protect_remove(position(pchSource, iLen, MATCH));
}

inline boolean SortedStringList::remove(const char *pszSource)
{
   return protect_remove(position(pszSource, MATCH));
}

inline boolean SortedStringList::find(const char *pchSource, int iLen)
{
   return (position(pchSource, iLen, MATCH) == -1)? FALSE : TRUE;
}

inline boolean SortedStringList::find(const char *pszSource)
{
   return (position(pszSource, MATCH) == -1)? FALSE : TRUE;
}
#endif /* _STRLST_H ***********************************************************/
