/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// syntax.h
///  Declarations for syntax
//   Pierre G. Richard 93/10/15 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
#ifndef _SYNTAX_H
#define _SYNTAX_H 1
//******************************************************************************

#include "sgmldcl.h"    // is a friend
#include "dlmvect.h"    // is a friend
#include "skmgr.h"      // this class is used by inlines

#include "syntax.eh"    // enums
#include "dlmstr.eh"
#include "rsvname.eh"
#include "lextoken.eh"
#include "externid.eh"

class EntityHandler;
class DelimiterVector;
class NewDelimiterVector;
class DelimiterString;
class NewDelimiterString;
class ReservedName;
class NewReservedName;
class Quantity;
class NewQuantity;
class Shortref;
class NewShortref;
class StringKeeper;
class SortArray;
class ExternId;

/*-----------------------------------------------------------------------------+
| Public structures and definitions                                            |
+-----------------------------------------------------------------------------*/
struct ReferChoiceEntry {     // Choice table entry. finding a Reference Name
   e_IxRsvnameRefer ix;       // Index in the Reference Reserved Names
   int  value;                // Value assigned by the caller
};

struct VarntChoiceEntry {     // Choice table entry. finding a Variant Name
   e_IxRsvnameVarnt ix;       // Index in the Variant Reserved Names table
   int  value;                // Value assigned by the caller
};

enum e_CharClass {
   NAMESTRT_CHCLS,
   NAME_CHCLS,
   DIGIT_CHCLS
};


/*-----------------------------------------------------------------------------+
| CharsetDescriptor: BASESET's + DESCSET's                                     |
| Saved in the OCS and ODT only to be able to check                            |
| charsets compatibility if SCOPE INSTANCE (13.3).                             |
+-----------------------------------------------------------------------------*/
struct CharsetDescriptor {      // DSET: indexed by the code # (d_chnb)
   boolean isDescribed :1;      // described
   boolean isBased     :1;      // described thru a based chnb
   boolean isParmLit   :1;      // described thru a parameter literal
   boolean isUnused    :1;      // described as UNUSED
   unsigned int s_chnb :8;      // s_chnb value (b_chnb -> system charset)
};                              // or pseudo s_chnb if parmlit

/*-----------------------------------------------------------------------------+
| FNCR table: Mapping of function characters.                                  |
+-----------------------------------------------------------------------------*/
struct FunctionCharAssoc {      // to map :
   union {
     const char *p;             // a function name (ex: "TAB")
     unsigned int ofs;          // (offset in OCS file)
   };
   unsigned int ch_nb;          // to a corresponding char number (ex: 9)
   e_FunctionClass fctcls;      // Function Class (ex: SEPCHAR_FCTCLS)
};

/*-----------------------------------------------------------------------------+
| Syntax Class                                                                 |
+-----------------------------------------------------------------------------*/
class Syntax
{
   friend void NewSGMLdcl::parse_function_parameter();
   friend void NewSGMLdcl::parse_basesets_and_descsets(boolean);
   friend void NewSGMLdcl::parse_shunchar_parameter();
   friend void NewSGMLdcl::parse_naming_parameter();
   friend void NewDelimiterVector::define_delimiter(
      const char *pszValue,
      const e_LexToken token,
      unsigned int allmode,
      const boolean isSingleDelimiter
   );
   friend ExternId;     // ExternId::formalize.. changes tv->IsCaseGeneral
public:
   /*-----------------+
   | Public Variables |
   +-----------------*/
   DelimiterVector *pDv;        // Delimiter vectors for this OCS
   DelimiterString *pDlmstr;    // Delimiter strings
   Quantity *pQuantity;
   ReservedName *pRsvname;
   Shortref *pShortref;

   /*-----------------+
   | Public Functions |
   +-----------------*/
   Syntax();                             // Minimal Syntax
   Syntax(ApplicationHandle ahlAccess);  // Constructor from OCS
   ~Syntax();                            // Destructor
                                      // According to this Syntax...
   int get_REchar();                  // what's a RE?
   int get_RSchar();                  // what's a RS?
   int get_SPACEchar();               // what's a SPACE?
   boolean is_nonsgml(const int c);   // is this a NONSGML?
   boolean is_digit(const int c);     // is this a digit?
   boolean is_alphanum(const int c);  // is this alphanumeric?
   boolean is_funchar(const int c);   // is this a FUNCHAR?
   boolean is_sepchar(const int c);   // is this a SEPCHAR?
   boolean is_REchar(const int c);    // is this a RE?
   boolean is_RSchar(const int c);    // is this a RS?
   boolean is_SPACEchar(const int c); // is this a SPACE?
   boolean is_blank(const int c);     // is this a Blank (SPACE+SEPCHAR)?
   boolean is_separator(const int c); // is this an S Separator? ([5])
   boolean is_sspchar(const int c);   // is this a RE, RS, or SPACE?
   boolean is_namestrt(const int c);  // can this starts a name? ([53])
   boolean is_name(const int c);      // is this a name character? ([52])
   boolean is_mindata(const int c);   // is this minimum data? [77,78]
   boolean is_minldat(const int c);   // is this LC/UC Letter+Digit+Special?
   boolean is_fctchar(const int c);   // is this a function character? ([54])
   boolean is_ms_char(const int c);   // is this one of the MSO, MSI, MSS?
   boolean is_msrchar(const int c);   // is this a MSO, or MSS?
   boolean is_msichar(const int c);   // is this a MSI?
   boolean is_msochar(const int c);   // is this a MSO?
   boolean is_msschar(const int c);   // is this a MSS?
   boolean is_cref(const int c);      // does this match the CREF constraint?
   int     uppercase(const int c);    // uppercase this character
   boolean is_case_general();         // is general case YES?
   boolean is_case_entity();          // is entity case YES?
   boolean is_char_class(const int c, e_CharClass chcls);


   const unsigned int max_functionchar();
   const char * functionchar_name(int ix);
   const char * functionchar_name(int ix, StringKeeper& sk);
   unsigned int functionchar_value(int ix);
   e_FunctionClass functionchar_class(int ix);

   void write_charset(ApplicationHandle ahlAccess);
   boolean read_and_check_charset(ApplicationHandle ahlAccess);

   unsigned int read_to_end_delimiter(
      EntityHandler *pEh,
      char * pchBuffer,
      e_LexToken token = NOT_FOUND_TK,  // useless for PI_GLC
      e_LiteralType glc = PI_GLC
   );
   unsigned long read_number(EntityHandler *pEh);
   boolean read_character_number(EntityHandler *pEh, unsigned int *i_p);
   unsigned int read_namechar(
      EntityHandler *pEh, char *pszName, unsigned int isNamecase
   );
   unsigned int read_general_name(EntityHandler *pEh, char *pszName);
   unsigned int read_entity_name(EntityHandler *pEh, char *pszName);
   unsigned int read_nametoken(EntityHandler *pEh, char *pszName);
   void read_refd_entity_name(EntityHandler *pEh, char *pszName);
   boolean match_rsvname(EntityHandler *pEh, e_IxRsvnameRefer ix);
   boolean match_rsvname(EntityHandler *pEh, e_IxRsvnameVarnt ix);
   void    match_rsvname_or_die(EntityHandler *pEh, e_IxRsvnameRefer ix);
   void    match_rsvname_or_die(EntityHandler *pEh, e_IxRsvnameVarnt ix);
   int match_rsvname_in_list(
      EntityHandler *pEh, const ReferChoiceEntry *choice_p
   );
   int match_rsvname_in_list(
      EntityHandler *pEh, const VarntChoiceEntry *choice_p
   );
   boolean match_string(EntityHandler *pEh, const char *pszString);

   // Skip to a given (ending) delimiter
   boolean skip_to_delimiter (
      EntityHandler *pEh, e_IxDlmstrVarnt vdl_ix
   );

   // Translate a named or numeric character reference
   e_CharRefType translate_character_ref(
      EntityHandler *pEh, int iLength
   );


protected:
   /*--------------------+
   | Protected Variables |
   +--------------------*/
   char *strbuf;                       // Buffer for all strings

   // Grouping tiny values into a structure is done to fix compilers
   // that don't handle enums properly (Borland C 3.1 bug, Jan 12, 93)
   struct TinyValues {
      TinyValues();
      char    chRE;                    // Record End Character
      char    chRS;                    // Record Start Character
      char    chSPACE;                 // Space Character
      boolean isCaseGeneral;           // if CASE GENERAL YES
      boolean isCaseEntity;            // if CASE ENTITY YES
   }tv;
   char    uctab[MAX_CHARSET];         // upper case table
   unsigned short gtype[MAX_CHARSET];  // classification of the character set
   CharsetDescriptor dsettab[MAX_CHARSET]; // Syntax Character Set

   FunctionCharAssoc *fncrtab;         // Function Character <-> Char Nmbr
   unsigned int fncr_max;              // count of entries in fncr table

   /*--------------------+
   | Protected Functions |
   +--------------------*/
   Syntax(int);                    // when building from the SGML declaration

private:
   /*------------------+
   | Private Functions |
   +------------------*/
   unsigned int eat_refc(EntityHandler *pEh);
};

/*-----------------------------------------------------------------------------+
| NewSyntax class  (while building it)                                         |
+-----------------------------------------------------------------------------*/
class NewSyntax : public Syntax
{
public:
   char chBSEQ;
   NewDelimiterVector *pDv;      // overload
   NewDelimiterString *pDlmstr;  // overload
   NewQuantity *pQuantity;       // overload
   NewReservedName *pRsvname;    // overload
   NewShortref *pShortref;       // overload

   NewSyntax(StringKeeper& sk);
   ~NewSyntax();
   const CharsetDescriptor& get_char_descriptor(const int c);
   void reset_functionchar_rsvname();
   boolean check_functionchar();
   void unload(ApplicationHandle ahlAccess);

private:
   StringKeeper& sk;
};

/*------------------------------------------------------------------get_XXchar-+
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
inline int Syntax::get_REchar()     { return tv.chRE; }
inline int Syntax::get_RSchar()     { return tv.chRS; }
inline int Syntax::get_SPACEchar()  { return tv.chSPACE; }

/*--------------------------------------------------------------------is_xxxxx-+
| Character classification functions for the current character read            |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
inline boolean Syntax::is_nonsgml(const int c) {
   return (!gtype[c])? TRUE : FALSE;
}
inline boolean Syntax::is_digit(const int c) {
   return (gtype[c] & (S_DIGIT))? TRUE : FALSE;
}
inline boolean Syntax::is_alphanum(const int c) {
   return (gtype[c] & (S_LETTER | S_DIGIT))? TRUE : FALSE;
}
inline boolean Syntax::is_funchar(const int c) {
   return (gtype[c] & (S_FUNCHAR))? TRUE : FALSE;
}
inline boolean Syntax::is_sepchar(const int c) {
   return (gtype[c] & (S_SEPCHAR))? TRUE : FALSE;
}
inline boolean Syntax::is_REchar(const int c) {
   return (c == tv.chRE)? TRUE : FALSE;
}
inline boolean Syntax::is_RSchar(const int c) {
   return (c == tv.chRS)? TRUE : FALSE;
}
inline boolean Syntax::is_SPACEchar(const int c) {
   return (c == tv.chSPACE)? TRUE : FALSE;
}
inline boolean Syntax::is_blank(const int c) {
   return ((c==tv.chSPACE) || (gtype[c] & (S_SEPCHAR)))? TRUE : FALSE;
}
inline boolean Syntax::is_separator(const int c) {
   return (gtype[c] & (S_SSPCHAR | S_SEPCHAR))? TRUE : FALSE;
}
inline boolean Syntax::is_sspchar(const int c) {
   return (gtype[c] & (S_SSPCHAR))? TRUE : FALSE;
}
inline boolean Syntax::is_namestrt(const int c) {
   return (gtype[c] & (S_LETTER | S_NMSTRT))? TRUE : FALSE;
}
inline boolean Syntax::is_name(const int c) {
   return (
      gtype[c] & (S_LETTER | S_NMSTRT | S_DIGIT | S_NMCHAR)
   )? TRUE : FALSE;
}
inline boolean Syntax::is_mindata(const int c) {
   return (
      gtype[c] & (S_SSPCHAR | S_LETTER | S_DIGIT | S_SPECIAL)
   )? TRUE : FALSE;
}
inline boolean Syntax::is_minldat(const int c) {
   return (gtype[c] & (S_LETTER | S_DIGIT | S_SPECIAL))? TRUE : FALSE;
}
inline boolean Syntax::is_fctchar(const int c) {
   return (
      gtype[c] & (
         S_SSPCHAR | S_SEPCHAR | S_FUNCHAR | S_MSICHAR | S_MSOCHAR | S_MSSCHAR
      )
   )? TRUE : FALSE;
}
inline boolean Syntax::is_ms_char(const int c) {
   return (gtype[c] & (S_MSICHAR | S_MSOCHAR | S_MSSCHAR))? TRUE : FALSE;
}
inline boolean Syntax::is_msrchar(const int c) {
   return (gtype[c] & (S_MSOCHAR | S_MSSCHAR))? TRUE : FALSE;
}
inline boolean Syntax::is_msichar(const int c) {
   return (gtype[c] & (S_MSICHAR))? TRUE : FALSE;
}
inline boolean Syntax::is_msochar(const int c) {
   return (gtype[c] & (S_MSOCHAR))? TRUE : FALSE;
}
inline boolean Syntax::is_msschar(const int c) {
   return (gtype[c] & (S_MSSCHAR))? TRUE : FALSE;
}
inline boolean Syntax::is_cref(const int c) {
   return (gtype[c] & (S_DIGIT | S_LETTER | S_NMSTRT))? TRUE : FALSE;
}
inline boolean Syntax::is_case_general() {
   return (tv.isCaseGeneral)? TRUE : FALSE;
}
inline boolean Syntax::is_case_entity() {
   return (tv.isCaseEntity)? TRUE : FALSE;
}

/*---------------------------------------------------------------is_char_class-+
| Address miscellaneous is_xxxx functions depending on a code                  |
+-----------------------------------------------------------------------------*/
inline boolean Syntax::is_char_class(const int c, e_CharClass chcls)
{
   switch (chcls) {
   case NAMESTRT_CHCLS:         return is_namestrt(c);
   case NAME_CHCLS:             return is_name(c);
   default: /* DIGIT_CHCLS: */  return is_digit(c);
   }
}

/*-------------------------------------------------------------------uppercase-+
| Uppercase a character                                                        |
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
inline int Syntax::uppercase(const int c) {
   return uctab[c];
}

/*-----------------------------------------------------------functionchar_xxxx-+
| -- INVARIANTS --                                                             |
| -- INPUT --                                                                  |
| -- PRECONDITION --                                                           |
| -- OUTPUT --                                                                 |
| -- POSTCONDITION --                                                          |
+-----------------------------------------------------------------------------*/
inline const unsigned int Syntax::max_functionchar() {
   return fncr_max;
}
inline const char * Syntax::functionchar_name(int ix) {
   return fncrtab[ix].p;
}
inline const char * Syntax::functionchar_name(int ix, StringKeeper& sk) {
   return sk.point(fncrtab[ix].ofs);
}
inline unsigned int Syntax::functionchar_value(int ix) {
   return fncrtab[ix].ch_nb;
}
inline e_FunctionClass Syntax::functionchar_class(int ix) {
   return fncrtab[ix].fctcls;
}
inline const CharsetDescriptor& NewSyntax::get_char_descriptor(const int c) {
   return dsettab[c];
}
#endif /* _SYNTAX_H ***********************************************************/
