/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// tag.c
///  Tag attribute specifications
//   Pierre G. Richard 88/08/04 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/
/*
| This module parses all that is related to element (tag) attributes.
| It is made of 4 main functions:
|
| att_val_spec:
|       Checks the validity of an Attribute Value Specification,
|       in conformance with its Attribute Definition (ATT)
|       It is called:
|       - by ODTATTL, for a default value assigned in an <ATTLIST> MD
|       - by att_spec_list (in this module) either when an attribute
|         specification is found in content (finish_parse_tag),
|         or when an attribute specification is given for a data
|         attribute spec of a data entity with a Notation (ENTDCL)
|       The _TGF flags tells what kind of ASL is processed:
|       - DEF_TGF:   Default AVS being parsed (DATA or TAG)
|       - TAG_TGF:   Tag Attribute Specification being parsed
|       - DATA_TGF:  Data Attribute Specification being parsed
|
| att_spec_list
|       Parses an attribute specification list: (attribute name = AVS)*
|       either for a data (DATA_TGF, at compile time) or a tag
|       (TAG_TGF, at doc instance time).  For each AVS, it mallocates
|       a string to be filled by the specified value, calling att_val_spec
|       to do the job.
|
| finish_parse_tag
|       Called when a tag is parsed.
|       Mallocates an Attribute array and fills it with the content
|       of the Attribute array created when the ATTLIST was parsed.
|       Then, calls att_spec_list.
|
| validate_attributes
|       Validates a TAG ASL, after all of its attributes have been
|       parsed.
|
| This can be resumed by the following diagram:
|
|                 | finish_parse_tag | att_spec_list | att_val_spec |
|                 +------------------+---------------+--------------+
| ASL of TAGS  -> |        X         |       X       |      X       |
| in CONTENT      |                  |               |  (TAG_TGF)   |
|                 +------------------+---------------+--------------+
| ASL of N/S/C DATA External .....-> |       X       |      X       |
| Entities (DTD Compile)             |               |  (DATA_TGF)  |
|                                    +---------------+--------------+
| AVS of default value ...........................-> |      X       |
| (ATTLIST for notation, or TAG)                     |   (DEF_TGF)  |
|                                                    +--------------+
|
| There is currently no distinction in processing DATA_TGF or TAG_TGF
| and we only consider if it is DEF_TGF or (DATA_TGF | TAG_TGF).
| If the distinction between DATA_TGF | TAG_TGF stays useless,
| in a next release, TGF will be a boolean: DEF_TGF or not.
*/
//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>

#include "gendef.h"
#include "externid.h"
#include "entity.h"
#include "dlmstr.h"
#include "rsvname.h"
#include "prsstate.h"
#include "dlmvect.h"
#include "sgmldcl.h"
#include "dtddcl.h"
#include "enthdlr.h"
#include "utilctl.h"
#include "quantity.h"
#include "capacity.h"
#include "notation.h"
#include "attrib.h"
#include "hash.h"
#include "prsmode.h"
#include "parser.h"


/*----------------------------------------------------------------att_val_spec-+
| Get (and check) an Attribute Value Specification                             |
|                                                                              |
|  [33] attribute value specification =                                        |
|       attribute value | attribute value literal                              |
|                                                                              |
|  [34] attribute value literal =                                              |
|       (lit, replaceable character data*, lit) |                              |
|       (lita, replaceable character data*, lita)                              |
|                                                                              |
|  [35] attribute value =                                                      |
|         character data |                                                     |
|         general entity name | general entity name list |                     |
|         id value |                                                           |
|         id reference value | id reference list |                             |
|         name | name list |                                                   |
|         name token | name token list |                                       |
|         notation name |                                                      |
|         number | number list |                                               |
|         number token | number token list                                     |
|                                                                              |
|                                                                              |
| On ENTRY:                                                                    |
|    `curr_token/curr_char' points the first character of the AVS,             |
|     pAtt points an Attribute that defines the attribute,                     |
|          or is a NULL pointer to tell "get rid of the AVS"                   |
|     pchBuf is pointing a buffer to receive the specified value.              |
|     pCompo references the caller address of an array of Component            |
|         structures describing what CDATA and SDATA entities pchBuf           |
|         if made of,  whenever it's needed.                                   |
|         This array is mallocated, it's up to the caller to free it.          |
|     uiNormedLenVal is where to put the normalized length of the AVS          |
|                                                                              |
| When RETURNING:                                                              |
|   - "curr_char" is at next unknown character.                                |
|   - Caller's buffer pointed by pchBuf has the attribute value,               |
|     its length (-1 if invalid), not counting the ending EOS, is returned.    |
|   - The reference to the pCompo address is NULL if it wasn't used.           |
|     Otherwhise, it points to a mallocated array of Component structures:     |
|      - pCompo[0].iCount field has the number-1 of components structure       |
|        in the array -- the first component being a header and not counted.   |
|      - Each component is made of:                                            |
|            - a pointer to the entity structure (0 means just text)           |
|            - the length of the concerned part of the string                  |
|     It's up to the caller to free this array.                                |
|                                                                              |
| The attribute value or attribute value literal is parsed as explained        |
| in 7.9.3 of ISO8879-1986(E).                                                 |
| `is.cdata' tells if we need to keep blanks sequences, leading and            |
| trailing blanks.                                                             |
|                                                                              |
| IMPLEMENTATION NOTES:                                                        |
|                                                                              |
| 1) Things are not exactly the same between:                                  |
|    - parsing of a default AVS (TAG or DATA attributes) (Compile time)        |
|    - parsing of a TAG Attribute Specification          (Compile Time)        |
|    - parsing of a DATA Attribute Specification         (Document Time)       |
|    To avoid two version of the almost_same code,                             |
|    the flag TGF makes the difference.                                        |
|                                                                              |
| 2) In most cases, the parse is kept going on to the end of the avs,          |
|    whatever error occurs.  Only one message is issued.                       |
|                                                                              |
| Recognized in LITC mode:                                                     |
|   CRO_TK        , CON_MO | LITC_MO | LITP_MO                                 |
|   LIT_TK        , GRP_MO | LITC_MO | LITP_MO | MD_MO | TAG_MO                |
|   LITA_TK       , GRP_MO | LITC_MO | LITP_MO | MD_MO | TAG_MO                |
|   ERO_TK        , CON_MO | LITC_MO                                           |
|   ERO_GRPO_TK   , CON_MO | LITC_MO                                           |
+-----------------------------------------------------------------------------*/
int ParseState::att_val_spec(
   char * const pchBuf,              // Where to put the AVL
   const Attribute *pAtt,            // Attribute or NULL
   e_AttSpecKind tgf,
   Component*& pCompo,               // Describes how the string was build
   unsigned int& uiNormedLenVal      // Where to put its normalized length
) {
   const int NORMSEP = pSyntax->pQuantity->value_of(NORMSEP_QTY_ix);
   const int NAMELEN = pSyntax->pQuantity->value_of(NAMELEN_QTY_ix);
   const int LITLEN = pSyntax->pQuantity->value_of(LITLEN_QTY_ix);
   e_CharClass chclsFirstChar = NAMESTRT_CHCLS;
   e_CharClass chclsOtherChar = NAME_CHCLS;
   boolean isNameCase = FALSE;
   boolean isPlural = FALSE;
   boolean isSpecifInvalid = FALSE;
   boolean isCdata = FALSE;
#if ((__BCPLUSPLUS__ == 0x310) || (__BCPLUSPLUS__ == 0x320))
   // Borland C++ 3.1 and 4.0 need a typedef (?)
   typedef boolean (Attribute::*CheckFunction)(ParseState *, const char *) const;
   CheckFunction check_item = 0;
#else
   boolean (Attribute::*check_item)(ParseState *, const char *) const = 0;
#endif

   pCompo = 0;
   uiNormedLenVal = NORMSEP;

   if (!pAtt) {                          // just get rid of AVS
      isSpecifInvalid = TRUE;
   }else {
      if (1 & pAtt->declared_value()) {  // isolate plural
         isPlural = TRUE;
      }else {
         isPlural = FALSE;
      }
      isNameCase = pSyntax->is_case_general();
      switch (pAtt->declared_value()) {
      case GROUP_ADV:
         chclsFirstChar = NAME_CHCLS;
         /* fall thru */
      case NOTATION_ADV:
#ifndef __GNUG__
         check_item = (
           (tgf==DEF_TGF) || (pAtt->default_value() == CURRENT_AFV)
         )
		  ? &Attribute::check_list_and_class 
		  : &Attribute::check_list ;
#else
	check_item = 0 ;
	if ( (tgf == DEF_TGF) || (pAtt->default_value() == CURRENT_AFV) ) {
		check_item = &Attribute::check_list_and_class ;
	}
	else {
		check_item = &Attribute::check_list ;
	}
#endif
         break;
      case ENTITY_ADV:
      case ENTITIES_ADV:
         isNameCase = pSyntax->is_case_entity();
         break;
      case CDATA_ADV:                   // nothing to check
         isCdata = TRUE;
         /* fall thru */
      case NMTOKEN_ADV:
      case NMTOKENS_ADV:
         chclsFirstChar = NAME_CHCLS;
         break;
      case NUTOKEN_ADV:
      case NUTOKENS_ADV:
         chclsFirstChar = DIGIT_CHCLS;
         break;
      case NUMBER_ADV:
      case NUMBERS_ADV:
         chclsFirstChar = chclsOtherChar = DIGIT_CHCLS;
         break;
      case ID_ADV:
         check_item = &Attribute::check_id;
         /* fall thru */
      default:   /* NAME(s)_ADV IDREF(s)_ADV */
         break;
      }
   }
   switch (get_curr_token()) {
   case LIT_TK:
   case LITA_TK:
      {
         /*------------------------------------------------------------+
         | Derive the attribute value from the attribute value literal |
         +------------------------------------------------------------*/
         int init_ent_level = pEh->curr_entity_level();
         unsigned int iListCount = 0;
         e_LexToken delim_token = get_curr_token();
         boolean isSpaceWaiting = FALSE;
         boolean isMssNow = FALSE;
         boolean isMssBefore = FALSE;
         char *pchCurr, *pszToken;
         char chNew;
         ComponentBuilder compobldr(pchBuf);

         pchCurr = pszToken = pchBuf;
         pPrsmode->push_mode(LITC_MODE);
         while (
            (isMssNow) ||
            (recognize_delimiter() != delim_token) ||
            (pEh->curr_entity_level() != init_ent_level)
         ) {
            isMssBefore = isMssNow;
            isMssNow = FALSE;        // curr_token is NOT_FOUND
            if (pEh->is_Ee()) {
               pPrsmode->enable_markup_recognition();
               if (pEh->curr_entity_level() == init_ent_level) {
                  errorh(
                     UBLNLIT, pEh, ERH_T_SEV,
                     pSyntax->pDlmstr->value_of(
                        (delim_token == LIT_TK)? LIT_VDL_ix : LITA_VDL_ix
                     )
                  );
               }else {               // Ignore Ee (7.9.3) and pursue
                  compobldr.record_Ee(pchCurr);
                  pEh->close();      // does read in previous file
               }
               continue;
            }
            switch (get_curr_token()) {
            case NOT_FOUND_TK:
            case CRO_TK:
            case LIT_TK:
            case LITA_TK:
               switch (get_curr_token()) {
               case CRO_TK:
                  switch (pSyntax->translate_character_ref(pEh, get_curr_count())) {
                  case INVALID_CHRF:
                     continue;       // has been discarded if bad
                  default:           // NAMED_CHRF, NUMERIC_CHRF
                     break;
                  }
                  break;
               case LIT_TK:
               case LITA_TK:
                  pEh->unread(get_curr_count());
                  break;
               default: /* NOT_FOUND_TK */
                  break;
               }
               if (!pEh->is_numchar()) {         // if *not* a numeric char ref
                  if (pEh->is_separator()) {     // RE, RS, SPACE, SEPCHAR
                     if (pEh->is_RSchar()) {
                        pEh->read();
                        continue;                // ignore RS
                     }
                     if (!isCdata) {
                        if (pchCurr > pchBuf) isSpaceWaiting = TRUE;
                        pEh->read();
                        continue;                // reduce SPACE sequences
                     }
                     chNew = (char)pSyntax->get_SPACEchar(); // don't reduce
                     break;
                  }
                  /*
                  | MSO/MSI/MSS CHAR recognition
                  */
                  if (pEh->is_ms_char()) {       // mso, msi, mss
                     if (pEh->is_msochar()) {    // 9.7:
                        if (!isMssBefore) pPrsmode->disable_markup_recognition();
                     }else if (pEh->is_msichar()) {
                        pPrsmode->enable_markup_recognition();
                     }else {                     // mss
                        isMssNow = TRUE;
                     }
                  }
               }
               chNew = (char)pEh->curr_char_val();
               break;

            case ERO_TK:                         // LITC_MODE only
               {
                  EntityDef *pEnt = open_general_entity_reference();

                  if (!pEnt) continue;
                  switch (pEnt->text_type()) {
                  case SDATA_EFK:
                  case CDATA_EFK:
                     if (!isPlural) uiNormedLenVal += NORMSEP; // 7.9.4.5, a + b
                     pPrsmode->disable_markup_recognition();
                     break;
                  default:
                     /*
                     | Describe components of only S/CDATA entities.
                     | To describe all entities, do a break instead, and...
                     | redo the Component class to handle nested entities.
                     */
                     continue;
                  }
                  /*
                  | To decompose others attributes than simply CDATA attribute,
                  | change the following if.  Warning: you can't do this for
                  | the AVS of a default value (at DEF_TGF time), that is a
                  | token in a group (name token or notation group): see attldcl.
                  | Only AVL of CDATA att are concerned at today (12/11/90).
                  */
                  if (isCdata) compobldr.record_ent_start(pchCurr, pEnt);
                  continue;
               }
            default:                             // ERO_GRPO
               /*
               | This cannot occur in start tag (see 9.4.4)
               | ERH_T makes the programs exits.
               | Recovery? try to skip the name-group
               */
               errorh(INVDLM, pEh, ERH_T_SEV);
               break;
            }
            if (!isSpecifInvalid) {
               /*--------------------------------------------------+
               | If a token just ended, make first specific checks |
               +--------------------------------------------------*/
               if (isSpaceWaiting) {             // never happens if isCdata
                  isSpaceWaiting = FALSE;
                  *pchCurr = '\0';
                  if (!isPlural) {
                     errorh(BIGTKLIST, pEh, ERH_E_SEV);
                     isSpecifInvalid = TRUE;
                  }else if (pchCurr > pszToken + NAMELEN) {
                     isSpecifInvalid = TRUE;
                     errorh(BIGNAME, pEh, ERH_E_SEV);
                  }else if (
                     (check_item != 0) && (!(pAtt->*check_item)(this, pszToken))
                  ) {
                     isSpecifInvalid = TRUE;
                  }else {
                     uiNormedLenVal += NORMSEP;   // check later for overflow
                     *pchCurr++ = (char)pSyntax->get_SPACEchar(); //space anyway
                     pszToken = pchCurr;
                     ++iListCount;
                  }
               }
               /*-----------------------------+
               | Then enter the new character |
               +-----------------------------*/
               if (++uiNormedLenVal > LITLEN) {
                  errorh(BIGAVL, pEh, ERH_S_SEV, uiNormedLenVal);
                  isSpecifInvalid = TRUE;
               }else if (isCdata) {
                  *pchCurr++ = chNew;
               }else {
                  if (!pSyntax->is_char_class(
                        chNew,
                        (pchCurr == pszToken)? chclsFirstChar : chclsOtherChar
                  )) {
                     errorh(INVATTSPC, pEh, ERH_E_SEV);
                     isSpecifInvalid = TRUE;
                  }else {
                     *pchCurr++ = (char)(
                        isNameCase? pSyntax->uppercase(chNew) : chNew
                     );
                  }
               }
            }
            pEh->read();          // get next_char ready, loop for-loop
         }                           // end of for-loop

         /*---------------------+
         | End of literal found |
         +---------------------*/
         if (!isSpecifInvalid) {
            *pchCurr = '\0';
            if (!isCdata) {
               ++iListCount;
               if (pchCurr == pchBuf) {
                  errorh(INVEMPTY, pEh, ERH_E_SEV); // 7.9.4.1
                  isSpecifInvalid = TRUE;
               }else if (pchCurr > pszToken + NAMELEN) {
                  errorh(BIGNAME, pEh, ERH_E_SEV);
                  isSpecifInvalid = TRUE;
               }else if (
                  (check_item != 0) && (!(pAtt->*check_item)(this, pszToken))
               ) {
                  isSpecifInvalid = TRUE;
               }else if (
                  (isPlural) &&
                  (uiNormedLenVal += NORMSEP, uiNormedLenVal > LITLEN)
               ) {
                  errorh(BIGAVL, pEh, ERH_E_SEV, uiNormedLenVal);
               }
            }
         }
         pCompo = compobldr.terminate(pchCurr);
         pPrsmode->pop_mode();
         if (!isSpecifInvalid) {
            return isCdata? 1 : iListCount;
         }
         if (pCompo) {
            free(pCompo);
            pCompo = 0;
         }
      }
      break;
   case NOT_FOUND_TK:
      {
         char *pchCurr = pchBuf;

         if (
            ((!pSgmldcl->features.isShorttag) && (tgf!=DEF_TGF)) || // 7.9.3.1 a
            (!pSyntax->is_char_class(pEh->curr_char_val(), chclsFirstChar))
         ) {
            if (!isSpecifInvalid) {
               errorh(INVATTSPC, pEh, ERH_E_SEV);
               isSpecifInvalid = TRUE;
            }
            chclsOtherChar = NAME_CHCLS;
         }
         do {
            *pchCurr++ = (char)(isNameCase ?
               pSyntax->uppercase(pEh->curr_char_val()) : pEh->curr_char_val()
            );
         }while (
            (++uiNormedLenVal <= LITLEN) &&
            (pSyntax->is_char_class(pEh->read(), chclsOtherChar))
         );
         if (!isSpecifInvalid) {
            if (uiNormedLenVal > LITLEN) {
               errorh(BIGAVL, pEh, ERH_S_SEV, uiNormedLenVal);
            }else if (pchCurr > pchBuf + (isCdata? LITLEN : NAMELEN)) {
               errorh(isCdata? BIGLIT : BIGNAME, pEh, ERH_E_SEV);
            }else {
               *pchCurr = '\0';
               if ((check_item == 0) || ((pAtt->*check_item)(this, pchBuf))) {
                  return 1;
               }
            }
         }
      }
      break;
   default:
      if (!isSpecifInvalid) errorh(INVATTSPC, pEh, ERH_E_SEV);
      pEh->unread(get_curr_count());        // Not for us!
      break;
   }
   uiNormedLenVal = 0;
   return 0;
}

/*---------------------------------------------------------------att_spec_list-+
| Parse an Attribute Specification List.                                       |
| TAG_MODE (or ASL_NMO) is assumed to be active!                               |
|                                                                              |
|  [31][32] attribute specification list =                                     |
|      (s*, (name, s*, vi, s*)?, attribute value specification)*               |
|                                                                              |
| When entering this routine:                                                  |
|    - curr_char after the GI + the S separators                               |
|      (or curr_token set, if not NOT_FOUND_TK).                               |
|    Arguments, pointed by arg_p:                                              |
|    - pAttList,                                                               |
|      - if not 0, points a copy of the attribute definition list.  Update     |
|        this attribute list, by changing the "value" pointers according to    |
|        the new specifications and doing malloc for these values.             |
|      - if 0, this means the ASL (if any) must be discarded                   |
|    - TGF is one of TAG_TGF / DATA_TGF                                        |
|      TAG_TGF: at run time, when encountering a Tag Attribute Specification   |
|      DATA_TGF: at compile time for a Data Attribute Specif (<!ENTITY MD)     |
|    - uiNormedLenVal: where to put the sum of of the normalized length        |
|      for each attribute value.                                               |
|                                                                              |
| Returns: One of the asl_rc values                                            |
+-----------------------------------------------------------------------------*/
e_RcAttSpecList ParseState::att_spec_list(
   const AttributeList *pAttlist,  // Input: 0 if must get rid of ASL
   e_AttSpecKind tgf,              // Input: TAG_TGF or DATA_TGF
   unsigned int& uiNormedLenVal    // Output: new normalized length
)
{
   const int NORMSEP = pSyntax->pQuantity->value_of(NORMSEP_QTY_ix);
   const int LITLEN = pSyntax->pQuantity->value_of(LITLEN_QTY_ix);
   uiNormedLenVal = 0;
   boolean isConrefSpecFound = FALSE;
   boolean isValidSpecifFound = FALSE;
   unsigned int uiNormedLenName = 0;
   char *pszLit;

   if ((!pAttlist) && (get_curr_token() == NOT_FOUND_TK)) {
      errorh(INVALS, pEh, ERH_E_SEV);            // tell: throwing the ASL
   }
   pszLit = (char *)malloc(LITLEN+1);
   for (; get_curr_token() == NOT_FOUND_TK; skip_S()) {
      if (pEh->is_Ee()) {
         errorh(UBLNTAG, pEh, ERH_T_SEV);
         break;
      }
      unsigned int iListCount;
      unsigned int uiNewNormedLenName;
      unsigned int uiNewNormedLenVal;
      int ixAtt;
      Attribute *pAtt;
      Component *pCompo;

      pSyntax->read_nametoken(pEh, pszLit);      // att_name, or avs ?
      skip_S();
      if (get_curr_token() == VI_TK) {           // it's an attribute name
         /*-------------------------------------+
         |Att name found: find if any attribute |
         |name of the ATTLIST could match it    |
         +-------------------------------------*/
         uiNewNormedLenName = strlen(pszLit) + NORMSEP;
         skip_S();

         if (!pAttlist) {
            pAtt = 0;
         }else {
            pAtt = pAttlist->find_by_name(pszLit, &ixAtt);
            if (ixAtt == -1) {                   // doesn't exist
               errorh(INVATTDEF, pEh, ERH_E_SEV, pszLit);
               pAtt = 0;                         // get rid of AVS
            }else if (pAtt->is_specified()) {    // else, if duplicated,
               errorh(DUPATT, pEh, ERH_E_SEV, pAtt->name());
               pAtt = 0;                         // get rid of AVS
            }
         }
         iListCount = att_val_spec(
            pszLit, pAtt, tgf, pCompo, uiNewNormedLenVal
         );
         if (!iListCount) continue;
      }else {                                    // No VI:
         pEh->unread(get_curr_count());          // this delimiter not for us
         if (!pSgmldcl->features.isShorttag) {
            errorh(INVATTSPC, pEh, ERH_E_SEV);   // see 7.1.9.2
            continue;                            // loop for next AVS
         }
         if (!pAttlist) continue;                // loop for next AVS
         /*-----------------------------------+
         | AVS found without att name, vi:    |
         | Find if any token in any GROUP_AKN |
         | or NOTATION_AKN could match it     |
         +-----------------------------------*/
         pAtt = pAttlist->find_by_value(pszLit, &ixAtt);
         if (ixAtt == -1) {                      // doesn't exist
            errorh(NOATKDEF, pEh, ERH_E_SEV,pszLit);
            continue;
         }
         if (pAtt->is_specified()) {             // Check if duplicated
            errorh(DUPATT, pEh, ERH_E_SEV, pAtt->name());
            continue;
         }
         iListCount = 1;
         uiNewNormedLenName = 0;
         uiNewNormedLenVal = strlen(pszLit);
         pCompo = 0;
      }
      if (
         (pAtt->default_value() == FIXED_AFV) &&
         (strcmp(pszLit, pAtt->value()))
      ) {
         errorh(INVFXATT, pEh, ERH_E_SEV);
         if (pCompo) free(pCompo);
         continue;
      }
      /*-----------------------------------------+
      | Here, the Attribute Spec is valid:       |
      | - pAtt points the ATT                    |
      | - iListCount is the list count ( >= 1)   |
      | - pszLit points the AVS                  |
      | - pCompo possibly points its composition |
      +-----------------------------------------*/
      isValidSpecifFound = TRUE;
      if (pAtt->default_value() == CONREF_AFV) isConrefSpecFound = TRUE;
      pAttlist->reset_spec(
         ixAtt, pszLit, iListCount, pCompo,
         uiNewNormedLenName? FALSE : TRUE,
         aulCapacityCount[ATTCHCAP_ix]
      );
      uiNormedLenName += uiNewNormedLenName;
      uiNormedLenVal += uiNewNormedLenVal;
      /* loop for next AVS */
   }
   free(pszLit);                                   // free temporary buffer
   if (uiNormedLenVal + uiNormedLenName > pSyntax->pQuantity->value_of(
         ATTSPLEN_QTY_ix
      )
   ) {
      errorh(BIGAVS, pEh, ERH_E_SEV, uiNormedLenVal + uiNormedLenName);
   }
   if (isConrefSpecFound) {
      return CONREF_CHANGED_ASLRC;
   }else if (isValidSpecifFound) {
      return CHANGED_ASLRC;
   }else {
      return NO_CHANGE_ASLRC;
   }
}

/*------------------------------------------------------------finish_parse_tag-+
| This routine is called to parse the rest of start-tag or end-tag.            |
|                                                                              |
| Set caller's pointer to an AttributeList that reflects the attribute spec    |
| list, or 0 if we must get rid of any AVS.                                    |
| Return TRUE if a CONREF attribute was assigned, making the content EMPTY.    |
|                                                                              |
| The purpose of this is to store the Attribute Specification in a convenient  |
| form while it is parsed.  It doesn't mean the list is fully valid: more      |
| checking is done later by validate_start_element (REQUIRED, for instance).   |
|                                                                              |
| When entering:                                                               |
|      - curr_char is the next char that follows the GI                        |
|      - "pAttList" points the address of the default attribute list, or it is |
|        0 to get a rid of any attributes: end tags, irrelevant start tags.    |
|                                                                              |
| When returning:                                                              |
|      - pAttlist is updated to reflect the specified attributes               |
|      - curr_char is:                                                         |
|          - either the character that follows a NET or a TAGC                 |
|          - or *AT* the ETAGOs, STAGOs delimiters                             |
|      - curr_token has still the last value found: this trick allows to       |
|        recognize a NET-enabling start tag                                    |
|      - if a CONREF attribute has been found, return TRUE.  Otherwise, FALSE. |
|                                                                              |
+-----------------------------------------------------------------------------*/
boolean ParseState::finish_parse_tag(
   const AttributeList *pAttlist          // zero if must get rid of the AVS
) {
   unsigned int uiNormedLenVal;

   pPrsmode->push_mode(TAG_MODE);
   skip_S();
   boolean isConrefSpecFound = (
      att_spec_list(pAttlist, TAG_TGF, uiNormedLenVal) == CONREF_CHANGED_ASLRC
   )? TRUE : FALSE;

   switch (get_curr_token()) {
   case TAGC_TK:
   case NET_TK:                           // a NET-enabling start_tag
      /*------------------------------------+
      | This is not allowed if SHORTTAG NO, |
      | for empty elements, and for end tag |
      | These errors are caught later.      |
      +------------------------------------*/
      break;
   case ETAGO_GRPO_TK:
   case STAGO_GRPO_TK:
   case ETAGO_TK:
   case ETAGO_TAGC_TK:
   case STAGO_TK:
   case STAGO_TAGC_TK:
      pEh->unread(get_curr_count());      // Could never be recognized:
      if (pSgmldcl->features.isShorttag) {
         break;                           // no contextual constraints
      }                                   // else, rule 7.4.1 and 7.5.1
   default:
      errorh(NOTAGC, pEh, ERH_E_SEV);
      break;
   }
   pPrsmode->pop_mode();
   return isConrefSpecFound;              // to check EMPTY content
}

/*------------------------------------------------------validate_start_element-+
| Validate an explicitely entered, or inferred start tag                       |
+-----------------------------------------------------------------------------*/
void ParseState::validate_start_element(
   const ElementType *pElem,
   const AttributeList& attlist,
   boolean isInferred,
   boolean isExplConref
)
{
   if (isInferred) {                         // Inferred, not parsed
      if (!pSgmldcl->features.isOmittag) {
         errorh(INVSTGOMI, pEh, ERH_E_SEV, pElem->name());
      }else if (pElem->is_reqd_open_tag()) {   // If it had to ...
         errorh(INVSTRTAG, pEh, ERH_E_SEV, pElem->name());
      }
   }
   /*
   | Check if the element was declared vs. generated by odtcomp
   */
   if ((pElem->is_any()) && (!pElem->is_pcdata())) {
      errorh(INVELEMGID, pEh, ERH_E_SEV, pElem->name());
   }
   if (pSgmldcl->features.isRank) {
      pElem->set_rank_suffices();
   }

   /*
   |  Validate the attribute list
   */
   const int GRPCNT = pSyntax->pQuantity->value_of(GRPCNT_QTY_ix);
   boolean isSpecifRequired;

   if ((pSgmldcl->features.isShorttag) || (pSgmldcl->features.isOmittag)) {
      isSpecifRequired = FALSE;
   }else {
      isSpecifRequired = TRUE;
   }

   int iNbrOfEntityNames = 0;
   int iNbrOfIdrefNames = 0;
   Attribute *pAtt;

   for (int ixAtt=0; (pAtt = attlist[ixAtt]) != 0; ++ixAtt) {
      if (!pAtt->is_link()) {                // LINKs are transparent
         switch (pAtt->default_value()) {
         case REQUIRED_AFV:
            if (!pAtt->is_specified()) {     // value not explicitely set
               errorh(NORQDATT, pEh, ERH_E_SEV, pAtt->name());
            }
            break;
         case CONREF_AFV:                    // CONREF is impliable
         case IMPLIED_AFV:
            break;                           // IMPLIED needs no specif
         case CURRENT_AFV:
            if (!pAtt->value()) {
               errorh(NOCURATT, pEh, ERH_E_SEV, pAtt->name());
               break;
            }
            /* Else, fall thru; */
         default:
            if ((!pAtt->is_specified()) && (isSpecifRequired)) {
               errorh(NOATTSPC, pEh, ERH_E_SEV, pAtt->name());
            }
            break;
         }
         switch (pAtt->declared_value()) {
         case ENTITIES_ADV:
         case ENTITY_ADV:
            if (pAtt->value()) {
               StringList strlst;
               char *psz;

               strlst.add_tokens(pAtt->value(), (char)pSyntax->get_SPACEchar());
               iNbrOfEntityNames += strlst.count();
               for (int i=0; psz = strlst[i], psz; ++i) {
                  EntityDef *pEnt = find_general_entity(psz);
                  if (pEnt != 0) {
                     if ((pEnt->is_rcdata()) || (pEnt->text_type() == PI_EFK)) {
                        errorh(INVENTATT, pEh, ERH_E_SEV, psz);  // by 7.9.4.3
                     }else if (pEnt->is_external()) {
                        pEnt->check_ref_validity(pEh, CON_MODE);
                     }
                  }
               }
            }
            break;
         case IDREFS_ADV:
         case IDREF_ADV:
            if (pAtt->value()) {
               /*
               |  Activate IDREF and IDREFS (specified or defaulted)
               */
               StringList strlst;
               char *psz;

               strlst.add_tokens(pAtt->value(), (char)pSyntax->get_SPACEchar());
               iNbrOfIdrefNames += strlst.count();
               for (int i=0; psz = strlst[i], psz; ++i) {
                   ExpnhashId.gimme_value(psz, sizeof(int));
               }
            }
            break;
         case ID_ADV:
            if (pAtt->is_specified()) ++(aulCapacityCount[IDCAP_ix]);
            break;
         case NOTATION_ADV:
            if (pAtt->value()) {
               NotationType *pNcb;

               pNcb = (NotationType *)pDtddcl->pHshNotation->search_value(
                  pAtt->value()
               );
               if (!pNcb) {
                  errorh(BADNOTA, pEh, ERH_E_SEV, pAtt->value());
               }else {
                  // 7.9.4.4:
                  if (isExplConref) errorh(INVNOTASP, pEh, ERH_E_SEV);
                  // cast: find violates the constness of ExternId
                  ((ExternId *)pNcb->external_id())->find(pEh);
               }
            }
            break;
         default:
            break;
         }
      }
   }
   aulCapacityCount[IDREFCAP_ix] += iNbrOfIdrefNames;
   if (iNbrOfIdrefNames > GRPCNT) {
      errorh(BIGASIDR, pEh, ERH_E_SEV, iNbrOfIdrefNames);
   }
   if ((unsigned int)iNbrOfEntityNames > GRPCNT) {
      errorh(BIGASENT, pEh, ERH_E_SEV, iNbrOfEntityNames);
   }
}

/*--------------------------------------------------------validate_end_element-+
| Validate an explicitely entered, or inferred end tag                         |
+-----------------------------------------------------------------------------*/
void ParseState::validate_end_element(
   const ElementType *pElem,
   boolean isInferred,
   boolean isPremature,
   boolean isBadEmpty
)
{
   if (isPremature) {                     // premature
      errorh(OOCETG, pEh, ERH_S_SEV, pElem->name());
   }
   if (isInferred) {                      // Inferred, not parsed
      if (!pSgmldcl->features.isOmittag) {
         errorh(INVETGOMI, pEh, ERH_E_SEV, pElem->name());
      }else if (pElem->is_reqd_end_tag()) { // if it had to ...
         errorh(INVENDTAG, pEh, ERH_E_SEV, pElem->name());
      }
   }
   if (isBadEmpty) {                      // Empty content not allowed
      errorh(INVEMPCON, pEh, ERH_E_SEV, pElem->name());
   }
}
/******************************************************************************/
