/*-------------------------------------------------------------------------
                PROJECT YAO OBJECT SGML PARSER (OSGML)
                     Version 1.0 Alpha level 4.0
   An SGML Parser conforming to International Standard ISO 8879 --
                 Standard Generalized Markup Language
  -------------------------------------------------------------------------
(C) 1987-1994 IBM Corporation

Licensed to Yuan-ze Institute of Technology (YUAN-ZE) for distribution
under the terms of the License and Disclaimer of Warranties for Project
YAO Materials, which includes the following text:

"YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
worldwide, non-exclusive license to use, execute, reproduce, display,
perform and distribute copies of, and to prepare derivative works
based upon these materials; and (2) the right to authorize others to
do any of the foregoing."

The full text of the License and Disclaimer of Warranties for Project YAO
Materials should be consulted before using these materials.
  -------------------------------------------------------------------------

               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE

Only a "final" version is subject to the YUAN-ZE license described
above.  A version is final if it is not identified as "Alpha" or
"Beta", or if no later version is released within 60 days.  Interim
releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// utilctl.c
///  Manages utility files above the entity handler
//   Pierre G. Richard 92/09/11 - Petrus Services and Consulting, Ltd.
  -------------------------------------------------------------------------*/

//******************************************************************************

/*--------------+
| Include Files |
+--------------*/
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <new.h>
#include "gendef.h"
#include "utilctl.h"
#include "strlst.h"
#include "sgmldcl.h"
#include "enthdlr.h"
#include "errorh.h"

/*-----------------------------------------------------------UtilityIdentifier-+
|                                                                              |
+-----------------------------------------------------------------------------*/
UtilityIdentifier::UtilityIdentifier(
   const char *pszNameArg,
   boolean isActiveArg,
   boolean isBaseDtdArg,
   e_LinkKind lkkindArg
) : pszName(duplicate_string(pszNameArg)), exid(pszName, INVALID_XFC)
{
   position = 0;
   isActive = isActiveArg;
   isBase = isBaseDtdArg;
   lkkind = lkkindArg;
   isExidSet = FALSE;
}

/*----------------------------------------------------------~UtilityIdentifier-+
|                                                                              |
+-----------------------------------------------------------------------------*/
UtilityIdentifier::~UtilityIdentifier()
{
   free(pszName);
}

/*-----------------------------------------------------------------Constructor-+
|                                                                              |
+-----------------------------------------------------------------------------*/
UtilityControl::UtilityControl()
{
   aUtidTable = (UtilityIdentifier *)calloc(2, sizeof(UtilityIdentifier));
   iUtidCount = 2;
   pUtidLast = aUtidTable + iUtidCount - 1;
}

/*------------------------------------------------------------------Destructor-+
| Destructor                                                                   |
+-----------------------------------------------------------------------------*/
UtilityControl::~UtilityControl()
{
   for (int ix=2; ix < iUtidCount; ++ix) {
      aUtidTable[ix].~UtilityIdentifier(); // delete each UtilityIdentifier
   }
   free(aUtidTable);                       // Free the utid table
}

/*--------------------------------------------------------------------new_utid-+
| Add a new entry to the utid table                                            |
| Check its name does not duplicate any other.                                 |
|                                                                              |
| If this DOCTYPE or LINKTYPE is listed in the sstrlActiveType,                |
| or if it is a base DOCTYPE -- which is always active                         |
|   - check that maximums are not reached,                                     |
|   - mark it as active (turn on the UTID.is.active bit)                       |
|   - remove it from the ActiveType list                                       |
|                                                                              |
| Return a pointer to the UtilityIdentifier, or 0 if something's wrong.        |
|                                                                              |
| The ActiveType list is a list to keep all requests from the application to   |
| activate a link (and later, a concurrent dtd).  At the time the Application  |
| requests it, the names are just entered in the active-list with              |
| no other checks.                                                             |
|                                                                              |
| When parsing the prolog, each time a LINKTYPE (or, later, DOCTYPE) is        |
| found, if the name also exists in the active-list, it is removed and added   |
| to the utid-table.                                                           |
|                                                                              |
| After the prolog ends:                                                       |
|   - all active links are listed in the utid-table,                           |
|   - the active-list must be empty, or the Application required to activate   |
|     an unknown link                                                          |
|                                                                              |
| IMPLEMENTATION NOTE:                                                         |
|                                                                              |
| A base DTD must always been entered first!                                   |
+-----------------------------------------------------------------------------*/
UtilityIdentifier * UtilityControl::new_utid(
   EntityHandler *pEh,
   SGMLdcl *pSgmldcl,
   SortedStringList& sstrlActiveType,
   boolean isBaseDtd,
   const char *pszNameType,
   e_LinkKind lkkind
)
{
   boolean isActive;

   if (isBaseDtd) {
      /*
      | Base do not need to be explicitely activated
      */
      sstrlActiveType.remove(pszNameType);  // be nice
      isActive = TRUE;
   }else {
      int iCountDtd = 0;
      int iCountSimple = 0;
      int iCountImplicit = 0;
      int iCountExplicit = 0;

      {
         UtilityIdentifier *pUtid = aUtidTable  + iUtidCount - 1;  // point last
         do {
            if (!(strcmp(pszNameType, pUtid->pszName))) {
               errorh(DUPNAMTYP, pEh, ERH_E_SEV, pszNameType);
               return 0;
            }
            switch (pUtid->lkkind) {
            case SIMPLE_LKKIND:
               if (pUtid->isActive) ++iCountSimple;
               break;
            case IMPLICIT_LKKIND:        // NOT YET IMPLEMENTED!
               if (pUtid->isActive) ++iCountImplicit;
               break;
            case EXPLICIT_LKKIND:
               if (pUtid->isActive) ++iCountExplicit;
               break;
            default:                     // Then, this must be a DTD
               ++iCountDtd;
               break;
            }
         }while (!(pUtid--)->isBase);
      }
      if (
         ((pSgmldcl->features.isSimple)   && (lkkind==SIMPLE_LKKIND)) ||
         ((pSgmldcl->features.isImplicit) && (lkkind==IMPLICIT_LKKIND)) ||
         ((pSgmldcl->features.isExplicit) && (lkkind==EXPLICIT_LKKIND)) ||
         ((pSgmldcl->features.isConcur)   && (lkkind==NOT_A_LINK_LKKIND))
      ) {
         isActive = sstrlActiveType.remove(pszNameType);
      }else {
         isActive = FALSE;   // So it stays in the ActiveType list, if any
      }
      switch (lkkind) {
      case SIMPLE_LKKIND:             // see 7.1, 12.1.1.1
         if ((isActive) && (iCountSimple > pSgmldcl->features.iMaxSimple)) {
            errorh(BIGLINK, pEh, ERH_E_SEV);
         }
         return 0;
      case IMPLICIT_LKKIND:           // NOT YET IMPLEMENTED!
         errorh(NIMPL, pEh, ERH_T_SEV, 1005); /*IMPLICIT*/
         return 0;
      case EXPLICIT_LKKIND:           // NOT YET IMPLEMENTED!
         errorh(NIMPL, pEh, ERH_T_SEV, 1006); /*EXPLICIT*/
         return 0;
      default:                        // Then, new is a DTD
         /*--------------------------------------------------------------+
         | Surprisingly, the number of DTD is limited by CONCUR (7.2.1), |
         | not by the number of activated concurrent types.              |
         +--------------------------------------------------------------*/
         if (iCountDtd > pSgmldcl->features.iMaxConcur) {  // see 7.1, 7.2.1
            errorh(BIGDTD, pEh, ERH_S_SEV);
         }
         return 0;
      }
   }
   aUtidTable = (UtilityIdentifier *)realloc(
      aUtidTable, ++iUtidCount * sizeof (UtilityIdentifier)
   );
   pUtidLast = aUtidTable + iUtidCount - 1;

   new(pUtidLast) UtilityIdentifier(pszNameType, isActive, isBaseDtd, lkkind);
   return pUtidLast;
}

/*---------------------------------------------------------build_compiled_exid-+
| Given the external id associated to a (source) DTD, or LPD this functions    |
| returns the corresponding external id associated to its compiled form:       |
| ODT, or OLT.                                                                 |
| This also works for SGML declarations (Local, Reference), except that these  |
| have no external id.  Therefore, this code does not expect `pExidSource'     |
| to be set when the xfcCompiled is one if UTSYNL or UTSYNR_XFC.               |
|                                                                              |
| -- INVARIANTS --                                                             |
| The initialization of the utility identifier is performed, if not already    |
| done by a previous call.                                                     |
|                                                                              |
| -- INPUT --                                                                  |
| `pExidSource' points to the external id describing the source form           |
| `xfcCompiled' is one of the UTxxxx_XFC of the text_class enum                |
|                                                                              |
| For DTD or OLT, the utility identifier of the compiled form *must* exist     |
| in the UtilityControl class: hence, created once by `new_utid' when the      |
| DTD or LPD declaration was parsed.  The initialization of the utility        |
| identifier is performed, if it was not already done by a previous call.      |
| Hence:                                                                       |
|                                                                              |
| -- PRECONDITION --                                                           |
|                                                                              |
|  - the UtilityControl class (`aUtidTable') knows the names of all DOCTYPE    |
|    and LINKTYPE found (so far) in prolog,                                    |
|  - build_compiled_exid is called with pExidSource having at least the        |
|    pszName member set to one of these names.                                 |
|                                                                              |
| -- OUTPUT --                                                                 |
|  pointer to the ExternId of the compiled form                                |
|                                                                              |
| -- POSTCONDITION --                                                          |
|  if no external id found, a NULL pointer is returned.                        |
|                                                                              |
+-----------------------------------------------------------------------------*/
ExternId * UtilityControl::build_compiled_exid(
   e_TextClass xfcCompile, ExternId *pExidSource
)
{
   UtilityIdentifier *pUtid;

   switch (xfcCompile) {
   case UTSYNL_XFC:               // Local OCS
   case UTSYNR_XFC:               // RefCon OCS
      pUtid = aUtidTable;         // utid_tbl[0] for Local or RefCon
      break;
   case UTSYNU_XFC:               // User OCS
      pUtid = aUtidTable + 1;     // utid_tbl[1] for the User OCS
      break;
   default:                       // doc type name or link type name
      /*
      | Do a global search
      */
      {
         pUtid = aUtidTable + 2;
         const char *pszName = pExidSource->name();
         for (int ix=2; ix < iUtidCount; ++ix, ++pUtid) {
            if (!strcmp(pszName, pUtid->pszName)) {
              break;              // Found
            }
         }
         if (ix == iUtidCount) return 0; // Cannot occur, but just in case...
      }
      break;
   }
   if (!pUtid->isExidSet) {
      if (pExidSource) {
         new(&(pUtid->exid)) ExternId(*pExidSource, pUtid->pszName, xfcCompile);
      }else {
         new(&(pUtid->exid)) ExternId(pUtid->pszName, xfcCompile);
      }
      pUtid->isExidSet = TRUE;
   }else {
      pUtid->exid.assign_class(0, xfcCompile, FALSE);  // reset class
   }
   return &(pUtid->exid);
}

/*--------------------------------------------------------------find_utid_base-+
| Find UtilityIdentifier of base DTD                                           |
| Return NULL if not found.                                                    |
+-----------------------------------------------------------------------------*/
UtilityIdentifier * UtilityControl::find_utid_base()
{
   UtilityIdentifier *pUtid;
   int i;

   for (
      i = iUtidCount, pUtid = aUtidTable  + iUtidCount - 1;  // point last
      !pUtid->isBase;
      --pUtid, --i
   ) {
     if (i <= 2) return 0;
   }
   return pUtid;
}

/*-------------------------------------------------------------------is_active-+
|Check if a document type or a link type is active                             |
+-----------------------------------------------------------------------------*/
boolean UtilityControl::is_active(char *pszName)
{
   for (
      UtilityIdentifier *pUtid = aUtidTable  + iUtidCount - 1;  // point last
      !pUtid->isBase;
      --pUtid
   ) {
      if ((pUtid->isActive) && (!strcmp(pszName, pUtid->pszName))) {
         return TRUE;
      }
   }
   return FALSE;
}
/******************************************************************************/
