/*-------------------------------------------------------------------------
         PROJECT YAO PORTABLE OBJECT-ORIENTED ENTITY MANAGER (POEM)
                         Version 1.0 Alpha level 2.0
  -------------------------------------------------------------------------
  (C) 1993-1994 Yuan-ze Institute of Technology
 
  Distributed under the terms of the License and Disclaimer of Warranties
  for Project YAO Materials, which includes the following text:
 
  "YUAN-ZE hereby grants to any user: (1) an irrevocable royalty-free,
  worldwide, non-exclusive license to use, execute, reproduce, display,
  perform and distribute copies of, and to prepare derivative works based
  upon these materials; and (2) the right to authorize others to do any of
  the foregoing."
 
  The full text of the License and Disclaimer of Warranties for Project YAO
  Materials should be consulted before using these materials.
  -------------------------------------------------------------------------
               INTERIM RELEASE RESTRICTED DISTRIBUTION NOTICE
 
     Only a "final" version is subject to the YUAN-ZE license described
       above. A version is final if it is not identified as "Alpha" or
     "Beta", or if no later version is released within 60 days.  Interim
     releases are confidential and restricted to the original recipient.
  -------------------------------------------------------------------------
//// poem.cc
///  Portable Object-oriented Entity Manager
//   Erik Naggum 1994-04-12
  -------------------------------------------------------------------------*/
 
#include <assert.h>
#include <string.h>
#include <stdlib.h>
 
#include <iostream.h>
 
#ifdef unix
#include <fcntl.h>
#include <unistd.h>
#endif
#ifdef __BCPLUSPLUS__
#include <io.h>
#endif
 
#include "poem.h"
 
 
POEM::StorageObject::StorageObject (const char *object)
: position(0), dimspec(0), error(0), openhandle(0), status(closed)
{
    objectname = new char[strlen (object) + 1];
    strcpy (objectname, object);
}
 
POEM::StorageObject::StorageObject (const StorageObject& copy)
: position(0), dimspec(0), error(0), openhandle(0), status(closed)
{
    objectname = new char[strlen (copy.objectname) + 1];
    strcpy (objectname, copy.objectname);
}
 
POEM::StorageObject::~StorageObject (void)
{
    delete [] objectname;
}
 
// open -- open the object, prepare for reading substring
int POEM::StorageObject::open (DimSpec *indimspec)
{
    assert (status != invalid);
    assert (indimspec);
 
    dimspec = indimspec;
 
    // open file if not already open
    if (status == closed && (openhandle = sys_open (objectname)) == -1)
        return status = invalid, 0;
    status = opened;
 
    // if file markers are already set, we do not need to recompute
    if (dimspec->getfile (1) == -1) {
        // internal copies
        long raw_1 = dimspec->getraw (1);
        long raw_2 = dimspec->getraw (2);
        long file_1;
        long file_2;
 
        // marker 1
        file_1 = sys_seek (openhandle, raw_1-(raw_1>0)+(raw_2+1)*(raw_2<0), 2*(raw_1<0));
        if (file_1 == -1) goto seek_failure_exception;
        position = file_1;
 
        // marker 2
        if (raw_2 == -1) file_2 = -1;
        else {
            // compute real end-marker
            if (raw_2 > 0) file_2 = file_1 + raw_2;
            else {
                file_2 = sys_seek (openhandle, raw_2 + 1, 2);
                if (file_2 == -1) goto seek_failure_exception;
            }
        }
        dimspec->setfile (file_1, file_2);
    }
 
    if (sys_seek (openhandle, dimspec->getfile(1), 0) == dimspec->getfile(1))
        return 1;
 
seek_failure_exception:
    status = invalid;
    close();
    return 0;
}
 
// read into the buffer
// update the position
int POEM::StorageObject::read (char * buffer, int maxread)
{
    assert (status != invalid);
 
    if (status == ateof) return 0;
 
    long end = dimspec->getfile (2);
 
    // TODO: handle non-seekable files
    if (end < 0 && end != -1) {
        error = "unsupported operation";
        return -1;
    }
 
    // re-open and seek if closed since last time
    if (status == closed
        && ((openhandle = sys_open (objectname)) == -1
            || sys_seek (openhandle, position, 0) != position)) {
        close ();
        status = invalid;
        error = "object won't reopen";
        return -1;
    }
 
    // compute how much to read
    long can_read = (end == -1 || (maxread < end - position))?
        maxread : end - position;
 
    // only one read operation.  if sys_read does more, fine
    int did_read = sys_read (openhandle, buffer, int(can_read));
    if (did_read == -1) {
        close ();
        status = invalid;
        error = "object won't read";
        return -1;
    }
 
    // special processing at end of physical file
    if (did_read == 0) {
        if (end == -1)
            dimspec->setfile (dimspec->getfile(1), position);
        status = ateof;
    }
 
    // update position and test for end of substring
    position += did_read;
    if (position == end)
        status = ateof;
 
    // return number of characters written
    return did_read;
}
 
int POEM::StorageObject::close (void)
{
    assert (status != invalid);
 
    if (status == closed)
        return 1;
 
    status = closed;
    return sys_close (openhandle);
}
 
long POEM::StorageObject::tell (void) const
{
    assert (status != invalid);
    return position;
}
 
long POEM::StorageObject::seek (long pos, int whence)
{
    assert (status != invalid);
    return position = sys_seek (openhandle, pos, whence);
}
 
static struct rbisymbol {
    char * name;                        // initial * means parse RE RS
    int RE;
    int RS;
} rbisym[] = {
    { "NONE", -1, -1 },
    { "DEFAULT", 0, 0 },                // will be filled in later
    { "LF", '\n', 0 },
    { "CR", '\r', 0 },
    { "CRLF", '\r', '\n' },
    { "LFCR", '\n', '\r' },
    { "*RERS", 0, 0 },                  // will be filled in when parsed
    { "RMS", -2, -2 },
    { 0, 0, 0 }
};
 
POEM::ObjSpec::ObjSpec (const char *storage, const char *sos, const char *rbi, const char * dimspeclist)
: object(0), dimspec(0), inRE(0), inRS(0), status(invalid), overrun (trunc)
{
    assert (sos);
 
    // default storage
    if (!storage) storage = "FILE";
 
    // storage and sos
    POEM::StorageSystem::creator create = POEM::StorageSystem::getsystem (storage);
    if (create == 0) return;
    object = create(sos);
 
    // rbi
    if (rbi) {
        for (rbisymbol * rbisymp = rbisym; rbisymp->name; rbisymp++) {
            int symlen = ::strlen (rbisymp->name);
            if (rbisymp->name[0] == '*'
                && !strncmp (rbi, rbisymp->name + 1, symlen - 1)
                && rbi[symlen - 1] == ' ') {
                char * start = (char *)rbi + symlen;
                if ((inRE = (int)::strtol (start, &start, 0)) == 0) goto unwind;
                inRS = (int)::strtol (start, &start, 0);
                break;
            }
            if (!strcmp (rbi, rbisymp->name)) {
                inRE = rbisymp->RE;
                inRS = rbisymp->RS;
                break;
            }
        }
        if (!rbisymp->name) goto unwind;
    }
 
    // dimspeclist
    if (dimspeclist) {
        for (int pass = 0; pass < 2; pass++) {
            long m1, m2;
            int count = 0;
            char * start = (char *)dimspeclist;
            while (1) {
                if ((m1 = ::strtol (start, &start, 0)) == 0) break;
                if ((m2 = ::strtol (start, &start, 0)) == 0) goto unwind;
                if (pass == 1)
                    dimspec[count] = new DimSpec (m1, m2);
                ++count;
            }
            if (pass == 0) {
                dimspec = new DimSpec *[count + 1];
                for (int i = 0; i <= count; i++)
                    dimspec[i] = 0;
            }
            while (start[0] == ' ') start++;
            if (!strcmp (start, "ERROR")) overrun = error;
        }
    } else {
        dimspec = new DimSpec *[2];
        dimspec[0] = new DimSpec (1, -1);
        dimspec[1] = 0;
    }
 
    status = parsed;
    return;
 
unwind:
    this->POEM::ObjSpec::~ObjSpec ();
}
 
POEM::ObjSpec::~ObjSpec (void)
{
    if (dimspec) {
        for (int i = 0; dimspec[i]; i++) {
            delete dimspec[i]; dimspec[i] = 0;
        }
        delete [] dimspec; dimspec = 0;
    }
    delete object; object = 0;
}
 
POEM::Entity::Entity (const char * entname, const char * object, int literal)
: name(entname), location(0), adjustring(0), status(invalid)
{
//  cerr << "In Entity constructor" << endl;
//  cerr << "   Parm entname is " << entname << endl;
//  cerr << "   Parm object  is " << object  << endl;
//  cerr << "   Parm literal is " << literal << endl;
 
    buffer[0] = 0;
    buffer[1] = 0;
 
    bufsize[0] = 0;
    bufsize[1] = 0;
 
    syntax.RE = 13;
    syntax.RS = 10;
    syntax.EE = 0;
    syntax.QUE = 127;
 
    // entity is internal string
    if (literal) {
        objspec = new ObjSpec *[2];
        objspec[0] = new ObjSpec ("LITERAL", object, "NONE");
        objspec[1] = 0;
        status = closed;
        return;
    }
 
    // object specification has quoted characters in it, so we lose
    if (strchr (object, '\'') ||
        strchr (object, '"')) {
        objspec = new ObjSpec *[2];
        objspec[0] = new ObjSpec ("LITERAL", "<?ERROR unsupported object specification syntax>", "NONE");
        objspec[1] = 0;
        status = closed;
        return;
    }
 
    // simple object specification
    if (strstr (object, "::") == 0) {
        objspec = new ObjSpec *[2];
        objspec[0] = new ObjSpec ("FILE", object);
        objspec[1] = 0;
        status = closed;
        return;
    }
 
    // complex object specification
    for (int pass = 0; pass < 2; pass++) {
        char * copy = new char[strlen (object) + 1];
        strcpy (copy, object);
        char * nextobj = copy;
        for (int count = 0; nextobj; count++) {
            char * storage = nextobj;   // storage system type
            char * sos;                 // storage object specification
            char * rbi = 0;             // record boundary indicator
            char * dimspeclist = 0;
            if ((nextobj = strstr (nextobj, "//")) != 0)
                *nextobj = 0, nextobj += 2;
            if ((sos = strstr (storage, "::")) != 0)
                *sos = 0, sos += 2;
            else return;                // syntax error
            if ((rbi = strstr (sos, "::")) != 0) {
                *rbi = 0, rbi += 2;
                if ((dimspeclist = strstr (rbi, "::")) != 0)
                    *dimspeclist = 0, dimspeclist += 2;
            }
            if (pass)
                objspec[count] = new ObjSpec (storage, sos, rbi, dimspeclist);
        }
        delete [] copy; copy = 0;
        if (pass == 0) {
            objspec = new ObjSpec *[count + 1];
            objspec[count] = 0;
        }
    }
 
    // sanity checking
    for (int i=0; objspec[i]; i++)
        if (!objspec[i]->valid()) {
            this->Entity::~Entity();
            return;
        }
 
    status = closed;
    return;
}
 
POEM::Entity::~Entity (void)
{
    if (status == opened)
        close ();
    if (objspec) {
        for (int i = 0; objspec[i]; i++) {
            delete objspec[i]; objspec[i] = 0;
        }
        delete [] objspec; objspec = 0;
    }
 
}
 
// internal character constants.  enum fails on itty-bitty machines
// do not touch these values.
 
const POEM::character cSPECIAL = 0177770;
const POEM::character cRSI     = 0177770; // -8 0001
const POEM::character cRS      = 0177771; // -7 0002
const POEM::character cRE      = 0177772; // -6 0004
const POEM::character cSOE     = 0177773; // -5 0010
const POEM::character cSSE     = 0177774; // -4 0020
const POEM::character cQUE     = 0177775; // -3 0040
const POEM::character cEE      = 0177776; // -2 0100
const POEM::character cEND     = 0177777; // -1 0200
 
int POEM::Entity::open (int maxlookahead)
{
    return open (0, maxlookahead);
}
 
int POEM::Entity::open (Location * at, int maxlookahead)
{
    assert (status != invalid);
    assert (!at || at->valid(this));
 
    // if already open, close is on the safe side
    if (status == opened)
        close ();
 
    // find first power of 2 >= maxlookahead
    while ((maxlookahead & -maxlookahead) != maxlookahead)
         maxlookahead += maxlookahead & -maxlookahead;
    lookahead = maxlookahead;
 
    if (at) location = new InternalLocation (at->location);
    else location = new InternalLocation ();
    adjustring = new AdjustRing (16);
 
    // fill initial buffers
    if (!fillbuffer (0) || !fillbuffer (1)) {
        close ();
        status = invalid;
        return 0;
    }
 
    exhausted = 0;
    status = opened;
    return 1;
}
 
void POEM::Entity::close (void)
{
    assert (status != invalid);
    if (status == closed)
        return;
 
    if (objspec[location->write.objspec])
        objspec[location->write.objspec]->getobject()->close();
 
    delete [] buffer[1]; buffer[1] = 0;
    delete [] buffer[0]; buffer[0] = 0;
    delete adjustring; adjustring = 0;
    delete location; location = 0;
 
    status = closed;
}
 
void POEM::Entity::setsyntax (int RE, int RS, int EE, int QUE)
{
    syntax.RE  = RE;
    syntax.RS  = RS;
    syntax.EE  = EE;
    syntax.QUE = QUE;
}
 
int POEM::Entity::swapbuffer (void)
{
    // swap buffers
    character * temp = buffer[1];
    buffer[1] = buffer[0];
    buffer[0] = temp;
 
    // swap buffer location info
    location->buffer[0] = location->buffer[1];
 
    // reload buffer
    if (!fillbuffer (1)) {
        close ();
        status = invalid;
        return 0;
    }
    return 1;
}
 
inline int POEM::Entity::stuff (int which, int where, int what)
{
    // cerr << "In Entity::stuff  which is " << which << endl;
    // cerr << "   where is " << where << " what is " << what << endl;
    // cerr << "   bufsize[which] is " << bufsize[which] << endl;
    if (where >= bufsize[which]) {
        // cerr << "   inside where >= bufsize[which]" << endl;
        if (buffer[which]) {
            // cerr << "   inside buffer[which]" << endl;
            character *oldbuffer = buffer[which];
            buffer[which] = new character[bufsize[which] <<= 1];
            memcpy (buffer[which], oldbuffer, bufsize[which]);
            delete [] oldbuffer;
        } else {
            // cerr << "   inside else " << endl;
            buffer[which] = new character[bufsize[which] = 32];
        }
    }
    // cerr << "   before return " << endl;
    return buffer[which][where] = what;
}
 
// TODO: error checking
int POEM::Entity::fillbuffer (int which)
{
    if (exhausted)                      // not an error
        return 1;
 
    char diskbuf[512];                  // read into this
    int payload = 0;                    // number of "real" characters
    int virgin = 0;                     // special treatment of first RS
    int needopen = 0;                   // we need to open the object
 
    // initialization; which == 0 only for first fillbuffer
    if (which == 0) {
        location->write = location->buffer[0];
        virgin = location->write.position == 0;
    }
 
    // shorthands
    int& currobjspec = location->write.objspec;
    int& currdimspec = location->write.dimspec;
    long& currfilepos = location->write.position;
    ObjSpec * curr = objspec[currobjspec];
    StorageObject * currobj = curr->getobject ();
 
    // initialization, part 2
    if (virgin)
        needopen = 1;
    else {
        currobj->open (curr->getdimspec (currdimspec));
        currobj->seek (currfilepos);
    }
 
    // for posterity
    location->buffer[which] = location->write;
 
    while (payload < lookahead) {
        if (needopen) {
 
            // cerr << "In Entity::fillbuffer needopen" << endl;
            // cerr << "     Before open() status is " << status << endl;
 
            currobj->open (curr->getdimspec (currdimspec));
 
            //status = opened;
            // cerr << "In Entity::fillbuffer" << endl;
            // cerr << "     After open() status is " << status << endl;
 
            currfilepos = currobj->tell ();
 
            // cerr << "In Entity::fillbuffer" << endl;
            // cerr << "     After tell() status is " << status << endl;
 
            needopen = 0;
        }
 
        // cerr << "In Entity::fillbuffer" << endl;
        // cerr << "     After needopen status is " << status << endl;
 
        int bytes = currobj->read (diskbuf, sizeof diskbuf);
 
        // cerr << "In Entity::fillbuffer" << endl;
        // cerr << "     After read status is " << status << endl;
        // cerr << "     Bytes read are " << bytes << endl;
 
        if (bytes) {
            int i = 0;
            currfilepos += bytes;
            // cerr << "In Entity::fillbuffer curr->RE is " << curr->RE() << endl;
            switch (curr->RE()) {
            case -2:                    // RMS
                // cerr << "In Entity::fillbuffer case RMS" << endl;
                if (!virgin++)
                    stuff (which, payload++, cRS);
                while (i < bytes)
                    stuff (which, payload++, diskbuf[i++]);
                stuff (which, payload++, cRE);
                break;
            case -1:                    // NONE
                // cerr << "In Entity::fillbuffer case NONE" << endl;
                while (i < bytes)
                    stuff (which, payload++, diskbuf[i++]);
                break;
            case 0:                     // DEFAULT
                // cerr << "In Entity::fillbuffer case DEFAULT" << endl;
                // cerr << "\t i is" << i << endl;
                // cerr << "\t bytes is" << bytes << endl;
                while (i < bytes) {
                    char c = diskbuf[i++];
                    // cerr << "\t working on char " << c << endl;
                    int nl0 = 0, nl1 = 0;
                    // test for widowed terminator
                    if (lastseen) {
                        // cerr << "\t in lastseen char " << c << endl;
                        nl0 = lastseen;
                        nl1 = c;
                        lastseen = 0;
                        if (nl0 == '\r')
                            if (nl1 != '\n')
                                nl1 = 0;
                        if (nl0 == '\n')
                            if (nl1 != '\r')
                                nl1 = 0;
                        curr->setRERS (nl0, nl1);
                        stuff (which, payload++, cRE);
                        stuff (which, payload++, nl1? cRS: cRSI);
                        if (!nl1) stuff (which, payload++, c);
                        goto know_rbi;
                    }
                    // possibly a line terminator
                    if (c == '\n' || c == '\r') {
                        lastseen = c;
                        continue;
                    }
                    // not a line terminator
                    // cerr << "\t before call to stuff" << endl;
                    stuff (which, payload++, c);
                    // cerr << "\t after call to stuff" << endl;
                }
                break;
            default:                    // the rest
            know_rbi:                   // continue here after DEFAULT
                // cerr << "In Entity::fillbuffer case know_rbi" << endl;
                int RE = curr->RE();
                int RS = curr->RS();
                while (i < bytes) {
                    int c = diskbuf[i++];
                    if (lastseen) {
                        lastseen = 0;
                        if (c == RS) {
                            stuff (which, payload++, cRE);
                            stuff (which, payload++, cRS);
                            continue;
                        }
                    }
                    if (c == RE) {
                        if (RS) {
                            lastseen = RE;
                            continue;
                        }
                        stuff (which, payload++, cRE);
                        stuff (which, payload++, cRSI);
                        continue;
                    }
                    stuff (which, payload++, c);
                }
            }
        }
 
        // cerr << "In Entity::fillbuffer" << endl;
        // cerr << "     Before checking eof() status is " << status << endl;
 
        if (currobj->eof()) {
            // something about overrun handling here, I guess.
            // substring end detected
            if (curr->getdimspec (++currdimspec)) {
                stuff (which, payload++, cSSE);
                needopen = 1;
                continue;
            }
            // storage object end detected
            currobj->close();
            currdimspec = 0;
            if (objspec[++currobjspec]) {
                curr = objspec[currobjspec];
                currobj = curr->getobject ();
                stuff (which, payload++, cSOE);
                needopen = 1;
                continue;
            }
            // entity end detected
            if (lastseen)
                if ((curr->RE() == 0 || curr->RE() == lastseen))
                    stuff (which, payload++, cRE);
                else stuff (which, payload++, lastseen);
            if (payload && (buffer[which][payload-1] == cRS ||
                            buffer[which][payload-1] == cRSI))
                --payload;
            stuff (which, payload++, cEE);
            exhausted = 1;
            break;
        }
    }
    stuff (which, payload, cEND);
    return 1;
}
 
POEM::character POEM::Entity::peek (int ahead)
{
    while (status == opened) {
        Adjust& top = adjustring->top();
        if (ahead >= top.pos) {
            character c = buffer[top.buf][ahead + top.adj];
            if (c < cSPECIAL) return c;
            if (c <= cRS) return syntax.RS;
            if (c <= cRE) return syntax.RE;
            if (c <= cSSE) {
                adjustring->push (ahead, 1);
                return peek (ahead);
            }
            if (c <= cQUE) return syntax.QUE;
            if (c <= cEE) {
                adjustring->push (ahead, -1);
                return syntax.EE;
            }
            // c == cEND
            adjustring->push (ahead, ahead, 1);
            continue;
        }
 
        // walk through adjust ring
        for (int i = 1; !adjustring->bottom (i); i++) {
            Adjust& adj = adjustring->item (i);
            if (ahead >= adj.pos)
                return buffer[adj.buf][ahead + adj.adj];
        }
 
        // if the adjust ring had lost it, it means the ring was too small for
        // our needs, so we double its size in the hopes that this will help.
        // initial size is small, so we waste as little memory as possible.
        // we have to walk through the buffer up to the point where the first
        // record we have is located.
 
        AdjustRing * oldring = adjustring;
        int newsize = 2 * oldring->size ();
        adjustring = new AdjustRing (newsize);
        adjustring->reset (location->point);
        Adjust& old = oldring->item (i - 1);
        for (int j = 0; j <= old.pos; j++) peek (j);
        top = adjustring->top ();
        assert (old.pos == top.pos && old.adj == top.adj);
        while (--i) adjustring->push (oldring->item (i));
        delete oldring;
    }
    return syntax.EE;
}
 
int POEM::Entity::step (int ahead)
{
    int buf = 0;                // always start in buf 0
    character c;
    InternalLocation& here = *location;
 
    // step forward, and count interesting things
    for (int i = 0; i < ahead; i++) {
        c = buffer[buf][here.point];
        int bit = 1<<((c < cSPECIAL)? 1 + ~cSPECIAL : c - cSPECIAL);
        if (bit & 0446) here.read.position++; // actually bytes
        if (bit & 0447) here.chars++;
        if (bit & 0440) here.columns++;
        if (bit & 0003) here.records++, here.columns = 0;
        if (bit & 0010) here.read.objspec++, here.read.dimspec = 0;
        if (bit & 0020) here.read.dimspec++;
        if (bit & 0230) ahead++;
        if (bit & 0100) break;
        if (bit & 0200) buf++, here.point = 0;
        if (bit & 0477) here.point++;
    }
 
    // switch buffers if necessary
    if (buf) swapbuffer ();
 
    // prepare for new peek
    adjustring->reset (here.point);
    return c != cEE;
}
 
POEM::InternalLocation::InternalLocation ()
{
    chars = columns = records = 1;
    point = 0;
    buffer[0].position = buffer[0].objspec = buffer[0].dimspec = 0;
    buffer[1].position = buffer[1].objspec = buffer[1].dimspec = 0;
    read.position = read.objspec = read.dimspec = 0;
    write.position = write.objspec = write.dimspec = 0;
}
 
POEM::InternalLocation::InternalLocation (InternalLocation *at)
{
    *this = *at;
}
 
POEM::Entity::Location::Location (const POEM::Entity * in)
: source(in)
{
    location = new InternalLocation (in->location);
}
 
POEM::StorageSystem **POEM::StorageSystem::list;
int POEM::StorageSystem::systems;
int POEM::StorageSystem::maxsystems = 32;
 
void POEM::StorageSystem::initsystem (void)
{
    systems = 0;
    list = new StorageSystem *[maxsystems];
    addsystem ("FILE", POEM::FileStorageObject::create);
    addsystem ("LITERAL", POEM::LiteralStorageObject::create);
    addsystem ("STREAM", POEM::StreamStorageObject::create);
}
 
int POEM::StorageSystem::addsystem (const char *name, creator create)
{
    if (list == 0) initsystem ();
 
    // TODO: make list grow, instead
    assert (systems + 1 < maxsystems);
 
    if (!getsystem (name)) {
        list[systems++] = new StorageSystem (name, create);
        return 1;
    }
    return 0;
}
 
POEM::StorageObject * (*POEM::StorageSystem::getsystem(const char *name))(const char *)
{
    if (list == 0) initsystem ();
 
    for (int i = 0; i < systems; i++)
        if (!strcmp (name, list[i]->name))
            return list[i]->create;
    return 0;
}
 
POEM::StorageObject * POEM::FileStorageObject::create (const char * object)
{
    return new FileStorageObject (object);
}
 
POEM::FileStorageObject::FileStorageObject (const char * object)
: POEM::StorageObject (object)
{
#ifdef OPENFILELIMIT
    for (int i = 0; i < OPENFILELIMIT; i++)
        openlist[i] = 0;
#endif
}
 
POEM::FileStorageObject::~FileStorageObject (void)
{
    if (status == opened || status == ateof)
        close ();
#ifdef OPENFILELIMIT
    for (int i = 0; i < OPENFILELIMIT; i++)
        if (openlist[i]) {
            openlist[i]->close ();
            openlist[i] = 0;
        }
#endif
}
 
int POEM::FileStorageObject::sys_open (const char *filename)
{
    // cerr << "In FileStorageObject::sys_open" << endl;
    // cerr << "     Trying to open file " << filename << endl;
 
    int fd = ::open (filename, 0, 0);
#ifdef OPENFILELIMIT
    for (int i = 0; i < OPENFILELIMIT; i++)
        if (openlist[i] == 0)
            break;
    if (openlist[i] != 0) {
        --i;
        openlist[i]->close ();
        openlist[i] = 0;
    }
    while (i > 0)
        openlist[i] = openlist[--i];
    openlist[0] = this;
#endif
    return fd;
}
 
long POEM::FileStorageObject::sys_seek (int fd, long pos, int whence)
{
    return ::lseek (fd, pos, whence);
}
 
int POEM::FileStorageObject::sys_read (int fd, char *buf, int len)
{
    return ::read (fd, buf, len);
}
 
int POEM::FileStorageObject::sys_close (int fd)
{
#ifdef OPENFILELIMIT
    for (int i = 0; i < OPENFILELIMIT; i++)
        if (openlist[i] == this) {
            openlist[i] = 0;
            break;
        }
    while (++i < OPENFILELIMIT)
        openlist[i-1] = openlist[i];
#endif
    return ::close (fd) == 0;
}
 
POEM::StorageObject * POEM::StreamStorageObject::create (const char * object)
{
    return new StreamStorageObject (object);
}
 
POEM::StreamStorageObject::StreamStorageObject (const char * object)
: POEM::StorageObject (object)
{ }
 
POEM::StreamStorageObject::~StreamStorageObject (void)
{ }
 
int POEM::StreamStorageObject::sys_open (const char *stream)
{
    int fd = (int)::strtol (stream, 0, 0);
    return fd;
}
 
long POEM::StreamStorageObject::sys_seek (int fd, long pos, int whence)
{
    return ::lseek (fd, pos, whence);   // may fail
}
 
int POEM::StreamStorageObject::sys_read (int fd, char *buf, int len)
{
    return ::read (fd, buf, len);
}
 
int POEM::StreamStorageObject::sys_close (int)
{
    return 1;                           // not closed
}
 
POEM::StorageObject * POEM::LiteralStorageObject::create (const char * object)
{
    return new LiteralStorageObject (object);
}
 
POEM::LiteralStorageObject::LiteralStorageObject (const char * object)
: POEM::StorageObject (object)
{ }
 
POEM::LiteralStorageObject::~LiteralStorageObject (void)
{ }
 
int POEM::LiteralStorageObject::sys_open (const char *)
{
    return 1;                   // we always open objectname
}
 
long POEM::LiteralStorageObject::sys_seek (int fd, long pos, int whence)
{
    assert (fd == 1);
 
    if (whence == 0) return pos;
    if (whence == 1) return position + pos;
    if (whence == 2) return strlen (objectname) + pos;
    return -1;
}
 
int POEM::LiteralStorageObject::sys_read (int fd, char *buf, int len)
{
    assert (fd == 1);
 
    for (long pos = position; len-- && ((*buf++ = objectname[int(pos)]) != 0); pos++)
        ;
    return int(pos - position);
}
 
int POEM::LiteralStorageObject::sys_close (int fd)
{
    assert (fd == 1);
 
    return 1;                   // true
}
 
POEM::AdjustRing::AdjustRing (int size)
{
    // find next larger power of 2
    while ((size & -size) != size)
        size += size & -size;
 
    ringsize = size;
    ring = new Adjust[size];
    ringtop = 0;
    reset (0);
}
 
POEM::AdjustRing::~AdjustRing (void)
{
    delete [] ring;
}
 
void POEM::AdjustRing::reset (int pos)
{
    ringbot = modsize (ringtop - 1);
    ring[ringtop].pos = 0;
    ring[ringtop].buf = 0;
    ring[ringtop].adj = pos;
}
 
void POEM::AdjustRing::push (int pos, int delta, int buf)
{
    int wasringtop = ringtop;
    if (ring[wasringtop].pos != pos) {
        ringtop = modsize (ringtop + 1);
        if (ringtop == ringbot)
            ringbot = modsize (ringbot + 1);
        ring[ringtop].pos = pos;
    }
    if (buf) {
        ring[ringtop].buf = buf;
        ring[ringtop].adj = -delta;
    } else {
        ring[ringtop].buf = ring[wasringtop].buf;
        ring[ringtop].adj = ring[wasringtop].adj + delta;
    }
}
 
void POEM::AdjustRing::push (Adjust old)
{
    ringtop = modsize (ringtop + 1);
    if (ringtop == ringbot)
        ringbot = modsize (ringbot + 1);
    ring[ringtop] = old;
}
 
// local variables:
// mode: c++
// end:
